#include "coloredit.h"

#include <spinner.h>
#include <interface/StringView.h>
#include <app/Message.h>

enum { COLOR_CHANGED };

static CPoint get_max_width( CPoint* pcFirst, ... )
{
	CPoint cSize = *pcFirst;
	va_list hList;

	va_start( hList, pcFirst );

	CPoint* pcTmp;
	while( (pcTmp = va_arg( hList, CPoint* )) != NULL )
	{
		if ( pcTmp->x > cSize.x )
		{
			cSize = *pcTmp;
		}
	}
	va_end( hList );
	return( cSize );
}

ColorEdit::ColorEdit( CRect cFrame, const char* pzName, const Color32_s& sColor ) : CView( cFrame, pzName )
{
	m_sColor = sColor;

	m_pcRedSpinner   = new Spinner( CRect(0,0,0,0), "red_spinner", sColor.red, new CMessage( COLOR_CHANGED ) );
	m_pcGreenSpinner = new Spinner( CRect(0,0,0,0), "green_spinner", sColor.green, new CMessage( COLOR_CHANGED ) );
	m_pcBlueSpinner  = new Spinner( CRect(0,0,0,0), "blue_spinner", sColor.blue, new CMessage( COLOR_CHANGED ) );

	m_pcRedStr   = new CStringView( CRect(0,0,0,0), "red_txt", "Red", ALIGN_CENTER );
	m_pcGreenStr = new CStringView( CRect(0,0,0,0), "green_txt", "Green", ALIGN_CENTER );
	m_pcBlueStr  = new CStringView( CRect(0,0,0,0), "blue_txt", "Blue", ALIGN_CENTER );

	m_pcRedSpinner->SetMinPreferredSize( 3 );

	AddChild( m_pcRedSpinner );
	AddChild( m_pcGreenSpinner );
	AddChild( m_pcBlueSpinner );

	m_pcRedSpinner->SetFormat( "%.0f" );
	m_pcRedSpinner->SetMinMax( 0.0f, 255.0f );
	m_pcRedSpinner->SetStep( 1.0f );
	m_pcRedSpinner->SetScale( 1.0f );

	m_pcGreenSpinner->SetFormat( "%.0f" );
	m_pcGreenSpinner->SetMinMax( 0.0f, 255.0f );
	m_pcGreenSpinner->SetStep( 1.0f );
	m_pcGreenSpinner->SetScale( 1.0f );

	m_pcBlueSpinner->SetFormat( "%.0f" );
	m_pcBlueSpinner->SetMinMax( 0.0f, 255.0f );
	m_pcBlueSpinner->SetStep( 1.0f );
	m_pcBlueSpinner->SetScale( 1.0f );

	AddChild( m_pcRedStr );
	AddChild( m_pcGreenStr );
	AddChild( m_pcBlueStr );

	FrameResized( 0.0f, 0.0f );
}

void ColorEdit::MessageReceived( CMessage* pcMessage )
{
	switch( pcMessage->what )
	{
		case COLOR_CHANGED:
			SetValue( Color32_s( m_pcRedSpinner->Value(), m_pcGreenSpinner->Value(), m_pcBlueSpinner->Value(), 255 ) );
			break;
		default:
			CView::MessageReceived( pcMessage );
			break;
	}
}

void ColorEdit::SetValue( const Color32_s& sColor )
{
	m_sColor = sColor;

	m_pcRedSpinner->SetValue( float( sColor.red ) );
	m_pcGreenSpinner->SetValue( float( sColor.green ) );
	m_pcBlueSpinner->SetValue( float( sColor.blue ) );

	Color32_s sOldCol = ViewColor();
	SetViewColor( m_sColor );
	DrawFrame( m_cTestRect, FRAME_RECESSED );
	SetViewColor( sOldCol );
}


const Color32_s& ColorEdit::GetValue() const
{
	return( m_sColor );
}


void ColorEdit::AllAttached()
{
	m_pcRedSpinner->SetTarget( this );
	m_pcGreenSpinner->SetTarget( this );
	m_pcBlueSpinner->SetTarget( this );
}


void ColorEdit::GetPreferredSize( float* outWidth, float* outHeight )
{
	m_pcRedSpinner->GetPreferredSize( outWidth, outHeight );

	if (outWidth)
	{
		*outWidth *= 3.0f;
		*outWidth += 8.0f;
	}

	if (outHeight)
	{
		*outHeight *= 5.0f;
	}
}


void ColorEdit::FrameResized( float inWidth, float inHeight )
{
	CRect cBounds = Bounds();
	float width, height;

	m_pcRedSpinner->GetPreferredSize(&width, &height);
	width = 16 + m_pcRedSpinner->StringWidth( "8888" );

	CPoint cRStrSize, cGStrSize, cBStrSize;
	m_pcRedStr->GetPreferredSize(&cRStrSize.x, &cRStrSize.y);
	m_pcGreenStr->GetPreferredSize(&cGStrSize.x, &cGStrSize.y);
	m_pcBlueStr->GetPreferredSize(&cBStrSize.x, &cBStrSize.y);
	CPoint cStrSize = get_max_width( &cRStrSize, &cGStrSize, &cBStrSize, NULL );

	CRect cStrRect( 0, 0, cStrSize.x - 1, cStrSize.y - 1 );
	CRect cSpinnerRect( 0, 0, width - 1.0f, height - 1.0f );

	if ( cStrSize.y < height )
	{
		cStrSize += CPoint( 0, (height - cStrSize.y) / 2  );
	}
	else
	{
		height += ((cStrSize.y - height) / 2 );
	}

	m_pcRedSpinner->SetFrame( cSpinnerRect + CPoint( 4, cStrSize.y + 8 ) );
	m_pcGreenSpinner->SetFrame( cSpinnerRect + CPoint( (cBounds.Width()+1.0f) / 2 - width / 2, cStrSize.y + 8 ) );
	m_pcBlueSpinner->SetFrame( cSpinnerRect + CPoint( (cBounds.Width()+1.0f) - width - 4, cStrSize.y + 8 ) );

	m_pcRedStr->SetFrame( cStrRect + CPoint( 4, 4 ) );
	m_pcGreenStr->SetFrame( cStrRect + CPoint( (cBounds.Width()+1.0f) / 2 - cStrSize.x / 2, 4 ));
	m_pcBlueStr->SetFrame( cStrRect + CPoint( cBounds.Width() + 1.0f - cStrSize.x - 4, 4 ) );

	m_cTestRect = CRect( 10, cSpinnerRect.right + 14, cBounds.right - 10, cBounds.bottom - 10 );
}


void ColorEdit::Draw( CRect cUpdateRect )
{
	SetViewColor( get_default_color( COL_NORMAL ) );
	DrawFrame( Bounds(), FRAME_RECESSED );

	SetViewColor( m_sColor );
	DrawFrame( m_cTestRect, FRAME_RECESSED );
}

