#include <assert.h>
#include <stdio.h>
#include <sys/stat.h>

#include <interface/Window.h>
#include <interface/Menu.h>

#include <app/Application.h>
#include <app/Message.h>

#include <cosmoe/kernel.h>

using namespace os;

#include "fontpanel.h"
#include "screenpanel.h"
#include "windowpanel.h"
#include "keyboardpanel.h"

#include <interface/TabView.h>


class MyWindow : public CWindow
{
public:
    MyWindow( const CRect& cFrame );
    ~MyWindow();

    virtual void        MessageReceived( CMessage* pcMessage );
      // From Window:
    virtual bool        OkToQuit();
private:
    void SetupMenus();
    CMenu* m_pcMenu;
    TabView* m_pcView;
};


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void MyWindow::SetupMenus()
{
	CRect cMenuFrame = Bounds();
	CRect cMainFrame = cMenuFrame;
	cMenuFrame.bottom = 16;

	m_pcMenu = new CMenu( cMenuFrame, "Menu",
						CF_FOLLOW_LEFT | CF_FOLLOW_RIGHT | CF_FOLLOW_TOP,
						WID_FULL_UPDATE_ON_H_RESIZE,
						ITEMS_IN_ROW );

	CMenu* pcItem1 = new CMenu( CRect( 0, 0, 100, 20 ), "File", ITEMS_IN_COLUMN, CF_FOLLOW_LEFT | CF_FOLLOW_TOP );

	pcItem1->AddItem( "Quit", new CMessage( M_QUIT ) );

	m_pcMenu->AddItem( pcItem1 );

	//cMenuFrame.bottom = m_pcMenu->GetPreferredSize( false ).y - 1;
	m_pcMenu->GetPreferredSize( NULL, &cMenuFrame.bottom );
	cMenuFrame.bottom -= 1.0f;
	cMainFrame.top = cMenuFrame.bottom + 1;

	m_pcMenu->SetFrame( cMenuFrame );

	m_pcMenu->SetTargetForItems( this );

	AddChild( m_pcMenu );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

MyWindow::MyWindow( const CRect& cFrame ) :
    CWindow( cFrame, "Preferences", WT_UNTYPED_WINDOW, 0 )
{
    SetupMenus();

    CRect cMainFrame = Bounds();
    CRect cMenuFrame = m_pcMenu->Frame();

    cMainFrame.top = cMenuFrame.bottom + 1.0f;

    cMainFrame.Resize( 5, 5, -5, -5 );
  
    m_pcView = new TabView( cMainFrame, "main_tab", CF_FOLLOW_ALL, WID_WILL_DRAW | WID_FULL_UPDATE_ON_RESIZE );
    CView* pcFontPanel = new FontPanel( cMainFrame );
    CView* pcScreenPanel = new ScreenPanel( cMainFrame );
    CView* pcWindowPanel = new WindowPanel( cMainFrame );
    CView* pcKeyboardPanel = new KeyboardPanel( cMainFrame );

    m_pcView->AppendTab( "Keyboard", pcKeyboardPanel );
    m_pcView->AppendTab( "Default Fonts", pcFontPanel );
    m_pcView->AppendTab( "Screen Mode", pcScreenPanel );
    m_pcView->AppendTab( "Appearance", pcWindowPanel );
  
    AddChild( m_pcView );
    Activate( true );
    pcFontPanel->MakeFocus( true );

	float width, height;

	m_pcView->GetPreferredSize(&width, &height);
	width  += 10.0f;
	height += cMenuFrame.bottom + 11.0f;
    //Point cMinSize = m_pcView->GetPreferredSize(false) + Point( 10.0f, cMenuFrame.bottom + 11.0f );
    //Point cMaxSize = m_pcView->GetPreferredSize(true);
    ResizeTo( width, height );
    SetSizeLimits( width, height, width, height );

    char* pzHome = getenv( "HOME" );
    if ( NULL != pzHome )
    {
        FILE*        hFile;
        char        zPath[ PATH_MAX ];
        struct stat sStat;
        CMessage        cArchive;
        
        strcpy( zPath, pzHome );
        strcat( zPath, "/config/guiprefs.cfg" );

        if ( stat( zPath, &sStat ) >= 0 ) {
            hFile = fopen( zPath, "rb" );
            if ( NULL != hFile ) {
                uint8* pBuffer = new uint8[sStat.st_size];
                fread( pBuffer, sStat.st_size, 1, hFile );
                fclose( hFile );
                cArchive.Unflatten( pBuffer );
                
                CRect cFrame;
                if ( cArchive.FindRect( "window_frame", &cFrame ) == 0 ) {
                    SetFrame( cFrame );
                }
            }
        }
    }
  
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

MyWindow::~MyWindow()
{
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void MyWindow::MessageReceived( CMessage* pcMessage )
{
  View*        pcSource;

  if ( pcMessage->FindPointer( "source", (void**) &pcSource ) != 0 ) {
    CWindow::MessageReceived( pcMessage );
    return;
  }
  
  switch( pcMessage->what )
  {
    case MID_ABOUT:      printf( "About...\n" );        break;
    case MID_HELP:       printf( "Help...\n" );                break;
    case MID_LOAD:       printf( "Load...\n" );                break;
    case MID_INSERT:     printf( "Insert...\n" );        break;
    case MID_SAVE:       printf( "Save\n" );                break;
                                 
    case MID_SAVE_AS:    printf( "Save as...\n" );        break;
    case MID_QUIT:       printf( "Quit...\n" );                break;
                                 
    case MID_DUP_SEL:    printf( "Duplicate\n" );        break;
    case MID_DELETE_SEL: printf( "Delete\n" );                break;
                                 
    case MID_RESTORE:    printf( "Restore...\n" );        break;
    case MID_SNAPSHOT:   printf( "Snapshot\n" );        break;
    case MID_RESET:      printf( "Reset...\n" );        break;
    default:
      CWindow::MessageReceived( pcMessage );
  }
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool MyWindow::OkToQuit()
{
    char* pzHome = getenv( "HOME" );
    if ( NULL != pzHome )
    {
        FILE*        hFile;
        char        zPath[ PATH_MAX ];
        CMessage        cArchive;
        
        strcpy( zPath, pzHome );
        strcat( zPath, "/config/guiprefs.cfg" );

        hFile = fopen( zPath, "wb" );

        if ( NULL != hFile ) {
            cArchive.AddRect( "window_frame", Frame() );

            int nSize = cArchive.FlattenedSize();
            uint8* pBuffer = new uint8[nSize];
            cArchive.Flatten( pBuffer, nSize );
            fwrite( pBuffer, nSize, 1, hFile );
            fclose( hFile );
        }
    }
    CApplication::GetInstance()->PostMessage( M_QUIT );
    return( true );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

int main()
{
	CApplication* pcApp = new CApplication( "application/x-vnd.KHS-font_prefs" );
	MyWindow* pcWindow = new MyWindow( CRect( 0, 0, 350 - 1, 300 - 1 ) + CPoint( 100, 50 ) );
	pcWindow->Show();
	pcApp->Run();
}
