/*
 *  The Cosmoe application server
 *  Copyright (C) 1999 - 2000  Kurt Skauen
 *  Copyright (C) 2002 Bill Hayden
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <time.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>

#include <support/SupportDefs.h>
#include <cosmoe/kernel.h>

#include <appserver/keymap.h>

#include <macros.h>

#include <interface/InterfaceDefs.h>

#include "ddriver.h"
#include "server.h"
#include "swindow.h"
#include "keyboard.h"
#include "config.h"

static uint32 g_nQual = 0;

/*
 * FIXME: Cosmoe translates key codes into the form described below
 * (in the "Raw key numbering" comment) before returning the codes to
 * userspace.  In addition, the kernel squelches the keyboard's native
 * repeat so key repeat is reimplemented here in userspace.  It does this
 * with the function ConvertKeyCode (and two translation tables).  We
 * reproduce these here and call the ConvertKeyCode function manually.
 *
 * Also note that the "windows" key has been added to the keymap.  During
 * debugging, it stops the server with kill(getpid(),SIGTERM).  I thought it
 * was an appropriate keybinding.  ;-)
 */

static const uint8 g_anRawKeyTab[] =
{
    0x00,	/*	NO KEY	*/
    0x01,	/* 1   ESC	*/
    0x12,	/* 2   1	*/
    0x13,	/* 3   2	*/
    0x14,	/* 4   3	*/
    0x15,	/* 5   4	*/
    0x16,	/* 6   5	*/
    0x17,	/* 7   6	*/
    0x18,	/* 8   7	*/
    0x19,	/* 9   8	*/
    0x1a,	/* 10   9	*/
    0x1b,	/* 11   0	*/
    0x1c,	/* 12   -	*/
    0x1d,	/* 13   =	*/
    0x1e,	/* 14   BACKSPACE	*/
    0x26,	/* 15   TAB	*/
    0x27,	/* 16   Q	*/
    0x28,	/* 17   W	*/
    0x29,	/* 18   E	*/
    0x2a,	/* 19   R	*/
    0x2b,	/* 20   T	*/
    0x2c,	/* 21   Y	*/
    0x2d,	/* 22   U	*/
    0x2e,	/* 23   I	*/
    0x2f,	/* 24   O	*/
    0x30,	/* 25   P	*/
    0x31,	/* 26   [ {	*/
    0x32,	/* 27   ] }	*/
    0x47,	/* 28   ENTER (RETURN)	*/
    0x5c,	/* 29   LEFT CONTROL	*/
    0x3c,	/* 30   A	*/
    0x3d,	/* 31   S	*/
    0x3e,	/* 32   D	*/
    0x3f,	/* 33   F	*/
    0x40,	/* 34   G	*/
    0x41,	/* 35   H	*/
    0x42,	/* 36   J	*/
    0x43,	/* 37   K	*/
    0x44,	/* 38   L	*/
    0x45,	/* 39   ; :	*/
    0x46,	/* 40   ' "	*/
    0x11,	/* 41   ` ~	*/
    0x4b,	/* 42   LEFT SHIFT	*/
    0x33,	/* 43 	NOTE : This key code was not defined in the original table! (' *)	*/
    0x4c,	/* 44   Z	*/
    0x4d,	/* 45   X	*/
    0x4e,	/* 46   C	*/
    0x4f,	/* 47   V	*/
    0x50,	/* 48   B	*/
    0x51,	/* 49   N	*/
    0x52,	/* 50   M	*/
    0x53,	/* 51   , <	*/
    0x54,	/* 52   . >	*/
    0x55,	/* 53   / ?	*/
    0x56,	/* 54   RIGHT SHIFT	*/
    0x24,	/* 55   *            (KEYPAD)	*/
    0x5d,	/* 56   LEFT ALT	*/
    0x5e,	/* 57   SPACEBAR	*/
    0x3b,	/* 58   CAPSLOCK	*/
    0x02,	/* 59   F1	*/
    0x03,	/* 60   F2	*/
    0x04,	/* 61   F3	*/
    0x05,	/* 62   F4	*/
    0x06,	/* 63   F5	*/
    0x07,	/* 64   F6	*/
    0x08,	/* 65   F7	*/
    0x09,	/* 66   F8	*/
    0x0a,	/* 67   F9	*/
    0x0b,	/* 68   F10	*/
    0x22,	/* 69   NUMLOCK      (KEYPAD)	*/
    0x0f,	/* 70   SCROLL LOCK	*/
    0x37,	/* 71   7 HOME       (KEYPAD)	*/
    0x38,	/* 72   8 UP         (KEYPAD)	*/
    0x39,	/* 73   9 PGUP       (KEYPAD)	*/
    0x25,	/* 74   -            (KEYPAD)	*/
    0x48,	/* 75   4 LEFT       (KEYPAD)	*/
    0x49,	/* 76   5            (KEYPAD)	*/
    0x4a,	/* 77   6 RIGHT      (KEYPAD)	*/
    0x3a,	/* 78   +            (KEYPAD)	*/
    0x58,	/* 79   1 END        (KEYPAD)	*/
    0x59,	/* 80   2 DOWN       (KEYPAD)	*/
    0x5a,	/* 81   3 PGDN       (KEYPAD)	*/
    0x64,	/* 82   0 INSERT     (KEYPAD)	*/
    0x65,	/* 83   . DEL        (KEYPAD)	*/
    0x7e,	/* 84		SYSRQ	*/
    0x00,	/* 85	*/
    0x69,	/* 86 	NOTE : This key code was not defined in the original table! (< >)	*/
    0x0c,	/* 87   F11	*/
    0x0d	/* 88   F12	*/
};

static const uint8 g_anExtRawKeyTab[] =
{
    0x00,	/*	0	*/
    0x00,	/*	1	*/
    0x00,	/*		2	*/
    0x00,	/*		3	*/
    0x00,	/*		4	*/
    0x00,	/*		5	*/
    0x00,	/*		6	*/
    0x00,	/*		7	*/
    0x00,	/*		8	*/
    0x00,	/*		9	*/
    0x00,	/*		10	*/
    0x00,	/*		11	*/
    0x00,	/*		12	*/
    0x00,	/*		13	*/
    0x00,	/*		14	*/
    0x00,	/*		15	*/
    0x00,	/*		16	*/
    0x00,	/*		17	*/
    0x00,	/*		18	*/
    0x00,	/*		19	*/
    0x00,	/*		20	*/
    0x00,	/*		21	*/
    0x00,	/*		22	*/
    0x00,	/*		23	*/
    0x00,	/*		24	*/
    0x00,	/*		25	*/
    0x00,	/*		26	*/
    0x00,	/*		27	*/
    0x5b,	/*		28   ENTER        (KEYPAD)	*/
    0x60,	/*		29   RIGHT CONTROL	*/
    0x00,	/*		30	*/
    0x00,	/*		31	*/
    0x00,	/*		32	*/
    0x00,	/*		33	*/
    0x00,	/*		34	*/
    0x00,	/*		35	*/
    0x00,	/*		36	*/
    0x00,	/*		37	*/
    0x00,	/*		38	*/
    0x00,	/*		39	*/
    0x00,	/*		40	*/
    0x00,	/*		41	*/
    0x00,	/*		42   PRINT SCREEN (First code)	*/
    0x00,	/*		43	*/
    0x00,	/*		44	*/
    0x00,	/*		45	*/
    0x00,	/*		46	*/
    0x00,	/*		47	*/
    0x00,	/*		48	*/
    0x00,	/*		49	*/
    0x00,	/*		50	*/
    0x00,	/*		51	*/
    0x00,	/*		52	*/
    0x23,	/*		53   /            (KEYPAD)	*/
    0x00,	/*		54	*/
    0x0e,	/*		55   PRINT SCREEN (Second code)	*/
    0x5f,	/*		56   RIGHT ALT	*/
    0x00,	/*		57	*/
    0x00,	/*		58	*/
    0x00,	/*		59	*/
    0x00,	/*		60	*/
    0x00,	/*		61	*/
    0x00,	/*		62	*/
    0x00,	/*		63	*/
    0x00,	/*		64	*/
    0x00,	/*		65	*/
    0x00,	/*		66	*/
    0x00,	/*		67	*/
    0x00,	/*		68	*/
    0x00,	/*		69	*/
    0x7f,	/*		70	 BREAK	*/
    0x20,	/*		71   HOME         (NOT KEYPAD)	*/
    0x57,	/*		72   UP           (NOT KEYPAD)	*/
    0x21,	/*		73   PAGE UP      (NOT KEYPAD)	*/
    0x00,	/*		74	*/
    0x61,	/*		75   LEFT         (NOT KEYPAD)	*/
    0x00,	/*		76	*/
    0x63,	/*		77   RIGHT        (NOT KEYPAD)	*/
    0x00,	/*		78	*/
    0x35,	/*		79   END          (NOT KEYPAD)	*/
    0x62,	/*		80   DOWN         (NOT KEYPAD)	*/
    0x36,	/*		81   PAGE DOWN    (NOT KEYPAD)	*/
    0x1f,	/*		82   INSERT       (NOT KEYPAD)	*/
    0x34,	/*		83   DELETE       (NOT KEYPAD)	*/
    0x00,	/*		84	*/
    0x00,	/*		85	*/
    0x00,	/*		86	*/
    0x00,	/*		87	*/
    0x00,	/*		88	*/
    0x00,	/*		89	*/
    0x00,	/*		90	*/
    0x00,	/*		91	 WINDOWS (First code) */
    0x00,	/*		92	*/
    0x00,	/*		93	*/
    0x00,	/*		94	*/
    0x00,	/*		95	*/
    0x00,	/*		96	*/
    0x00,	/*		97	*/
    0x00,	/*		98	*/
    0x00,	/*		99	*/
    0x00,	/*		100	*/
    0x00,	/*		101	*/
    0x00,	/*		102	*/
    0x00,	/*		103	*/
    0x00,	/*		104	*/
    0x00,	/*		105	*/
    0x00,	/*		106	*/
    0x00,	/*		107	*/
    0x00,	/*		108	*/
    0x00,	/*		109	*/
    0x00,	/*		110	*/
    0x00,	/*		111   MACRO	*/
    0x00,	/*		112	*/
    0x00,	/*		113	*/
    0x00,	/*		114	*/
    0x00,	/*		115	*/
    0x00,	/*		116	*/
    0x00,	/*		117	*/
    0x00,	/*		118	*/
    0x00,	/*		119	*/
    0x00,	/*		120	*/
    0x00,	/*		121	*/
    0x00,	/*		122	*/
    0x00,	/*		123	*/
    0x00,	/*		124	*/
    0x00,	/*		125	*/
    0x00,	/*		126	*/
    0x00,	/*		127	*/
    0x00,	/*		128	*/
    0x00,	/*		129	*/

    0x00,	/*		130	*/
    0x00,	/*		131	*/
    0x00,	/*		132	*/
    0x00,	/*		133	*/
    0x00,	/*		134	*/
    0x00,	/*		135	*/
    0x00,	/*		136	*/
    0x00,	/*		137	*/
    0x00,	/*		138	*/
    0x00,	/*		139	*/
    0x00,	/*		130	*/

    0x00,	/*		1	*/
    0x00,	/*		2	*/
    0x00,	/*		3	*/
    0x00,	/*		4	*/
    0x00,	/*		5	*/
    0x00,	/*		6	*/
    0x00,	/*		7	*/
    0x00,	/*		8	*/
    0x00,	/*		9	*/

    0x00,	/*		150	*/
    0x00,	/*		151	*/
    0x00,	/*		152	*/
    0x00,	/*		153	*/
    0x00,	/*		154	*/
    0x00,	/*		155	*/
    0x00,	/*		156	*/
    0x00,	/*		157	*/
    0x00,	/*		158	*/
    0x00,	/*		159	*/

    0x00,	/*		0	*/
    0x00,	/*		1	*/
    0x00,	/*		2	*/
    0x00,	/*		3	*/
    0x00,	/*		4	*/
    0x00,	/*		5	*/
    0x00,	/*		6	*/
    0x00,	/*		7	*/
    0x00,	/*		8	*/
    0x00,	/*		9	*/

    0x00,	/*		170	*/
    0x00,	/*		171	*/
    0x00,	/*		172	*/
    0x00,	/*		173	*/
    0x00,	/*		174	*/
    0x00,	/*		175	*/
    0x00,	/*		176	*/
    0x00,	/*		177	*/
    0x00,	/*		178	*/
    0x00,	/*		179	*/

    0x00,	/*		0	*/
    0x00,	/*		1	*/
    0x00,	/*		2	*/
    0x00,	/*		3	*/
    0x00,	/*		4	*/
    0x00,	/*		5	*/
    0x00,	/*		6	*/
    0x00,	/*		7	*/
    0x00,	/*		8	*/
    0x00,	/*		9	*/

    0x00,	/*		190	*/
    0x00,	/*		191	*/
    0x00,	/*		192	*/
    0x00,	/*		193	*/
    0x00,	/*		194	*/
    0x00,	/*		195	*/
    0x00,	/*		196	*/
    0x00,	/*		197	*/
    0x00,	/*		198	*/
    0x00,	/*		199	*/

    0x00,	/*		0	*/
    0x00,	/*		1	*/
    0x00,	/*		2	*/
    0x00,	/*		3	*/
    0x00,	/*		4	*/
    0x00,	/*		5	*/
    0x00,	/*		6	*/
    0x00,	/*		7	*/
    0x00,	/*		8	*/
    0x00,	/*		9	*/

    0x00,	/*		210	*/
    0x00,	/*		211	*/
    0x00,	/*		212	*/
    0x00,	/*		213	*/
    0x00,	/*		214	*/
    0x00,	/*		215	*/
    0x00,	/*		216	*/
    0x00,	/*		217	*/
    0x00,	/*		218	*/
    0x7d,	/*		219	 WINDOWS (Second code) */

    0x00,	/*		0	*/
    0x00,	/*		1	*/
    0x00,	/*		2	*/
    0x00,	/*		3	*/
    0x00,	/*		4	*/
    0x00,	/*		5	*/
    0x00,	/*		6	*/
    0x00,	/*		7	*/
    0x00,	/*		8	*/
    0x00,	/*		9	*/

    0x00,	/*		230	*/
    0x00,	/*		231	*/
    0x00,	/*		232	*/
    0x00,	/*		233	*/
    0x00,	/*		234	*/
    0x00,	/*		235	*/
    0x00,	/*		236	*/
    0x00,	/*		237	*/
    0x00,	/*		238	*/
    0x00,	/*		239	*/

    0x00,	/*		0	*/
    0x00,	/*		1	*/
    0x00,	/*		2	*/
    0x00,	/*		3	*/
    0x00,	/*		4	*/
    0x00,	/*		5	*/
    0x00,	/*		6	*/
    0x00,	/*		7	*/
    0x00,	/*		8	*/
    0x00,	/*		9	*/

    0x00,	/*		250	*/
    0x00,	/*		251	*/
    0x00,	/*		252	*/
    0x00,	/*		253	*/
    0x00,	/*		254	*/
    0x00,	/*		255	*/
    0x00,	/*		256	*/
};


static uint8 KernelConvertKeyCode( uint8 nCode )
{
	uint nKey;
	uint nFlg;

	static uint nLastCode      = 0;
	static uint nLastKey       = 0;
	static int  nPauseKeyCount = 0;

	if ( nPauseKeyCount > 0 )
	{
		nPauseKeyCount++;

		if ( nPauseKeyCount == 6 )
		{
			nPauseKeyCount	=	0;
			return( 0x10 );		/* PAUSE	*/
		}
		else
		{
			return( 0 );
		}
	}

	if ( 0xe1 == nCode )
	{
		nPauseKeyCount = 1;
		return( 0 );
	}

	nFlg = nCode & 0x80;

	if ( 0xe0 == nLastCode )
	{
		nKey = g_anExtRawKeyTab[ nCode & ~0x80 ];
	}
	else
	{
		if ( 0xe0 != nCode )
		{
			nKey = g_anRawKeyTab[ nCode & ~0x80 ];
		}
		else
		{
			nKey = 0;
		}
	}

	nLastCode = nCode;

	if ( (nKey | nFlg) == nLastKey || 0 == nKey )
	{
		return( 0 );
	}
	nLastKey = nKey | nFlg;

	return( nKey | nFlg );
}

/*
 *     Raw key numbering for 101 keyboard:
 *
 * [win] = 0x7d
 *                                                                                       [sys]       [brk]
 *                                                                                        0x7e        0x7f
 *[esc]       [ f1] [ f2] [ f3] [ f4] [ f5] [ f6] [ f7] [ f8] [ f9] [f10] [f11] [f12]    [prn] [scr] [pau]
 * 0x01        0x02  0x03  0x04  0x05  0x06  0x07  0x08  0x09  0x0a  0x0b  0x0c  0x0d     0x0e  0x0f  0x10              KEYPAD
 *
 *[ ` ] [ 1 ] [ 2 ] [ 3 ] [ 4 ] [ 5 ] [ 6 ] [ 7 ] [ 8 ] [ 9 ] [ 0 ] [ - ] [ = ] [bck]    [ins] [hme] [pup]  [num] [ / ] [ * ] [ - ]
 * 0x11  0x12  0x13  0x14  0x15  0x16  0x17  0x18  0x19  0x1a  0x1b  0x1c  0x1d  0x1e     0x1f  0x20  0x21   0x22  0x23  0x24  0x25
 *
 *[tab] [ q ] [ w ] [ e ] [ r ] [ t ] [ y ] [ u ] [ i ] [ o ] [ p ] [ [ ] [ ] ] [ \ ]    [del] [end] [pdn]  [ 7 ] [ 8 ] [ 9 ] [ + ]
 * 0x26  0x27  0x28  0x29  0x2a  0x2b  0x2c  0x2d  0x2e  0x2f  0x30  0x31  0x32  0x33     0x34  0x35  0x36   0x37  0x38  0x39  0x3a
 *
 *[cap] [ a ] [ s ] [ d ] [ f ] [ g ] [ h ] [ j ] [ k ] [ l ] [ ; ] [ ' ] [  enter  ]                       [ 4 ] [ 5 ] [ 6 ]
 * 0x3b  0x3c  0x3d  0x3e  0x3f  0x40  0x41  0x42  0x43  0x44  0x45  0x46     0x47                           0x48  0x49  0x4a
 *
 *[shift]     [ z ] [ x ] [ c ] [ v ] [ b ] [ n ] [ m ] [ , ] [ . ] [ / ]     [shift]         [ up]         [ 1 ] [ 2 ] [ 3 ] [ent]
 * 0x4b       0x4c  0x4d  0x4e  0x4f  0x50  0x51  0x52  0x53  0x54  0x55       0x56            0x57         0x58  0x59  0x5a  0x5b
 *
 *[ctr]    {opt}      [cmd]      [  space  ]      [cmd]    {opt}    {menu}    [ctr]     [lft] [dwn] [rgt]   [ 0 ] [ . ]
 * 0x5c     0x66       0x5d          0x5e          0x5f     0x67     0x68      0x60      0x61  0x62  0x63    0x64  0x65
 */


// Legend:
//   n = Normal
//   s = Shift
//   c = Control
//   C = CapsLock
//   o = Option

// Key n        s        c        o        os       C        Cs       Co       Cos

keymap g_sKeymap;

void SetQualifiers( int nKeyCode )
{
	switch( nKeyCode )
	{
		case 0x4b:
			g_nQual |= QUAL_LSHIFT;
			break;
		case 0x4b | 0x80:
			g_nQual &= ~QUAL_LSHIFT;
			break;

		case 0x56:
			g_nQual |= QUAL_RSHIFT;
			break;
		case 0x56 | 0x80:
			g_nQual &= ~QUAL_RSHIFT;
			break;

		case 0x5c:
			g_nQual |= QUAL_LCTRL;
			break;
		case 0x5c | 0x80:
			g_nQual &= ~QUAL_LCTRL;
			break;

		case 0x60:
			g_nQual |= QUAL_RCTRL;
			break;
		case 0x60 | 0x80:
			g_nQual &= ~QUAL_RCTRL;
			break;

		case 0x5d:
			g_nQual |= QUAL_LALT;
			break;
		case 0x5d | 0x80:
			g_nQual &= ~QUAL_LALT;
			break;

		case 0x5f:
			g_nQual |= QUAL_RALT;
			break;
		case 0x5f | 0x80:
			g_nQual &= ~QUAL_RALT;
			break;
	}
}


uint32 GetQualifiers()
{
	return( g_nQual );
}


int convert_key_code( char* pzDst, int nRawKey, int nQual )
{
	int nTable = 0;
	int nQ = 0;

	if ( nRawKey < 0 || nRawKey >= 128 )
	{
		dbprintf( "convert_key_code() invalid keycode: %d\n", nRawKey );
		return( -1 );
	}

	if ( nQual & QUAL_SHIFT ) nQ |= 0x01;
	if ( nQual & QUAL_CTRL )  nQ |= 0x02;
	if ( nQual & QUAL_ALT )   nQ |= 0x04;

	switch ( nQ )
	{
		case 0x0:  nTable = 0; break;
		case 0x1:  nTable = 1; break;
		case 0x2:  nTable = 2; break;
		case 0x3:  nTable = 6; break;
		case 0x4:  nTable = 3; break;
		case 0x5:  nTable = 4; break;
		case 0x6:  nTable = 7; break;
		case 0x7:  nTable = 8; break;
	}
//    int nLen = unicode_to_utf8( pzDst, g_sKeymap.m_anMap[nRawKey][nTable] );
	int nLen;
	uint nChar = g_sKeymap.m_anMap[nRawKey][nTable];
	if ( nChar < 0x100 )
	{
		nLen = 1;
	}
	else if ( nChar < 0x10000 )
	{
		nLen = 2;
	}
	else if ( nChar < 0x1000000 )
	{
		nLen = 3;
	}
	else
	{
		nLen = 4;
	}

	for ( int i = 0 ; i < nLen ; ++i )
	{
		pzDst[i] = (nChar >> ((nLen-i-1)*8)) & 0xff;
	}
	pzDst[nLen] = '\0';
	return( nLen );
}


void HandleKeyboard()
{
	uint        nKeyDownTime = 0;
	uint8       nLastKey     = 0;
	uint8       nRawKeyCode;
	uint8       nKeyCode;

//    signal( SIGINT,  SIG_IGN );
//    signal( SIGQUIT, SIG_IGN );

	// If you use the escape-key debugging mentioned below, you can
	// comment the following line.

    //signal( SIGTERM, SIG_IGN );

	int nKbdDev = init_keyboard(); //open( "/dev/keybd", O_RDONLY );

	if ( nKbdDev < 0 )
	{
		dbprintf( "Panic : Could not open keyboard device!\n" );
	}

	for (;;)
	{
		snooze( 10000 );
		if ( read( nKbdDev, &nRawKeyCode, 1 ) == 1 )
		{
			// If you are debugging the appserver and want the ability
			// to easily kill it, uncomment the following line.  This
			// will allow you to kill it with the escape key.  Don't plan
			// on using vi editing with this uncommented...  :-)

			if( nRawKeyCode == 1 ) kill( getpid(), SIGTERM );

			nKeyCode = KernelConvertKeyCode( nRawKeyCode );
			SetQualifiers( nKeyCode );

			if ( 0 != nKeyCode )
			{
				if( nKeyCode == 0x7d )
				{
					kill( getpid(), SIGTERM );
				}

				if ( nKeyCode & 0x80 )
				{
					nKeyDownTime = 0;
					nLastKey     = 0;
				}
				else
				{
					nKeyDownTime = clock() + (AppserverConfig::GetInstance()->GetKeyDelay() * CLOCKS_PER_SEC / 1000);
					nLastKey     = nKeyCode;
				}
				AppServer::GetInstance()->SendKeyCode( nKeyCode, g_nQual );
			}
			continue;
		}

		if ( 0 != nKeyDownTime )
	{
			uint nTime = clock();
			if ( nTime > nKeyDownTime )
		{
				nKeyDownTime = nTime + (AppserverConfig::GetInstance()->GetKeyRepeat() * CLOCKS_PER_SEC / 1000);
				AppServer::GetInstance()->SendKeyCode( nLastKey, g_nQual | QUAL_REPEAT );
			}
		}
	}
}


void InitKeyboard( void )
{
	thread_id hKbdThread;

	FILE* hFile = fopen( AppserverConfig::GetInstance()->GetKeymapPath().c_str(), "r" );

	if ( hFile != NULL )
	{
		if ( load_keymap( hFile, &g_sKeymap ) != 0 )
		{
			dbprintf( "Error: failed to load keymap\n" );
			g_sKeymap = g_sDefaultKeyMap;
		}
		fclose( hFile );
	}
	else
	{
		dbprintf( "Failed to open keymap file '%s'\n", AppserverConfig::GetInstance()->GetKeymapPath().c_str() );
		g_sKeymap = g_sDefaultKeyMap;
	}
	hKbdThread = spawn_thread( "keyb_thread", (void*)HandleKeyboard, 120, 0, NULL );
	resume_thread( hKbdThread );
}
