/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */

 
#ifndef	__APPLICATION_H__
#define	__APPLICATION_H__

#include <app/Looper.h>
#include <app/AppDefs.h>
#include <interface/Window.h>
#include <interface/Font.h>
#include <app/Clipboard.h>
#include <app/Messenger.h>
#include <app/Roster.h>

#include <vector>
#include <string>

#define BApplication		os::CApplication
#define be_app				os::CApplication::c_app
#define be_clipboard		os::CApplication::c_clipboard
#define be_app_messenger	#error

namespace os {
    class CWindow;
    class CRect;
    class View;
}

/**  \brief os:: This namespace wraps all Cosmoe API classes
 *
 *	All functions and classes defined by the Cosmoe API are contained
 *	in the 'os' namespace.
 *
 * \author	Kurt Skauen (kurt@atheos.cx), Bill Hayden (hayden@haydentech.com)
 *//////////////////////////////////////////////////////////////////////////////

namespace os
{

/** Singleton class representing an application
 * \ingroup app
 * \par Description:
 *	Before you can use any of the classes in the Cosmoe API you must instantiate
 *	an application object. The application object establishes the connection with
 *	the appserver, and is responsible for all application level communication
 *	after the connection is established.
 *
 *	os::CApplication inherits from CLooper, but it does behave different from
 *	other loopers in that it does not spawn a new thread when the
 *	Run() member is called.  Insetead it steals the thread calling the run member
 *	(normally the main thread) and lets it run the message loop.
 *
 * \sa os::CWindow, os::CView, os::CLooper
 * \author	Kurt Skauen (kurt@atheos.cx), Bill Hayden (hayden@haydentech.com)
 *//////////////////////////////////////////////////////////////////////////////

class CApplication : public CLooper
{
public:

	static CApplication*	c_app;
	static CClipboard*		c_clipboard;
	static CRoster*			c_roster;

							CApplication( const char* pzMimeType );
	virtual					~CApplication();
	static CApplication*	GetInstance();
	uint32					GetQualifiers();

	virtual void			MessageReceived( CMessage* pcMessage );

	void					PushCursor( mouse_ptr_mode eMode, void* pImage, int nWidth, int nHeight, const IPoint& cHotSpot = IPoint(0,0)  );
	void					PopCursor();
	void					HideCursor();
	void					ShowCursor();
	bool					IsCursorHidden() const;

	bigtime_t				GetIdleTime();
	void					GetKeyboardConfig( std::string* pcKeymapName, int* pnKeyDelay, int* pnKeyRepeat );
	status_t				SetKeymap( const char* pzName );
	status_t				SetKeyboardTimings( int nDelay, int nRepeat );

	int						GetScreenModeCount();
	int						GetScreenModeInfo( int nIndex, screen_mode* psMode );

	void					SetWindowDecorator( const char* pzPath );
	void					CommitColorConfig();

	thread_id				Run();

	virtual	void			ReadyToRun()				 {}

	virtual	void			Pulse()						 {}
	void					SetPulseRate(bigtime_t inRate);
	virtual void			TimerTick( int inID );

	CWindow*				WindowAt(int32 index);
	int32					CountWindows();


private:
	friend class CWindow;
	friend class CScreen;
	friend class CBitmap;
	friend class Sprite;
	friend class CView;
	friend class CFont;
	friend class AppserverConfig;
	friend class CRoster;

	port_id					GetServerPort() const;
	port_id					GetAppPort() const;

	// Support for the Window class:
	port_id 				CreateWindow( CView* pcTopView, const CRect& cFrame,
										const std::string& cTitle,
										uint32 nFlags, uint32 nDesktopMask,
										port_id hEventPort, int* phTopView );
	port_id 				CreateWindow( CView* pcTopView, int hBitmapHandle,
										int* phTopView );

	void					AddWindow( CWindow* pcWindow );
	void					RemoveWindow( CWindow* pcWindow );

	// Support for the Font class:
	int						GetFontConfigNames( std::vector<std::string>* pcTable ) const;
	int						GetDefaultFont( const std::string& cName,
											font_properties* psProps ) const;
	int						SetDefaultFont( const std::string& cName,
											const font_properties& sProps );
	int						AddDefaultFont( const std::string& cName,
											const font_properties& sProps );

	int						GetFontFamilyCount();
	status_t				GetFontFamily( int nIndex, char* pzFamily );
	int						GetFontStyleCount( const char* pzFamily );
	status_t				GetFontStyle( const char* pzFamily,
										  int nIndex,
										  char* pzStyle,
										  uint32* pnFlags = NULL );

	int   					LockDesktop( int nDesktop, CMessage* pcDesktopParams );
	int	  					UnlockDesktop( int nCookie );
	void  					SetScreenMode( int nDesktop, uint32 nValidityMask, screen_mode* psMode );

	int						CreateBitmap( int nWidth, int nHeight,
										  color_space eColorSpc,
										  bool inShareFramebuffer,
										  int* phHandle,
										  int* phArea );
	int						DeleteBitmap( int nBitmap );
	int						CreateSprite( const CRect& cFrame,
										  int nBitmapHandle,
										  uint32* pnHandle );
	void					DeleteSprite( uint32 nSprite );
	void					MoveSprite( uint32 nSprite, const CPoint& cNewPos );

	virtual void			FBC_Application1();
	virtual void			FBC_Application2();
	virtual void			FBC_Application3();
	virtual void			FBC_Application4();
	virtual void			FBC_Application5();
	virtual void			FBC_Application6();
	virtual void			FBC_Application7();
	virtual void			FBC_Application8();
	virtual void			FBC_Application9();
	virtual void			FBC_Application10();

    class Private;
    Private* m;
//    uint32   __reserved__[5];
};

}

#endif	// __APPLICATION_H__
