/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */


#ifndef __OS_H__
#define __OS_H__

#include <support/SupportDefs.h>
#include <sys/types.h>

#ifdef __cplusplus
extern "C" {
#endif

// ***
// Area Definitions
// ***

#define PAGE_SHIFT	12
#define PAGE_SIZE	(1UL << PAGE_SHIFT)
#define PAGE_MASK	(~(PAGE_SIZE-1))


typedef struct
{
    char    name[OS_NAME_LENGTH];
    area_id area;
    size_t  size;
    int	    lock;
    uint32  protection;
    proc_id process;
    uint32  allocSize;
    void*   address;
} AreaInfo_s;

  // Locking
#define AREA_NO_LOCK			0
#define AREA_LAZY_LOCK			1
#define AREA_FULL_LOCK			2
#define AREA_CONTIGUOUS			3

  // Protection
#define	AREA_READ				0x00000001
#define	AREA_WRITE				0x00000002
#define	AREA_EXEC				0x00000004
#define	AREA_FULL_ACCESS		(AREA_READ | AREA_WRITE | AREA_EXEC)

  // Creation/cloning allocation policy
#define AREA_ANY_ADDRESS		0x00000000
#define AREA_EXACT_ADDRESS		0x00000100
#define AREA_BASE_ADDRESS		0x00000200
#define AREA_CLONE_ADDRESS		0x00000300
#define AREA_ANY_KERNEL_ADDRESS	0x00001000

  // Area IDs
#define AREA_ID_MAX  256
#define AREA_ID_FREE 0xFFFFFFFF
#define AREA_ID_FB   0xFFFFFFFE


void	init_area_map( void );

area_id	create_area( const char* pzName, void** ppAddress, size_t nSize, uint32 nProtection, int nLock );

status_t resize_area( area_id hArea, uint32 nNewSize, bool bAtomic );
status_t remap_area( area_id nArea, void* pPhysAddress );

status_t delete_area( area_id hArea );
area_id	 clone_area( const char* pzName, void** ppAddress, uint32 nProtection, int nLockMode, area_id hSrcArea );

status_t get_area_info( area_id hArea, AreaInfo_s* psInfo );


// ***
// Semaphore Definitions
// ***

typedef struct
{
    sem_id 		si_sema_id;
    proc_id 	si_owner;
    int32		si_count;
    thread_id	si_latest_holder;
    char		si_name[OS_NAME_LENGTH];
} sem_info;



#define	SEM_RECURSIVE	    0x0001	/* Allow same thread to lock semaphore multiple times	*/
#define SEM_WARN_DBL_LOCK   0x0002	/* Warn if a thread takes the semaphore twice		*/
#define SEM_WARN_DBL_UNLOCK 0x0004	/* Warn if the count gets higher than 1			*/
#define SEM_GLOBAL			0x0008	/* Alloc the semaphore in the global namespace		*/

sem_id    create_sem( int32 nCount, const char* pzName );
sem_id    create_sem_etc( int32 nCount, const char* pzName, uint32 nFlags );
status_t  delete_sem( sem_id hSema );

status_t  lock_mutex( sem_id hSema, bool bIgnoreSignals );
status_t  unlock_mutex( sem_id hSema );

status_t  acquire_sem( sem_id hSema );
status_t  acquire_sem_etc( sem_id hSema, int32 nCount, uint32 nFlags, bigtime_t nTimeOut );

status_t  release_sem( sem_id hSema );
status_t  release_sem_etc( sem_id hSema, int32 nCount, uint32 nFlags );

status_t  sleep_on_sem( sem_id hSema, bigtime_t nTimeOut );
status_t  wakeup_sem( sem_id hSema, bool bAll );
status_t  unlock_and_suspend( sem_id hWaitQueue, sem_id hSema );

int 	  get_semaphore_count( sem_id hSema );
thread_id get_semaphore_owner( sem_id hSema );

status_t  get_semaphore_info( sem_id hSema, proc_id hOwner, sem_info* psInfo );
status_t  get_next_semaphore_info( sem_info* psInfo );

int		  is_semaphore_locked( sem_id hSema );



// ***
// Thread Definitions
// ***

#define	TF_DEADLOCK	0x0001
enum
{
    TS_INVALID,	/* Bad luck					*/
    TS_RUN,		/* The lucky ones 			*/
    TS_READY,	/* I want time NOW			*/
    TS_WAIT,	/* Stay off 				*/
    TS_SLEEP,	/* Just a minute			*/
    TS_STOPPED,	/* LET ME OUT 				*/
    TS_ZOMBIE	/* waiting for waitpid		*/
};

typedef	struct
{
    thread_id	ti_thread_id;
    proc_id		ti_process_id;
    char		ti_process_name[ OS_NAME_LENGTH ];
    char		ti_thread_name[ OS_NAME_LENGTH ];		/* Thread name				*/

    int32  		ti_state;		/* Current task state.			*/
    uint32		ti_flags;
    sem_id		ti_blocking_sema;	/* The semaphore we wait's for, or -1	*/

      /*** Timing members	***/

    int			priority;
    int			ti_dynamic_pri;
    bigtime_t	kernel_time;	/*	Micros in kernel mode		*/
    bigtime_t	user_time;		/*	Micros in user mode		*/
    bigtime_t	ti_real_time;	/*	Total micros of execution	*/
    bigtime_t	ti_quantum;		/*	Maximum allowed micros of execution before preemption	*/

    void*		ti_stack;
    uint32		ti_stack_size;
    void*		ti_kernel_stack;		/* Top of kernel stack						*/
    uint32		ti_kernel_stack_size;	/* Size (in bytes) of kernel stack				*/
} thread_info;

enum
{
    IDLE_PRIORITY	    	= -100,
    LOW_PRIORITY	    	= -50,
    NORMAL_PRIORITY	    	=  0,
    DISPLAY_PRIORITY	    =  50,
    URGENT_DISPLAY_PRIORITY =  100,
    REALTIME_PRIORITY	    =  150
};


status_t	get_thread_info( const thread_id hThread, thread_info* const psInfo );
status_t	get_next_thread_info( thread_info* psInfo );
proc_id		get_thread_proc( const thread_id hThread );
thread_id	get_thread_id( const char* const pzName );
proc_id		get_process_id( const char* pzName );
status_t	rename_process( int nProcessID, const char* pzName );
status_t	rename_thread( thread_id hThread, const char* const pzName );
status_t	send_data( const thread_id hThread, const uint32 nCode, void* const pData, const uint32 nSize );
uint32		receive_data( thread_id* const phSender, void* const pData, const uint32 nSize );
int	 		thread_data_size( void );
int	 		has_data( const thread_id hThread );
thread_id	spawn_thread( const char* const pzName, void* const pfEntry,
							const int nPri, int nStackSize, void* const pData );

int 		set_strace_level( thread_id hThread, int nLevel );

int	 		set_thread_priority( const thread_id hThread, const int nPriority );
status_t	suspend_thread( const thread_id hThread );
status_t	resume_thread( const thread_id hThread );
int	  		wait_for_thread( const thread_id hThread );
int			exit_thread( int nCode );

status_t  	suspend( void );

// ***
// Time Definitions
// ***

bigtime_t	system_time( void );
bigtime_t	get_real_time( void );
bigtime_t	get_idle_time( int nProcessor );
status_t	snooze( bigtime_t nTimeout );


#ifdef __cplusplus
}
#endif

#endif /* ndef __OS_H__ */
