/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */


#ifndef	__BITMAP_H__
#define	__BITMAP_H__

#include <support/SupportDefs.h>
#include <cosmoe/Debug.h>
#include <interface/InterfaceDefs.h>
#include <interface/Region.h>

#define BBitmap	os::CBitmap


namespace os
{

class CWindow;
class CView;

/** Container for bitmap-image data.
 * \ingroup interface
 * \par Description:
 *	The Bitmap class make it possible to render bitmap graphics
 *	into views and to make views render into an offscreen buffer
 *	to implement things like double-buffering.
 *
 *	The bitmap class has two different ways to communicate with
 *	the application server. If the inShareFramebuffer parameter is
 *	set the bitmap's raster memory is created in a memory-area
 *	shared between the application and the appserver. This makes
 *	it possible for the appserver to blit graphics written
 *	directly to the bitmaps raster buffer by the application into
 *	views on the screen (or inside other bitmaps). If the
 *	inAcceptViews parameter is set the bitmap will accept views to be
 *	added much like a os::CWindow object. All rendering performed
 *	by the views will then go into the bitmap's offscreen buffer
 *	rather than the screen. The rendered image can then be read
 *	out by the application (requires the \b inShareFramebuffer
 *	parameter to be set to true) or it can be blited into other views.
 *
 * \sa CView, CWindow
 * \author	Kurt Skauen (kurt@atheos.cx), Bill Hayden (hayden@haydentech.com)
 *****************************************************************************/

class CBitmap
{
public:
//	enum { ACCEPT_VIEWS = 0x0001, SHARE_FRAMEBUFFER = 0x0002 };
					CBitmap( int nWidth,
							 int nHeight,
							 color_space eColorSpc,
							 bool inAcceptViews = false,
							 bool inShareFramebuffer = true );
					CBitmap( CRect inBounds,
							 color_space eColorSpc,
							 bool inAcceptViews = false,
							 bool inShareFramebuffer = true );
	virtual			~CBitmap();

	status_t		InitCheck( void ) const { return mInitCheck;	}
	bool			IsValid( void ) const	{ return mBits != NULL;	}

	area_id			Area() const			{ return mArea;			}
	color_space		ColorSpace() const		{ return mColorSpace;	}
	CRect			Bounds( void ) const	{ return mBounds;		}
	int32			BitsLength() const;
	int32			BytesPerRow() const;
	
	virtual void	AddChild( CView* pcView );
	virtual bool	RemoveChild( CView* pcView );
	CView*			FindView( const char* pzName ) const;

	void			Sync( void );
	void			Flush( void );
	
	bool			IsLocked( void ) const;
	bool			Lock( void );
	void			Unlock( void );

	// Will be deprecated:
	uint8*			LockRaster( void )		{ return( mBits );		}
	
	// Replacement for LockRaster:
	void*			Bits( void ) const		{ return( (void*)mBits );}

private:
	friend class CView;
	friend class CWindow;
	friend class Sprite;

	CWindow*		mWindow;
	int	  			mHandle;
	area_id			mArea;
	color_space		mColorSpace;
	CRect			mBounds;
	uint8*  		mBits;
	status_t		mInitCheck;
};



inline int BitsPerPixel( color_space eColorSpc )
{
    switch( eColorSpc )
    {
	case CS_RGB32:
	case CS_RGBA32:
	    return( 32 );
	case CS_RGB24:
	    return( 24 );
	case CS_RGB16:
	case CS_RGB15:
	case CS_RGBA15:
	    return( 16 );
	case CS_CMAP8:
	case CS_GRAY8:
	    return( 8 );
	case CS_GRAY1:
	    return( 1 );
/*      
	case CS_YUV422:
	case CS_YUV411:
	case CS_YUV420:
	return( 8 );
	case CS_YUV444:
	return( 16 );
	case CS_YUV9:
	case CS_YUV12:
	*/
	default:
	    dbprintf( "BitsPerPixel() invalid color space %d\n", eColorSpc );
	    return( 8 );
    }
}


extern const uint8 __5_to_8_bit_table[];
extern const uint8 __6_to_8_bit_table[];

inline uint32 COL_TO_RGB32( const Color32_s& col )
{
    return( ( ((col).red << 16) | ((col).green << 8) | (col).blue | ((col).alpha << 24) ) );
}

inline uint16 COL_TO_RGB16( const Color32_s& col )
{
    return( ( (((col).red >> 3) << 11) | (((col).green >> 2) << 5) | ((col).blue >> 3) ) );
}

inline uint16 COL_TO_RGB15( const Color32_s& col )
{
    return( ( (((col).red >> 3) << 10) | (((col).green >> 3) << 5) | ((col).blue >> 3) ) );
}

inline Color32_s RGB32_TO_COL( uint32 pix )
{
    return( Color32_s( ((pix) >> 16) & 0xff, ((pix) >> 8) & 0xff, (pix) & 0xff, ((pix) >> 24) & 0xff ) );
}

inline Color32_s RGB16_TO_COL( uint16 pix )
{
    return( Color32_s( __5_to_8_bit_table[ ((pix) >> 11) & 0x1f ],
		       __6_to_8_bit_table[ ((pix) >> 5) & 0x3f ],
		       __5_to_8_bit_table[ (pix) & 0x1f ], 255 ) );
}

inline Color32_s RGB15_TO_COL( uint16 pix )
{
    return( Color32_s( __5_to_8_bit_table[ ((pix) >> 10) & 0x1f ],
		       __5_to_8_bit_table[ ((pix) >> 5) & 0x1f ],
		       __5_to_8_bit_table[ (pix) & 0x1f ], 255 ) );
}

inline Color32_s RGBA15_TO_COL( uint16 pix )
{
    return( Color32_s( __5_to_8_bit_table[ ((pix) >> 10) & 0x1f ],
		       __5_to_8_bit_table[ ((pix) >> 5) & 0x1f ],
		       __5_to_8_bit_table[ (pix) & 0x1f ],
		       ((pix) & 0x8000) ? 0x00 : 0xff ) );
}




}

#endif // __BITMAP_H__
