/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */


#ifndef __LISTVIEW_H__
#define __LISTVIEW_H__

#include <interface/Control.h>

#include <vector>
#include <string>
#include <functional>

namespace os
{

class CScrollBar;
class Button;
class CMessage;

class ListViewContainer;


/** 
 * \ingroup interface
 * \par Description:
 *
 * \sa
 * \author	Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

class ListViewRow
{
public:
    ListViewRow();
    virtual ~ListViewRow();

    virtual void	AttachToView( CView* pcView, int nColumn ) = 0;
    virtual void	SetRect( const CRect& cRect, int nColumn ) = 0;
    
    virtual float	GetWidth( CView* pcView, int nColumn ) = 0;
    virtual float	GetHeight( CView* pcView ) = 0;
    virtual void	Draw( const CRect& cFrame, CView* pcView, uint nColumn,
						  bool bSelected, bool bHighlighted, bool bHasFocus ) = 0;
    virtual bool	HitTest( CView* pcView, const CRect& cFrame, int nColumn, CPoint cPos );
    virtual bool	IsLessThan( const ListViewRow* pcOther, uint nColumn ) const = 0;
    
    void    		SetIsSelectable( bool bSelectable );
    bool			IsSelectable() const;
    bool			IsSelected() const;
    bool			IsHighlighted() const;
private:
    friend class ListView;
    friend class ListViewContainer;
    friend class ListViewCol;
    friend class std::vector<ListViewRow>;
    friend struct RowPosPred;
    float	m_vYPos;
    float	m_vHeight;
    bool	m_bIsSelectable;
    bool	m_bSelected;
    bool	m_bHighlighted;
};

/** 
 * \ingroup gui
 * \par Description:
 *
 * \sa
 * \author	Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

class ListViewStringRow : public ListViewRow
{
public:
	ListViewStringRow() {}
	virtual ~ListViewStringRow() {}

	void				AttachToView( CView* pcView, int nColumn );
	void				SetRect( const CRect& cRect, int nColumn );
	void				AppendString( const std::string& cString );
	const std::string&	GetString( int nIndex ) const;
	virtual float 		GetWidth( CView* pcView, int nColumn );
	virtual float  		GetHeight( CView* pcView );
	virtual void 		Draw( const CRect& cFrame, CView* pcView, uint nColumn,
								bool bSelected, bool bHighlighted, bool bHasFocus );
	virtual bool		IsLessThan( const ListViewRow* pcOther, uint nColumn ) const;

private:
    std::vector< std::pair<std::string,float> > m_cStrings;
};


/** Flexible multicolumn list view
 * \ingroup gui
 * \par Description:
 *
 * \sa
 * \author	Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

class ListView : public CControl
{
public:
	enum scroll_direction { SCROLL_UP, SCROLL_DOWN };
	enum { AUTOSCROLL_BORDER = 20 };
	enum {	F_MULTI_SELECT = 0x0001,
			F_NO_AUTO_SORT = 0x0002,
			F_RENDER_BORDER = 0x0004,
			F_DONT_SCROLL = 0x0008,
			F_NO_HEADER = 0x0010,
			F_NO_COL_REMAP = 0x0020 };

	enum {	INV_HEIGHT = 0x01,
			INV_WIDTH  = 0x02,
			INV_VISUAL = 0x04 };

	typedef std::vector<ListViewRow*>::const_iterator const_iterator;
	typedef std::vector<int>			      column_map;

	ListView( CRect inFrame, const char* inTitle, uint32 inModeFlags = F_MULTI_SELECT | F_RENDER_BORDER,
		uint32 inResizeMask = CF_FOLLOW_LEFT | CF_FOLLOW_TOP,
		uint32 inFlags = WID_WILL_DRAW | WID_FULL_UPDATE_ON_RESIZE );

	~ListView();

	virtual void	LabelChanged( const std::string& cNewLabel );
	virtual void	EnableStatusChanged( bool bIsEnabled );
	virtual bool	Invoked( CMessage* pcMessage );

	virtual void	Invoked( int nFirstRow, int nLastRow );
	virtual void	SelectionChanged( int nFirstRow, int nLastRow );
	virtual bool	DragSelection( const CPoint& cPos );

	virtual void __reserved1__();
	virtual void __reserved2__();
	virtual void __reserved3__();
	virtual void __reserved4__();

	void			StartScroll( scroll_direction eDirection, bool bSelect );
	void			StopScroll();

	bool			IsMultiSelect() const;
	void			SetMultiSelect( bool bMulti );

	bool			IsAutoSort() const;
	void			SetAutoSort( bool bAuto );

	bool			HasBorder() const;
	void			SetRenderBorder( bool bRender );

	bool			HasColumnHeader() const;
	void			SetHasColumnHeader( bool bFlag );

	void			MakeVisible( int nRow, bool bCenter = true );
	int				InsertColumn( const char* pzTitle, int nWidth, int nPos = -1 );

	const column_map&	GetColumnMapping() const;
	void			SetColumnMapping( const column_map& cMap );

	void			InsertRow( int nPos, ListViewRow* pcRow, bool bUpdate = true );
	void			InsertRow( ListViewRow* pcRow, bool bUpdate = true );
	ListViewRow*	RemoveRow( int nIndex, bool bUpdate = true );
	void			InvalidateRow( int nRow, uint32 nFlags );
	uint			GetRowCount() const;
	ListViewRow*	GetRow( const CPoint& cPos ) const;
	ListViewRow*	GetRow( uint nIndex ) const;
	int				HitTest( const CPoint& cPos ) const;
	float			GetRowPos( int nRow );
	void			Clear();
	bool			IsSelected( uint nRow ) const;
	void			Select( int nFirst, int nLast, bool bReplace = true, bool bSelect = true );
	void			Select( int nRow, bool bReplace = true, bool bSelect = true );
	void			ClearSelection();

	void			Highlight( int nFirst, int nLast, bool bReplace, bool bHighlight = true );
	void			Highlight( int nRow, bool bReplace, bool bHighlight = true );

	void			SetCurrentRow( int nRow );
	void			Sort();
	int				GetFirstSelected() const;
	int				GetLastSelected() const;
	void	 		SetSelChangeMsg( CMessage* pcMsg );
	void	 		SetInvokeMsg( CMessage* pcMsg );
	CMessage*		GetSelChangeMsg() const;
	CMessage*		GetInvokeMsg() const;
	virtual void	Draw( CRect cUpdateRect );
	virtual void	FrameResized( float inWidth, float inHeight );
	virtual void	KeyDown( const char* pzString, const char* pzRawString, uint32 nQualifiers );
	virtual void	AllAttached();
	virtual bool	IsFocus( void ) const;

	// STL iterator interface to the rows.
	const_iterator	begin() const;
	const_iterator	end() const;

private:
	friend class ListViewContainer;
	friend class ListViewHeader;
	void	Layout();
	void  	AdjustScrollBars( bool bOkToHScroll = true );

	ListViewContainer* m_pcMainView;
	ListViewHeader*    m_pcHeaderView;
	CScrollBar*	       m_pcVScroll;
	CScrollBar*	       m_pcHScroll;
	CMessage*	       m_pcSelChangeMsg;
	CMessage*	       m_pcInvokeMsg;
//    uint32	       m_nModeFlags;
};

} // end of namespace


#endif // __F_GUI_LISTVIEW_H__
