/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */

 
#ifndef __ENTRY_H__
#define __ENTRY_H__

#include <sys/stat.h>
#include <storage/Directory.h>
#include <storage/Path.h>

#include <string>

#define BEntry		os::CEntry
#define entry_ref	os::EntryRef

namespace os {


/** Semi-persistent reference to a file
 * \ingroup storage
 * \par Description:
 *	CEntry's serve much the same functionality as a plain
 *	path.  It uniquely identifies a file within the
 *	filesystem. The main advantage of a CEntry is that it
 *	can partially "track" the file it references. While moving a file
 *	to another directory or renaming the directory it lives in or
 *	any of its subdirectories would break a path, it will not
 *	affect a CEntry.  Renaming the file itself however from
 *	outside the CEntry (ie. not using the CEntry::Rename() member)
 *	will break a CEntry.
 *
 *  \sa
 *  \author	Kurt Skauen (kurt@atheos.cx), Bill Hayden (hayden@haydentech.com)
 *****************************************************************************/

struct EntryRef
{
					EntryRef();
  					EntryRef(dev_t dev, ino_t dir, const char *inName);
					EntryRef(const EntryRef &inRef);
					~EntryRef();

  	status_t 		set_name(const char *name);

	bool			operator==(const EntryRef &inRef) const;
	bool			operator!=(const EntryRef &inRef) const;
//	EntryRef&		operator=(const EntryRef &ref);

	dev_t			device;
	ino_t			directory;
	std::string		name;
};


class CEntry : public CStatable
{
public:
					CEntry();
					CEntry( const char* cPath,
							bool bFollowLinks = false );
					CEntry( const CDirectory* cDir,
							const char* cName,
							bool bFollowLinks = false );
					CEntry( const CEntry& cRef );
					CEntry( const EntryRef* inEntryRef, bool inFollowLinks = false );
	virtual			~CEntry();


	status_t		SetTo( const char* cPath,
						   bool bFollowLinks = false );
	status_t		SetTo( const CDirectory* cDir,
						   const char* cName,
						   bool bFollowLinks = false );
	status_t		SetTo( const CEntry& cRef );
	void			Unset();
	
	status_t 		InitCheck() const;

	bool			Exists() const;

	status_t		GetRef( EntryRef* ioRef) const;
	status_t 		GetPath( CPath* pcPath ) const;
	status_t		GetName( char *ioBuffer ) const;
	std::string		GetName() const;					// Deprecated
	status_t		GetParent( CEntry *ioEntry ) const;
	status_t		GetParent( CDirectory *ioDirectory ) const;

	int				Rename( const char* cNewName );
	status_t		Delete();

	status_t		GetStat( struct stat* psStat ) const;

	const CDirectory& CEntry::GetDirectory() const;

private:
    CDirectory		m_cDirectory;
    std::string 	m_cName;
};


} // end of namespace

#endif // __ENTRY_H__
