/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */


#ifndef __NODE_H__
#define __NODE_H__

#include <support/SupportDefs.h>
#include <sys/types.h>
#include <cosmoe/fs_info.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>

#include <string>

#include <util/exceptions.h>
#include <storage/Statable.h>

#define BNode		os::CNode
#define node_ref	os::NodeRef

namespace os {

class CEntry;
class CDirectory;

struct NodeRef
{
				NodeRef();
				NodeRef(const NodeRef &ref);

	bool		operator==(const NodeRef &ref) const;
	bool		operator!=(const NodeRef &ref) const;
//	NodeRef&	operator=(const NodeRef &ref);

	dev_t		device;
	ino_t		node;
};


/** Lowlevel filesystem node class.
 * \ingroup storage
 * \par Description:
 *	This class give access to the lowest level of a filesystem node.
 *	A node can be a directory, regular file, symlink or named pipe.
 *
 *	It gives you access to stats that are common to all nodes in the
 *	filesystem like time-stamps, access-rights, inode and device
 *	numbers, and most imporantly the file-attributes.
 *
 *	The native AtheOS filesystem (AFS) supports "attributes", which are
 *	extra data-streams associated with filesystem nodes. An attribute
 *	can have a specific type like int, float, string, etc., or it
 *	can be a untyped stream of data. Attributes can be used to store
 *	information associated by the file but that doesn't belong to file
 *	content itself (for example the file's icon-image and mime-type).
 *
 * \sa os::CEntry, os::CFile, os::CDirectory
 * \author	Kurt Skauen (kurt@atheos.cx), Bill Hayden (hayden@haydentech.com)
 *****************************************************************************/


class CNode : public CStatable
{
public:
						CNode();
						CNode( const char* inPath,
							   int nOpenMode = O_RDONLY );
						CNode( const CDirectory* inDirectory,
							   const char* cName,
							   int nOpenMode = O_RDONLY );
						CNode( const CEntry* iNEntry,
							   int nOpenMode = O_RDONLY );
						CNode( const CNode& cNode );
	
	virtual				~CNode();

	status_t			InitCheck() const;

	virtual status_t	FDChanged( int nNewFD );

	virtual status_t	SetTo( const char* cPath, int nOpenMode = O_RDONLY );
	virtual status_t	SetTo( const CDirectory* cDir, const char* cPath, int nOpenMode = O_RDONLY );
	virtual status_t	SetTo( const CEntry* cRef, int nOpenMode = O_RDONLY );
	virtual status_t	SetTo( const CNode& cNode );
	virtual void		Unset();

	virtual status_t	GetStat( struct stat* psStat ) const;
//	virtual dev_t		GetDev() const;

	virtual status_t	GetNextAttrName( std::string* pcName );
	virtual status_t	RewindAttrdir();

	virtual ssize_t		WriteAttr( const char* cAttrName, int nFlags, int nType,
									const void* pBuffer, off_t nPos, size_t nLen );
	virtual ssize_t		ReadAttr( const char* cAttrName, int nType,
									void* pBuffer, off_t nPos, size_t nLen );

	virtual status_t	RemoveAttr( const char* cName );
	virtual status_t	StatAttr( const char* cName, struct ::attr_info* psBuffer );

	virtual int			GetFileDescriptor() const;

	std::string			GetPathName( void ) const { return m_cPath; }

private:
friend class CDirectory;

	void				_SetFD( int nFD );

	std::string     	m_cPath;
	int             	m_nFD;
	DIR*            	m_hAttrDir;
	status_t			mInitCheck;
};



} // end of namespace

#endif // __NODE_H__
