/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */


#ifndef __LOCKER_H__
#define __LOCKER_H__


#include <stdio.h>
#include <errno.h>
#include <support/SupportDefs.h>
#include <cosmoe/OS.h>
#include <cosmoe/atomic.h>

#define BLocker		os::CLocker

namespace os
{

/**
 * \ingroup support
 * \par Description:
 *
 * \sa
 * \author	Kurt Skauen (kurt@atheos.cx), Bill Hayden (hayden@haydentech.com)
 *****************************************************************************/

class CLocker
{
public:
	CLocker( const char* inName="" )
	{
		mSemaphore = create_sem_etc( 0, inName, /*(inRecursive) ? SEM_RECURSIVE :*/ 0 );
		mLock = 0;
		// FIXME: Check for errors.
	}
	
	CLocker( bool inRecursive )
	{
		mSemaphore = create_sem_etc( 0, "", /*(inRecursive) ? SEM_RECURSIVE :*/ 0 );
		mLock = 0;
		// FIXME: Check for errors.
	}
	
	CLocker( const char* inName, bool inRecursive )
	{
		mSemaphore = create_sem_etc( 0, inName, /*(inRecursive) ? SEM_RECURSIVE :*/ 0 );
		mLock = 0;
		// FIXME: Check for errors.
	}

					~CLocker()				{ delete_sem( mSemaphore ); }

	inline bool		Lock() const;
	inline int		LockWithTimeout( bigtime_t inTimeout ) const;
	inline int		Unlock() const;
	inline bool 	IsLocked() const;
	
	int				CountLocks() const		{ return( mNestCount ); }
//	sem_id			Sem() const				( return( mSemaphore ); }
	thread_id		LockingThread() const	{ return( mOwner );     }

private:
					CLocker( const CLocker& );
					CLocker& operator=(const CLocker&);
					
	sem_id			mSemaphore;
	mutable thread_id mOwner;
	mutable int		mLock;
	mutable int		mNestCount;
};

/** 
 * \ingroup util
 * \par Description:
 *
 * \sa
 * \author	Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/


bool CLocker::Lock() const
{
	thread_id nMyThread = get_thread_id( NULL );
	int nError;

	if ( mOwner == nMyThread )
	{
		mNestCount++;
		nError = 0;
	}
	else
	{
		if ( ath_atomic_add( &mLock, 1 ) > 0 )
		{
			for (;;)
			{
				nError = acquire_sem( mSemaphore );
				if ( nError >= 0 || errno != EINTR )
				{
					break;
				}
			}
			if ( nError < 0 )
			{
				ath_atomic_add( &mLock, -1 );
			}
			else
			{
				mOwner = nMyThread;
				mNestCount = 1;
			}
		}
		else
		{
			nError = 0;
			mOwner = nMyThread;
			mNestCount = 1;
		}
	}

	return( 0 == nError );
}


int CLocker::LockWithTimeout( bigtime_t inTimeout ) const
{
	thread_id nMyThread = get_thread_id( NULL );
	int nError;

	if ( mOwner == nMyThread )
	{
		mNestCount++;
		nError = 0;
	}
	else
	{
		if ( ath_atomic_add( &mLock, 1 ) > 0 )
		{
			for (;;)
			{
				nError = acquire_sem_etc( mSemaphore, 1, 0, inTimeout );
				if ( nError >= 0 || errno != EINTR )
				{
					break;
				}
			}
			
			if ( nError < 0 )
			{
				ath_atomic_add( &mLock, -1 );
			}
			else
			{
				mOwner = nMyThread;
				mNestCount = 1;
			}
		}
		else
		{
			nError = 0;
			mOwner = nMyThread;
			mNestCount = 1;
		}
    }

    return( nError );
}


int CLocker::Unlock() const
{
#if 0
    return( release_sem( mSemaphore ) );
#endif
    int nError;

    if ( mNestCount > 1 )
	{
		mNestCount--;
		return( 0 );
    }
	else
	{
		mOwner = -1;
		mNestCount = 0;

		if ( ath_atomic_add( &mLock, -1 ) > 1 )
		{
			nError = release_sem( mSemaphore );
		}
		else
		{
			nError = 0;
		}
		return( nError );
    }
}


inline bool CLocker::IsLocked() const
{
    return( mNestCount > 0 && mOwner == get_thread_id(NULL) );
}


/**
 * \ingroup support
 * \par Description:
 *
 * \sa
 * \author	Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

class Gate
{
public:
	Gate( const char* pzName, bool bRecursive = true )
	{
	    mSemaphore = create_sem_etc( 100000000, pzName, (bRecursive) ? SEM_RECURSIVE : 0 );
	      // FIXME: Check for errors.
	}
		~Gate()			{ delete_sem( mSemaphore );					}
	int Lock() const	{ return( acquire_sem( mSemaphore ) );		}
	int Unlock() const	{ return( release_sem( mSemaphore ) );		}

	int Close() const	{ return( acquire_sem_etc( mSemaphore, 100000000, 0, INFINITE_TIMEOUT ) ); }
	int Open() const	{ return( release_sem_etc( mSemaphore, 100000000, 0 ) ); }
  
private:
	sem_id				mSemaphore;
};

}

#endif // __LOCKER_H__


