/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */


#include <assert.h>
#include <errno.h>
#include <string.h>
#include <support/SupportDefs.h>
#include <cosmoe/OS.h>
#include <cosmoe/kernel.h>
#include <app/Application.h>
#include <app/Message.h>
#include <app/Messenger.h>
#include <appserver/protocol.h>

#include <algorithm>

using namespace os;

CApplication*	CApplication::c_app = NULL;
CClipboard*		CApplication::c_clipboard = NULL;
CRoster*		CApplication::c_roster = NULL;

#define kAppPulseTimerID	-1973

class CApplication::Private
{
public:
	port_id               m_hServerPort;
	port_id               m_hSrvAppPort;
	port_id               m_hReplyPort;
	std::vector<CWindow*> m_cWindows;
};

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CApplication::CApplication( const char* inMimeName ) : CLooper( inMimeName )
{
	m = new Private;
	CMessage cReq( DR_CREATE_APP );
	assert( NULL == c_app );

	m->m_hReplyPort = create_port( "app_reply", DEFAULT_PORT_SIZE );

	assert( m->m_hReplyPort >= 0 );

	m->m_hServerPort = get_app_server_port();

	cReq.AddInt32( "process_id", get_process_id(NULL) );
	cReq.AddInt32( "event_port", GetMsgPort() );
	cReq.AddString( "app_name", inMimeName );

	CMessage cReply;

	if ( CMessenger(m->m_hServerPort).SendMessage( &cReq, &cReply ) >= 0 )
	{
		cReply.FindInt( "app_cmd_port", &m->m_hSrvAppPort );
	}
	else
	{
		dbprintf( "CApplication::CApplication() failed to send message to server: %s\n", strerror( errno ) );
	}

	for ( int i = 0 ; i < COL_COUNT ; ++i )
	{
		Color32_s sColor;
		if ( cReply.FindColor32( "cfg_colors", &sColor, i ) != 0 )
		{
			break;
		}
		__set_default_color( static_cast<default_color_t>(i), sColor );
	}

	assert( m->m_hSrvAppPort >= 0 );

	c_app = this;
	c_clipboard = new CClipboard();
	c_roster = new CRoster();

	// FIXME: Probably the wrong place for this, but...
	this->PostMessage( M_READY_TO_RUN );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CApplication::~CApplication()
{
	Lock();
	for ( uint i = 0 ; i < m->m_cWindows.size() ; ++i )
	{
		m->m_cWindows[i]->PostMessage( M_QUIT );
	}

	while ( m->m_cWindows.size() > 0 )
	{
		Unlock();
		snooze( 20000 );	// Give the windows some time to get rid of themselves
		Lock();
	}

	if ( send_msg( m->m_hSrvAppPort, M_QUIT, NULL, 0 ) < 0 )
	{
		dbprintf( "Error: CApplication::~CApplication() failed to send M_QUIT request to server\n" );
	}

	delete_port( m->m_hReplyPort );

	c_app = NULL;
	delete c_clipboard;
	delete c_roster;

	Unlock();
	delete m;
}

/** \brief Get the (one and only) instance of the CApplication class.
 *
 *        When the CApplication class is instantiated it will assign a pointer to
 *        the instance to a static member that can be obtained through this function.
 *        Many other classes expect this function to return a valid pointer, so it
 *        is important that you instantiate the CApplication class once (and only once)
 *        before calling any other functions in the native AtheOS API.
 *  \return
 *        Pointer to the global instance of the CApplication drived class.
 *  \sa
 *  \author        Kurt Skauen (kurt.skauen@c2i.net)
 *//////////////////////////////////////////////////////////////////////////////

CApplication* CApplication::GetInstance()
{
    return( c_app );
}

/** \brief Return the appserver message port.
 *
 *        When contructed the CApplication class will establish a connection to the
 *        CApplication server. The appserver will create an message port and spawn
 *        a thread that will handle application-level operation on behave of the
 *        CApplication class. This function returnes the message port that connect
 *        the CApplication object with the corresponding thread in the application
 *        server
 *  \return
 *        The message port id for the server side message port connecting this
 *        object to the application server.
 *
 *  \sa
 *  \author        Kurt Skauen (kurt.skauen@c2i.net)
 *//////////////////////////////////////////////////////////////////////////////

port_id CApplication::GetAppPort() const
{
	return( m->m_hSrvAppPort );
}


port_id CApplication::GetServerPort() const
{
	return( m->m_hServerPort );
}


void CApplication::SetPulseRate(bigtime_t inRate)
{
	if (inRate == -1LL)
	{
		RemoveTimer( this, kAppPulseTimerID );
	}
	else if (inRate > 0LL)
	{
		AddTimer( this, kAppPulseTimerID, inRate, false );
	}
}


void CApplication::TimerTick( int inID )
{
	if (inID == kAppPulseTimerID)
	{
		this->PostMessage( M_PULSE );
	}
}


CWindow* CApplication::WindowAt(int32 index)
{
	if ( (index < 0L) || (index >= (int32)m->m_cWindows.size()) )
		return NULL;

	return (m->m_cWindows[index]);
}


int32 CApplication::CountWindows()
{
	return (m->m_cWindows.size());
}


bigtime_t CApplication::GetIdleTime()
{
	CMessage cReq( AR_GET_IDLE_TIME );
	CMessage cReply;

	CMessenger( m->m_hSrvAppPort ).SendMessage( &cReq, &cReply );

	int64 nTime;

	cReply.FindInt64( "idle_time", &nTime );

	return( nTime );
}


void CApplication::PushCursor( mouse_ptr_mode eMode, void* pImage, int nWidth, int nHeight, const IPoint& cHotSpot )
{
	CMessage cReq( WR_PUSH_CURSOR );
	cReq.AddData( "image", T_ANY_TYPE, pImage, nWidth * nHeight );
	cReq.AddIPoint( "size", IPoint( nWidth, nHeight ) );
	cReq.AddInt32( "mode", eMode );
	cReq.AddIPoint( "hot_spot", cHotSpot );

	CMessenger( m->m_hSrvAppPort ).SendMessage( &cReq );
}


void CApplication::PopCursor()
{
	CMessage cReq( WR_POP_CURSOR );
	CMessenger( m->m_hSrvAppPort ).SendMessage( &cReq );
}


int CApplication::GetScreenModeCount()
{
	CMessage cReq( AR_GET_SCREENMODE_COUNT );
	CMessage cReply;

	CMessenger( m->m_hSrvAppPort ).SendMessage( &cReq, &cReply );

	int nCount;

	cReply.FindInt( "count", &nCount );

	return( nCount );
}


int CApplication::GetScreenModeInfo( int nIndex, screen_mode* psMode )
{
	CMessage cReq( AR_GET_SCREENMODE_INFO );
	CMessage cReply;

	cReq.AddInt32( "index", nIndex );

	CMessenger( m->m_hSrvAppPort ).SendMessage( &cReq, &cReply );

	IPoint cScrRes;
	int    nColorSpace;
	int    nError;

	cReply.FindIPoint( "resolution", &cScrRes );
	cReply.FindInt( "bytes_per_line", &psMode->m_nBytesPerLine );
	cReply.FindInt( "color_space", &nColorSpace );
	cReply.FindInt( "error", &nError );

	psMode->m_nWidth  = cScrRes.x;
	psMode->m_nHeight = cScrRes.y;
	psMode->m_eColorSpace = (color_space) nColorSpace;

	if ( nError != 0 )
	{
		errno = nError;
		return( -1 );
	}

	return( 0 );
}


int CApplication::GetFontConfigNames( std::vector<std::string>* pcTable ) const
{
	CMessage cReq( DR_GET_DEFAULT_FONT_NAMES );
	CMessage cReply;

	if ( CMessenger(m->m_hServerPort).SendMessage( &cReq, &cReply ) >= 0 )
	{
		const char* pzName;
		for ( int i = 0 ; cReply.FindString( "config_names", &pzName, i ) == 0 ; ++i )
		{
			pcTable->push_back( pzName );
		}
		return( 0 );
	}

	dbprintf( "CApplication::GetFontConfigNames() failed to send message to server: %s\n", strerror( errno ) );
	return( -1 );
}


int CApplication::GetDefaultFont( const std::string& cName, font_properties* psProps ) const
{
	CMessage cReq( DR_GET_DEFAULT_FONT );
	CMessage cReply;

	cReq.AddString( "config_name", cName );
	if ( CMessenger(m->m_hServerPort).SendMessage( &cReq, &cReply ) >= 0 )
	{
		int nError = -EINVAL;
		cReply.FindInt( "error", &nError );
		if ( nError < 0 )
		{
			errno = -nError;
			return( -1 );
		}

		if ( psProps != NULL )
		{
			cReply.FindString( "family",  &psProps->m_cFamily );
			cReply.FindString( "style",   &psProps->m_cStyle );
			cReply.FindFloat( "size",     &psProps->m_vSize );
			cReply.FindFloat( "shear",    &psProps->m_vShear );
			cReply.FindFloat( "rotation", &psProps->m_vRotation );
		}
		return( 0 );
	}
	else
	{
		dbprintf( "CApplication::GetFontConfigNames() failed to send message to server: %s\n", strerror( errno ) );
		return( -1 );
	} 
}

int CApplication::SetDefaultFont( const std::string& cName, const font_properties& sProps )
{
	CMessage cReq( DR_SET_DEFAULT_FONT );
	CMessage cReply;

	cReq.AddString( "config_name", cName );
	cReq.AddString( "family",  sProps.m_cFamily );
	cReq.AddString( "style",   sProps.m_cStyle );
	cReq.AddFloat( "size",     sProps.m_vSize );
	cReq.AddFloat( "shear",    sProps.m_vShear );
	cReq.AddFloat( "rotation", sProps.m_vRotation );

	if ( CMessenger(m->m_hServerPort).SendMessage( &cReq, &cReply ) >= 0 )
	{
		int nError = -EINVAL;
		cReply.FindInt( "error", &nError );
		if ( nError < 0 )
		{
			errno = -nError;
			return( -1 );
		}
		else
		{
			return( 0 );
		}
	}
	else
	{
		dbprintf( "CApplication::GetDefaultFont() failed to send message to server: %s\n", strerror( errno ) );
		return( -1 );
	}
}


int CApplication::AddDefaultFont( const std::string& cName, const font_properties& sProps )
{
	CMessage cReq( DR_ADD_DEFAULT_FONT );
	CMessage cReply;

	cReq.AddString( "config_name", cName );
	cReq.AddString( "family",  sProps.m_cFamily );
	cReq.AddString( "style",   sProps.m_cStyle );
	cReq.AddFloat( "size",     sProps.m_vSize );
	cReq.AddFloat( "shear",    sProps.m_vShear );
	cReq.AddFloat( "rotation", sProps.m_vRotation );

	if ( CMessenger(m->m_hServerPort).SendMessage( &cReq, &cReply ) >= 0 )
	{
		int nError = -EINVAL;
		cReply.FindInt( "error", &nError );
		if ( nError < 0 )
		{
			errno = -nError;
			return( -1 );
		}
		else
		{
			return( 0 );
		}
	}
	else
	{
		dbprintf( "CApplication::GetDefaultFont() failed to send message to server: %s\n", strerror( errno ) );
		return( -1 );
	}
}

/** \brief Get the number of font families installed on the system.
 *
 *  \return
 *        Number of font families that can be obtained through CApplication::GetFontFamily()
 *  \sa CApplication::GetFontFamily()
 *  \author        Kurt Skauen (kurt.skauen@c2i.net)
 *//////////////////////////////////////////////////////////////////////////////

int CApplication::GetFontFamilyCount()
{
	CMessage cReq( AR_GET_FONT_FAMILY_COUNT );

	CMessage cReply;

	if ( CMessenger( m->m_hSrvAppPort ).SendMessage( &cReq, &cReply ) < 0 )
	{
		dbprintf( "Error: CApplication::GetFontFamilyCount() failed to send message: %s\n", strerror( errno ) );
		return( -1 );
	}
	int nCount = -1;
	cReply.FindInt( "count", &nCount );
	return( nCount );
}

/** \brief Obtaine the name of a font family.
 *
 *        Call this function to learn the name of the nIndex'th font family.
 *  \param nIndex - An integer between 0 and CApplication::GetFontFamilyCount() - 1
 *                       of the font familiy you are interrested in.
 *  \param pzFamily - Pointer to a buffer of at least FONT_FAMILY_LENGTH bytes
 *                         that will receive the family name
 *  \return
 *        If everyting when well 0 will be returned. Otherwise a negative number
 *        is returned, and the <errno> variable is set.
 *  \sa CApplication::GetFontFamilyCount(), CApplication::GetFontStyleCount(), CApplication::GetFontStyle()
 *  \author        Kurt Skauen (kurt.skauen@c2i.net)
 *//////////////////////////////////////////////////////////////////////////////

status_t CApplication::GetFontFamily( int nIndex, char* pzFamily )
{
	CMessage cReq( AR_GET_FONT_FAMILY );

	cReq.AddInt32( "index", nIndex );

	CMessage cReply;

	if ( CMessenger( m->m_hSrvAppPort ).SendMessage( &cReq, &cReply ) < 0 )
	{
		dbprintf( "Error: CApplication::GetFontFamily() failed to send message: %s\n", strerror( errno ) );
		return( -1 );
	}

	int nError = -1;
	cReply.FindInt( "error", &nError );

	if ( nError >= 0 )
	{
		const char* pzName;
		cReply.FindString( "family", &pzName );
		strcpy( pzFamily, pzName );
	}
	else
	{
		errno = -nError;
	}
	return( nError );
}

/** \brief Obtaine the number of styles belonging to a font family
 *
 *        Call this function to learn how many styles are embedded in a font family.
 *  \param pzFamily - Name of the family to investigate.
 *  \return The style count
 *  \sa CApplication::GetFontStyleCount(), GetFontFamilyCount(), GetFontFamily()
 *  \author        Kurt Skauen (kurt.skauen@c2i.net)
 *//////////////////////////////////////////////////////////////////////////////

int CApplication::GetFontStyleCount( const char* pzFamily )
{
	CMessage cReq( AR_GET_FONT_STYLE_COUNT );
	CMessage cReply;

	cReq.AddString( "family", pzFamily );

	if ( CMessenger( m->m_hSrvAppPort ).SendMessage( &cReq, &cReply ) < 0 )
	{
		dbprintf( "Error: CApplication::GetFontStyleCount() failed to send message: %s\n", strerror( errno ) );
		return( -1 );
	}
	int nCount = 0;

	cReply.FindInt( "count", &nCount );

	if ( nCount < 0 )
	{
		errno = -nCount;
		nCount = -1;
	}
	return( nCount );
}

/** \brief Get a font style.
 *
 *        Call this function to get the name and properties of a font style based
 *        on the family name, and an index into the family's list of styles.
 *
 *  \param pzFamily - Name of the family containing the style.
 *  \param nIndex - The index into the family's list of styles.
 *  \param pnFlags - Pointer to an uint32 that if not NULL will receive a
 *                     bitmask describing the font (see font_style_flags)
 *  \return
 *        If everyting when well 0 will be returned. Otherwise a negative number
 *        is returned, and the <errno> variable is set.
 *  \sa CApplication::GetFontStyleCount(), CApplication::GetFontFamilyCount()
 *  \sa CApplication::GetFontFamily()
 *  \author        Kurt Skauen (kurt.skauen@c2i.net)
 *//////////////////////////////////////////////////////////////////////////////

status_t CApplication::GetFontStyle( const char* pzFamily, int nIndex, char* pzStyle, uint32* pnFlags )
{
	CMessage cReq( AR_GET_FONT_STYLE );
	CMessage cReply;


	cReq.AddString( "family", pzFamily );
	cReq.AddInt32( "index", nIndex );

	if ( CMessenger( m->m_hSrvAppPort ).SendMessage( &cReq, &cReply ) < 0 )
	{
		dbprintf( "Error: CApplication::GetFontStyle() failed to send message: %s\n", strerror( errno ) );
		return( -1 );
	}

	int nError = -EINVAL;

	cReply.FindInt( "error", &nError );

	if ( nError >= 0 )
	{
		const char* pzStyleName = "";

		cReply.FindString( "style", &pzStyleName );
		cReply.FindInt( "flags", pnFlags );
		strcpy( pzStyle, pzStyleName );
		return( 0 );
	}
	else
	{
		errno = -nError;
		return( -1 );
	}
}

/** \brief Returnes the current state of keyboards qualifiers.
 *
 *        Sends a requester to the appserver to obtain the current
 *        state of keyboard qualifiers.
 *  \return
 *        A bit mask composed from os::qualifiers decribing which qualifiers
 *        currently pressed
 *  \sa
 *  \author        Kurt Skauen (kurt.skauen@c2i.net)
 *//////////////////////////////////////////////////////////////////////////////

uint32 CApplication::GetQualifiers()
{
	AR_GetQualifiers_s sReq( m->m_hReplyPort );
	AR_GetQualifiersReply_s sReply;

	if ( send_msg( m->m_hSrvAppPort, AR_GET_QUALIFIERS, &sReq, sizeof( sReq ) ) != 0 )
	{
		dbprintf( "Error: CApplication::GetQualifiers() failed to send request to server\n" );
		return( 0 );
	}

	if ( get_msg( m->m_hReplyPort, NULL, &sReply, sizeof( sReply ) ) < 0 )
	{
		dbprintf( "Error: CApplication::GetQualifiers() failed to read reply from server\n" );
		return( 0 );
	}

	return( sReply.m_nQualifiers );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

port_id CApplication::CreateWindow( CView* pcTopView, const CRect& cFrame, const std::string& cTitle,
                                   uint32 nFlags, uint32 nDesktopMask, port_id hEventPort, int* phTopView )
{
	CMessage cReq( AR_OPEN_WINDOW );

	cReq.AddString( "title", cTitle.c_str() );
	cReq.AddInt32( "event_port", hEventPort );
	cReq.AddInt32( "flags", nFlags );
	cReq.AddInt32( "desktop_mask", nDesktopMask );
	cReq.AddRect( "frame", cFrame );
	cReq.AddPointer( "user_obj", pcTopView );

	CMessage cReply;

	CMessenger( m->m_hSrvAppPort ).SendMessage( &cReq, &cReply );

	port_id hCmdPort;
	cReply.FindInt( "top_view", phTopView );
	cReply.FindInt( "cmd_port", &hCmdPort );
	return( hCmdPort );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

port_id CApplication::CreateWindow( CView* pcTopView, int hBitmapHandle, int* phTopView )
{
	CMessage cReq( AR_OPEN_BITMAP_WINDOW );

	cReq.AddInt32( "bitmap_handle", hBitmapHandle );
	cReq.AddPointer( "user_obj", pcTopView );

	CMessage cReply;

	CMessenger( m->m_hSrvAppPort ).SendMessage( &cReq, &cReply );

	port_id hCmdPort;
	cReply.FindInt( "top_view", phTopView );
	cReply.FindInt( "cmd_port", &hCmdPort );

	return( hCmdPort );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

int CApplication::CreateBitmap( int nWidth,
								int nHeight,
								color_space eColorSpc,
								bool inShareFramebuffer,
								int* phHandle,
								int* phArea )
{
	AR_CreateBitmap_s      sReq;
	AR_CreateBitmapReply_s sReply;

	sReq.hReply            = m->m_hReplyPort;
	sReq.bShareFramebuffer = inShareFramebuffer;
	sReq.nWidth            = nWidth;
	sReq.nHeight           = nHeight;
	sReq.eColorSpc         = eColorSpc;

	if ( send_msg( m->m_hSrvAppPort, AR_CREATE_BITMAP, &sReq, sizeof( sReq ) ) != 0 )
	{
		dbprintf( "Error: CApplication::CreateBitmap() failed to send request to server\n" );
		return( -1 );
	}

	if ( get_msg( m->m_hReplyPort, NULL, &sReply, sizeof( sReply ) ) < 0 )
	{
		dbprintf( "Error: CApplication::CreateBitmap() failed to read reply from server\n" );
		return( -1 );
	}

	if ( sReply.m_hHandle < 0 )
	{
		return( sReply.m_hHandle );
	}
	*phHandle = sReply.m_hHandle;
	*phArea   = sReply.m_hArea;
	return( 0 );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

int CApplication::DeleteBitmap( int nBitmap )
{
	AR_DeleteBitmap_s sReq( nBitmap );

	if ( send_msg( m->m_hSrvAppPort, AR_DELETE_BITMAP, &sReq, sizeof( sReq ) ) != 0 )
	{
		dbprintf( "Error: CApplication::DeleteBitmap() failed to send request to server\n" );
		return( -1 );
	}
	
	return( 0 );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

int CApplication::CreateSprite( const CRect& cFrame, int nBitmapHandle, uint32* pnHandle )
{
	AR_CreateSprite_s sReq( m->m_hReplyPort, cFrame, nBitmapHandle );

	if ( send_msg( m->m_hSrvAppPort, AR_CREATE_SPRITE, &sReq, sizeof( sReq ) ) != 0 )
	{
		dbprintf( "Error: CApplication::CreateSprite() failed to send request to server\n" );
		return( -1 );
	}

	AR_CreateSpriteReply_s sReply;

	if ( get_msg( m->m_hReplyPort, NULL, &sReply, sizeof( sReply ) ) < 0 )
	{
		dbprintf( "Error: CApplication::CreateSprite() failed to read reply from server\n" );
		return( -1 );
	}
	*pnHandle = sReply.m_nHandle;
	return( sReply.m_nError);
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CApplication::DeleteSprite( uint32 nSprite )
{
	AR_DeleteSprite_s sReq( nSprite );

	if ( send_msg( m->m_hSrvAppPort, AR_DELETE_SPRITE, &sReq, sizeof( sReq ) ) != 0 )
	{
		dbprintf( "Error: CApplication::DeleteSprite() failed to send request to server\n" );
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CApplication::MoveSprite( uint32 nSprite, const CPoint& cNewPos )
{
	AR_MoveSprite_s sReq( nSprite, cNewPos );

	if ( send_msg( m->m_hSrvAppPort, AR_MOVE_SPRITE, &sReq, sizeof( sReq ) ) != 0 )
	{
		dbprintf( "Error: CApplication::MoveSprite() failed to send request to server\n" );
	}
}

int CApplication::LockDesktop( int nDesktop, CMessage* pcDesktopParams )
{
	CMessage cReq( AR_LOCK_DESKTOP );
	cReq.AddInt32( "desktop", nDesktop );
	return( CMessenger( m->m_hSrvAppPort ).SendMessage( &cReq, pcDesktopParams ) );
}

int CApplication::UnlockDesktop( int nCookie )
{
	return( 0 );
}

void CApplication::SetScreenMode( int inDesktop, uint32 nValidityMask, screen_mode* psMode )
{
	CMessage cReq( AR_SET_SCREEN_MODE );

	cReq.AddInt32( "desktop", inDesktop );
	cReq.AddIPoint( "resolution", IPoint( psMode->m_nWidth, psMode->m_nHeight ) );
	cReq.AddInt32( "color_space", psMode->m_eColorSpace );
	cReq.AddFloat( "refresh_rate", psMode->m_vRefreshRate );
	cReq.AddFloat( "h_pos", psMode->m_vHPos );
	cReq.AddFloat( "v_pos", psMode->m_vVPos );
	cReq.AddFloat( "h_size", psMode->m_vHSize );
	cReq.AddFloat( "v_size", psMode->m_vVSize );

	CMessage cReply;
	CMessenger( m->m_hSrvAppPort ).SendMessage( &cReq, &cReply );
}

void CApplication::SetWindowDecorator( const char* pzPath )
{
	DR_SetWindowDecorator_s sReq;
	strcpy( sReq.m_zDecoratorPath, pzPath );
	if ( send_msg( m->m_hServerPort, DR_SET_WINDOW_DECORATOR, &sReq, sizeof( sReq ) ) != 0 )
	{
		dbprintf( "Error: CApplication::SetWindowDecorator() failed to send request to server\n" );
	}
}

void CApplication::CommitColorConfig()
{
	CMessage cMsg( DR_SET_COLOR_CONFIG );

	for ( int i = 0 ; i < COL_COUNT ; ++i )
	{
		cMsg.AddColor32( "_colors", get_default_color( static_cast<default_color_t>(i) ) );
	}
	CMessenger( m->m_hServerPort ).SendMessage( &cMsg );
}

void CApplication::GetKeyboardConfig( std::string* pcKeymapName, int* pnKeyDelay, int* pnKeyRepeat )
{
	CMessage cReq( DR_GET_KEYBOARD_CFG );
	CMessage cReply;

	CMessenger( m->m_hServerPort ).SendMessage( &cReq, &cReply );

	const char* pzKeymap;
	if ( pcKeymapName != NULL && cReply.FindString( "keymap", &pzKeymap ) == 0 )
	{
		*pcKeymapName = pzKeymap;
	}

	if ( pnKeyDelay != NULL )
	{
		cReply.FindInt( "delay", pnKeyDelay );
	}

	if ( pnKeyRepeat != NULL )
	{
		cReply.FindInt( "repeat", pnKeyRepeat );
	}
}

status_t CApplication::SetKeymap( const char* pzName )
{
	CMessage cReq( DR_SET_KEYBOARD_CFG );
	cReq.AddString( "keymap", pzName );
	CMessenger( m->m_hServerPort ).SendMessage( &cReq );
	return( 0 );
}

status_t CApplication::SetKeyboardTimings( int nDelay, int nRepeat )
{
	CMessage cReq( DR_SET_KEYBOARD_CFG );
	cReq.AddInt32( "delay", nDelay );
	cReq.AddInt32( "repeat", nRepeat );
	CMessenger( m->m_hServerPort ).SendMessage( &cReq );
	return( 0 );
}

void CApplication::MessageReceived( CMessage* pcMessage )
{
	switch( pcMessage->what )
	{
		case M_COLOR_CONFIG_CHANGED:
			for ( int i = 0 ; i < COL_COUNT ; ++i )
			{
				Color32_s sColor;
				if ( pcMessage->FindColor32( "_colors", &sColor, i ) != 0 )
				{
					break;
				}
				__set_default_color( static_cast<default_color_t>(i), sColor );
			}
//            for ( uint i = 0 ; i < m_cWindows.size() ; ++i ) {
//                m_cWindows[i]->_DefaultColorsChanged();
//            }
			break;

		case M_READY_TO_RUN:
			ReadyToRun();
			break;

		case M_PULSE:
			Pulse();
			break;

		default:
			CLooper::MessageReceived( pcMessage );
			break;
	}
}

/** \brief Entry point for the message loop.
 *
 *        The CApplication class is different from other loopers in that it does not
 *        spawn a new thread when the Run() member is called. Instead the Run() member
 *        directly enters the message loop, and does not return until the message loop
 *        quits.
 *
 *  \return The thread id of the thread calling it.
 *  \sa
 *  \author        Kurt Skauen (kurt.skauen@c2i.net)
 *//////////////////////////////////////////////////////////////////////////////

thread_id CApplication::Run()
{
	thread_id hThread = get_thread_id( NULL );
	_SetThread( hThread );
	Unlock();
	_Loop();
	return( hThread );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CApplication::AddWindow( CWindow* pcWindow )
{
	Lock();
	m->m_cWindows.push_back( pcWindow );
	Unlock();
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CApplication::RemoveWindow( CWindow* pcWindow )
{
	Lock();
	std::vector<CWindow*>::iterator i = std::find( m->m_cWindows.begin(), m->m_cWindows.end(), pcWindow );

	if ( i == m->m_cWindows.end() )
	{
		dbprintf( "Error: CApplication::RemoveWindow() could not find window %p\n", pcWindow );
	}

	if ( i != m->m_cWindows.end() )
	{
		m->m_cWindows.erase( i );
	}

	Unlock();
}



void CApplication::FBC_Application1() {}
void CApplication::FBC_Application2() {}
void CApplication::FBC_Application3() {}
void CApplication::FBC_Application4() {}
void CApplication::FBC_Application5() {}
void CApplication::FBC_Application6() {}
void CApplication::FBC_Application7() {}
void CApplication::FBC_Application8() {}
void CApplication::FBC_Application9() {}
void CApplication::FBC_Application10() {}

