/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */

 
#include <string.h>

#include <app/Handler.h>
#include <app/Looper.h>
#include <app/MessageFilter.h>
#include <cosmoe/kernel.h>

using namespace os;


/** Construct a handler.
 * \par Description:
 *	Initialize the handler and give it a name.
 *	The name does not need to be unique unless you want to be
 *	able to look it up with os::CLooper::FindHandler()
 * \param cName
 *	A name identifying the handler. This can be passed to os::CLooper::FindHandler()
 *	on a looper to find the handler object when it have been added to the
 *	looper with os::CLooper::AddHandler().
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CHandler::CHandler( const char* cName ) : mName( cName )
{
	static int nLastToken = 1;
	
	m_nToken              = nLastToken++;
	m_pcNextHandler       = NULL;
	m_pcLooper            = NULL;
}


/** Destructor
 * \par Description:
 *	Free all resources allocated by the handler.
 *	If the handler is member of a looper it will detach itself
 *	from the looper and delete all message filters associated
 *	with the handler.
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CHandler::~CHandler()
{
	if ( m_pcLooper != NULL )
	{
		if ( m_pcLooper->RemoveHandler( this ) == false )
		{
			dbprintf( "CHandler::~CHandler() failed to remove handler %p from looper\n", this );
		}
	}

	for ( MsgFilterList::iterator i = mFilterList.begin() ; i != mFilterList.end() ; ++i )
	{
		(*i)->mHandler = NULL;
		delete (*i);
	}

	m_nToken = ~m_nToken;
}


/** Handle a message targeted at this handler.
 * \par Description:
 *	Overload this member to dispatch messages sent to this handler.
 *	When a looper receives a message for one of its handlers it will
 *	call the taget handlers MessageReceived() to allow the handler to
 *	dispatch the message.
 * \par
 *	The message passed in \p pcMessage is also available through
 *	os::CLooper::CurrentMessage() and os::CLooper::DetachCurrentMessage()
 *	until this member returns. This is normally not very useful for
 *	MessageReceived() itself but it can be convinient for other members
 *	called from MessageReceived() in case they need data from the message
 *	that was not passed on from MessageReceived().
 * \par
 *	The looper will be locked when this member is called. The default
 *	implementation of this member will pass the message on to the
 *	next handler if one was set with SetNextHandler().
 * \note
 *	Never do any lenthy operations in any hook members that are called
 *	from the looper thread if the looper is involved with the GUI (for
 *	example if the looper is a os::CWindow). The looper will not be
 *	able to dispatch messages until the hook returns so spending a
 *	long time in this members will make the GUI feel unresponsive.
 * \param pcMessage
 *	The message that should be handled. This message will be deleted by
 *	the looper when MessageReceived() returns unless you detach it with
 *	os::CLooper::DetachCurrentMessage(),
 * \sa os::CLooper::DispatchMessage(), os::CLooper::DetachCurrentMessage()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CHandler::MessageReceived( CMessage* pcMessage )
{
	if ( NULL != m_pcNextHandler )
	{
		m_pcLooper->PostMessage( pcMessage, m_pcNextHandler );
	}
}


/** Get the loopers name.
 * \par Description:
 *	Get the name assigned to the handler in the constructor or with
 *	the SetName() member.
 * \return The current name of the handler.
 * \sa SetName(), os::CLooper::FindHandler()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

const char* CHandler::Name() const
{
	return( mName.c_str() );
}


/** Get a pointer to the looper this handler belongs to.
 * \par Description:
 *	GetLooper() returns a pointer to the looper this handler have been
 *	added to with os::CLooper::AddHandler(). If the handler has not been
 *	added to any looper NULL will be returned.
 * \par
 *	If this is called on a os::CLooper object (os::CLooper is a subclass of os::CHandler)
 *	this member will always return a pointer to itself.
 * \return Pointer to the looper this handler is a member to or NULL if it have
 *	not yet been added to a looper.
 * \sa os::CLooper::AddHandler()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CLooper* CHandler::Looper() const
{
	return( m_pcLooper );
}


/** Set a handler that should handle messages this handler is not interrested in.
 * \par Description:
 *	The default implementation of HandleMessage() will check if there is
 *	set an alternative handler through SetNextHandler() and if so the
 *	message will be forwarded to this handler.
 * \param pcNextHandler
 *	Pointer to the handler that should receive messages after we are
 *	done with them or NULL to disable the forwarding.
 * \sa GetNextHandler(), HandleMessage()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CHandler::SetNextHandler( CHandler* pcNextHandler )
{
	m_pcNextHandler = pcNextHandler;
}


/** Get the next handler in a handler chain.
 * \par Description:
 *	Get the next handler in a handler chain (as set with SetNextHandler()).
 * \return Pointer to the next handler in the handler chain or NULL if no
 *	"next" handler is set.
 * \sa SetNextHandler(), MessageReceived()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CHandler* CHandler::NextHandler() const
{
	return( m_pcNextHandler );
}


/** Rename the handler.
 * \par Description:
 *	Rename the handler
 * \param cName
 *	The new handler name.
 * \sa GetName(), os::Looper::FindHandler()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CHandler::SetName( const char* inName )
{
	mName = inName;
}


/** Timer dispatch member.
 * \par Description:
 *	This member will be called by the looper thread when a timer targeting
 *	this handler expires.
 * \par
 *	When a timer created with os::CLooper::AddTimer() expires the looper
 *	thread will lock the looper and call this member on the target for
 *	the expired timer. If more than one timer is created it is possible
 *	to distinguish them by the timer ID that is assigned to the timer
 *	with os::CLooper::AddTimer() and that is passed to TimerTick()
 *	through the \p inID parameter.
 * \note
 *	Never do any lenthy operations in any hook members that are called
 *	from the looper thread if the looper is involved with the GUI (for
 *	example if the looper is a os::CWindow). The looper will not be
 *	able to dispatch messages until the hook returns so spending a
 *	long time in this members will make the GUI feel unresponsive.
 * \par Warning:
 * \param
 * \par Error codes:
 * \sa
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CHandler::TimerTick( int inID )
{
}


void CHandler::AddFilter( CMessageFilter* pcFilter )
{
	if ( pcFilter == NULL )
	{
		dbprintf( "Error: CHandler::AddFilter() called with a NULL pointer\n" );
		return;
	}

	if ( pcFilter->mHandler != NULL )
	{
		if ( pcFilter->mGlobalFilter )
		{
			static_cast<CLooper*>(pcFilter->mHandler)->RemoveCommonFilter( pcFilter );
		}
		else
		{
			pcFilter->mHandler->RemoveFilter( pcFilter );
		}
	}
	pcFilter->mIterator     = mFilterList.insert( mFilterList.end(), pcFilter );
	pcFilter->mHandler      = this;
	pcFilter->mGlobalFilter = false;
}


void CHandler::RemoveFilter( CMessageFilter* pcFilter )
{
	if ( pcFilter->mHandler != this )
	{
		dbprintf( "Error: CHandler::RemoveFilter() attempt to remove filter not belonging to us\n" );
		return;
	}

	if ( pcFilter->mGlobalFilter )
	{
		dbprintf( "Error: CHandler::RemoveFilter() attempt to remove a common filter using RemoveFilter()\n" );
		return;
	}

	mFilterList.erase( pcFilter->mIterator );
	pcFilter->mHandler = NULL;
}


const MsgFilterList& CHandler::FilterList() const
{
	return( mFilterList );
}


bool CHandler::LockLooper()
{
	return m_pcLooper->Lock();
}


status_t CHandler::LockLooperWithTimeout(bigtime_t inTimeout)
{
	return m_pcLooper->LockWithTimeout(inTimeout);
}


void CHandler::UnlockLooper()
{
	m_pcLooper->Unlock();
}



void CHandler::FBC_Handler1() {}
void CHandler::FBC_Handler2() {}
void CHandler::FBC_Handler3() {}
void CHandler::FBC_Handler4() {}
void CHandler::FBC_Handler5() {}

