/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */

 
#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <support/SupportDefs.h>

#include <cosmoe/OS.h>
#include <cosmoe/kernel.h>
#include <cosmoe/fs_info.h>
#include <app/AppDefs.h>
#include <app/Looper.h>
#include <app/Message.h>
#include <app/MessageQueue.h>
#include <app/MessageFilter.h>
#include <support/String.h>
#include <support/Autolock.h>

using namespace os;

class CLooper::Private
{
public:
	Private( const char* cName ) : m_cName( cName ), m_cDefaultMutex( cName ) { m_pcMutex = &m_cDefaultMutex; }

	std::string					m_cName;
	TimerNode*					m_pcFirstTimer;

	std::map<int,CHandler*>		m_cHandlerMap;
	CHandler*					m_pcDefaultHandler;
	int							m_nHandlerCount;

	CMessage*					m_pcCurrentMessage;
	CMessageQueue*				m_pcMsgQueue;

	int							m_nID;
	int							m_nPriority;
	thread_id					m_hThread;
	port_id						m_hPort;
	mutable CLocker				m_cDefaultMutex;
	CLocker*					m_pcMutex;
	CLooper*					m_pcNextLooper;
	bigtime_t					m_nNextEvent;
	std::list<CMessageFilter*>	m_cCommonFilterList;
};

//-------------------- Static members        --------------------------------------

static CLocker                      g_cLopperListMutex( "looper_list_lock" );
static std::map<CLooper*,int>       g_cLooperPtrMap;
static std::map<thread_id,CLooper*> g_cLooperThreadMap;
static std::map<port_id,CLooper*>   g_cLooperPortMap;

CLooper::TimerNode::TimerNode( CHandler* pcHandler, int nID, bigtime_t nPeriod, bool bOneShot )
{
	m_nPeriod   = nPeriod;
	m_nTimeout  = system_time() + nPeriod;
	m_pcHandler = pcHandler;
	m_nID       = nID;
	m_bOneShot  = bOneShot;
}

/** \brief The looper constructor.
 *
 *        This is the only CLooper constructor. It initiates the looper and
 *        create the needed message port and message queue, but it does not
 *        spawn a new thread. To actually start the message loop you must
 *        call Run(). Before the constructor returns it will lock the looper
 *        by calling Lock(). This means that before any other threads are able
 *        to access it you must call Unlock().
 *
 *  \param pzName - Passed down to the CHandler::CHandler() constructor.
 *  \param nPriority - Stored for later usage as the looper-thread priority.
 *  \param nPortSize - Maximum number of pending messages.
 *  \sa CHandler::CHandler()
 *  \author        Kurt Skauen (kurt@atheos.cx)
 *//////////////////////////////////////////////////////////////////////////////

CLooper::CLooper( const char* cName, int nPriority, int nPortSize ) : CHandler( cName )
{
	static atomic_t nLastID = 1;

	m = new Private( cName );
	m->m_nID = ath_atomic_add( &nLastID, 1 );
	
	m->m_nPriority        = nPriority;
	m->m_hThread          = -1;
	m->m_hPort            = create_port( "looper_port", nPortSize );
	m->m_pcDefaultHandler = NULL;
	m->m_pcFirstTimer     = NULL;
	m->m_nNextEvent       = INFINITE_TIMEOUT;

	m->m_pcCurrentMessage = NULL;
	m->m_pcMsgQueue       = new CMessageQueue();

	Lock();
	AddHandler( this );
	_AddLooper( this );
}


/** Looper destructor
 * \par Description:
 *	Free all resources allocated by the looper.
 * \note
 *	You should normally not delete a looper with the "delete" operator.
 *	The looper thread will delete the object itself before it terminates.
 *	To get rid of a looper you should terminate it by calling Quit() or
 *	by sending it a M_QUIT message.
 * \par
 *	In some very rare cases it is still possible to delete the looper
 *	which is why the destructor is not private. If you just create the
 *	looper but never call Run() it is safe to delete the object. After
 *	calling Run() the looper thread are responsible for the looper object
 *	and deleting it will cause the application to crash!
 *  \sa CLooper::CLooper()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CLooper::~CLooper()
{
	_RemoveLooper( this );
	RemoveHandler( this );
	delete_port( m->m_hPort );

	while( m->m_pcFirstTimer != NULL )
	{
		TimerNode* pcTimer = m->m_pcFirstTimer;
		m->m_pcFirstTimer = pcTimer->m_pcNext;
		delete pcTimer;
	}

	delete m;
}


void CLooper::_SetThread( thread_id hThread )
{
	m->m_hThread = hThread;
	if ( hThread != -1 )
	{
		g_cLopperListMutex.Lock();
		g_cLooperThreadMap[hThread] = this;
		g_cLopperListMutex.Unlock();
	}
}


port_id CLooper::_GetPort() const
{
	return( m->m_hPort );
}


/** Rename the looper.
 * \par Description:
 *	Rename the looper. If the looper is already started the looper-thread
 *	will be renamed aswell. If there is no looper-thread yet the name
 *	will be remembered and will be used as the thread-name when the
 *	looper is started.
 * \par
 *	The looper name is mostly used for debugging and does not need to
 *	be unique.
 * \param inName
 *	The new looper name
 * \sa Name()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CLooper::SetName( const char* inName )
{
	m->m_cName = inName;
	if ( m->m_hThread != -1 )
	{
		rename_thread( m->m_hThread, inName );
	}
}


/** Get the looper's name.
 * \par Description:
 *	Get the looper name as set by the constructor or the SetName()
 *	member.
 * \return The looper's name.
 * \sa SetName()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

const char* CLooper::Name() const
{
	return( m->m_cName.c_str() );
}


/** \brief Add a timer to the looper.
 *
 *  \par Description:
 *        Timers should be employed when you have a small task that should be
 *        run periodically, or that you would like to schedule some point ahead
 *        in time. A timer is targeted at a specific handler, and will cause
 *        the CHandler::TimerTick() function to be called each time the timer
 *        fires. You can add an arbritary number of timers to a looper.
 *        The CHandler::TimerTick() will be called by the looper thread, with
 *        the looper locked.
 *  \par Note:
 *        As with every callbacks from the looper thread, you should not do any
 *        long-lasting computing in the CHandler::TimerTick() function. The looper
 *        will not be able to handle any messages until the function returns.
 *        This can cause the application to feel unresponsive. If you have any
 *        long-lasting tasks, you hould spawn a new thread to get the job done.
 *
 *  \param pcTarget - Must be a valid pointer to a handler belonging to this looper.
 *                      This is the handler that will get it's TimerTick() member called.
 *  \param nID - An user defined ID that is passed to the TimerTick() function
 *                 to make it possible to distinguish different timers.
 *  \param nPeriod - Time in micro seconds before the timer fires. If not in
 *                      one-shot mode, the timer will be rescheduled with the
 *                      same delay until manually removed.
 *  \param bOneShot - If true the timer will fire once, and then be removed.
 *                      If false the timer will be continually rescheduled until
 *                      removed by CLooper::RemoveTimer().
 *
 *  \sa CLooper::RemoveTimer(),CHandler::TimerTick()
 *  \author        Kurt Skauen (kurt@atheos.cx)
 *//////////////////////////////////////////////////////////////////////////////

void CLooper::AddTimer( CHandler* pcTarget, int nID, bigtime_t nPeriod, bool bOneShot )
{
	RemoveTimer( pcTarget, nID );

	TimerNode* pcTimer = new TimerNode( pcTarget, nID, nPeriod, bOneShot );
	_InsertTimer( pcTimer );
}

/** \brief Delete a timer.
 *
 *        When creating a timer in repeate mode, you will propably at some time
 *        want to get rid of it. This is the way to go.
 *
 *  \param - pcTarget - The handler for which a timer is to be removed.
 *  \param - nID - The ID of the timer to be removed.
 *
 *  \return True if the timer was found, false otherwise.
 *  \sa CLooper::AddTimer(), CHandler::TimerTick()
 *  \author        Kurt Skauen (kurt@atheos.cx)
 *//////////////////////////////////////////////////////////////////////////////

bool CLooper::RemoveTimer( CHandler* pcTarget, int nID )
{
	TimerNode** ppcTmp;

	for ( ppcTmp = &m->m_pcFirstTimer ; NULL != *ppcTmp ; ppcTmp = &(*ppcTmp)->m_pcNext )
	{
		TimerNode* pcTimer = *ppcTmp;
		if ( pcTimer->m_pcHandler == pcTarget && pcTimer->m_nID == nID )
		{
			*ppcTmp = pcTimer->m_pcNext;
			delete pcTimer;
			return( true );
		}
	}
	return( false );
}


/** Obtain the low-level message port used by this looper.
 *  \par Description:
 *	Obtain the low-level message port used by this looper.
 *  \return The port ID of the internal message port.
 *  \sa MessageQueue(), CurrentMessage()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

port_id CLooper::GetMsgPort() const
{
	return( m->m_hPort );
}


/** Obtain the thread id of the looper thread.
 *  \par Description:
 *	Obtain the thread id of the looper thread.
 *  \return A valid thread ID if the loop is running, -1 otherwise.
 *  \sa GetProcess()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

thread_id CLooper::GetThread() const
{
	return( m->m_hThread );
}


/** Obtain the process ID of the loopers thread.
 *  \par Description:
 *	Obtain the process ID of the loopers thread.
 *  \return A valid process ID if the loop is running, -1 otherwise.
 *  \sa GetThread()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

proc_id CLooper::GetProcess() const
{
	if (m->m_hThread == -1)
		return -1;

	return( get_thread_proc( m->m_hThread ) );
}

/** Lock the looper object.
 * \par Description:
 *	Before calling almost any other function on a looper or one of its derived
 *	classes, you must lock it by calling this member. You can nest calls
 *	to Lock() within the same thread, each call to Lock() will then require a
 *	corresponding call to Unlock() to release the looper.
 * \par
 *	CLooper::Lock() will not be affected by signals that don't kill the thread.
 *	It will keep on retrying until it succed, or the failure was from someting
 *	else than a signal.
 * \note
 *	You should avoid keeping the looper locked for an extended periode of time
 *	since that will prevent it from handling any messages and can cause the
 *	application to feel unresponsive.
 *  \return Returns true, unless someting really bad happened.
 *  \sa SafeLock(), Unlock()
 * \author Kurt Skauen (kurt@atheos.cx), Bill Hayden (hayden@haydentech.com)
 *****************************************************************************/

bool CLooper::Lock()
{
	status_t        nError;

	for ( ; ; )
	{
		nError = m->m_pcMutex->LockWithTimeout(INFINITE_TIMEOUT);

		if ( nError < 0 )
		{
			dbprintf( "CLooper::Lock() - failed to lock looper! Err = %s\n", strerror( errno ) );
			if ( EINTR == errno )
			{
				continue;
			}
		}
		break;
	}

	return( 0 == nError );
}


/** Lock the looper with a timeout.
 * \par Description:
 *	Lock the looper with a timeout. Like Lock() this member will attempt
 *	to acquire the loopers mutex but this version will only wait \p nTimeout
 *	micro seconds before timing out and fail with errno==ETIME.
 * \note
 *	Unlike the regular Lock() member this member will fail with errno==EINTR
 *	if the thread receive a POSIX signal while waiting to aquire the mutex
 *	or with errno==ETIME if the timeout is reached.
 * \par Warning:
 * \param nTimeout
 *	Maximum number of micro seconds to wait for the looper mutex
 *	to be available.
 * \return On success 0 is returned. On failure -1 is returned and a error
 *	code is written to the global variable \b errno.
 * \par Error codes:
 *	<dl>
 *	<dd>EINTR	An unignored POSIX signal was received.
 *	<dd>ETIME	The timeout expired before the mutex could be aquired.
 *	</dl>
 * \sa Lock(), SafeLock(), Unlock()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

status_t CLooper::LockWithTimeout( bigtime_t nTimeout )
{
	return (m->m_pcMutex->LockWithTimeout( nTimeout ));
}


/** Attempt to lock the looper.
 * \par Description:
 *	SafeLock() have the same semantics as Lock() except that it verify
 *	that the looper object is valid before attempting to lock it.
 * \par
 *	A common problem when locking a looper from an external thread is
 *	that the looper might terminate an become invalid between the
 *	time you obtain a pointer to it and the time you call Lock().
 *	This makes it impossible to lock a looper with Lock()
 *	from anything but the loopers own thread unless you have some other
 *	mechanism in place that guarantee that the looper will not go away
 *	before it is properly locked.
 * \par
 *	SafeLock() will verify that the looper is valid and lock it in
 *	a atomic fasion thus solving this proble. Note that this
 *	is much more expensive than the regular Lock() method so you
 *	should think twice before calling SafeLock().
 * \note
 *	When communicating between two loopers it is often better to
 *	do so asyncronously through messages than by locking the
 *	peer and do direct member calls.
 * \return On success 0 is returned. On error -1 is returned and a error code
 *	is written to the global variable \b errno.
 * \par Error codes:
 *	<dl>
 *	<dd>EINVAL	SafeLock() was called on an invalid looper.
 *	</dl>
 * \sa Lock(), Unlock()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

status_t CLooper::SafeLock()
{
	status_t nError;

	g_cLopperListMutex.Lock();
	for (;;)
	{
		std::map<CLooper*,int>::iterator i = g_cLooperPtrMap.find( this );
		if ( i != g_cLooperPtrMap.end() && (*i).second == m->m_nID )
		{
			nError = LockWithTimeout( 5000LL );
			if ( nError < 0 )
			{
				if ( errno == ETIME || errno == EINTR )
				{
					g_cLopperListMutex.Unlock();
					snooze( 5000LL );
					g_cLopperListMutex.Lock();
					continue;
				}
				else
				{
					break;
				}
			}
			else
			{
				break;
			}
		}
		else
		{
			nError = -1;
			errno = EINVAL;
			break;
		}
	}
	g_cLopperListMutex.Unlock();
	return( nError );
}



/** Unlock the looper.
 * \par Description:
 *	Unlock the looper.
 * \return Return 0, unless someting really bad happened.
 * \sa Lock(), SafeLock()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

status_t CLooper::Unlock()
{
	return( m->m_pcMutex->Unlock() );
}


/** Set a custom mutex to be used by the looper.
 * \par Description:
 *	Each looper own a mutex that is used to protect the looper object
 *	against other threads. Normally this mutex is handled entirely
 *	internally by the looper and is only accesses indirectly through
 *	the various Lock() and Unlock() members. In some very rare cases
 *	it might be necessarry to control this mutex from the application.
 * \par
 
 *	One possible usage for this is to "link" two loopers together
 *	to give the impression that they are both run by a single
 *	thread.  By fetching the mutex from one looper with GetMutex()
 *	and passing it to another looper through SetMutex() (or by
 *	creating a new semaphore and give it to both the loopers) you
 *	will assure that the two looper threads will never run
 *	simultanously. This should very rarly be used in a Cosmoe
 *	application as it will cripple the concurrency you normally
 *	gain by using multiple loopers but it might sometimes be
 *	defendable. For example when porting applications from a
 *	single-threaded environment it can often be very hard to
 *	rewrite all the code to work with multiple threads. Then
 *	SetMutex() is a possible way to chicken-out of the heavy
 *	multithreading normally employed by Cosmoe.
 * \par
 *	In any case you should think twice before using this method.
 *	If you still think calling SetMutex() is a exellent idea
 *	think again.
 * \note
 *	The looper will automatically lock the looper before calling
 *	any event handlers (like os::CLooper::DispatchMessage(),
 *	os::CHandler::MessageReceived, and os::CHandler::TimerTick) and
 *	unlock it again when the event handler return. If you replace
 *	the mutex from one of these members you must make sure the
 *	new mutex is locked exactly once so it will be unlocked
 *	correctly when the method returns.
 * \par
 *	You must also make sure that you don't introduce any
 *	race conditions with other threads as a result of the swap.
 * \note
 *	The mutex object will not be deleted when the looper dies. You are
 *	responsible to get rid of it after (but only after) the looper is
 *	dead. Also note that if you lend the internal mutex from one looper
 *	to another the second looper will be left high and dry if the first
 *	looper dies and delete its internal semaphore.
 * \param pcMutex
 *	Pointer to the new mutex object or NULL to restore the default
 *	mutex
 * \sa GetMutex(), Lock(), Unlock(), os::CLocker
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CLooper::SetMutex( CLocker* pcMutex )
{
	if ( pcMutex == NULL )
	{
		m->m_pcMutex = &m->m_cDefaultMutex;
	}
	else
	{
		m->m_pcMutex = pcMutex;
	}
}


/** Get a pointer to the loopers mutex.
 * \par Description:
 *	Returns the currently active mutex. This is normally a internal os::CLocker
 *	object allocated by the looper itself but if the mutex has been replaced
 *	with SetMutex() the new mutex will be returned.
 * \return Pointer to the mutex currently used to protect the looper.
 * \sa SetMutex(), Lock(), Unlock(), os::CLocker
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CLocker* CLooper::GetMutex() const
{
	return( m->m_pcMutex );
}


/** Obtain the number of locks held on the looper.
 * \par Description:
 *	Return the nest count (number of locks held by the owner) from the
 *	internal mutex used to protect the looper.
 * \par Warning:
 *	This function is dangerous in that it is very hard to use without
 *	encountering race condition and is generally not very useful.
 *	The reason for including it is that it in some cases can be a useful
 *	help when debugging a looper.
 * \return Loopers lock nest count.
 * \sa Lock(), SafeLock(), Unlock()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

int CLooper::CountLocks() const
{
	return( m->m_pcMutex->CountLocks() );
}


/** Obtaing the owner of the looper lock.
 * \par Description:
 *	Returns the thread ID of the thread currently holding the looper lock.
 *	If the looper is not locked, -1 is returned.
 * \par Warning:
 *	This function is dangerous in that it is very hard to use without
 *	encountering race condition and are generally not very useful.
 *	The reason for including it is that it in some cases can be a useful
 *	help when debuggin a looper.
 * \return The thread ID of the lock owner or -1 if the looper is not locked.
 * \sa CountLocks(), Lock(), Unlock()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

thread_id CLooper::LockingThread() const
{
	return( m->m_pcMutex->LockingThread() );
}


/** Check if the looper lock is owned by the calling thread.
 * \par Description:
 *	Check if the looper lock is owned by the calling thread.
 * \par Warning:
 *	I can't think of any safe usage of this function, other than as a
 *	debugging aid.
 *  \return
 *  \sa Lock(), SafeLock(), Unlock()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

bool CLooper::IsLocked() const
{
	return( m->m_pcMutex->IsLocked() );
}

/** Obtain the message currently being processed.
 * \par Description:
 *	This member can be called from within os::CLooper::DispatchMessage() or the
 *	active handler's os::CHandler::MessageReceived() to learn which message is
 *	currently being processed. Given the fact that the very same message
 *	is passed to both these members you may wonder why you should ever
 *	need to call this function? The reason is that many of the classes
 *	derived from either os::CLooper or os::CHandler will convert known messages
 *	to specialized callbacks. These callbacks are normally not passed
 *	the entire message, only the most used elements. If you should find
 *	yourself in the unlucky situation of needing one of the not so much
 *	used elements this member is the solution.
 * \note
 *	The message will be automatically deleted when you return to the message
 *	loop. To avoid this you should use Looper::DetachCurrentMessage() instead.
 * \par
 *	Called outside the message loop this member will return NULL.
 *
 * \return Pointer to a message, or NULL if no message is currently being processed.
 * \sa DetachCurrentMessage(), DispatchMessage(), CHandler::MessageReceived()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CMessage* CLooper::CurrentMessage() const
{
	return( m->m_pcCurrentMessage );
}


/** Steal the current message.
 * \par Description:
 *	DetachCurrentMessage() does basically the same job as CurrentMessage()
 *	except that it detachs the message from the looper, preventing it from being
 *	automatically deleted. You are responsible of getting rid of the message
 *	when it is no longer needed.
 * \return Pointer to a message, or NULL if no message is currently being processed.
 * \sa CurrentMessage()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CMessage* CLooper::DetachCurrentMessage()
{
	CMessage* pcMsg = m->m_pcCurrentMessage;
	m->m_pcCurrentMessage = NULL;
	return( pcMsg );
}

/** The loopers message handling callback.
 * \par Description:
 *	When a message arrives, the looper extracts it from the queue and calls
 *	this member with the message as a parameter. The default implementation
 *	will attempt to pass the message on to a Handler through it's
 *	os::CHandler::HandleMessage() member. The Handler that should receive the
 *	message is selected as follows:
 * \par
 *	If the message targets a specific Handler, that Handler will receive the
 *	message. DispatchMessage() can determine wether the message had a final
 *	destination by examining the pcHandler argument. If not NULL it points at
 *	the handler targeted by the message.
 * \par
 *	If the message destination however is not fully qualified DispatchMessage()
 *	attempts to pass the message on to the Default handler (as set through the
 *	SetDefaultHandler()).
 * \par
 *	If there is no final destination and no default handler the Looper will
 *	handle the message itself by calling its own version of HandleMessage()
 *	(The looper is itself a os::CHandler)
 * \par
 *	Not all messages are passed on to a handler. If the message code is M_QUIT
 *	the CLooper::QuitRequested() member is called instead and if it return true
 *	looper object will be deleted and the looper thread terminated.
 * \par
 *	If you would like to handle certain messages directly by the looper, bypassing
 *	the normal scheduling you can overload DispatchMessage() to process messages
 *	before they are passed on to any handler. If you do so, you should call the
 *	loopers version of DispatchMessage() for each message you don't know how to
 *	handle.
 * \par
 *	Please note however that you should very rarely overload this member.
 *	It is normaly better to overload the MessageReceived() member and let
 *	the looper handle the message as any other handlers if you want to
 *	pass messages to the looper itself.
 * \note
 *	The looper is locked when DispatchMessage() is called.
 * \note
 *	Never do any lengthy operations in any hook members that are called
 *	from the looper thread if the looper is involved with the GUI (for
 *	example if the looper is a os::CWindow). The looper will not be
 *	able to dispatch messages until the hook returns so spending a
 *	long time in this member will make the GUI feel unresponsive.
 * \param pcMsg - Pointer to the received messge. Note that this message will be
 *		  deleted when DispatchMessage() returns, unless detatched from
 *		  the looper through DetachCurrentMessage().
 * \param pcHandler - Pointer to the handler targeted by this message. If the
 *		      message was not targeted at any spesific handler this
 *		      argument is NULL.
 * \sa SetDefaultHandler(), GetDefaultHandler(), PostMessage(), CurrentMessage()
 * \sa DetachCurrentMessage(), CHandler::MessageReceived()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CLooper::DispatchMessage( CMessage* pcMsg, CHandler* pcHandler )
{
	switch( pcMsg->what )
	{
		case M_QUIT:
			if ( QuitRequested() )
			{
				Unlock();
				delete this;
				exit_thread( 0 );
			}
			break;

		default:
		{
			if ( NULL != pcHandler )
			{
				pcHandler->MessageReceived( pcMsg );
			}
			else
			{
				if ( NULL != m->m_pcDefaultHandler )
				{
					m->m_pcDefaultHandler->MessageReceived( pcMsg );
				}
				else
				{
					MessageReceived( pcMsg );
				}
			}
		}
	}
}


/** Obtain the internal message queue used by the looper.
 * \par Description:
 *	You should rarely need to examine the message queue yourself but it can
 *	be useful in certain situations. For example some messages might come in such
 *	an overwelming amount and are of a nature that you can miss some of them
 *	without any harm. It is then useful to be able to look through the message
 *	queue when such a message arrives to check if there is more of them waiting
 *	and if so throw away the current message, and wait for the next to crawl its
 *	way down the queue. One example of such messages are the M_MOUSE_MOVED message.
 *	You will probably call SpoolMessages() before calling this message (but after
 *	locking the looper) to make sure the queue contains all messages sent to the
 *	looper.
 * \return A pointer to the internal MessageQueue object used by the looper.
 * \sa DispatchMessage(), PostMessage()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CMessageQueue* CLooper::MessageQueue() const
{
	return( m->m_pcMsgQueue );
}


/** Get the internal handler list
 * \par Description:
 *	You can use this member to gain read-only access to the internal stl
 *	map of handlers. The map key is an unique ID that is assigned to each
 *	handler in the system (unique accross process boundaries). This ID
 *	should be considered a implemntation detail and are of no value outside
 *	the looper/handler implemntation itself. You can still traverse the map
 *	to examine the handlers that currently populate the looper. You should
 *	always keep the looper locked while examining the map.
 * \return const reference to the internal stl map of handlers.
 * \sa FindHandler(), AddHandler(), RemoveHandler()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

const CLooper::handler_map& CLooper::GetHandlerMap() const
{
	return( m->m_cHandlerMap );
}


/** Add a handler to the looper.
 * \par Description:
 *	Adds a handler to the looper. In order to receive messages and before
 *	it can be set as a default handler the handler must be added to a looper.
 *	The handler must not be member of any looper when added.
 * \param pcHandler - Pointer to the handler to be added.
 * \sa RemoveHandler(), SetDefaultHandler()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CLooper::AddHandler( CHandler* pcHandler )
{
	CAutolock _lock_( m->m_pcMutex );
	
	if ( NULL == pcHandler->m_pcLooper )
	{
		m->m_cHandlerMap[pcHandler->m_nToken] = pcHandler;
		pcHandler->m_pcLooper = this;
	}
	else
	{
		dbprintf( "Warning: Attempt to add handler %d twice\n", pcHandler->m_nToken );
	}
}

/** Remove a handler previously added by AddHandler()
 * \par Description:
 *	Removes a handler from the looper. If the default handler is removed
 *	the default looper pointer as returned by GetDefaultHandler() will be
 *	set to NULL.
 * \param pcHandler
 *	The handler to remove
 * \return True if the Handler in fact was a member of the looper, false otherwise.
 * \sa AddHandler(), SetDefaultHandler(), GetDefaultHandler()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

bool CLooper::RemoveHandler( CHandler* pcHandler )
{
	if ( this != pcHandler->m_pcLooper )
	{
		dbprintf( "Error : Attempt to remove handler not belonging to this looper\n" );
		return( false );
	}

	handler_map::iterator i;

	CAutolock _lock_( m->m_pcMutex );

	if ( pcHandler == m->m_pcDefaultHandler )
	{
		m->m_pcDefaultHandler = NULL;
	}

	i = m->m_cHandlerMap.find( pcHandler->m_nToken );

	if ( i == m->m_cHandlerMap.end() )
	{
		dbprintf( "Error: Looper::RemoveHandler() could not find handler %d in map!\n", pcHandler->m_nToken );
		return( false );
	}
	else
	{
		m->m_cHandlerMap.erase( i );
		pcHandler->m_pcLooper = NULL;
		return( true );
	}
	return( false );
}


/** Search the looper for a named handler
 * \par Description:
 *	Return the handler named <pzName>, or NULL if there is no such handler
 *	added to the looper.
 * \param pzName
 *	The name to search for.
 * \return Pointer to the handler if one was found, NULL otherwise
 * \sa GetHandler(), GetHandlerCount(), GetHandlerIndex()
 * \sa AddHandler(), RemoveHandler()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CHandler* CLooper::FindHandler( const std::string& cName ) const
{
	handler_map::const_iterator i;

	CAutolock _lock_( m->m_pcMutex );
	
	for ( i = m->m_cHandlerMap.begin() ; i != m->m_cHandlerMap.end() ; ++i )
	{
		CHandler* pcHandler = (*i).second;
		if ( pcHandler->mName == cName )
		{
			return( pcHandler );
		}
	}

	dbprintf( "Error: Looper::FindHandler() Unable to find handler %s\n", cName.c_str() );
	return( NULL );
}


/** Set the default target for incomming messages.
 * \par Description:
 *	Call this method to assign a handler as the default handler for the looper.
 *      When a handler become the default handler it will receive all messages
 *	that is not target directly at another handler.
 *	You can remove the current default handler by passing in a NULL pointer.
 * \note
 *	The handler must already be added to this looper with AddHandler()
 *	before it can be made the default handler of the looper.
 * \param pcHandler
 *	The handler to set as default, or NULL
 * \sa GetDefaultHandler(), DispatchMessage(), Handler::HandleMessage()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CLooper::SetPreferredHandler( CHandler* pcHandler )
{
	CAutolock _lock_( m->m_pcMutex );
	
	if ( NULL == pcHandler || pcHandler->m_pcLooper == this )
	{
		m->m_pcDefaultHandler        = pcHandler;
	}
}


/** Obtain the default handler for the looper.
 * \par Description:
 *	Return the handler last set by SetPreferredHandler() or NULL if there
 *	currently is no default handler
 * \return Pointer to the default handler.
 * \sa SetPreferredHandler()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CHandler* CLooper::PreferredHandler() const
{
	return( m->m_pcDefaultHandler );
}


/** Obtain the count of handlers added to this looper
 * \return Number of handlers currently attached to the looper
 * \sa GetHandler(), AddHandler(), RemoveHandler()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

int32 CLooper::CountHandlers() const
{
	return( m->m_cHandlerMap.size() );
}

/** Deliver a message to the looper.
 * \par Description:
 *	The message is posted to the looper with no absolute destination. This
 *	means that it will be handled by the default handler or the looper
 *	itself if there currently is no default handler.
 * \par Note:
 *	The caller is responsible for deleting the message (It is not deleted
 *	or kept by the looper)
 * \param pcMsg
 *	The message to post.
 * \return 0 if everyting went ok. If someting went wrong, a negative number is
 *	   retuned, and the errno variable is set.
 * \sa DispatchMessage()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

status_t CLooper::PostMessage( CMessage* pcMsg )
{
	status_t nError;
	nError = pcMsg->Post( GetMsgPort(), ~0 );
	if ( nError < 0 )
	{
		dbprintf( "Error: Looper::PostMessage:1() failed to send message\n" );
	}
	return( nError );
}


/** Deliver a message to the looper.
 * \par Description:
 *	Construct a message with the given code, and posts it to the looper
 *	by calling PostMessage( CMessage* pcMsg )
 * \param nCode
 *	The code that should be assigned to the sent message
 * \return 0 if everyting went ok. If someting went wrong, a negative number is
 *	   retuned, and the errno variable is set.
 * \sa PostMessage( CMessage* pcMsg ),DispatchMessage()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

status_t CLooper::PostMessage( uint32 nCode )
{
	CMessage cMsg( nCode );
	return( PostMessage( &cMsg ) );
}


/** Deliver a message to the looper.
 * \par Description:
 *	Post a message to a specific handler. The message is targeted to the
 *	handler pointed at by pcHandler, and the reply is set to the handler
 *	pointed at by pcReplyTo.
 * \par Note:
 *	The caller is responsible for deleting the message (It is not deleted
 *	or kept by the looper)
 * \par
 *	You may not pass NULL as the target, but you can send a pointer to the
 *	looper itself to force it to handle the message.
 *
 * \param nCode
 *	The message that should be posted to the message queue.
 * \param pcHandler
 *	Must point to a valid handler that will receive the message.
 * \param pcReplyTo
 *	If not NULL should point to the handler that should receive
 *	the reply sent to the message.
 *
 * \return 0 if everyting went ok. If someting went wrong, a negative number is
 *	   retuned, and the errno variable is set.
 * \sa PostMessage( CMessage* pcMsg ), DispatchMessage()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

status_t CLooper::PostMessage( CMessage *pcMsg, CHandler *pcHandler, CHandler* pcReplyHandler )
{
	CLooper* pcLooper    = NULL;
	port_id  hReplyPort  = -1;
	int      nReplyToken = -1;

	if ( NULL != pcReplyHandler )
	{
		pcLooper = pcReplyHandler->Looper();
		if ( NULL != pcLooper )
		{
			hReplyPort  = pcLooper->GetMsgPort();
			nReplyToken = pcReplyHandler->m_nToken;
		}
	}
	status_t nError = pcMsg->Post( GetMsgPort(), pcHandler->m_nToken, hReplyPort, nReplyToken );

	if ( nError < 0 )
	{
		dbprintf( "Looper::PostMessage:2() failed to send message\n" );
	}

	return( nError );
}


/** Deliver a message to the looper.
 * \par Description:
 *	Construct a message with the given code, and posts it to a specific handler
 *	by calling PostMessage( Message *pcMsg, Handler *pcHandler, Handler* pcReplyHandler )
 *
 * \param nCode
 *	The code that should be assigned to the sent message
 * \param pcHandler
 *	Must point to a valid handler that will receive the message.
 * \param pcReplyTo
 *	If not NULL should point to the handler that should receive
 *	the reply sent to the message.
 *
 * \return 0 if everyting went ok. If someting went wrong, a negative number is
 *	   retuned, and the errno variable is set.
 * \sa PostMessage( CMessage *pcMsg, CHandler *pcHandler, CHandler* pcReplyHandler )
 * \sa DispatchMessage()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

status_t CLooper::PostMessage( uint32 nCode, CHandler* pcHandler, CHandler* pcReplyTo )
{
	CMessage cMsg( nCode );
	return( PostMessage( &cMsg, pcHandler, pcReplyTo ) );
}


/** Check if it is ok to break the loop.
 * \par Description:
 *	You can overload this function to affect how the looper will react to
 *	M_QUIT messages. When an M_QUIT message arrive the looper will call
 *	this function to figure out what to do. If it returns false the message
 *	is ignored. If it returns true the loop is terminated, the looper object
 *	deleted and the looper thread will exit.
 *
 * \return true if it is ok to terminate the looper, false otherwise.
 * \sa Quit(), PostMessage()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

bool CLooper::QuitRequested()
{
	return( true );
}

/** Unconditionally terminate the looper.
 * \par Description:
 *	Calling quit will unconditionally terminates the looper (As opposed to
 *	post an M_QUIT message that will politly call OkToQuit() before terminating
 *	the looper). If the call is made by the looper thread itself this member
 *	will delete the looper object, and call exit_thread(). If an external thread
 *	made the call, a M_TERMINATE message is sent, and the caller thread is
 *	blocked until the looper thread is dead. Either way, the looper object
 *	gets deleted, and the looper thread dies.
 * \sa QuitRequested(), PostMessage()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CLooper::Quit()
{
	if ( get_thread_id( NULL ) == m->m_hThread )
	{
		delete this;
		exit_thread( 0 );
	}
	else
	{
		CMessage cMsg( M_TERMINATE );

		PostMessage( &cMsg );

		while( wait_for_thread( m->m_hThread ) < 0)
		{
			if ( errno != EINTR )
			{
				break;
			}
		}
	}
}


/** Drain the low-level message port.
 * \par Description:
 *	SpoolMessage() will fetch all messages from the low-level message port
 *	unflatten the message objects, and add the os::CMessage objects to the
 *	internal message queue
 * \sa MessageQueue(), PostMessage()
 * \author Kurt Skauen (kurt@atheos.cx)
 * \internal Should have used the same temporary message buffer as the main loop!!
 *****************************************************************************/

void CLooper::SpoolMessages()
{
	uint8* pBuffer = new uint8[8192];
	uint32 nCode;

	while( get_msg_x( m->m_hPort, &nCode, pBuffer, 8192, 0 ) >= 0 )
	{
		_DecodeMessage( nCode, pBuffer );
	}
	delete[] pBuffer;
}


/** Start and unlock the looper.
 * \par Description:
 *	As mention in the description of the constructor, no thread are spawned
 *	there and the looper is locked. Calling Run() will spawn a new thread
 *	to run the message loop, and unlock the looper.
 * \par Note:
 *	Not all loopers will spawn a new thread. F.eks. the Application class
 *	will overload the Run() member and lead the calling thread into the
 *	message loop. This means that when calling Run() on an application object
 *	it will not return until the loop terminates.
 * \return The thread ID of the looper thread.
 * \sa Quit(), QuitRequested(), PostMessage(), Lock(), Unlock()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

thread_id CLooper::Run()
{
	m->m_hThread = spawn_thread( m->m_cName.c_str(), (void*)_Entry, m->m_nPriority, 0, this );
	if ( m->m_hThread != -1 )
	{
		resume_thread( m->m_hThread );
		g_cLopperListMutex.Lock();
		g_cLooperThreadMap[m->m_hThread] = this;
		g_cLopperListMutex.Unlock();
	}
	Unlock();
	return( m->m_hThread );
}


/** Wait for the looper thread to die
 * \par Description:
 *	This member can be called from an external thread to wait for the
 *	looper thread to die.
 * \note
 *	If this member return successfully it means that the looper is dead
 *	and no further members should be called on the looper object.
 * \return On success 0 is returned. On failure a negative value is returned
 *	and a error code is written to the global variable \b errno.
 * \sa Run(), Quit()
 * \author Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

int CLooper::Wait() const
{
	if ( m->m_hThread < 0 )
	{
		errno = EINVAL;
		return( -1 );
	}
	else
	{
		return( wait_for_thread( m->m_hThread ) );
	}
}


/** Hook called each time the message queue is drained.
 * \par Description:
 *	Normally the looper thread is blocked on the internal message
 *	queue waiting for messages. When one or more messages
 *	arrive it will wake up and process the messages. When all
 *	messages is processed it will first call Idle() and then
 *	go back looking for more messages. If Idle() returned \b false
 *	the thread will block until the next message arrive and if
 *	Idle() returned \b true it will just look for new messages
 *	without blocking and call Idle() again (after processing new
 *	messages if any).
 * \par Note:
 *	The looper will not be able to process new messages until
 *	Idle() returns so it is very important to not do anything
 *	lengthy if the looper is part of for example an os::CWindow
 *	since that will make the application feel unresponsive
 *	to the user.
 *
 * \return
 *	The Idle() function should return \b true if it want to be
 *	called again immediatly after the looper has polled for new
 *	messages, or \b false if it don't want to be called until
 *	at least one more message has been processed.
 * \sa MessageReceived(), DispatchMessage()
 * \author	Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

bool CLooper::Idle()
{
	return( false );
}


/** Called by the looper thread before entering the message loop.
 * \par Description:
 *	This hook can be overloaded to do one-shot initialization
 *	of the looper that can not be done in the contructor.
 *	Started() is called from the loopers thread just before
 *	it start processing messages.
 * \sa Idle(), Run()
 * \author	Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CLooper::Started()
{
}


void CLooper::AddCommonFilter( CMessageFilter* pcFilter )
{
	if ( pcFilter == NULL )
	{
		dbprintf( "Error: Looper::AddCommonFilter() called with a NULL pointer\n" );
		return;
	}

	if ( pcFilter->mHandler != NULL )
	{
		if ( pcFilter->mGlobalFilter )
		{
			static_cast<CLooper*>(pcFilter->mHandler)->RemoveCommonFilter( pcFilter );
		}
		else
		{
			pcFilter->mHandler->RemoveFilter( pcFilter );
		}
	}
	pcFilter->mIterator     = m->m_cCommonFilterList.insert( m->m_cCommonFilterList.end(), pcFilter );
	pcFilter->mHandler      = this;
	pcFilter->mGlobalFilter = true;
}


void CLooper::RemoveCommonFilter( CMessageFilter* pcFilter )
{
    if ( pcFilter->mHandler != this )
	{
        dbprintf( "Error: Looper::RemoveCommonFilter() attempt to remove filter not belonging to us\n" );
        return;
    }

    if ( pcFilter->mGlobalFilter )
	{
        dbprintf( "Error: Looper::RemoveCommonFilter() attempt to remove a local filter using RemoveCommonFilter()\n" );
        return;
    }

    m->m_cCommonFilterList.erase( pcFilter->mIterator );
    pcFilter->mHandler = NULL;
}


const MsgFilterList& CLooper::CommonFilterList() const
{
    return( m->m_cCommonFilterList );
}

/******************************************************************************/
/****** Private functions *****************************************************/
/******************************************************************************/

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CLooper::_AddLooper( CLooper* pcLooper )
{
	g_cLopperListMutex.Lock();
	g_cLooperPtrMap[pcLooper] = pcLooper->m->m_nID;
	g_cLooperPortMap[pcLooper->m->m_hPort] = pcLooper;
//    pcLooper->m->m_pcNextLooper = s_pcFirstLooper;
//    s_pcFirstLooper = pcLooper;
	g_cLopperListMutex.Unlock();
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CLooper::_RemoveLooper( CLooper* pcLooper )
{
//    CLooper** ppcPtr;

	g_cLopperListMutex.Lock();

	std::map<CLooper*,int>::iterator i = g_cLooperPtrMap.find( pcLooper );
	if ( i != g_cLooperPtrMap.end() )
	{
		g_cLooperPtrMap.erase( i );
	}

	std::map<port_id,CLooper*>::iterator j = g_cLooperPortMap.find( pcLooper->m->m_hPort );
	if ( j != g_cLooperPortMap.end() )
	{
		g_cLooperPortMap.erase( j );
	}

	if ( pcLooper->m->m_hThread != -1 ) {
		std::map<thread_id,CLooper*>::iterator k = g_cLooperThreadMap.find( pcLooper->m->m_hThread );
		if ( k != g_cLooperThreadMap.end() )
		{
			g_cLooperThreadMap.erase( k );
		}
	}


/*    for ( ppcPtr = &s_pcFirstLooper ; NULL != *ppcPtr ; ppcPtr = &((*ppcPtr)->m->m_pcNextLooper) )
	{
		if ( *ppcPtr == pcLooper ) {
			*ppcPtr = pcLooper->m->m_pcNextLooper;
			break;
		}
	}*/
	g_cLopperListMutex.Unlock();
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool CLooper::_IsLooperValid( CLooper* pcLooper )
{
	bool bValid = false;

	g_cLopperListMutex.Lock();

	std::map<CLooper*,int>::iterator i = g_cLooperPtrMap.find( pcLooper );
	if ( i != g_cLooperPtrMap.end() && (*i).second == pcLooper->m->m_nID )
	{
		bValid = true;
	}
	g_cLopperListMutex.Unlock();

	return( bValid );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CLooper* CLooper::_GetLooperForThread( thread_id hThread )
{
	CLooper* pcLooper = NULL;

	g_cLopperListMutex.Lock();

	std::map<thread_id,CLooper*>::iterator i = g_cLooperThreadMap.find( hThread );
	if ( i != g_cLooperThreadMap.end() )
	{
		pcLooper = (*i).second;
	}
	g_cLopperListMutex.Unlock();
	return( pcLooper );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CLooper* CLooper::_GetLooperForPort( port_id hPort )
{
	CLooper* pcLooper = NULL;

	g_cLopperListMutex.Lock();
	std::map<port_id,CLooper*>::iterator i = g_cLooperPortMap.find( hPort );
	if ( i != g_cLooperPortMap.end() )
	{
		pcLooper = (*i).second;
	}
	g_cLopperListMutex.Unlock();
	return( pcLooper );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CHandler* CLooper::_FindHandler( int nToken ) const
{
	handler_map::const_iterator i;

	i = m->m_cHandlerMap.find( nToken );

	if ( i == m->m_cHandlerMap.end() )
	{
		dbprintf( "Error: Looper::_FindHandler() could not find handler %d\n", nToken );
		return( NULL );
	}

	return( (*i).second );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CLooper::_InsertTimer( TimerNode* pcTimer )
{
	TimerNode** ppcTmp;
	bool bWakeup = m->m_pcFirstTimer == NULL || pcTimer->m_nTimeout < m->m_pcFirstTimer->m_nTimeout;

	for ( ppcTmp = &m->m_pcFirstTimer ; NULL != *ppcTmp ; ppcTmp = &(*ppcTmp)->m_pcNext )
	{
		if ( pcTimer->m_nTimeout < (*ppcTmp)->m_nTimeout )
		{
			pcTimer->m_pcNext = *ppcTmp;
			*ppcTmp = pcTimer;
			goto done;
		}
	}
	*ppcTmp = pcTimer;
	pcTimer->m_pcNext = NULL;
done:
	if ( bWakeup )
	{
		resume_thread( m->m_hThread );
	}
}


void CLooper::_DecodeMessage( uint32 nCode, const void* pData )
{
	try {
		CMessage* pcMsg;
		if ( nCode != M_NODE_MONITOR )
		{
			pcMsg = new CMessage( pData, true );
		}
		else
		{
			pcMsg = new CMessage( M_NODE_MONITOR );
			const NodeWatchEvent_s* psEvent = static_cast<const NodeWatchEvent_s*>(pData);

			pcMsg->AddInt32( "event", psEvent->nw_nEvent );
			pcMsg->AddInt32( "device", psEvent->nw_nDevice );
			pcMsg->AddInt64( "node", psEvent->nw_nNode );
			pcMsg->m_nTargetToken = int(psEvent->nw_pUserData);

			switch( psEvent->nw_nEvent )
			{
				case NWEVENT_CREATED:
				case NWEVENT_DELETED:
					pcMsg->AddInt64( "dir_node", psEvent->nw_nOldDir );
					pcMsg->AddString( "name", std::string( NWE_NAME(psEvent), psEvent->nw_nNameLen ) );
					break;
				case NWEVENT_MOVED:
					pcMsg->AddInt64( "old_dir", psEvent->nw_nOldDir );
					pcMsg->AddInt64( "new_dir", psEvent->nw_nNewDir );
					pcMsg->AddString( "name", std::string( NWE_NAME(psEvent), psEvent->nw_nNameLen ) );
					if ( psEvent->nw_nPathLen > 0 ) {
						if ( psEvent->nw_nWhichPath == NWPATH_NEW )
						{
							pcMsg->AddString( "new_path", std::string( NWE_PATH(psEvent), psEvent->nw_nPathLen ) );
						}
						else
						{
							pcMsg->AddString( "old_path", std::string( NWE_PATH(psEvent), psEvent->nw_nPathLen ) );
						}
					}
					break;
				case NWEVENT_STAT_CHANGED:
					pcMsg->AddString( "name", std::string( NWE_NAME(psEvent), psEvent->nw_nNameLen ) );
					break;
				case NWEVENT_ATTR_WRITTEN:
				case NWEVENT_ATTR_DELETED:
					pcMsg->AddString( "attr_name", std::string( NWE_NAME(psEvent), psEvent->nw_nNameLen ) );
					break;
				case NWEVENT_FS_MOUNTED:
				case NWEVENT_FS_UNMOUNTED:
					break;
			}
		}
		if ( m->m_pcMsgQueue->Lock() )
		{
			m->m_pcMsgQueue->AddMessage( pcMsg );
			m->m_pcMsgQueue->Unlock();
		}
		else
		{
			delete pcMsg;
		}
	}
	catch ( std::exception& )
	{
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CLooper::_Loop()
{
	void* pBuffer;

	pBuffer = malloc( 8192 );

	Lock();
	Started();
	bool bDoBlock = !Idle();
	Unlock();
	for (;;)
	{
		bigtime_t nTimeOut;

		if ( bDoBlock )
		{
			Lock();
			if ( m->m_pcFirstTimer == NULL )
			{
				nTimeOut = 500000;
			}
			else
			{
				bigtime_t nCurTime = system_time();
				if ( nCurTime < m->m_pcFirstTimer->m_nTimeout )
				{
					nTimeOut = m->m_pcFirstTimer->m_nTimeout - nCurTime;
				}
				else
				{
					nTimeOut = 0;
				}
			}
			Unlock();
		}
		else
		{
			nTimeOut = 0;
		}
		uint32 nCode;
		if( get_msg_x( m->m_hPort, &nCode, pBuffer, 8192, nTimeOut ) >= 0 )
		{
			_DecodeMessage( nCode, pBuffer );
		}

		Lock();
		bigtime_t nCurTime = system_time();

		TimerNode* pcFirstRestarted = NULL;

		while( m->m_pcFirstTimer != NULL && nCurTime >= m->m_pcFirstTimer->m_nTimeout )
		{
			TimerNode* pcTimer = m->m_pcFirstTimer;
			m->m_pcFirstTimer = pcTimer->m_pcNext;

			CHandler* pcTarget = pcTimer->m_pcHandler;
			int      nID      = pcTimer->m_nID;

			if ( pcTimer->m_bOneShot )
			{
				delete pcTimer;
			}
			else
			{
				pcTimer->m_nTimeout = nCurTime + pcTimer->m_nPeriod;
				pcTimer->m_pcNext = pcFirstRestarted;
				pcFirstRestarted = pcTimer;
			}
			pcTarget->TimerTick( nID );
		}

		while( pcFirstRestarted != NULL )
		{
			TimerNode* pcTimer = pcFirstRestarted;
			pcFirstRestarted = pcTimer->m_pcNext;
			_InsertTimer( pcTimer );
		}
		Unlock();

		while( get_msg_x( m->m_hPort, &nCode, pBuffer, 8192, 0 ) >= 0 )
		{
			_DecodeMessage( nCode, pBuffer );
		}

		for (;;)
		{
			if ( m->m_pcMsgQueue->Lock() == false )
			{
				dbprintf( "Error: Looper::_Loop() failed to lock message queue\n" );
				break;
			}
			m->m_pcCurrentMessage = m->m_pcMsgQueue->NextMessage();
			m->m_pcMsgQueue->Unlock();

			if ( m->m_pcCurrentMessage == NULL )
			{
				break;
			}

			if ( m->m_pcCurrentMessage->what == M_TERMINATE )
			{
				free( pBuffer );
				return;
			}

			if ( m->m_pcCurrentMessage->what == M_QUIT )
			{
				if ( QuitRequested() )
				{
					free( pBuffer );
					return;
				}
				else
				{
					continue;
				}
			}

			CHandler* pcTarget;

			if ( -1 != m->m_pcCurrentMessage->m_nTargetToken )
			{
				pcTarget = _FindHandler( m->m_pcCurrentMessage->m_nTargetToken );
			}
			else
			{
				pcTarget = NULL;
			}
			Lock();
			try {
				DispatchMessage( m->m_pcCurrentMessage, pcTarget );
			}
			catch (...)
			{
				dbprintf( "Error: Looper::_Loop() uncaught exception in Looper::DispatchMessage()\n" );
			}
			Unlock();
			delete m->m_pcCurrentMessage;
			m->m_pcCurrentMessage = NULL;
		}
		Lock();
		try
		{
			bDoBlock = !Idle();
		} catch (...) {
			dbprintf( "Error: Looper::_Loop() uncaught exception in Looper::Idle()\n" );
		}
		Unlock();
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

uint32        CLooper::_Entry( void* pData )
{
	CLooper*        pcThis = (CLooper*) pData;

//        signal( SIGHUP, SIG_IGN );
//        signal( SIGINT, SIG_IGN );
//        signal( SIGQUIT, SIG_IGN );
	signal( SIGALRM, SIG_IGN );
	signal( SIGCHLD, SIG_IGN );
//        signal( SIGINT, SIG_IGN );
//        signal( SIGINT, SIG_IGN );

	pcThis->_Loop();

	delete pcThis;

	return( 0 );
}


void CLooper::FBC_Looper1() {}
void CLooper::FBC_Looper2() {}
void CLooper::FBC_Looper3() {}
void CLooper::FBC_Looper4() {}
void CLooper::FBC_Looper5() {}
void CLooper::FBC_Looper6() {}
void CLooper::FBC_Looper7() {}
void CLooper::FBC_Looper8() {}

