/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *  Portions Copyright (C) 2001 Tom Marshall
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */


#include <unistd.h>
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/shm.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>

#include <cosmoe/Debug.h>
#include <support/SupportDefs.h>
#include <storage/StorageDefs.h>	// Just because BeOS apps expect this here
#include <cosmoe/OS.h>

#ifndef min
#define min(a,b) ( ((a)<(b)) ? (a) : (b) )
#endif



/*
 * We put the area functions here because they need access to the framebuffer
 * info.
 */

AreaInfo_s* g_pAreaMap = NULL;

#ifndef _NO_FB_
extern int fb_fd;
extern struct fb_fix_screeninfo fb_fixsi;
extern struct fb_var_screeninfo fb_varsi;
extern size_t fb_len;
extern void* fb_ptr;
#endif


typedef void* (*pfn_pthread_entry)(void*);


void init_area_map( void )
{
	int  shmid;
	uint n;

	shmid = shmget( AREA_ID_MAX, sizeof(AreaInfo_s)*AREA_ID_MAX, IPC_CREAT|0700 );
	if( shmid == -1 )
	{
		dbprintf( "linux_compat: init_area_map failed in shmget: %s\n", strerror(errno) );
		return;
	}

	g_pAreaMap = shmat( shmid, NULL, 0 );
	if( g_pAreaMap == (void*)(-1) )
	{
		dbprintf( "linux_compat: init_area_map failed in shmat: %s\n", strerror(errno) );
		return;
	}

	for( n = 0; n < AREA_ID_MAX; n++ )
	{
		g_pAreaMap[n].area = AREA_ID_FREE;
	}
}


//XXX: need to sync access to area map with mutex
area_id create_area( const char* pzName, void** ppAddress, size_t nSize, uint32 nProtection, int nLock )
{
	uint n;

	/* check for vesa framebuffer: this is already mmap()'d */
	if( strcmp( pzName, "vesa_io" ) == 0 )
	{
		if( fb_ptr == NULL )
		{
			/* the framebuffer is not yet mapped or mapping failed */
			return -1;
		}
		if( ppAddress != NULL )
		{
			*ppAddress = fb_ptr;
		}
		return AREA_ID_FB;
	}

	/* find a suitable area */
	if( g_pAreaMap == NULL )
	{
		int shmid = shmget( AREA_ID_MAX, sizeof(AreaInfo_s)*AREA_ID_MAX, 0700 );
		if( shmid == -1 )
		{
			dbprintf( "linux_compat: cannot get area map: %s\n", strerror(errno) );
			return -ENOMEM;
		}
		g_pAreaMap = shmat( shmid, NULL, 0 );
		if( g_pAreaMap == (void*)(-1) )
		{
			dbprintf( "linux_compat: cannot attach to area map: %s\n", strerror(errno) );
			g_pAreaMap = NULL;
			return -ENOMEM;
		}
	}

	for( n = 0; n < AREA_ID_MAX; n++ )
	{
		if( g_pAreaMap[n].area == AREA_ID_FREE )
		{
			int iShmID = shmget( n, nSize, IPC_CREAT|0700 );
			if( iShmID == -1 )
			{
				dbprintf( "linux_compat: shmget(%u,%u) failed (%s)\n", n, nSize, strerror(errno) );
				return -ENOMEM;
			}
			g_pAreaMap[n].address = shmat( iShmID, NULL, 0 );
			if( g_pAreaMap[n].address == (void*)(-1) )
			{
				dbprintf( "linux_compat: shmat(%d) failed (%s)\n", iShmID, strerror(errno) );
				return -ENOMEM;
			}
			if( ppAddress != NULL )
			{
				*ppAddress = g_pAreaMap[n].address;
			}
			strncpy( g_pAreaMap[n].name, pzName, 63 );
			g_pAreaMap[n].name[63] = '\0';
			g_pAreaMap[n].area = iShmID;
			g_pAreaMap[n].size = nSize;
			g_pAreaMap[n].lock = nLock;
			g_pAreaMap[n].protection = nProtection;
			g_pAreaMap[n].process = getpid();
			g_pAreaMap[n].allocSize = nSize;
			return n;
		}
	}
	return -ENOMEM;
}


status_t remap_area( area_id nArea, void* pPhysAddress )
{
	dbprintf( "linux_compat: UNIMPLEMENTED: remap_area( %d, ... )\n", nArea );

	if( nArea != AREA_ID_FB )
	{
		return -EPERM;
	}
	return -1;
}


area_id clone_area( const char* pzName, void** ppAddress, uint32 nProtection, int nLock, area_id hSrcArea )
{
	uint n;

	if( g_pAreaMap == NULL )
	{
		int shmid = shmget( AREA_ID_MAX, sizeof(AreaInfo_s)*AREA_ID_MAX, 0700 );
		if( shmid == -1 )
		{
			dbprintf( "linux_compat: cannot get area map: %s\n", strerror(errno) );
			return -ENOMEM;
		}
		g_pAreaMap = shmat( shmid, NULL, 0 );
		if( g_pAreaMap == (void*)(-1) )
		{
			dbprintf( "linux_compat: cannot attach to area map: %s\n", strerror(errno) );
			g_pAreaMap = NULL;
			return -ENOMEM;
		}
	}


	if( hSrcArea < 0 || hSrcArea >= AREA_ID_MAX || g_pAreaMap == NULL ||
		g_pAreaMap[hSrcArea].area == AREA_ID_FREE )
	{
		dbprintf( "linux_compat: clone_area: AREA IS FREE\n" );
		return -EPERM;
	}

	for( n = 0; n < AREA_ID_MAX; n++ )
	{
		size_t nSize = g_pAreaMap[hSrcArea].size;
		uint32 nProtection = g_pAreaMap[hSrcArea].protection;

		if( g_pAreaMap[n].area == AREA_ID_FREE )
		{
			int iShmID = shmget( hSrcArea, nSize, IPC_CREAT|0700 );
			if( iShmID == -1 )
			{
				dbprintf( "linux_compat: shmget(%u,%u) failed (%s)\n", n, nSize, strerror(errno) );
				return -ENOMEM;
			}
			g_pAreaMap[n].address = shmat( iShmID, NULL, 0 );
			if( g_pAreaMap[n].address == (void*)(-1) )
			{
				dbprintf( "linux_compat: shmat(%d) failed (%s)\n", iShmID, strerror(errno) );
				return -ENOMEM;
			}
			if( ppAddress != NULL )
			{
				*ppAddress = g_pAreaMap[n].address;
			}
			strncpy( g_pAreaMap[n].name, pzName, 63 );
			g_pAreaMap[n].name[63] = '\0';
			g_pAreaMap[n].area = iShmID;
			g_pAreaMap[n].size = nSize;
			g_pAreaMap[n].lock = nLock;
			g_pAreaMap[n].protection = nProtection;
			g_pAreaMap[n].process = getpid();
			g_pAreaMap[n].allocSize = nSize;
			return n;
		}
	}

	return -ENOMEM;
}


status_t delete_area( area_id hArea )
{
#ifndef _NO_FB_
	if( hArea == AREA_ID_FB )
	{
		if( fb_fd == -1 || fb_ptr == NULL )
		{
			return ENXIO;
		}
		munmap( fb_ptr, fb_len );
		close( fb_fd );
		return 0;
	}
#endif

	if( hArea < 0 || hArea >= AREA_ID_MAX || g_pAreaMap == NULL ||
		g_pAreaMap[hArea].area == AREA_ID_FREE )
	{
		return -ENXIO;
	}
	/* XXX: refcount the shmseg and shmdt() when zero */
	g_pAreaMap[hArea].area = AREA_ID_FREE;

	return 0;
}


status_t get_area_info( area_id hArea, AreaInfo_s* psInfo )
{
#ifndef _NO_FB_
	if( hArea == AREA_ID_FB )
	{
		if( fb_ptr == NULL )
		{
			return ENXIO;
		}
		strcpy( psInfo->name, "vesa_io" );
		psInfo->area = AREA_ID_FB;
		psInfo->size = fb_len;
		psInfo->lock = 1;
		psInfo->protection = AREA_FULL_ACCESS;
		psInfo->process = get_process_id(NULL);
		psInfo->allocSize = fb_len;
		psInfo->address = fb_ptr;
		return 0;
	}
#endif

	if( hArea < 0 || hArea >= AREA_ID_MAX || g_pAreaMap == NULL ||
		g_pAreaMap[hArea].area == AREA_ID_FREE )
	{
		return -ENXIO;
	}
	*psInfo = g_pAreaMap[hArea];
	return 0;
}


/*** threads and processes ***/

thread_id get_thread_id( const char* const pzName )
{
	/* cosmoe/kernel.h always returns the current thread_id and ignores pzName */
	return (thread_id)pthread_self();
}


thread_id spawn_thread( const char* const pzName, void* const pfEntry, const int nPri, int nStackSize, void* const pData )
{
	pthread_t tid;

	if( 0 != pthread_create( &tid, NULL, (pfn_pthread_entry)pfEntry, pData ) )
	{
		tid = 0;
	}

	return (thread_id)tid;
}


status_t resume_thread( const thread_id hThread )
{
	pthread_kill( hThread, SIGCONT );

	return 0;
}


status_t rename_thread( thread_id hThread, const char* const pzName )
{
	return 0;
}


int wait_for_thread( const thread_id hThread )
{
	void* pvoid;

	pthread_join( hThread, &pvoid );

	return 0;
}


int exit_thread( int nCode )
{
	pthread_exit( (void*)&nCode );
	return 0;
}


int set_thread_priority( const thread_id hThread, const int nPriority )
{
	return 0;
}


proc_id get_thread_proc( const thread_id hThread )
{
	return 0;
}


proc_id get_process_id( const char* pzName )
{
	if( pzName == NULL )
	{
		return (proc_id)getpid();
	}
dbprintf( "linux_compat: UNIMPLEMENTED: get_process_id( %s )\n", pzName );
	return 0;
}

/* cosmoe semaphore info struct */
typedef struct
{
    char            name[64];
    pthread_mutex_t mutex;
    pthread_cond_t  cond;
/*    proc_id         owner; */
    int32           count;
    thread_id       holder;
    uint32          flags;
} SemInfo_s;

#define SEM_RECURSIVE 0x0001


status_t lock_mutex( sem_id hSema, bool bIgnoreSignals )
{
	return acquire_sem( hSema );
}


status_t unlock_mutex( sem_id hSema )
{
	return release_sem( hSema );
}


sem_id create_sem( int32 inCount, const char* inName )
{
	return create_sem_etc(inCount, inName, 0);
}


sem_id create_sem_etc( int32 nCount, const char* pzName, uint32 nFlags )
{
	SemInfo_s* semptr = (SemInfo_s*)malloc( sizeof(SemInfo_s) );
	if( semptr != NULL )
	{
		strncpy( semptr->name, pzName, 64 );
		semptr->name[63] = '\0';
		pthread_mutex_init( &semptr->mutex, NULL );
		pthread_cond_init( &semptr->cond, NULL );
		semptr->count = nCount;
		semptr->holder = (thread_id)pthread_self();
		semptr->flags = nFlags;
	}
	return (sem_id)semptr;
}


status_t delete_sem( sem_id hSema )
{
	free( (void*)hSema );
	return 0;
}


status_t acquire_sem( sem_id hSema )
{
	return acquire_sem_etc( hSema, 1, 0, INFINITE_TIMEOUT );
}


status_t release_sem( sem_id hSema )
{
	return release_sem_etc( hSema, 1, 0 );
}


status_t acquire_sem_etc( sem_id hSema, int32 nCount, uint32 nFlags, bigtime_t nTimeOut )
{
	SemInfo_s* semptr = (SemInfo_s*)hSema;
	struct timespec tsTimeout;
	struct timeval  tvNow;

	if( semptr == NULL )
	{
		return -1;
	}

	if( nTimeOut == INFINITE_TIMEOUT )
	{
		pthread_mutex_lock( &semptr->mutex );
		while( semptr->count < nCount &&
			!( (semptr->flags & SEM_RECURSIVE) &&
				(semptr->holder == (thread_id)pthread_self()) ) )
		{
			pthread_cond_wait( &semptr->cond, &semptr->mutex );
		}
		semptr->holder = (thread_id)pthread_self();
		semptr->count -= nCount;
		pthread_mutex_unlock( &semptr->mutex );
		return 0;
	}

	gettimeofday( &tvNow, NULL );
	tsTimeout.tv_sec  = tvNow.tv_sec + nTimeOut / 1000000LL;
	tsTimeout.tv_nsec = (tvNow.tv_usec + nTimeOut % 1000000LL) * 1000L;
	while( tsTimeout.tv_nsec >= 1000000000L )
	{
		tsTimeout.tv_sec++;
		tsTimeout.tv_nsec -= 1000000000L;
	}

	pthread_mutex_lock( &semptr->mutex );
	while( semptr->count < nCount &&
		!( (semptr->flags & SEM_RECURSIVE) &&
			(semptr->holder == (thread_id)pthread_self()) ) )
	{
		if( pthread_cond_timedwait( &semptr->cond, &semptr->mutex, &tsTimeout ) == ETIMEDOUT )
		{
			pthread_mutex_unlock( &semptr->mutex );
			return -ETIMEDOUT;
		}
	}
	semptr->holder = (thread_id)pthread_self();
	semptr->count -= nCount;
	pthread_mutex_unlock( &semptr->mutex );

	return 0;
}


status_t release_sem_etc( sem_id hSema, int32 nCount, uint32 nFlags )
{
	SemInfo_s* semptr = (SemInfo_s*)hSema;

	if( semptr == NULL )
	{
		return -1;
	}

	pthread_mutex_lock( &semptr->mutex );
	semptr->count += nCount;
	pthread_cond_signal( &semptr->cond );
	pthread_mutex_unlock( &semptr->mutex );

	return 0;
}

