/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *  Portions Copyright (C) 2001 Tom Marshall
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */


#include <unistd.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/signal.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <pthread.h>

#include <cosmoe/Debug.h>
#include <support/SupportDefs.h>
#include <cosmoe/msgport.h>
#include <cosmoe/OS.h>

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef min
#define min(a,b) ( ((a)<(b)) ? (a) : (b) )
#endif

/******** socket based msgport implementation ********************************
 *
 * Each message consists of a 4-byte type, 4-byte datalen, and data.
 *
 * Messages are assumed to be passed on a reliable channel.  If this is not
 * the case, an additional protocol layer will need to be imposed.
 *
 * The type field allows for a variety of things including urgent data,
 * encryption, etc.  It is currently unused and ignored.
 */

static void* do_msg_io( void* );

int     g_servsock = -1;
int     g_clientsock = -1;

pthread_t*  g_pthrMessageIO = NULL;
int         g_fdmax;
fd_set      g_fdsr;

typedef struct _port_msg port_msg;
struct _port_msg
{
    uint32 len;
    uint32 type;
    char   data[1];
};

typedef struct _MessageNode MessageNode_s;
struct _MessageNode
{
    MessageNode_s*  mn_psNext;
    int             mn_nSize;
    uint32          mn_nCode;
};

typedef struct _MsgPort MsgPort_s;
struct _MsgPort
{
    char            mp_zName[32];
    port_id         mp_hPortID;
    uint32          mp_nFlags;
    int             mp_nMaxCount;
    int             mp_nCount;
    sem_id          mp_hSyncSema;
    int             mp_fd;
    MessageNode_s*  mp_psFirstMsg;
    MessageNode_s*  mp_psLastMsg;
};

#define MAX_MSGPORT 1024 /* XXX: make this dynamic */

static  pthread_mutex_t g_mutexPortList = PTHREAD_MUTEX_INITIALIZER;
static  int             g_nMsgPortCount = 0;
static  MsgPort_s*      g_sMsgPorts[MAX_MSGPORT];
static  port_id         g_hDisplayServerPort;

#define MSG_REQ_CREATE_PORT             0x0100
#define MSG_RSP_CREATE_PORT             0x0101
#define MSG_REQ_DELETE_PORT             0x0102
#define MSG_RSP_DELETE_PORT             0x0103
#define MSG_REQ_GET_APP_SERVER_PORT     0x0104
#define MSG_RSP_GET_APP_SERVER_PORT     0x0105
#define MSG_APP                         0x0106

static pthread_mutex_t g_mutexSpecial = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t  g_condSpecial = PTHREAD_COND_INITIALIZER;
static int             g_nSpecialCount;
static uint32          g_nSpecialValue;

static void get_special_msg( uint32* pval )
{
	pthread_mutex_lock( &g_mutexSpecial );
	while( g_nSpecialCount < 1 )
	{
		pthread_cond_wait( &g_condSpecial, &g_mutexSpecial );
	}
	*pval = g_nSpecialValue;
	g_nSpecialCount--;
	pthread_mutex_unlock( &g_mutexSpecial );
}


static void put_special_msg( uint32 val )
{
	pthread_mutex_lock( &g_mutexSpecial );
	while( g_nSpecialCount > 0 )
	{
		pthread_mutex_unlock( &g_mutexSpecial );
		sched_yield();
		pthread_mutex_lock( &g_mutexSpecial );
	}
	g_nSpecialValue = val;
	g_nSpecialCount++;
	pthread_cond_signal( &g_condSpecial );
	pthread_mutex_unlock( &g_mutexSpecial );
}


static status_t msgport_remove( int i )
{
	MsgPort_s* psPort;

	pthread_mutex_lock( &g_mutexPortList );
	if( i < 0 || i >= MAX_MSGPORT || g_sMsgPorts[i] == NULL )
	{
		pthread_mutex_unlock( &g_mutexPortList );
		return -1;
	}
	psPort = g_sMsgPorts[i];
	g_sMsgPorts[i] = NULL;
	g_nMsgPortCount--;
	pthread_mutex_unlock( &g_mutexPortList );

	while( psPort->mp_psFirstMsg != NULL )
	{
		MessageNode_s* psNode = psPort->mp_psFirstMsg;
		psPort->mp_psFirstMsg = psPort->mp_psFirstMsg->mn_psNext;
		psPort->mp_nCount--;
		free( psNode );
	}
	free( psPort );

	return 0;
}


static status_t local_send_msg_x( port_id hPort, uint32 nCode, const void* pBuffer,
                                  int nSize, bigtime_t nTimeOut )
{
	status_t nError;
	MsgPort_s* psPort;
	MessageNode_s* psNode;

	pthread_mutex_lock( &g_mutexPortList );
	psPort = g_sMsgPorts[hPort];
	if( psPort == NULL )
	{
		dbprintf( "linux_compat: local_send_msg_x got msg for bad port\n" );
		nError = -EINVAL;
		goto error1;
	}

	psNode = (MessageNode_s*)malloc( sizeof(MessageNode_s) + nSize );
	if( psNode == NULL )
	{
		dbprintf( "linux_compat: out of memory\n" );
		nError = -ENOMEM;
		goto error1;
	}
	psNode->mn_psNext = NULL;
	psNode->mn_nCode = nCode;
	if( pBuffer == NULL )
	{
		psNode->mn_nSize = 0;
	}
	else
	{
		psNode->mn_nSize = nSize;
		memcpy( psNode+1, pBuffer, nSize );
	}

	if( psPort->mp_nCount >= psPort->mp_nMaxCount )
	{
		dbprintf( "linux_compat: warning: overflow on message port %u, ignoring\n", hPort );
	}
	if( psPort->mp_psFirstMsg == NULL )
	{
		psPort->mp_psFirstMsg = psNode;
		psPort->mp_psLastMsg = psNode;
	}
	else
	{
		psPort->mp_psLastMsg->mn_psNext = psNode;
		psPort->mp_psLastMsg = psNode;
	}
	psPort->mp_nCount++;
	pthread_mutex_unlock( &g_mutexPortList );
	release_sem( psPort->mp_hSyncSema );
	nError = 0;

error1:
	pthread_mutex_unlock( &g_mutexPortList );
	return nError;
}

#if 0
static void put_short( char** p, uint32* prem, uint16 n )
{
    if( *prem >= 2 )
    {
        uint16 tmp = htons( n );
        memcpy( *p, &tmp, 2 );
        *p = *p + 2;
        *prem -= 2;
    }
}
#endif

static void put_long( char** p, uint32* prem, uint32 n )
{
	if( *prem >= 4 )
	{
		uint32 tmp = htonl( n );
		memcpy( *p, &tmp, 4 );
		*p = *p + 4;
		*prem = *prem - 4;
	}
}

static void put_string( char** p, uint32* prem, const char* sz )
{
	uint32 szlen = strlen(sz);
	char* pcur = *p;
	if( *prem >= 4+szlen )
	{
		uint32 tmp = htonl( szlen );
		memcpy( pcur, &tmp, 4 );
		memcpy( pcur+4, sz, szlen );
		*p = *p + (4 + szlen);
		*prem = *prem - (4 + szlen);
	}
}

static void put_buffer( char** p, uint32* prem, const void* buf, uint32 buflen )
{
	char* pcur = *p;
	if( *prem >= 4+buflen )
	{
		uint32 tmp = htonl( buflen );
		memcpy( pcur, &tmp, 4 );
		memcpy( pcur+4, buf, buflen );
		*p = *p + (4 + buflen);
		*prem = *prem - (4 + buflen);
	}
}

#if 0
static void get_short( char** p, uint32* prem, uint16* pn )
{
    if( *prem >= 2 )
    {
        uint16 tmp;
        memcpy( &tmp, *p, 2 );
        *pn = ntohs( tmp );
        *p = *p + 2;
        *prem = *prem - 2;
    }
}
#endif

static void get_long( char** p, uint32* prem, uint32* pn )
{
	if( *prem >= 4 )
	{
		uint32 tmp;
		memcpy( &tmp, *p, 4 );
		*pn = ntohl( tmp );
		*p = *p + 4;
		*prem = *prem - 4;
	}
}

static void get_string( char** p, uint32* prem, char** psz )
{
	uint32 tmp;
	uint32 szlen;
	char* pcur = *p;
	*psz = NULL;
	if( *prem >= 4 )
	{
		memcpy( &tmp, pcur, 4 );
		szlen = ntohl( tmp );
		if( *prem >= 4 + szlen )
		{
			*psz = (char*)malloc( szlen+1 );
			memcpy( *psz, pcur+4, szlen );
			*(*psz + szlen) = '\0';
			*p = *p + (4 + szlen);
			*prem = *prem - (4 + szlen);
		}
	}
}

#if 0
static void get_buffer( char** p, uint32* prem, void** ppbuf, uint32* pbuflen )
{
    uint32 tmp;
    char* pcur = *p;
    *ppbuf = NULL;
    if( *prem >= 4 )
    {
        memcpy( &tmp, pcur, 4 );
        *pbuflen = ntohl( tmp );
        if( *prem >= 4 + *pbuflen )
        {
            *ppbuf = (void*)malloc( *pbuflen );
            memcpy( *ppbuf, pcur+4, *pbuflen );
            *p = *p + (p + *pbuflen);
            *prem = *prem - (4 + *pbuflen);
        }
    }
}
#endif

static void tcp_remove_socket( int fd )
{
	int i;

	dbprintf( "linux_compat: tcp_remove_socket( %d )\n", fd );
	for( i = 0; i < MAX_MSGPORT; i++ )
	{
		pthread_mutex_lock( &g_mutexPortList );
		if( g_sMsgPorts[i] != NULL && g_sMsgPorts[i]->mp_fd == fd )
		{
			pthread_mutex_unlock( &g_mutexPortList );
			msgport_remove( i );
		}
		else
		{
			pthread_mutex_unlock( &g_mutexPortList );
		}
	}
	close( fd );
	FD_CLR( fd, &g_fdsr );
}

static int tcp_send_msg( int fd, port_msg* pmsg )
{
	char* p = (char*)pmsg;
	uint32 bufrem = pmsg->len;
	pmsg->len = htonl( pmsg->len );
	pmsg->type = htonl( pmsg->type );
	while( bufrem > 0 )
	{
		ssize_t res = send( fd, p, bufrem, 0 );
		if( res <= 0 )
		{
			dbprintf( "linux_compat: tcp_send_msg failed (fd=%d), send returned %d\n", fd, res );
			tcp_remove_socket( fd );
			return -1;
		}
		p += res;
		bufrem -= res;
	}
	return 0;
}


static int tcp_recv_msg( int fd, port_msg** ppmsg )
{
	ssize_t  res;
	port_msg tmpmsg;
	uint32   bufrem;
	char*    p;
	res = recv( fd, &tmpmsg, 8, 0 );
	if( res != 8 )
	{
		dbprintf( "linux_compat: tcp_recv_msg requested header, got %d\n", res );
		tcp_remove_socket( fd );
		return -1;
	}
	tmpmsg.len = ntohl(tmpmsg.len);
	tmpmsg.type = ntohl(tmpmsg.type);
	*ppmsg = (port_msg*)malloc( tmpmsg.len );
	if( *ppmsg == NULL )
	{
		dbprintf( "linux_compat: out of memory\n" );
		return -1;
	}
	(*ppmsg)->len = tmpmsg.len;
	(*ppmsg)->type = tmpmsg.type;
	p = (*ppmsg)->data;
	bufrem = tmpmsg.len - 8;
	while( bufrem > 0 )
	{
		res = recv( fd, p, bufrem, 0 );
		if( res <= 0 )
		{
			dbprintf( "linux_compat: tcp_recv_msg requested %u, got %d\n", bufrem, res );
			tcp_remove_socket( fd );
			return -1;
		}
		p += res;
		bufrem -= res;
	}
	return 0;
}


bool init_server_sock( uint16 dpy )
{
	struct sockaddr_in sa;
	socklen_t salen;
	int tmp;

	if( g_servsock != -1 )
	{
		return false;
	} 

	g_servsock = socket( AF_INET, SOCK_STREAM, IPPROTO_IP );
	if( g_servsock < 0 )
	{
		return false;
	}

	tmp = 1;
	setsockopt( g_servsock, SOL_SOCKET, SO_REUSEADDR, &tmp, sizeof(tmp) );

	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = INADDR_ANY;
	sa.sin_port = htons( 6100 + dpy );
	salen = sizeof(sa);
	if( bind( g_servsock, (struct sockaddr*)&sa, salen ) != 0 )
	{
		close( g_servsock );
		g_servsock = -1;
		return false;
	}

	if( listen( g_servsock, 5 ) != 0 )
	{
		close( g_servsock );
		g_servsock = -1;
		return false;
	}
	g_fdmax = g_servsock;
	FD_ZERO( &g_fdsr );
	FD_SET( g_servsock, &g_fdsr );

	g_pthrMessageIO = (pthread_t*)malloc( sizeof(pthread_t) );
	if( g_pthrMessageIO == NULL )
	{
		dbprintf( "linux_compat: out of memory\n" );
		close( g_servsock );
		g_servsock = -1;
		return false;
	}
	
	if( pthread_create( g_pthrMessageIO, NULL, do_msg_io, NULL ) != 0 )
	{
		free( g_pthrMessageIO );
		g_pthrMessageIO = NULL;
		close( g_servsock );
		g_servsock = -1;
		return false;
	}
	dbprintf( "linux_compat: created server i/o thread\n" );

	return true;
}

bool init_client_sock( void )
{
	const char* disp;
	char*       p;
	char*       disphost;
	ushort      dispport;
	struct hostent* phent;
	struct sockaddr_in sa;
	socklen_t salen;

	g_clientsock = socket( AF_INET, SOCK_STREAM, IPPROTO_IP );
	if( g_clientsock == -1 )
	{
		return -1;
	}

	if( (disp = getenv( "ADISPLAY" )) != NULL )
	{
		disphost = strdup( disp );
		dispport = 6100;
		if( (p = strchr( disphost, ':' )) != NULL )
		{
			dispport += atoi( p+1 );
			*p = '\0';
		}
	}
	else
	{
		disphost = strdup( "127.0.0.1" );
		dispport = 6100;
	}
	phent = gethostbyname( disphost );
	free( disphost );
	if( phent == NULL )
	{
		return false;
	}

	sa.sin_family = AF_INET;
	memcpy( &sa.sin_addr.s_addr, phent->h_addr, sizeof(sa.sin_addr.s_addr) );
	sa.sin_port = htons( dispport );
	salen = sizeof(sa);
	dbprintf( "connecting to %s:%hu\n", inet_ntoa( sa.sin_addr ), dispport );
	if( connect( g_clientsock, (struct sockaddr*)&sa, salen ) != 0 )
	{
		close( g_clientsock );
		g_clientsock = -1;
		return false;
	}
	g_fdmax = g_clientsock;
	FD_ZERO( &g_fdsr );
	FD_SET( g_clientsock, &g_fdsr );

	g_pthrMessageIO = (pthread_t*)malloc( sizeof(pthread_t) );
	if( g_pthrMessageIO == NULL )
	{
		dbprintf( "linux_compat: out of memory\n" );
		close( g_clientsock );
		g_clientsock = -1;
		return false;
	}
	if( pthread_create( g_pthrMessageIO, NULL, do_msg_io, NULL ) != 0 )
	{
		free( g_pthrMessageIO );
		g_pthrMessageIO = NULL;
		close( g_clientsock );
		g_servsock = -1;
		return false;
	}
	dbprintf( "linux_compat: created client i/o thread\n" );

	return true;
}

void close_server_sock( void )
{
	if( g_servsock != -1 )
	{
		close( g_servsock );
		FD_CLR( g_servsock, &g_fdsr );
	}
}

static port_id server_create_port( const char* const pzName, int nMaxCount, int client_fd )
{
	port_id nError;
	int i;
	MsgPort_s* psPort;

	psPort = (MsgPort_s*)malloc( sizeof(MsgPort_s) );
	if ( psPort == NULL )
	{
		nError = -ENOMEM;
		goto error1;
	}

	strncpy( psPort->mp_zName, pzName, 32 );
	psPort->mp_zName[31] = '\0';

	psPort->mp_hSyncSema = create_sem_etc( 0, "port_sync", 0 );
	if( psPort->mp_hSyncSema < 0 )
	{
		dbprintf( "linux_compat: failed to create semaphore\n" );
		nError = psPort->mp_hSyncSema;
		goto error2;
	}
	psPort->mp_fd = client_fd;

	psPort->mp_hPortID = -1;
	psPort->mp_nFlags = 0;
	psPort->mp_nMaxCount = 1024;
	psPort->mp_nCount = 0;
	psPort->mp_psFirstMsg = NULL;
	psPort->mp_psLastMsg = NULL;

	pthread_mutex_lock( &g_mutexPortList );
	for( i = 0; i < MAX_MSGPORT; i++ )
	{
		if( g_sMsgPorts[i] == NULL )
		{
			g_sMsgPorts[i] = psPort;
			psPort->mp_hPortID = i;
			g_nMsgPortCount++;
			break;
		}
	}
	pthread_mutex_unlock( &g_mutexPortList );

	if(  psPort->mp_hPortID < 0 )
	{
		dbprintf( "Failed to alloc descriptor for msg port %s\n", pzName );
		nError = psPort->mp_hPortID;
		goto error3;
	}

	return( psPort->mp_hPortID );

error3:
	delete_sem( psPort->mp_hSyncSema );
error2:
	free( psPort );
error1:
	return( nError );
}


static port_id client_create_port( const char* const pzName, int nMaxCount )
{
	port_id    port;
	port_msg*  pmsg;
	uint32     msglen;
	uint32     msgrem;
	char*      p;
	MsgPort_s* psPort;

	if( g_clientsock == -1 )
	{
		if( !init_client_sock() )
		{
			port = -1;
			goto error1;
		}
	}

	/* type, len, name_len, name, maxcount */
	msglen = 4 + 4 + 4 + strlen(pzName) + 4;
	pmsg = (port_msg*)malloc( msglen );
	if( pmsg == NULL )
	{
		port = -1;
		goto error1;
	}
	pmsg->type = MSG_REQ_CREATE_PORT;
	pmsg->len = msglen;
	p = pmsg->data;
	msgrem = msglen - 8;

	put_string( &p, &msgrem, pzName );
	put_long( &p, &msgrem, nMaxCount );

	if( tcp_send_msg( g_clientsock, pmsg ) != 0 )
	{
		port = -1;
		goto error2;
	}

	get_special_msg( &port );

	psPort = (MsgPort_s*)malloc( sizeof(MsgPort_s) );
	if( psPort == NULL )
	{
		dbprintf( "linux_compat: out of memory\n" );
		port = -ENOMEM;
		goto error2;
	}
	strcpy( psPort->mp_zName, pzName );
	psPort->mp_hPortID = port;
	psPort->mp_nFlags = 0; /* XXX */
	psPort->mp_nMaxCount = 0xFFFF; /* XXX */
	psPort->mp_nCount = 0;
	psPort->mp_hSyncSema = create_sem_etc( 0, "port_sync", 0 );
	psPort->mp_fd = -1;
	psPort->mp_psFirstMsg = NULL;
	psPort->mp_psLastMsg = NULL;
	if( psPort->mp_hSyncSema < 0 )
	{
		dbprintf( "linux_compat: failed to create semaphore\n" );
		port = psPort->mp_hSyncSema;
		goto error2;
	}

	pthread_mutex_lock( &g_mutexPortList );
	if( g_sMsgPorts[port] != NULL )
	{
		dbprintf( "linux_compat: server reused port %d, trashing old one\n", port );
		pthread_mutex_unlock( &g_mutexPortList );
		msgport_remove( port );
		pthread_mutex_lock( &g_mutexPortList );
	}
	g_sMsgPorts[port] = psPort;
	g_nMsgPortCount++;
	pthread_mutex_unlock( &g_mutexPortList );

error2:
	free( pmsg );
error1:

	return port;
}

port_id create_port( const char* const pzName, int nMaxCount )
{
	port_id port;
	int i;

	pthread_mutex_lock( &g_mutexPortList );
	if( g_nMsgPortCount == 0 )
	{
		for( i = 0; i < MAX_MSGPORT; i++ )
		{
			g_sMsgPorts[i] = NULL;
		}
	}
	pthread_mutex_unlock( &g_mutexPortList );
	if( g_servsock != -1 )
	{
		/* server should never send over its port so use fd=-1 */
		port = server_create_port( pzName, nMaxCount, -1 );
	}
	else
	{
		/* client just requests a port from the server */
		port = client_create_port( pzName, nMaxCount );
	}
	pthread_mutex_unlock( &g_mutexPortList );

	return port;
}

static status_t server_delete_port( port_id hPort )
{
	return msgport_remove( hPort );
}

static status_t client_delete_port( port_id hPort )
{
	port_msg* pmsg;
	uint32    msglen;
	uint32    msgrem;
	char*     p;
	status_t  res;

	if( g_clientsock == -1 )
	{
		res = -1;
		goto error1;
	}

	if( hPort < 0 || hPort >= MAX_MSGPORT )
	{
		dbprintf( "linux_compat: client_delete_port got invalid port %d\n", hPort );
		res = -1;
		goto error1;
	}

	/* type, len, port */
	msglen = 4 + 4 + 4;
	pmsg = (port_msg*)malloc( msglen );
	if( pmsg == NULL )
	{
		res = -1;
		goto error1;
	}
	pmsg->type = MSG_REQ_DELETE_PORT;
	pmsg->len = msglen;
	p = pmsg->data;
	msgrem = msglen - 8;

	put_long( &p, &msgrem, hPort );

	if( tcp_send_msg( g_clientsock, pmsg ) != 0 )
	{
		res = -1;
		goto error2;
	}

	get_special_msg( &res );

	msgport_remove( hPort );

error2:
	free( pmsg );
error1:

	return res;
}

status_t delete_port( port_id hPort )
{
	status_t res;
	if( g_servsock != -1 )
	{
		res = server_delete_port( hPort );
	}
	else
	{
		res = client_delete_port( hPort );
	}
	return res;
}

static port_id server_get_app_server_port( void )
{
	return g_hDisplayServerPort;
}

static port_id client_get_app_server_port( void )
{
	port_msg*  pmsg;
	uint32     msglen;
	uint32     msgrem;
	char*      p;
	port_id    port;

	if( g_clientsock == -1 )
	{
		port = -1;
		goto error1;
	}

	/* type, len */
	msglen = 4 + 4;
	pmsg = (port_msg*)malloc( msglen );
	if( pmsg == NULL )
	{
		port = -1;
		goto error1;
	}
	pmsg->type = MSG_REQ_GET_APP_SERVER_PORT;
	pmsg->len = msglen;
	p = pmsg->data;
	msgrem = msglen - 8;

	if( tcp_send_msg( g_clientsock, pmsg ) != 0 )
	{
		port = -1;
		goto error2;
	}

	get_special_msg( &port );

	g_nMsgPortCount++;

error2:
	free( pmsg );
error1:

	return port;
}


port_id get_app_server_port( void )
{
	port_id port;
	if( g_servsock != -1 )
	{
		port = server_get_app_server_port();
	}
	else
	{
		port = client_get_app_server_port();
	}
	return port;
}


int set_app_server_port( port_id hPort )
{
	int res;
	if( g_servsock != -1 )
	{
		g_hDisplayServerPort = hPort;
		res = 0;
	}
	else
	{
		res = -1;
	}
	return res;
}


static void handle_create_port_request( port_msg* pmsg, int client_fd )
{
	uint32    msglen;
	uint32    msgrem;
	char*     p;
	char*     pzName;
	int       nMaxCount;
	port_id   port;

	if( g_servsock == -1 )
	{
		dbprintf( "linux_compat: client got create_port request\n" );
		return;
	}

	msglen = pmsg->len;
	p = pmsg->data;
	msgrem = msglen - 8;

	get_string( &p, &msgrem, &pzName );
	get_long( &p, &msgrem, &nMaxCount );
	port = server_create_port( pzName, nMaxCount, client_fd );
	free( pzName );

	/* type, len, port */
	msglen = 4 + 4 + 4;
	pmsg = (port_msg*)malloc( msglen );
	if( pmsg == NULL )
	{
		dbprintf( "linux_compat: out of memory\n" );
		return;
	}
	pmsg->type = MSG_RSP_CREATE_PORT;
	pmsg->len = msglen;
	p = pmsg->data;
	msgrem = msglen - 8;

	put_long( &p, &msgrem, port );
	tcp_send_msg( client_fd, pmsg );
	free( pmsg );
}


static void handle_create_port_response( port_msg* pmsg )
{
	port_id    port;
	uint32     msgrem;
	char*      p;

	if( g_servsock != -1 )
	{
		dbprintf( "linux_compat: server got create_port response\n" );
		return;
	}

	if( pmsg->len != 8+4 )
	{
		dbprintf( "linux_compat: handle_create_port_response got bogus message\n" );
		return;
	}

	p = pmsg->data;
	msgrem = pmsg->len - 8;
	get_long( &p, &msgrem, &port );

	put_special_msg( port );
}


static void handle_delete_port_request( port_msg* pmsg, int client_fd )
{
	uint32    msglen;
	uint32    msgrem;
	char*     p;
	port_id   port;
	status_t  res;

	if( g_servsock == -1 )
	{
		dbprintf( "linux_compat: client got delete_port request\n" );
		return;
	}

	msglen = pmsg->len;
	p = pmsg->data;
	msgrem = msglen - 8;

	get_long( &p, &msgrem, &port );
	if( port != g_hDisplayServerPort )
	{
		res = server_delete_port( port );
	}
	else
	{
		dbprintf( "linux_compat: client tried to delete g_hDisplayServerPort (%d)\n", port );
		res = -EPERM;
	}

	/* type, len, status */
	msglen = 4 + 4 + 4;
	pmsg = (port_msg*)malloc( msglen );
	if( pmsg == NULL )
	{
		dbprintf( "linux_compat: out of memory\n" );
		return;
	}
	pmsg->type = MSG_RSP_DELETE_PORT;
	pmsg->len = msglen;
	p = pmsg->data;
	msgrem = msglen - 8;

	put_long( &p, &msgrem, res );
	tcp_send_msg( client_fd, pmsg );
	free( pmsg );
}


static void handle_delete_port_response( port_msg* pmsg )
{
	status_t   res;
	uint32     msgrem;
	char*      p;

	if( g_servsock != -1 )
	{
		dbprintf( "linux_compat: server got delete_port response\n" );
		return;
	}

	if( pmsg->len != 8+4 )
	{
		dbprintf( "linux_compat: handle_delete_port_response got bogus message\n" );
		return;
	}

	p = pmsg->data;
	msgrem = pmsg->len - 8;
	get_long( &p, &msgrem, &res );

	put_special_msg( res );
}


static void handle_get_app_server_port_request( port_msg* pmsg, int client_fd )
{
	uint32    msglen;
	uint32    msgrem;
	char*     p;
	port_id   port;

	if( g_servsock == -1 )
	{
		dbprintf( "linux_compat: client got get_app_server_port request\n" );
		return;
	}

	/* no payload */

	port = server_get_app_server_port();

	/* type, len, port */
	msglen = 4 + 4 + 4;
	pmsg = (port_msg*)malloc( msglen );
	if( pmsg == NULL )
	{
		dbprintf( "linux_compat: out of memory\n" );
		return;
	}
	pmsg->type = MSG_RSP_GET_APP_SERVER_PORT;
	pmsg->len = msglen;
	p = pmsg->data;
	msgrem = msglen - 8;

	put_long( &p, &msgrem, port );
	tcp_send_msg( client_fd, pmsg );
	free( pmsg );
}


static void handle_get_app_server_port_response( port_msg* pmsg )
{
	status_t   res;
	uint32     msgrem;
	char*      p;

	if( g_servsock != -1 )
	{
		dbprintf( "linux_compat: server got get_app_server_port response\n" );
		return;
	}

	if( pmsg->len != 8+4 )
	{
		dbprintf( "linux_compat: handle_get_app_server_port_response got bogus message\n" );
		return;
	}

	p = pmsg->data;
	msgrem = pmsg->len - 8;
	get_long( &p, &msgrem, &res );

	put_special_msg( res );
}


static void handle_incoming_msg( port_msg* pmsg )
{
	uint32          msglen;
	uint32          msgrem;
	char*           p;
	uint32          port;
	uint32          code;
	uint32          buflen;
	MessageNode_s*  psNode;
	MsgPort_s*      psPort;

	msglen = pmsg->len;
	p = pmsg->data;
	msgrem = msglen - 8;

	/* port, code, bufferlen, buffer */
	get_long( &p, &msgrem, &port );
	get_long( &p, &msgrem, &code );
	get_long( &p, &msgrem, &buflen );
	if( buflen != msgrem )
	{
		dbprintf( "linux_compat: received bad msg, buflen=%u, msgrem=%u\n", buflen, msgrem );
		goto error1;
	}
	if( port >= MAX_MSGPORT )
	{
		dbprintf( "linux_comapt: got msg for bad port %u\n", port );
		goto error1;
	}

	pthread_mutex_lock( &g_mutexPortList );
	psPort = g_sMsgPorts[port];
	if( psPort == NULL )
	{
		dbprintf( "linux_compat: got msg for bad port\n" );
		goto error2;
	}

	/*
	* If the server gets a message for a port that the client created,
	* reflect the message back to the client.
	*/
	if( g_servsock != -1 && psPort->mp_fd != -1 )
	{
dbprintf( "linux_compat: reflect msg to client port %d, fd=%d\n", port, psPort->mp_fd );
		tcp_send_msg( psPort->mp_fd, pmsg );
		goto error2;
	}

	psNode = (MessageNode_s*)malloc( sizeof(MessageNode_s) + buflen );
	if( psNode == NULL )
	{
		dbprintf( "linux_compat: out of memory\n" );
		goto error2;
	}
	psNode->mn_psNext = NULL;
	psNode->mn_nSize = buflen;
	psNode->mn_nCode = code;
	memcpy( psNode+1, p, buflen );

	if( psPort->mp_nCount >= psPort->mp_nMaxCount )
	{
		dbprintf( "linux_compat: warning: overflow on message port %u, ignoring\n", port );
	}
	if( psPort->mp_psFirstMsg == NULL )
	{
		psPort->mp_psFirstMsg = psNode;
		psPort->mp_psLastMsg = psNode;
	}
	else
	{
		psPort->mp_psLastMsg->mn_psNext = psNode;
		psPort->mp_psLastMsg = psNode;
	}
	psPort->mp_nCount++;
	pthread_mutex_unlock( &g_mutexPortList );
	release_sem( psPort->mp_hSyncSema );
	return;

error2:
	pthread_mutex_unlock( &g_mutexPortList );
error1:
	return;
}


/*
 * This is the thread that handles socket i/o
 */
static void* do_msg_io( void* pv )
{
	int             i;
	port_msg*       pmsg;
	int             res;
	fd_set          fdsr;

dbprintf( "linux_compat: do_msg_io enter pid=%d, thread=%ld\n", getpid(), pthread_self() );
	while( 1 )
	{
		fdsr = g_fdsr;
		res = select( g_fdmax+1, &fdsr, NULL, NULL, NULL );
		if( res == -1 )
		{
			dbprintf( "linux_compat: select() failed\n" );
			break;
		}
		if( res == 0 )
		{
			dbprintf( "linux_compat: select() timed out???\n" );
			continue;
		}

		if( g_servsock != -1 && FD_ISSET( g_servsock, &fdsr ) )
		{
			int                client_sock;
			struct sockaddr_in client_addr;
			socklen_t          client_len;

			FD_CLR( g_servsock, &fdsr );
			client_len = sizeof(client_addr);
			client_sock = accept( g_servsock, (struct sockaddr*)&client_addr, &client_len );
			if( client_sock != -1 )
			{
				if( client_sock > g_fdmax ) g_fdmax = client_sock;
				FD_SET( client_sock, &g_fdsr );
				dbprintf( "linux_compat: connection from %s:%hu = %d, new g_fdmax = %d\n",
						inet_ntoa( client_addr.sin_addr ),
						ntohs( client_addr.sin_port ),
						client_sock, g_fdmax );
			}
		}
		for( i = 0; i <= g_fdmax; i++ )
		{
			if( FD_ISSET( i, &fdsr ) )
			{
				if( tcp_recv_msg( i, &pmsg ) != 0 )
				{
					dbprintf( "linux_compat: tcp_recv_msg failed on %d\n", i );
					continue;
				}
				switch( pmsg->type )
				{
				case MSG_REQ_CREATE_PORT:
					handle_create_port_request( pmsg, i );
					break;
				case MSG_RSP_CREATE_PORT:
					handle_create_port_response( pmsg );
					break;
				case MSG_REQ_DELETE_PORT:
					handle_delete_port_request( pmsg, i );
					break;
				case MSG_RSP_DELETE_PORT:
					handle_delete_port_response( pmsg );
					break;
				case MSG_REQ_GET_APP_SERVER_PORT:
					handle_get_app_server_port_request( pmsg, i );
					break;
				case MSG_RSP_GET_APP_SERVER_PORT:
					handle_get_app_server_port_response( pmsg );
					break;
				case MSG_APP:
					handle_incoming_msg( pmsg );
					break;
				default:
					dbprintf( "linux_compat: unknown message type %d\n", pmsg->type );
				}
				free( pmsg );
			}
		}
	}
	return NULL; /* avoid compiler warning */
}

status_t get_msg( port_id hPort, uint32* pnCode, void* pBuffer, int nSize )
{
	return get_msg_x( hPort, pnCode, pBuffer, nSize, INFINITE_TIMEOUT );
}

status_t send_msg( port_id hPort, uint32 nCode, const void* pBuffer, int nSize )
{
	return send_msg_x( hPort, nCode, pBuffer, nSize, INFINITE_TIMEOUT );
}

/*
 * this is probably the trickiest part of the whole thing.  the strategy is
 * to select() on the fd's and read messages until a data message arrives on
 * the desired port or the timer expires.  on the server, we have the added
 * task of accepting new connections.
 *
 * note that multiple threads will be calling get_msg_x simultaneously so we
 * need to release and reacquire the port mutex after each message is read,
 * regardless whether it is on our desired port.
 */
status_t get_msg_x( port_id hPort, uint32* pnCode, void* pBuffer, int nSize, bigtime_t nTimeOut )
{
	MsgPort_s*	 psPort = NULL;
	MessageNode_s* psNode;
	int 		 nRealSize;
	int	     	 nError;

	pthread_mutex_lock( &g_mutexPortList );
	if( hPort < 0 || hPort >= MAX_MSGPORT )
	{
		nError = -EINVAL;
		goto error1;
	}
	
	psPort = g_sMsgPorts[hPort];
	if( g_servsock != -1 )
	{
		/*
		* we are authoritative so make sure the port exists.
		*/
		if( psPort == NULL )
		{
			dbprintf( "Error: get_msg_x() attempt to receive message from invalid port %d\n", hPort );
			nError = -EINVAL;
			goto error1;
		}
	}
	else
	{
		/*
		* the port may or may not be valid on the server, so we need to
		* ensure we have a queue for it.
		*/
		if( psPort == NULL )
		{
			/* XXXTODO: make a GET_PORT command */
			psPort = (MsgPort_s*)malloc( sizeof(MsgPort_s) );
			if( psPort == NULL )
			{
				dbprintf( "linux_compat: out of memory\n" );
				nError = -ENOMEM;
				goto error1;
			}
			strcpy( psPort->mp_zName, "__auto" );
			psPort->mp_hPortID = hPort;
			psPort->mp_nFlags = 0;
			psPort->mp_nMaxCount = 0xFFFF;
			psPort->mp_nCount = 0;
			psPort->mp_hSyncSema = create_sem_etc( 0, "port_sync", 0 );
			psPort->mp_fd = -1;
			psPort->mp_psFirstMsg = NULL;
			psPort->mp_psLastMsg = NULL;
			if( psPort->mp_hSyncSema < 0 )
			{
				dbprintf( "linux_compat: failed to create semaphore\n" );
				nError = psPort->mp_hSyncSema;
				free( psPort );
				goto error1;
			}
			g_sMsgPorts[hPort] = psPort;
			g_nMsgPortCount++;
			dbprintf( "linux_compat: created auto port for %d\n", hPort );
		}
	}

again:
	if ( psPort->mp_psFirstMsg == NULL )
	{
		sem_id hSyncSema;
		// We take a backup here since the port might get deleted
		// when we release the g_hPortListSema semaphore.
		// If the port is deleted so is the semaphore and acquire_sem()
		// will return an error. It's then very important not to
		// touch the port again until we has revalidated the port-handle.

		hSyncSema = psPort->mp_hSyncSema;
		pthread_mutex_unlock( &g_mutexPortList );
		nError = acquire_sem_etc( hSyncSema, 1, 0, nTimeOut );
		pthread_mutex_lock( &g_mutexPortList );
		if ( nError < 0 )
		{
			goto error1;
		}

		// We must revalidate the handle after unlocking g_hPortListSema
		psPort = g_sMsgPorts[hPort];
		if ( NULL == psPort )
		{
			nError = -EINVAL;
			dbprintf( "Message port %d deleted while we was wating for messages\n", hPort );
			goto error1;
		}
	}
	else
	{
		if ( acquire_sem_etc( psPort->mp_hSyncSema, 1, 0, 0LL ) < 0 )
		{
			dbprintf( "panic: get_msg_x() failed to lock mp_hSyncSema even though there was pending messages\n" );
		}
	}
	psNode = psPort->mp_psFirstMsg;
	if ( psNode == NULL )
	{
		dbprintf( "Error: We got lock on mp_hSyncSema even though the message list was empty on port %d\n", hPort );
		goto again;
	}

	nRealSize = min( nSize, psNode->mn_nSize );

	psPort->mp_psFirstMsg = psNode->mn_psNext;
	if ( NULL == psPort->mp_psFirstMsg ) {
		psPort->mp_psLastMsg = NULL;
	}
	psPort->mp_nCount--;

	if ( NULL != pBuffer ) {
		memcpy( pBuffer, psNode + 1, nRealSize );
	}
	if ( NULL != pnCode ) {
		*pnCode = psNode->mn_nCode;
	}
	nError = ( nSize >= psNode->mn_nSize ) ? 0 : -E2BIG;
	free( psNode );

error1:
	pthread_mutex_unlock( &g_mutexPortList );
	return( nError );
}


/*
 * the server uses the global port list to find the socket.  it must be
 * valid as the server keeps the authoritative port list.
 */
static status_t server_send_msg_x( port_id hPort, uint32 nCode, const void* pBuffer,
                                   int nSize, bigtime_t nTimeOut )
{
	MsgPort_s*     psPort;
	int            fd;
	int            nError;

	port_msg* pmsg;
	uint32    msglen;
	uint32    msgrem;
	char*     p;

	pthread_mutex_lock( &g_mutexPortList );
	psPort = g_sMsgPorts[hPort];
	if( hPort < 0 || hPort >= MAX_MSGPORT || psPort == NULL ) {
		dbprintf( "Error: send_msg_x() attempt to send to invalid port %d\n", hPort );
		nError = EINVAL;
		goto error2;
	}
	fd = psPort->mp_fd;
	pthread_mutex_unlock( &g_mutexPortList );

	if( fd == -1 )
	{
		dbprintf( "linux_compat: reflect internal server msg\n" );
		return local_send_msg_x( hPort, nCode, pBuffer, nSize, nTimeOut );
	}

	/* type, len, port, code, bufferlen, buffer */
	msglen = 4 + 4 + 4 + 4 + 4 + nSize;
	pmsg = (port_msg*)malloc( msglen );
	if( pmsg == NULL )
	{
		nError = -ENOMEM;
		goto error1;
	}
	pmsg->type = MSG_APP;
	pmsg->len = msglen;
	p = pmsg->data;
	msgrem = msglen - 8;

	put_long( &p, &msgrem, hPort );
	put_long( &p, &msgrem, nCode );
	put_buffer( &p, &msgrem, pBuffer, nSize );

	nError = 0;
	if( tcp_send_msg( psPort->mp_fd, pmsg ) != 0 )
	{
		dbprintf( "linux_compat: server_send_msg_x failed in tcp_send_msg() - broken connection?\n" );
		nError = -EPIPE;
	}
	free( pmsg );
	return nError;

error2:
	pthread_mutex_unlock( &g_mutexPortList );
error1:
	return( nError );
}

/*
 * the client may send messages on any port since the server has the
 * authoritative port list.  we just require a valid client socket.
 */
static status_t client_send_msg_x( port_id hPort, uint32 nCode, const void* pBuffer,
                                   int nSize, bigtime_t nTimeOut )
{
	int       nError;
	port_msg* pmsg;
	uint32    msglen;
	uint32    msgrem;
	char*     p;

	/* type, len, port, code, bufferlen, buffer */
	msglen = 4 + 4 + 4 + 4 + 4 + nSize;
	pmsg = (port_msg*)malloc( msglen );
	if( pmsg == NULL )
	{
		nError = -ENOMEM;
		goto error1;
	}
	pmsg->type = MSG_APP;
	pmsg->len = msglen;
	p = pmsg->data;
	msgrem = msglen - 8;

	put_long( &p, &msgrem, hPort );
	put_long( &p, &msgrem, nCode );
	put_buffer( &p, &msgrem, pBuffer, nSize );

	if( tcp_send_msg( g_clientsock, pmsg ) != 0 )
	{
		dbprintf( "linux_compat: client_send_msg_x failed in tcp_send_msg() - broken connection?\n" );
		nError = -EPIPE;
		goto error2;
	}

	nError = 0;

error2:
	free( pmsg );
error1:
	return( nError );
}


status_t send_msg_x( port_id hPort, uint32 nCode, const void* pBuffer, int nSize, bigtime_t nTimeOut )
{
	status_t res;
	if( g_servsock != -1 )
	{
		res = server_send_msg_x( hPort, nCode, pBuffer, nSize, nTimeOut );
	}
	else
	{
		res = client_send_msg_x( hPort, nCode, pBuffer, nSize, nTimeOut );
	}
	return res;
}
