/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */


#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <errno.h>
#include <string.h>

#include <cosmoe/Debug.h>
#include <support/SupportDefs.h>
#include <cosmoe/fs_attr.h>

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>

#ifndef min
#define min(a,b) ( ((a)<(b)) ? (a) : (b) )
#endif


int dbprintf( const char* pzFmt, ... )
{
	int rc;
	va_list ap;

	va_start( ap, pzFmt );
	rc = vfprintf( stderr, pzFmt, ap );
	va_end(ap);

	return rc;
}


/*** shared libs ***/

#include <dlfcn.h>

int load_library( const char* pzPath, uint32 nFlags )
{
	void* hdll = dlopen( pzPath, RTLD_LAZY );
	if( !hdll )
	{
		fprintf( stderr, "linux_compat: dlopen( '%s', RTLD_LAZY ) failed: %s\n", pzPath, dlerror() );
	}
	else
	{
		fprintf( stderr, "linux_compat: dlopen( '%s', RTLD_LAZY ) succeeded\n", pzPath );
	}
	return (int)hdll;
}

int unload_library( int nLibrary )
{
	void* hdll = (void*)nLibrary;
	return dlclose( hdll );
}

int get_symbol_address( int nLibrary, const char* pzName, int nIndex, void** pPtr )
{
	void* hdll = (void*)nLibrary;
	const char* err = NULL;

	*pPtr = dlsym( hdll, pzName );
	err = dlerror();
	if( err )
	{
		fprintf( stderr, "linux_compat: dlsym( '%s' ) failed: %s\n", pzName, err );
		return -1;
	}

	return 0;
}


status_t snooze( bigtime_t nTimeout )
{
	usleep( (unsigned long)nTimeout );
	return 0;
}


/* return usecs since boot */
bigtime_t system_time( void )
{
	FILE*       fp;
	char        buf[80];
	double      utime;
	bigtime_t   systime;

	systime = 0;
	if( (fp = fopen( "/proc/uptime", "r" )) != NULL )
	{
		if( fgets( buf, sizeof(buf), fp ) != NULL )
		{
			sscanf( buf, "%lf", &utime );
			systime = (bigtime_t)( utime * 1000000.0f );
		}
		fclose( fp );
	}

	return systime;
}


/* return usecs since epoch */
bigtime_t get_real_time( void )
{
	struct timeval tv;
	struct timezone tz;

	gettimeofday( &tv, &tz ); /* timezone unused but cant pass NULL */

	return (bigtime_t)tv.tv_sec * (bigtime_t)(1000*1000) + (bigtime_t)tv.tv_usec;
}


/*
 * return usecs idle since boot: this is the last field in the main cpu
 * entry in /proc/stat, but that's in jiffies.  scaling factor is wired
 * to HZ=100 here, which is not always correct.  note that the logic is
 * complicated by the fact that uniprocessor /proc/stat does not have a
 * "cpu0" entry.
 */
bigtime_t get_idle_time( int nProcessor )
{
	FILE*         fp;
	char          cpuid[32];
	char          buf[80];
	unsigned long n1, n2, n3, nidle, nidle_all;
	bigtime_t     idletime;

	sprintf( cpuid, "cpu%d ", nProcessor );
	nidle = 0;
	nidle_all = 0;
	idletime = 0;
	if( (fp = fopen( "/proc/stat", "r" )) != NULL )
	{
		while( fgets( buf, sizeof(buf), fp ) != NULL )
		{
			if( strncmp( buf, "cpu ", 4 ) == 0 )
			{
				sscanf( buf+4, "%lu %lu %lu %lu", &n1, &n2, &n3, &nidle_all );
			}
			if( strncmp( buf, cpuid, strlen(cpuid) ) == 0 )
			{
				sscanf( buf+strlen(cpuid), "%lu %lu %lu %lu", &n1, &n2, &n3, &nidle );
				break;
			}
		}
		fclose( fp );
	}

	if( nProcessor == 0 && nidle == 0 )
	{
		/* use the "cpu" line */
		idletime = (bigtime_t)nidle_all * 10000LL;
	}
	else
	{
		/* use the "cpuN" line */
		idletime = (bigtime_t)nidle * 10000LL;
	}

	return idletime;
}


/* helper for get_system_info */
static void get_cpu_info( system_info* psInfo )
{
	FILE*         fp;
	int           ncpu;
	char          buf[80];
	char*         p;
	bigtime_t     systime;
	bigtime_t     idletime;
	unsigned long n1, n2, n3, nidle;

	systime = system_time();
	psInfo->nBootTime = get_real_time() - systime;
	psInfo->bus_clock_speed = 66;	/* FIXME */
	ncpu = 0;
	if( (fp = fopen( "/proc/cpuinfo", "r" )) != NULL )
	{
		while( fgets( buf, sizeof(buf), fp ) != NULL )
		{
			if( strncmp( buf, "processor\t", 10 ) == 0 )
			{
				ncpu++;
			}

			if( strncmp( buf, "cpu MHz\t", 8 ) == 0 &&
				ncpu < MAX_CPU_COUNT )
			{
				p = strchr( buf, ':' );
				if( p != NULL )
				{
					psInfo->cpu_clock_speed = atoi( p+2 );
				}
			}
		}
		fclose( fp );
	}
	psInfo->cpu_count = ncpu;

	if( (fp = fopen( "/proc/stat", "r" )) != NULL )
	{
		while( fgets( buf, sizeof(buf), fp ) != NULL )
		{
			if( ncpu == 1 && strncmp( buf, "cpu ", 4 ) == 0 )
			{
				/* there are no cpuN lines, use the overall stat */
				sscanf( buf+4, "%lu %lu %lu %lu", &n1, &n2, &n3, &nidle );
				idletime = (bigtime_t)nidle * 10000LL;
				psInfo->cpu_infos[0].active_time = systime - idletime;
				break;
			}

			if( strncmp( buf, "cpu", 3 ) == 0 )
			{
				sscanf( buf+3, "%d %lu %lu %lu %lu", &ncpu, &n1, &n2, &n3, &nidle );
				if( ncpu < psInfo->cpu_count )
				{
					idletime = (bigtime_t)nidle * 10000LL;
					psInfo->cpu_infos[ncpu].active_time = systime - idletime;
				}
			}
		}
		fclose( fp );
	}
}

/* helper for get_system_info */
static void get_mem_info( system_info* psInfo )
{
}


/* helper for get_system_info */
static void get_fs_info( system_info* psInfo )
{
}


status_t get_system_info( system_info* psInfo )
{
	psInfo->nKernelVersion = 1;
	strcpy( psInfo->zKernelName, "unknown" );
	strcpy( psInfo->zKernelBuildDate, "unknown" );
	strcpy( psInfo->zKernelBuildTime, "unknown" );
	get_cpu_info( psInfo ); /* set boot time and cpu info */
	get_mem_info( psInfo ); /* set various mem info */
	get_fs_info( psInfo );  /* set various fs info */

	return 0;
}

/*** XXX: totally unimplemented ***/

DIR* open_attrdir( int nFile )
{
	dbprintf( "Cosmoe: UNIMPLEMENTED: open_attrdir\n" );
	return NULL;
}

int close_attrdir( DIR* pDir )
{
	dbprintf( "Cosmoe: UNIMPLEMENTED: close_attrdir\n" );
	return -1;
}

void rewind_attrdir( DIR* pDir )
{
	dbprintf( "Cosmoe: UNIMPLEMENTED: rewind_attrdir\n" );
}

struct dirent* read_attrdir( DIR* pDir )
{
	dbprintf( "Cosmoe: UNIMPLEMENTED: read_attrdir\n" );
	return NULL;
}

ssize_t write_attr( int nFile, const char *pzAttrName, int nFlags, int nType,
					const void* pBuffer, off_t nPos, size_t nLen )
{
	dbprintf( "Cosmoe: UNIMPLEMENTED: write_attr\n" );
	return -1;
}

ssize_t read_attr( int nFile, const char *pzAttrName, int nType,
				void* pBuffer, off_t nPos, size_t nLen )
{
	dbprintf( "Cosmoe: UNIMPLEMENTED: read_attr\n" );
	return -1;
}

int remove_attr( int nFile, const char* pzName )
{
	dbprintf( "Cosmoe: UNIMPLEMENTED: remove_attr\n" );
	return -1;
}

int stat_attr( int nFile, const char* pzName, struct attr_info* psBuffer )
{
	dbprintf( "Cosmoe: UNIMPLEMENTED: stat_attr\n" );
	return -1;
}

int watch_node( int nNode, int nPort, uint32 nFlags, void* pDate )
{
	dbprintf( "Cosmoe: UNIMPLEMENTED: watch_node\n" );
	return -1;
}

status_t delete_node_monitor( int nMonitor )
{
	dbprintf( "Cosmoe: UNIMPLEMENTED: delete_node_monitor\n" );
	return -1;
}

status_t get_directory_path( int nFD, char* pzBuffer, int nBufLen )
{
	dbprintf( "Cosmoe: UNIMPLEMENTED: get_directory_path\n" );
	return -1;
}

int freadlink( int nFile, char* pzBuffer, size_t nBufSize )
{
	dbprintf( "Cosmoe: UNIMPLEMENTED: freadlink\n" );
	return -1;
}

ssize_t read_pos( int nFile, off_t nPos, void* pBuffer, size_t nLength )
{
	dbprintf( "Cosmoe: UNIMPLEMENTED: read_pos\n" );
	return -1;
}

ssize_t write_pos( int nFile, off_t nPos, const void* pBuffer, size_t nLength )
{
	dbprintf( "Cosmoe: UNIMPLEMENTED: write_pos\n" );
	return -1;
}

ssize_t readv_pos( int nFile, off_t nPos, const struct iovec* psVector, int nCount )
{
	dbprintf( "Cosmoe: UNIMPLEMENTED: readv_pos\n" );
	return -1;
}

ssize_t writev_pos( int nFile, off_t nPos, const struct iovec* psVector, int nCount )
{
	dbprintf( "Cosmoe: UNIMPLEMENTED: writev_pos\n" );
	return -1;
}

