/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *  Portions Copyright (C) 2001 Tom Marshall
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */


#include <linux/fb.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <stdlib.h>

#include <cosmoe/Debug.h>
#include <support/SupportDefs.h>
#include <cosmoe/vesa_gfx.h>

#ifndef _NO_FB_
int fb_fd = -1;
struct fb_fix_screeninfo fb_fixsi;
struct fb_var_screeninfo fb_varsi;
size_t fb_len;
void* fb_ptr = NULL;
#endif

#ifndef _NO_FB_
int get_vesa_mode_info( VESA_Mode_Info_s* psVesaModeInfo, uint32 nModeNr )
{
	memset( psVesaModeInfo, 0, sizeof(VESA_Mode_Info_s) );
	if( nModeNr != 1 || !fb_ptr )
	{
		dbprintf( "linux_compat: get_vesa_mode_info for unknown mode %u\n", nModeNr );
		return EINVAL;
	}

	psVesaModeInfo->ModeAttributes = 0; /* XXX */
	psVesaModeInfo->WinAAttributes = 0; /* XXX */
	psVesaModeInfo->WinBAttributes = 0; /* XXX */
	psVesaModeInfo->WinGranularity = 0; /* XXX */
	psVesaModeInfo->WinSize = 0;        /* XXX */
	psVesaModeInfo->WinASegment = 0;    /* XXX */
	psVesaModeInfo->WinBSegment = 0;    /* XXX */
	psVesaModeInfo->WinFuncPtr = 0;     /* XXX */

	psVesaModeInfo->BytesPerScanLine = fb_fixsi.line_length;
	psVesaModeInfo->XResolution = fb_varsi.xres;
	psVesaModeInfo->YResolution = fb_varsi.yres;
	psVesaModeInfo->XCharSize = 0; /* XXX */
	psVesaModeInfo->YCharSize = 0; /* XXX */
	psVesaModeInfo->NumberOfPlanes = 1;
	psVesaModeInfo->BitsPerPixel = fb_varsi.bits_per_pixel;
	psVesaModeInfo->NumberOfBanks = 1;
	psVesaModeInfo->MemoryModel = 0;
	psVesaModeInfo->BankSize = 0;
	psVesaModeInfo->NumberOfImagePages = 0;
	psVesaModeInfo->Reserved = 0;

	/* TODO: figure out other depth mappings .. probably need a table */
	switch( fb_varsi.bits_per_pixel )
	{
	case 16:
		psVesaModeInfo->RedMaskSize = 5; psVesaModeInfo->RedFieldPosition = 11;
		psVesaModeInfo->GreenMaskSize = 6; psVesaModeInfo->GreenFieldPosition = 5;
		psVesaModeInfo->BlueMaskSize = 5; psVesaModeInfo->BlueFieldPosition = 0;
		psVesaModeInfo->RsvdMaskSize = 0; psVesaModeInfo->RsvdFieldPosition = 0;
		psVesaModeInfo->DirectColorModeInfo = 0; /* XXX */
		break;
	case 32:
		psVesaModeInfo->RedMaskSize = 8; psVesaModeInfo->RedFieldPosition = 16;
		psVesaModeInfo->GreenMaskSize = 8; psVesaModeInfo->GreenFieldPosition = 8;
		psVesaModeInfo->BlueMaskSize = 8; psVesaModeInfo->BlueFieldPosition = 0;
		psVesaModeInfo->RsvdMaskSize = 8; psVesaModeInfo->RsvdFieldPosition = 24;
		psVesaModeInfo->DirectColorModeInfo = 0; /* XXX */
		break;
	default:
		dbprintf( "linux_compat: unknown color depth %i\n", fb_varsi.bits_per_pixel );
		return EINVAL;
	}
	psVesaModeInfo->DirectColorModeInfo = 0;
	psVesaModeInfo->PhysBasePtr = (int32)fb_ptr;
	psVesaModeInfo->OffScreenMemOffset = 0;
	psVesaModeInfo->OffScreenMemSize = 0;

	return 0;
}

int get_vesa_info( Vesa_Info_s* psVesaInfo, uint16* pnModeList, int nMaxModeCount )
{
	int n = 0;

	if( nMaxModeCount < 1 )
	{
		return 0;
	}

	memset( psVesaInfo, 0, sizeof(Vesa_Info_s) );
	for( n = 0; n < nMaxModeCount; n++ )
	{
		pnModeList[n] = 0;
	}

	/* open framebuffer device */
	if( fb_fd == -1 )
	{
		const char* fbdev = getenv("FRAMEBUFFER");
		if( !fbdev ) fbdev = "/dev/fb0";
		fb_fd = open( fbdev, O_RDWR );
		if( -1 == fb_fd )
		{
			dbprintf( "linux_compat: open( %s ) failed: %s\n", fbdev, strerror(errno) );
			return 0;
		}
	}

	if( fb_ptr == NULL )
	{
		/* get relevant info */
		if( 0 == ioctl( fb_fd, FBIOGET_FSCREENINFO, &fb_fixsi ) &&
			0 == ioctl( fb_fd, FBIOGET_VSCREENINFO, &fb_varsi ) )
		{
			fb_len = fb_fixsi.line_length * fb_varsi.yres_virtual;

			/* okay now map the buffer */
			fb_ptr = mmap( NULL, fb_len, PROT_READ|PROT_WRITE, MAP_FILE|MAP_SHARED, fb_fd, 0 );
			if( MAP_FAILED == fb_ptr )
			{
				dbprintf( "mmap() failed: %s\n", strerror(errno) );
				fb_ptr = NULL;
				return 0;
			}
			/* we win */
			pnModeList[0] = 1;
			/* the psVesaInfo struct appears totally unused on return */
			memcpy( psVesaInfo->VesaSignature, "VBE2", 4 );
			psVesaInfo->VesaVersion = 2;
		}
	}

	return 1;
}

int set_vesa_mode( uint32 nMode )
{
	dbprintf( "linux_compat: set_vesa_mode( %u )\n", nMode );

	/* we only support one mode: 0 */
	return ( nMode == 0 );
}
#endif
