/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */


#include <interface/Alert.h>
#include <interface/Font.h>
#include <interface/Screen.h>
#include <cosmoe/Debug.h>

using namespace os;

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CAlert::CAlert( const std::string& cTitle, const std::string& cText, int nFlags, ... ) :
    CWindow( CRect( 100, 50, 100, 50 ), cTitle.c_str() )
{
	va_list pArgs;

	va_start( pArgs, nFlags );

	mAlertView = new AlertView( cText, pArgs );
	mAlertView->SetHighColor( 0, 0, 0 );
	mAlertView->SetLowColor( 200, 200, 200 );

	float width, height;

	mAlertView->GetPreferredSize( &width, &height );
	mAlertView->ResizeTo( width, height );
	ResizeTo( width, height );

	CScreen cDesktop;
	MoveTo( (cDesktop.GetResolution().x - width) / 2,
			(cDesktop.GetResolution().y - height) / 2 );

	AddChild( mAlertView );
	va_end( pArgs );
	mInvoker = NULL;
	mMsgPort  = -1;
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CAlert::CAlert( const char* inTitle,
				const char* inText,
				const char* inButtonOne,
				const char* inButtonTwo,
				const char* inButtonThree) :
    CWindow( CRect( 100, 50, 100, 50 ), inTitle )
{
	std::string	cText(inText);

	mAlertView = new AlertView( cText, inButtonOne, inButtonTwo, inButtonThree );
	mAlertView->SetHighColor( 0, 0, 0 );
	mAlertView->SetLowColor( 200, 200, 200 );

	float width, height;

	mAlertView->GetPreferredSize( &width, &height );
	mAlertView->ResizeTo( width, height );
	ResizeTo( width, height );

	CScreen cDesktop;
	MoveTo( (cDesktop.GetResolution().x - width) / 2,
			(cDesktop.GetResolution().y - height) / 2 );

	AddChild( mAlertView );

	mInvoker = NULL;
	mMsgPort  = -1;
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CAlert::~CAlert()
{
	if ( mMsgPort != -1 )
	{
		delete_port( mMsgPort );
	}
	
	delete mInvoker;
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CAlert::MessageReceived( CMessage* pcMessage )
{
	if ( pcMessage->what < uint32(mAlertView->mButtons.size()) )
	{
		if ( mMsgPort >= 0 )
		{
			send_msg( mMsgPort, pcMessage->what, NULL, 0 );
		}
		else
		{
			if ( mInvoker != NULL )
			{
				CMessage* pcMsg = mInvoker->Message();
				if ( pcMsg == NULL )
				{
					dbprintf( "Error: Invoker registered with this Alert requester does not have a message!\n" );
				}
				else
				{
					pcMsg->AddInt32( "which", pcMessage->what );
					mInvoker->Invoke();
				}
			}
			PostMessage( M_QUIT );
		}
	}
	else
	{
		CHandler::MessageReceived( pcMessage );
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

int32 CAlert::Go()
{
	uint32	nCode;
	int		nError;

	mMsgPort = create_port( "alert_port", DEFAULT_PORT_SIZE );

	Show();
	Activate();

	if ( mMsgPort < 0 )
	{
		dbprintf( "Alert::WaitForSelection() failed to create message port\n" );
		nError = -EINVAL;
	}
	else
	{
		nError = get_msg( mMsgPort, &nCode, NULL, 0 );
	}

	PostMessage( M_QUIT );

	if ( nError < 0 )
	{
		return -1;
	}

	return( nCode );
}


void CAlert::Go( CInvoker* inInvoker )
{
	mInvoker = inInvoker;
	Show();
	Activate();
}


CButton* CAlert::ButtonAt(int32 inButtonIndex) const
{
	return mAlertView->ButtonAt(inButtonIndex);
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

AlertView::AlertView( const std::string& cText,va_list pButtons ) : CView( CRect(0,0,1,1), "alert", CF_FOLLOW_ALL )
{
	const char*	pzButName;
	uint32		i = 0UL;

	while( (pzButName = va_arg( pButtons, const char* )) != NULL )
	{
		AddButton(pzButName, i++);
	}

	PrepareAlert(cText);
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

AlertView::AlertView( const std::string& cText,
					  const char* inButtonOne,
					  const char* inButtonTwo,
					  const char* inButtonThree ) : CView( CRect(0,0,1,1), "alert", CF_FOLLOW_ALL )
{
	uint32		i = 0UL;

	AddButton(inButtonOne, i++);

	if (inButtonTwo != NULL)
		AddButton(inButtonTwo, i++);

	if (inButtonThree != NULL)
		AddButton(inButtonThree, i++);

	PrepareAlert(cText);
}


void AlertView::AddButton(const char* inButtonName, uint32 inButtonNumber)
{
	CButton* pcButton = new CButton( CRect(0,0,1,1), inButtonName, inButtonName,
									 new CMessage( inButtonNumber ) );
	mButtons.push_back( pcButton );
	AddChild( pcButton );

	float width, height;

	pcButton->GetPreferredSize( &width, &height );

	if ( width > mButtonSize.x )
	{
		mButtonSize.x = width;
	}

	if ( height > mButtonSize.y )
	{
		mButtonSize.y = height;
	}
}


void AlertView::PrepareAlert(const std::string& cText)
{
	uint	i;
	float	nWidth = (mButtonSize.x + 5) * mButtons.size();

	int nStart = 0;
	for ( i = 0 ; i <= cText.size() ; ++i )
	{
		if ( i == cText.size() || cText[i] == '\n' )
		{
			int nLen = i - nStart;
			std::string cLine( cText, nStart, nLen );
			m_cLines.push_back( cLine );
			float nStrLen = StringWidth( cLine.c_str() ) + 20;
			if ( nStrLen > nWidth )
			{
				nWidth = nStrLen;
			}
			nStart = i + 1;
		}
	}

	font_height sFontHeight;
	GetFontHeight( &sFontHeight );
	m_vLineHeight = sFontHeight.ascender + sFontHeight.descender + sFontHeight.line_gap;

	float nHeight = m_cLines.size() * m_vLineHeight;
	nHeight += mButtonSize.y + 30;

	float y = nHeight - mButtonSize.y - 5;
	float x = nWidth - mButtons.size() * (mButtonSize.x + 5) - 5;

	for ( i = 0 ; i < mButtons.size() ; ++i )
	{
		mButtons[i]->SetFrame( CRect( x, y, x + mButtonSize.x - 1, y + mButtonSize.y - 1 ) );
		x += mButtonSize.x + 5;
	}
	m_cMinSize.x = nWidth;
	m_cMinSize.y = nHeight;
}


void AlertView::AllAttached()
{
	mButtons[0]->MakeFocus();
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void AlertView::GetPreferredSize( float* outWidth, float* outHeight )
{
	if (outWidth)
		*outWidth = m_cMinSize.x;

	if (outHeight)
		*outHeight = m_cMinSize.y;
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void AlertView::Draw( CRect cUpdateRect )
{
	FillRect( cUpdateRect, LowColor() );

	float y = 20.0f;
	for ( uint i = 0 ; i < m_cLines.size() ; ++i )
	{
		MovePenTo( 10, y );
		DrawString( m_cLines[i].c_str() );
		y += m_vLineHeight;
	}
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CButton*	AlertView::ButtonAt(int32 inButtonIndex) const
{
	return mButtons[inButtonIndex];
}

