/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */


#include <app/Application.h>
#include <interface/Bitmap.h>
#include <interface/Window.h>
#include <exceptions.h>

#include <cosmoe/OS.h>

using namespace os;


CBitmap::CBitmap( int inWidth,
				  int inHeight,
				  color_space eColorSpc,
				  bool inAcceptViews,
				  bool inShareFramebuffer )
{
	area_id hArea;
	
	mBounds     = CRect( 0, 0, inWidth-1, inHeight-1 );
	mHandle     = -1;
	mBits       = NULL;
	mColorSpace = eColorSpc;

	int nError = CApplication::GetInstance()->CreateBitmap( inWidth, inHeight, eColorSpc, inShareFramebuffer, &mHandle, &hArea );

	if ( nError < 0 )
	{
		throw( GeneralFailure( "Application server failed to create bitmap", -nError ) );
	}

	if ( inShareFramebuffer )
	{
		mArea = clone_area( "bm_clone", (void**) &mBits, AREA_FULL_ACCESS, AREA_NO_LOCK, hArea );
	}
	else
	{
		mArea = -1;
	}

	if ( inAcceptViews )
	{
		mWindow = new CWindow( this );
		mWindow->Unlock();
	}
	else
	{
		mWindow = NULL;
	}
}


CBitmap::CBitmap( CRect			inBounds,
				  color_space	eColorSpace,
				  bool			inAcceptViews,
				  bool			inShareFramebuffer )
{
	area_id hArea;
	
	mBounds     = inBounds;
	mHandle     = -1;
	mBits       = NULL;
	mColorSpace = eColorSpace;
	mWindow     = NULL;
	mArea       = -1;

	mInitCheck = CApplication::c_app->CreateBitmap( inBounds.IntegerWidth(),
													inBounds.IntegerHeight(),
													eColorSpace,
													inShareFramebuffer,
													&mHandle,
													&hArea );

	if ( mInitCheck == EOK )
	{
		if ( inShareFramebuffer )
		{
			mArea = clone_area( "bm_clone", (void**) &mBits, AREA_FULL_ACCESS, AREA_NO_LOCK, hArea );
		}

		if ( inAcceptViews )
		{
			mWindow = new CWindow( this );
			mWindow->Unlock();
		}
	}
}


CBitmap::~CBitmap()
{
	if ( mArea != -1 )
		delete_area( mArea );
	
	if (mWindow)
		delete mWindow;
	
	CApplication::GetInstance()->DeleteBitmap( mHandle );
}


void CBitmap::AddChild( CView* inView )
{
	if ( NULL != mWindow )
	{
		mWindow->AddChild( inView );
	}
}


bool CBitmap::RemoveChild( CView* inView )
{
	if ( NULL != mWindow )
	{
		mWindow->RemoveChild( inView );
		return( true );
	}

	return( false );
}


int32	CBitmap::BitsLength() const
{
	AreaInfo_s	info;
	status_t	err;
	int32		length;
	 
	err = get_area_info( mArea, &info );
	if (err == EOK)
		length = info.size;
	else
		length = 0L;
	
	return (length);	
}


CView* CBitmap::FindView( const char* inName ) const
{
	if ( NULL != mWindow )
	{
		return( mWindow->FindView( inName ) );
	}
	
	return( NULL );
}


void CBitmap::Flush( void )
{
	if ( NULL != mWindow )
	{
		mWindow->Flush();
	}
}


void CBitmap::Sync( void )
{
	if ( NULL != mWindow )
	{
		mWindow->Sync();
	}
}


bool CBitmap::IsLocked( void ) const
{
	if ( NULL != mWindow )
	{
		return mWindow->IsLocked();
	}
	
	return false;
}


bool CBitmap::Lock( void )
{
	if ( NULL != mWindow )
	{
		return mWindow->Lock();
	}
	
	return false;
}


void CBitmap::Unlock( void )
{
	if ( NULL != mWindow )
	{
		mWindow->Unlock();
	}
}


int CBitmap::BytesPerRow() const
{
	int nBitsPerPix = BitsPerPixel( mColorSpace );
	if ( nBitsPerPix == 15 )
	{
		nBitsPerPix = 16;
	}
	return( int(mBounds.right - mBounds.left + 1.0f) * nBitsPerPix / 8 );
}




const uint8 os::__5_to_8_bit_table[] = {
    0x00,0x08,0x10,0x18,0x20,0x29,0x31,0x39,
    0x41,0x4a,0x52,0x5a,0x62,0x6a,0x73,0x7b,
    0x83,0x8b,0x94,0x9c,0xa4,0xac,0xb4,0xbd,
    0xc5,0xcd,0xd5,0xde,0xe6,0xee,0xf6,0xff
};

const uint8 os::__6_to_8_bit_table[] = {
    0x00,0x04,0x08,0x0c,0x10,0x14,0x18,0x1c,
    0x20,0x24,0x28,0x2c,0x30,0x34,0x38,0x3c,
    0x40,0x44,0x48,0x4c,0x50,0x55,0x59,0x5d,
    0x61,0x65,0x69,0x6d,0x71,0x75,0x79,0x7d,
    0x81,0x85,0x89,0x8d,0x91,0x95,0x99,0x9d,
    0xa1,0xa5,0xaa,0xae,0xb2,0xb6,0xba,0xbe,
    0xc2,0xc6,0xca,0xce,0xd2,0xd6,0xda,0xde,
    0xe2,0xe6,0xea,0xee,0xf2,0xf6,0xfa,0xff
};
