/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */

 
#include <interface/Box.h>
#include <interface/Font.h>
#include <macros.h>

using namespace os;



CBox::CBox( CRect inFrame, const char* inName, const char* cLabel,
            uint32 inResizeMask, uint32 inFlags )
			: LayoutView( inFrame, inName, NULL, inResizeMask, inFlags ), m_cLabelString( cLabel )
{
	m_pcLabelView  = NULL;
	_CalcStringLabelSize();
}


void CBox::SetLabel( const char* cLabel )
{
	m_cLabelString = cLabel;
	_CalcStringLabelSize();
	CRect cDamage = Bounds();
	cDamage.bottom = m_cLabelSize.y;
	Invalidate( cDamage );
	Flush();
}


CView* CBox::SetLabel( CView* pcLabel, bool bResizeToPreferred )
{
	CView* pcOldLabel = m_pcLabelView;
	m_pcLabelView = pcLabel;

	if ( pcOldLabel != NULL )
	{
		RemoveChild( pcOldLabel );
	}

	if ( pcLabel != NULL )
	{
		pcLabel->Show( false );
		AddChild( pcLabel );
		if ( bResizeToPreferred )
		{
			pcLabel->ResizeToPreferred();
		}
		pcLabel->MoveTo( 10.0f, 0.0f );
		pcLabel->Show( true );
		m_cLabelSize = pcLabel->Frame().Size();
	}

	return( pcOldLabel );
}


const char* CBox::Label() const
{
	return( m_cLabelString.c_str() );
}


CView* CBox::LabelView() const
{
	return( m_pcLabelView );
}


void CBox::AttachedToWindow()
{
	CView* parentView = Parent();

	if ( parentView != NULL )
	{
		SetViewColor( parentView->ViewColor() );
		SetLowColor( parentView->ViewColor() );
	}
}


void CBox::FrameResized( float inWidth, float inHeight )
{
	CRect cBounds( GetNormalizedBounds() );
	bool bNeedFlush = false;

	LayoutNode* pcRoot = GetRoot();
	if ( pcRoot != NULL )
	{
		CRect cClientFrame = cBounds;
		cClientFrame.Resize( 3.0f, m_cLabelSize.y + 1.0f, -3.0f, -3.0f );
		pcRoot->SetFrame( cClientFrame );
	}

	if ( inWidth != 0.0f )
	{
		CRect cDamage = cBounds;

		cDamage.left = cDamage.right - max( 2.0f, inWidth + 1.0f );
		Invalidate( cDamage );
		bNeedFlush = true;
	}

	if ( inHeight != 0.0f )
	{
		CRect cDamage = cBounds;

		cDamage.top = cDamage.bottom - max( 2.0f, inHeight + 1.0f );
		Invalidate( cDamage );
		bNeedFlush = true;
	}

	if ( bNeedFlush )
		Flush();
}


void CBox::FontChanged( CFont* pcNewFont )
{
	if ( m_pcLabelView != NULL )
	{
		return;
	}
	float vOldHeight = m_cLabelSize.y;
	_CalcStringLabelSize();

	CRect cDamage = Bounds().Bounds();

	cDamage.bottom = cDamage.top + max( vOldHeight, m_cLabelSize.y ) + 1.0f;

	Invalidate( cDamage );
	Flush();
}


void CBox::_CalcStringLabelSize()
{
	if ( m_cLabelString.size() == 0 )
	{
		return;
	}
	GetFontHeight( &m_sFontHeight );

	m_cLabelSize = CPoint( ceil( StringWidth( m_cLabelString.c_str(), m_cLabelString.size() ) ),
						ceil( m_sFontHeight.ascender + m_sFontHeight.descender ) );
}


void CBox::Draw( CRect cUpdateRect )
{
	CRect cBounds = GetNormalizedBounds();
	cBounds.Floor();

	bool bHasStrLabel = m_pcLabelView == NULL && m_cLabelString.size() > 0;
	bool bHasLabel    = bHasStrLabel || m_pcLabelView != NULL;

	cBounds.top = m_cLabelSize.y * 0.5f - 1.0f;
	float x1 = 5.0f;
	float x2 = x1 + m_cLabelSize.x + 10.0f;

	CRect cInnerBounds( (bHasLabel) ? GetNormalizedBounds() : cBounds );
	cInnerBounds.Resize( 2.0f, (bHasLabel) ? 0.0f : 2.0f, -2.0f, -2.0f );
	FillRect( cInnerBounds, LowColor() );

	if ( bHasLabel )
	{
		FillRect( CRect( 0.0f, 0.0f, 1.0f, cBounds.top - 1.0f ), LowColor() );
		FillRect( CRect( cBounds.right - 1.0f, 0.0f, cBounds.right, cBounds.top - 1.0f ), LowColor() );
	}

	SetHighColor( 0, 0, 0 );

	if ( bHasStrLabel )
	{
		DrawString( m_cLabelString.c_str(), CPoint( x1 + 5.0f, m_sFontHeight.ascender ) );
	}

	Color32_s sFgCol = get_default_color( COL_SHINE );
	Color32_s sBgCol = get_default_color( COL_SHADOW );

	SetHighColor( sBgCol );

	// Left
	StrokeLine( CPoint( cBounds.left, cBounds.bottom ), CPoint( cBounds.left, cBounds.top ) );
	// Top
	if ( bHasLabel )
	{
		StrokeLine( CPoint( cBounds.left, cBounds.top ), CPoint( x1, cBounds.top ) );
		StrokeLine( CPoint( x2, cBounds.top ), CPoint( cBounds.right, cBounds.top ) );
	}
	else
	{
		StrokeLine( CPoint( cBounds.left, cBounds.top ), CPoint( cBounds.right, cBounds.top ) );
	}
	// Right
	StrokeLine( CPoint( cBounds.right - 1.0f, cBounds.top ), CPoint( cBounds.right - 1.0f, cBounds.bottom ) );
	// Bottom
	StrokeLine( CPoint( cBounds.left + 2.0f, cBounds.bottom - 1.0f ), CPoint( cBounds.right - 1.0f, cBounds.bottom - 1.0f ) );


	SetHighColor( sFgCol );
	// Left
	StrokeLine( CPoint( cBounds.left + 1.0f, cBounds.bottom ), CPoint( cBounds.left + 1.0f, cBounds.top + 1.0f ) );
	// Top
	if ( bHasLabel )
	{
		StrokeLine( CPoint( 1.0f, cBounds.top + 1.0f ), CPoint( x1, cBounds.top + 1.0f ) );
		StrokeLine( CPoint( x2, cBounds.top + 1.0f ), CPoint( cBounds.right, cBounds.top + 1.0f ) );
	}
	else
	{
		StrokeLine( CPoint( 1.0f, cBounds.top + 1.0f ), CPoint( cBounds.right, cBounds.top + 1.0f ) );
	}
	// Right
	StrokeLine( CPoint( cBounds.right, cBounds.top + 1.0f ), CPoint( cBounds.right, cBounds.bottom ) );
	// Bottom
	StrokeLine( CPoint( cBounds.left + 1.0f, cBounds.bottom ), CPoint( cBounds.right, cBounds.bottom ) );    
}


void CBox::GetPreferredSize( float* outWidth, float* outHeight )
{
	LayoutView::GetPreferredSize( outWidth, outHeight );

	if (outWidth)
		*outWidth += 6.0f;

	if (outHeight)
		*outHeight += m_cLabelSize.y + 4.0f;
}
