/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */


#include <stdio.h>

#include <interface/ListView.h>

#include <interface/ScrollBar.h>
#include <interface/Button.h>
#include <interface/Font.h>
#include <app/Message.h>
#include <app/Application.h>
#include <macros.h>

#include <limits.h>
#include <algorithm>

using namespace os;


struct RowContentPred { // : public binary_function<ListViewRow,ListViewRow, bool> {
    RowContentPred( int nColumn ) { m_nColumn = nColumn; }
    bool operator()(const ListViewRow* x, const ListViewRow* y) const { return( x->IsLessThan( y, m_nColumn ) ); }
    int m_nColumn;
};

namespace os {
    struct RowPosPred { // : public binary_function<ListViewRow,ListViewRow, bool> {
        bool operator()(const ListViewRow* x, const ListViewRow* y) const { return( x->m_vYPos < y->m_vYPos ); }
    };
}

/*
  ListView -+
            |
            +-- ListViewHeader -+
                                |
                                +-- ListViewContainer -+
                                                       |
                                                       +-- ListViewCol
                                                       |
                                                       +-- ListViewCol
                                                       |
                                                       +-- ListViewCol


  +-----------------------------------------------------+
  |                      ListView                       |
  |+---------------------------------------------------+|
  ||               |  ListViewHeader      |            ||
  || +-----------------------------------------------+ ||
  || |               ListViewContainer               | ||
  || | +-------------++-------------++-------------+ | ||
  || | | ListViewCol || ListViewCol || ListViewCol | | ||
  || | |             ||             ||             | | ||
  || | |             ||             ||             | ^ ||
  || | |             ||             ||             | | ||
  || | |             ||             ||             | | ||
  || | |             ||             ||             | | ||
  || | |             ||             ||             | | ||
  || | |             ||             ||             | | ||
  || | |             ||             ||             | | ||
  || | |             ||             ||             | | ||
  || | |             ||             ||             | | ||
  || | |             ||             ||             | - ||
  || | |             ||             ||             | | ||
  || | |             ||             ||             | | ||
  || | |             ||             ||             | | ||
  || | |             ||             ||             | | ||
  || | +-------------++-------------++-------------+ | ||
  || +-----------------------------------------------+ ||
  |+------<------------------------------------->------+|
  +-----------------------------------------------------+
 */

namespace os
{

class ListViewCol : public CView
{
public:
						ListViewCol( ListViewContainer* pcParent,
									 const CRect& cFrame,
									 const std::string& cTitle );
						~ListViewCol();

	virtual void		Draw( CRect cUpdateRect );
	void				Refresh( const CRect& cUpdateRect );
	
private:
	friend class ListViewHeader;
	friend class ListViewContainer;

	ListViewContainer*	m_pcParent;
	std::string			m_cTitle;
	float				m_vContentWidth;
};


class ListViewHeader : public CView
{
	friend class ListView;
	friend class ListViewRow;
	friend class ListViewCol;

						ListViewHeader( ListView* pcParent,
										const CRect& cFrame,
										uint32 nModeFlags );
										
	void				DrawButton( const char* pzTitle, const CRect& cFrame, CFont* pcFont, font_height* psFontHeight );
	virtual void        Draw( CRect cUpdateRect );

	virtual void        MouseDown( CPoint cPosition );
	virtual void        MouseUp( CPoint cPosition );
	virtual void        MouseMoved( CPoint cNewPos, uint32 nCode, CMessage* pcData );
	virtual void        FrameResized( float inWidth, float inHeight );
	virtual void        ViewScrolled( const CPoint& cDelta );

	virtual bool        IsFocus( void ) const;

	void                Layout();

	ListView*           m_pcParent;
	ListViewContainer*  m_pcMainView;

	int                 m_nSizeColumn;
	int                 m_nDragColumn;
	CPoint              m_cHitPos;
	float               m_vHeaderHeight;
};


class ListViewContainer : public CView
{
	friend class ListView;
	friend class ListViewHeader;
	friend class ListViewRow;
	friend class ListViewCol;

	enum { AUTOSCROLL_TIMER = 1 };

	ListViewContainer( ListView* pcListView, CRect inFrame, uint32 nModeFlags );
	int                        InsertColumn( const char* pzTitle, int nWidth, int nPos = -1 );
	int                        InsertRow( int nPos, ListViewRow* pcRow, bool bUpdate );
	ListViewRow*        RemoveRow( int nIndex, bool bUpdate );
	void                InvalidateRow( int nRow, uint32 nFlags, bool bImidiate = false );
	void                Clear();
	int                 GetRowIndex( float y ) const;
	void                MakeVisible( int nRow, bool bCenter );
	void                StartScroll( ListView::scroll_direction eDirection, bool bSelect );
	void                StopScroll();

	virtual void        FrameResized( float inWidth, float inHeight );
	virtual void        MouseDown( CPoint cPosition );
	virtual void        MouseUp( CPoint cPosition );
	virtual void        MouseMoved( CPoint cNewPos, uint32 nCode, CMessage* pcData );
	virtual void        Draw( CRect cUpdateRect );
	virtual void        TimerTick( int nID );
	virtual void        DetachedFromWindow( void );
	virtual bool        IsFocus( void ) const;

	bool                HandleKey( char nChar, uint32 nQualifiers );
	void                LayoutColumns();
	ListViewCol*        GetColumn( int nCol ) const { return( m_cCols[m_cColMap[nCol]] ); }


	bool        ClearSelection();
	bool        SelectRange( int nStart, int nEnd, bool bSelect );
	void        InvertRange( int nStart, int nEnd );
	void        ExpandSelect( int nNumRow, bool bInvert, bool bClear );

	std::vector<ListViewCol*> m_cCols;
	std::vector<int>              m_cColMap;
	std::vector<ListViewRow*> m_cRows;
	ListView*                 m_pcListView;
	uint32                    m_nModeFlags;
	CRect                      m_cSelectRect;
	bool                      m_bIsSelecting;
	bool                      m_bMouseMoved;
	bool                      m_bDragIfMoved;
	int                       m_nBeginSel;
	int                       m_nEndSel;
	bigtime_t                 m_nMouseDownTime;
	int                       m_nLastHitRow;
	int                       m_nFirstSel;
	int                       m_nLastSel;
	float                     m_vVSpacing;
	float                     m_vTotalWidth; // Total with of columns
	float                     m_vContentHeight;
	int                       m_nSortColumn;
	bool                      m_bAutoScrollUp;
	bool                      m_bAutoScrollDown;
	bool                      m_bMousDownSeen;
	bool                      m_bAutoScrollSelects;
};

class DummyRow : public ListViewRow
{
public:
	DummyRow() {}

	virtual void AttachToView( CView* pcView, int nColumn ) {}
	virtual void SetRect( const CRect& cRect, int nColumn ) {}

	virtual float GetWidth( CView* pcView, int nColumn ) { return( 0.0f ); }
	virtual float GetHeight( CView* pcView ) { return( 0.0f ); }
	virtual void  Draw( const CRect& cFrame, CView* pcView, uint nColumn,
						bool bSelected, bool bHighlighted, bool bHasFocus ) {};
	virtual bool  IsLessThan( const ListViewRow* pcOther, uint nColumn ) const { return( false ); }
};

} // end of namespace


ListViewRow::ListViewRow()
{
	m_vHeight       = 0.0f;
	m_bSelected     = false;
	m_bHighlighted  = false;
	m_bIsSelectable = true;
}

ListViewRow::~ListViewRow()
{

}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool ListViewRow::HitTest( CView* pcView, const CRect& cFrame, int nColumn, CPoint cPos )
{
	return( true );
}


void ListViewRow::SetIsSelectable( bool bSelectable )
{
	m_bIsSelectable = bSelectable;
}

bool ListViewRow::IsSelectable() const
{
	return( m_bIsSelectable );
}

bool ListViewRow::IsSelected() const
{
	return( m_bSelected );
}

bool ListViewRow::IsHighlighted() const
{
	return( m_bHighlighted );
}

void ListViewStringRow::AttachToView( CView* pcView, int nColumn )
{
	m_cStrings[nColumn].second = pcView->StringWidth( m_cStrings[nColumn].first ) + 5.0f;
}

void ListViewStringRow::SetRect( const CRect& cRect, int nColumn )
{
}

void ListViewStringRow::AppendString( const std::string& cString )
{
	m_cStrings.push_back( std::make_pair( cString, 0.0f ) );
}

const std::string& ListViewStringRow::GetString( int nIndex ) const
{
	return( m_cStrings[nIndex].first );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

float ListViewStringRow::GetWidth( CView* pcView, int nIndex )
{
	return( m_cStrings[nIndex].second );
//    return( pcView->StringWidth( m_cStrings[nIndex] ) + 5.0f );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

float ListViewStringRow::GetHeight( CView* pcView )
{
	font_height sHeight;
	pcView->GetFontHeight( &sHeight );

	return( sHeight.ascender + sHeight.descender );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListViewStringRow::Draw( const CRect& cFrame, CView* pcView, uint nColumn,
                               bool bSelected, bool bHighlighted, bool bHasFocus )
{
	font_height sHeight;
	pcView->GetFontHeight( &sHeight );

	pcView->SetHighColor( 255, 255, 255 );
	pcView->FillRect( cFrame );

	float vFontHeight = sHeight.ascender + sHeight.descender;
	float vBaseLine = cFrame.top + (cFrame.Height()+1.0f) / 2 - vFontHeight / 2 + sHeight.ascender;
	pcView->MovePenTo( cFrame.left + 3.0f, vBaseLine );

	if ( bHighlighted && nColumn == 0 )
	{
		pcView->SetHighColor( 255, 255, 255 );
		pcView->SetLowColor( 0, 50, 200 );
	}
	else if ( bSelected && nColumn == 0 )
	{
		pcView->SetHighColor( 255, 255, 255 );
		pcView->SetLowColor( 0, 0, 0 );
	}
	else
	{
		pcView->SetLowColor( 255, 255, 255 );
		pcView->SetHighColor( 0, 0, 0 );
	}

	if ( bSelected && nColumn == 0 )
	{
		CRect cRect = cFrame;
		cRect.right  = cRect.left + pcView->StringWidth( m_cStrings[nColumn].first.c_str() ) +          4;
		cRect.top    = vBaseLine - sHeight.ascender - 1;
		cRect.bottom = vBaseLine + sHeight.descender + 1;
		pcView->FillRect( cRect, Color32_s( 0, 0, 0, 0 ) );
	}
	pcView->DrawString( m_cStrings[nColumn].first.c_str() );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool ListViewStringRow::IsLessThan( const ListViewRow* pcOther, uint nColumn ) const
{
	const ListViewStringRow* pcRow = dynamic_cast<const ListViewStringRow*>(pcOther);
	if ( NULL == pcRow || nColumn >= m_cStrings.size() || nColumn >= pcRow->m_cStrings.size() )
	{
		return( false );
	}
	return( m_cStrings[nColumn].first < pcRow->m_cStrings[nColumn].first );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

ListViewCol::ListViewCol( ListViewContainer* pcParent, const CRect& cFrame, const std::string& cTitle ) :
    CView( cFrame, "_lv_column" ), m_cTitle(cTitle)
{
	m_pcParent = pcParent;
	m_vContentWidth = 0.0f;
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

ListViewCol::~ListViewCol()
{
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListViewCol::Refresh( const CRect& cUpdateRect )
{
	if ( m_pcParent->m_cRows.empty() )
	{
		SetHighColor( 255, 255, 255 );
		FillRect( cUpdateRect );
		return;
	}
	CRect cBounds = Bounds();

	ListViewRow* pcLastRow = m_pcParent->m_cRows[m_pcParent->m_cRows.size()-1];
	if ( cUpdateRect.top > pcLastRow->m_vYPos + pcLastRow->m_vHeight + m_pcParent->m_vVSpacing ) {
		SetHighColor( 255, 255, 255 );
		FillRect( cUpdateRect );
		return;
	}

	int nFirstRow  = m_pcParent->GetRowIndex( cUpdateRect.top );

	if ( nFirstRow < 0 )
	{
		nFirstRow = 0;
	}

	uint nColumn = 0;
	for ( nColumn = 0 ; nColumn < m_pcParent->m_cCols.size() ; ++nColumn )
	{
		if ( m_pcParent->m_cCols[nColumn] == this )
		{
			break;
		}
	}

	std::vector<ListViewRow*>& cList = m_pcParent->m_cRows;
	bool bHasFocus = m_pcParent->m_pcListView->IsFocus();

	CRect cFrame( cBounds.left, 0.0f, cBounds.right, 0.0f );

	for ( int i = nFirstRow ; i < int(cList.size()) ; ++i )
	{
		if ( cList[i]->m_vYPos > cUpdateRect.bottom )
		{
			break;
		}
		cFrame.top = cList[i]->m_vYPos;
		cFrame.bottom = cFrame.top + cList[i]->m_vHeight + m_pcParent->m_vVSpacing - 1.0f;

		cList[i]->Draw( cFrame, this, nColumn, cList[i]->m_bSelected, cList[i]->m_bHighlighted, bHasFocus );
	}
	if ( cFrame.bottom < cUpdateRect.bottom )
	{
		cFrame.top = cFrame.bottom + 1.0f;
		cFrame.bottom = cUpdateRect.bottom;
		SetHighColor( 255, 255, 255 );
		FillRect( cFrame );
	}
}

//----------------------------------------------------------------------------
// Name:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListViewCol::Draw( CRect cUpdateRect )
{
	if ( cUpdateRect.IsValid() == false )
	{
		return; // FIXME: Workaround for appserver bug. Fix appserver.
	}

	if ( m_pcParent->m_bIsSelecting )
	{
		m_pcParent->SetDrawingMode( DM_INVERT );
		m_pcParent->DrawFrame( m_pcParent->m_cSelectRect, FRAME_TRANSPARENT | FRAME_THIN );
		m_pcParent->SetDrawingMode( DM_COPY );
	}

	Refresh( cUpdateRect );
	if ( m_pcParent->m_bIsSelecting )
	{
		m_pcParent->SetDrawingMode( DM_INVERT );
		m_pcParent->DrawFrame( m_pcParent->m_cSelectRect, FRAME_TRANSPARENT | FRAME_THIN );
		m_pcParent->SetDrawingMode( DM_COPY );
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

ListViewContainer::ListViewContainer( ListView* pcListView, CRect inFrame, uint32 nModeFlags ) :
    CView( inFrame, "main_view", CF_FOLLOW_NONE, WID_WILL_DRAW | WID_DRAW_ON_CHILDREN )
{
	m_pcListView     = pcListView;
	m_vVSpacing      = 3.0f;
	m_nSortColumn    = 0;
	m_vTotalWidth    = 0.0f;
	m_vContentHeight = 0.0f;
	m_nBeginSel      = -1;
	m_nEndSel        = -1;
	m_nFirstSel      = -1;
	m_nLastSel       = -1;
	m_nMouseDownTime = 0;
	m_nLastHitRow    = -1;
	m_bMousDownSeen  = false;
	m_bDragIfMoved   = false;
	m_bAutoScrollSelects = false;
	m_bIsSelecting       = false;
	m_nModeFlags         = nModeFlags;
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool ListViewContainer::IsFocus( void ) const
{
	if ( CView::IsFocus() )
	{
		return( true );
	}
	
	for ( uint i = 0 ; i < m_cColMap.size() ; ++i )
	{
		if ( m_cCols[m_cColMap[i]]->IsFocus() )
		{
			return( true );
		}
	}
	
	return( false );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListViewContainer::DetachedFromWindow( void )
{
	StopScroll();
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListViewContainer::StartScroll( ListView::scroll_direction eDirection, bool bSelect )
{
	m_bAutoScrollSelects = bSelect;

	if ( eDirection == ListView::SCROLL_DOWN )
	{
		if ( m_bAutoScrollDown == false )
		{
			m_bAutoScrollDown = true;
			if ( m_bAutoScrollUp == false )
			{
				CLooper* pcLooper = Looper();
				if ( pcLooper != NULL )
				{
					pcLooper->AddTimer( this, AUTOSCROLL_TIMER, 50000, false );
				}
			}
			else
			{
				m_bAutoScrollUp = false;
			}
		}
	}
	else
	{
		if ( m_bAutoScrollUp == false )
		{
			m_bAutoScrollUp = true;
			if ( m_bAutoScrollDown == false )
			{
				CLooper* pcLooper = Looper();
				if ( pcLooper != NULL )
				{
					pcLooper->AddTimer( this, AUTOSCROLL_TIMER, 50000, false );
				}
			}
			else
			{
				m_bAutoScrollDown = false;
			}
		}
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListViewContainer::StopScroll()
{
	if ( m_bAutoScrollUp || m_bAutoScrollDown )
	{
		m_bAutoScrollUp = m_bAutoScrollDown = false;
		CLooper* pcLooper = Looper();
		if ( pcLooper != NULL )
		{
			pcLooper->RemoveTimer( this, AUTOSCROLL_TIMER );
		}
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListViewContainer::MouseDown( CPoint cPosition )
{
	MakeFocus( true );

	if ( m_cRows.empty() )
	{
		return;
	}

	m_bMousDownSeen = true;

	int        nHitCol = -1;
	int nHitRow;
	if ( cPosition.y >= m_vContentHeight )
	{
		nHitRow = m_cRows.size() - 1;
	}
	else
	{
		nHitRow = GetRowIndex( cPosition.y );
		if ( nHitRow < 0 )
		{
			nHitRow = 0;
		}
	}
	bool  bDoubleClick = false;

	bigtime_t nCurTime = system_time();

	if ( nHitRow == m_nLastHitRow && nCurTime - m_nMouseDownTime < 500000 )
	{
		bDoubleClick = true;
	}
	m_nLastHitRow = nHitRow;
	if ( nHitRow < 0 )
	{
		return;
	}

	ListViewRow* pcHitRow = (nHitRow>=0) ? m_cRows[nHitRow] : NULL;

	m_nMouseDownTime = nCurTime;

	uint32 nQualifiers = CApplication::GetInstance()->GetQualifiers();

	if ( (nQualifiers & QUAL_SHIFT) && (m_nModeFlags & ListView::F_MULTI_SELECT) )
	{
		m_nEndSel   = m_nBeginSel;
	}
	else
	{
		m_nBeginSel = nHitRow;
		m_nEndSel   = nHitRow;
	}

	if ( bDoubleClick )
	{
		m_pcListView->Invoked( m_nFirstSel, m_nLastSel );
		return;
	}
	if ( pcHitRow != NULL )
	{
		for ( uint i = 0 ; i < m_cColMap.size() ; ++i )
		{
			CRect cFrame = GetColumn(i)->Frame();
			if ( cPosition.x >= cFrame.left && cPosition.x < cFrame.right )
			{
				CRect cFrame = GetColumn(i)->Frame();
				cFrame.top    = pcHitRow->m_vYPos;
				cFrame.bottom = cFrame.top + pcHitRow->m_vHeight + m_vVSpacing;
				if ( pcHitRow->HitTest( GetColumn(i), cFrame, i, cPosition ) )
				{
					nHitCol = i;
				}
				break;
			}
		}
	}
	m_bMouseMoved = false;

	if ( nHitCol == 1 && (nQualifiers & QUAL_CTRL) == 0 && m_cRows[nHitRow]->m_bSelected ) {
		m_bDragIfMoved = true;
	}
	else
	{
		ExpandSelect( nHitRow, (nQualifiers & QUAL_CTRL),
					((nQualifiers & (QUAL_CTRL | QUAL_SHIFT)) == 0 || (m_nModeFlags & ListView::F_MULTI_SELECT) == 0) );
		if ( m_nModeFlags & ListView::F_MULTI_SELECT )
		{
			m_bIsSelecting = true;
			m_cSelectRect = CRect( cPosition.x, cPosition.y, cPosition.x, cPosition.y );
			SetDrawingMode( DM_INVERT );
			DrawFrame( m_cSelectRect, FRAME_TRANSPARENT | FRAME_THIN );
			SetDrawingMode( DM_COPY );
		}
	}
	Flush();
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListViewContainer::MouseUp( CPoint cPosition )
{
	m_bMousDownSeen = false;

	CLooper* pcLooper = Looper();
	if ( pcLooper != NULL )
	{
		pcLooper->RemoveTimer( this, AUTOSCROLL_TIMER );
	}

	if ( m_bDragIfMoved )
	{
		ExpandSelect( m_nEndSel, false, true );
		m_bDragIfMoved = false;
	}

	if ( m_bIsSelecting )
	{
		SetDrawingMode( DM_INVERT );
		DrawFrame( m_cSelectRect, FRAME_TRANSPARENT | FRAME_THIN );
		SetDrawingMode( DM_COPY );
		m_bIsSelecting = false;
		Flush();
	}

	// We may have drag data for one of our parents
	
	CMessage cData((uint32)0);
	
	if ( Window()->CurrentMessage()->FindMessage( "_drag_message", &cData ) == EOK )
	{
		CView::MouseUp( cPosition );
		return;
	}
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListViewContainer::MouseMoved( CPoint cNewPos, uint32 nCode, CMessage* pcData )
{
	m_bMouseMoved = true;
	if ( pcData != NULL )
	{
		CView::MouseMoved( cNewPos, nCode, pcData );
		return;
	}

	if ( m_bMousDownSeen == false )
	{
		return;
	}

	if ( m_bDragIfMoved )
	{
		m_bDragIfMoved = false;
		if ( m_pcListView->DragSelection( m_pcListView->ConvertFromScreen( ConvertToScreen( cNewPos ) ) ) ) {
			m_bMousDownSeen = false;
			if ( m_bIsSelecting )
			{
				SetDrawingMode( DM_INVERT );
				DrawFrame( m_cSelectRect, FRAME_TRANSPARENT | FRAME_THIN );
				SetDrawingMode( DM_COPY );
				m_bIsSelecting = false;
			}
			return;
		}
	}

	if ( m_bIsSelecting )
	{
		SetDrawingMode( DM_INVERT );
		DrawFrame( m_cSelectRect, FRAME_TRANSPARENT | FRAME_THIN );
		SetDrawingMode( DM_COPY );
		m_cSelectRect.right = cNewPos.x;
		m_cSelectRect.bottom = cNewPos.y;

		int nHitRow;
		if ( cNewPos.y >= m_vContentHeight )
		{
			nHitRow = m_cRows.size() - 1;
		}
		else
		{
			nHitRow = GetRowIndex( cNewPos.y );
			if ( nHitRow < 0 )
			{
				nHitRow = 0;
			}
		}

		if ( nHitRow != m_nEndSel )
		{
			ExpandSelect( nHitRow, true, false );
		}
		SetDrawingMode( DM_INVERT );
		DrawFrame( m_cSelectRect, FRAME_TRANSPARENT | FRAME_THIN );
		SetDrawingMode( DM_COPY );
		Flush();
		CRect cBounds = Bounds();

		if ( cNewPos.y < cBounds.top + ListView::AUTOSCROLL_BORDER )
		{
			StartScroll( ListView::SCROLL_DOWN, true );
		}
		else if ( cNewPos.y > cBounds.bottom - ListView::AUTOSCROLL_BORDER )
		{
			StartScroll( ListView::SCROLL_UP, true );
		}
		else
		{
			StopScroll();
		}
	}
}


int ListViewContainer::GetRowIndex( float y ) const
{
	if ( y < 0.0f || m_cRows.empty() )
	{
		return( -1 );
	}

	DummyRow cDummy;
	cDummy.m_vYPos = y;

	std::vector<ListViewRow*>::const_iterator cIterator = std::lower_bound( m_cRows.begin(),
																			m_cRows.end(),
																			&cDummy,
																			RowPosPred() );

	int nIndex = cIterator - m_cRows.begin() - 1;
	if ( nIndex >= 0 && y > m_cRows[nIndex]->m_vYPos + m_cRows[nIndex]->m_vHeight + m_vVSpacing )
	{
		nIndex = -1;
	}
	return( nIndex );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListViewContainer::FrameResized( float inWidth, float inHeight )
{
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListViewContainer::TimerTick( int nID )
{
	if ( nID != AUTOSCROLL_TIMER )
	{
		CView::TimerTick( nID );
		return;
	}

	CRect cBounds = Bounds();

	if ( cBounds.Height() >= m_vContentHeight )
	{
		return;
	}
	CPoint cMousePos;
	GetMouse( &cMousePos, NULL );

	float vPrevScroll = GetScrollOffset().y;
	float vCurScroll = vPrevScroll;

	if ( m_bAutoScrollDown )
	{
		float nScrollStep = (cBounds.top + ListView::AUTOSCROLL_BORDER) - cMousePos.y;
		nScrollStep /= 2.0f;
		nScrollStep++;

		vCurScroll += nScrollStep;

		if ( vCurScroll > 0 )
		{
			vCurScroll = 0;
		}
	}
	else if ( m_bAutoScrollUp )
	{
		float vMaxScroll = -(m_vContentHeight - Bounds().Height());

		float vScrollStep = cMousePos.y - (cBounds.bottom - ListView::AUTOSCROLL_BORDER);
		vScrollStep *= 0.5f;
		vScrollStep = vScrollStep + 1.0f;

		vCurScroll  -= vScrollStep;
		cMousePos.y += vScrollStep;

		if ( vCurScroll < vMaxScroll )
		{
			vCurScroll = vMaxScroll;
		}
	}
	if ( vCurScroll != vPrevScroll )
	{
		if ( m_bIsSelecting )
		{
			SetDrawingMode( DM_INVERT );
			DrawFrame( m_cSelectRect, FRAME_TRANSPARENT | FRAME_THIN );
			SetDrawingMode( DM_COPY );
		}
		ScrollTo( 0, vCurScroll );
		if ( m_bIsSelecting ) {
			m_cSelectRect.right  = cMousePos.x;
			m_cSelectRect.bottom = cMousePos.y;

			int nHitRow;
			if ( cMousePos.y >= m_vContentHeight ) {
				nHitRow = m_cRows.size() - 1;
			} else {
				nHitRow = GetRowIndex( cMousePos.y );
				if ( nHitRow < 0 ) {
					nHitRow = 0;
				}
			}

			if ( nHitRow != m_nEndSel ) {
				ExpandSelect( nHitRow, true, false );
			}
			SetDrawingMode( DM_INVERT );
			DrawFrame( m_cSelectRect, FRAME_TRANSPARENT | FRAME_THIN );
			SetDrawingMode( DM_COPY );
		}
		Flush();
	}
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool ListViewContainer::HandleKey( char nChar, uint32 nQualifiers )
{
	if ( m_cRows.empty() )
	{
		return( false );
	}
	
	switch( nChar )
	{
		case 0:
			return( false );
		case VK_PAGE_UP:
		case VK_UP_ARROW:
		{
			int nPageHeight = 0;

			if ( m_nEndSel == 0 ) {
				return( true );
			}
			if ( m_cRows.empty() == false ) {
				nPageHeight = int(Bounds().Height() / (m_vContentHeight / float(m_cRows.size())));
			}

			if ( m_nEndSel == -1 ) {
				m_nEndSel = m_nBeginSel;
			}
			if ( m_nBeginSel == -1 ) {
				m_nBeginSel = m_nEndSel = m_cRows.size() - 1;
				ExpandSelect( m_nEndSel, false, (nQualifiers & QUAL_SHIFT) == 0 || (m_nModeFlags & ListView::F_MULTI_SELECT) == 0 );
			} else {
				if ( (nQualifiers & QUAL_SHIFT) == 0 || (m_nModeFlags & ListView::F_MULTI_SELECT) == 0 ) {
					m_nBeginSel = -1;
				}
				ExpandSelect( m_nEndSel - ((nChar == VK_UP_ARROW) ? 1 : nPageHeight),
							false, (nQualifiers & QUAL_SHIFT) == 0 || (m_nModeFlags & ListView::F_MULTI_SELECT) == 0 );
			}
			if ( (nQualifiers & QUAL_SHIFT) == 0 ) {
				m_nBeginSel = m_nEndSel;
			}
			if ( m_cRows[m_nEndSel]->m_vYPos < Bounds().top ) {
				MakeVisible( m_nEndSel, false );
			}
			Flush();
			return( true );
		}
		case VK_PAGE_DOWN:
		case VK_DOWN_ARROW:
		{
			int nPageHeight = 0;

			if ( m_nEndSel == int(m_cRows.size()) - 1 ) {
				return( true );
			}

			if ( m_cRows.size() > 0 ) {
				nPageHeight = int(Bounds().Height() / (m_vContentHeight / float(m_cRows.size())));
			}

			if ( m_nEndSel == -1 ) {
				m_nEndSel = m_nBeginSel;
			}
			if ( m_nBeginSel == -1 ) {
				m_nBeginSel = m_nEndSel =0;
				ExpandSelect( m_nEndSel, false, (nQualifiers & QUAL_SHIFT) == 0 || (m_nModeFlags & ListView::F_MULTI_SELECT) == 0 );
			} else {
				if ( (nQualifiers & QUAL_SHIFT) == 0 || (m_nModeFlags & ListView::F_MULTI_SELECT) == 0 ) {
					m_nBeginSel = -1;
				}
				ExpandSelect( m_nEndSel + ((nChar == VK_DOWN_ARROW) ? 1 : nPageHeight),
							false, (nQualifiers & QUAL_SHIFT) == 0 || (m_nModeFlags & ListView::F_MULTI_SELECT) == 0 );
			}
			if ( (nQualifiers & QUAL_SHIFT) == 0 )
			{
				m_nBeginSel = m_nEndSel;
			}
			if ( m_cRows[m_nEndSel]->m_vYPos + m_cRows[m_nEndSel]->m_vHeight + m_vVSpacing > Bounds().bottom ) {
				MakeVisible( m_nEndSel, false );
			}
			Flush();
			return( true );
		}
		case '\n':
			if ( m_nFirstSel >= 0 )
			{
				m_pcListView->Invoked( m_nFirstSel, m_nLastSel );
			}
			return( true );
		default:
			return( false );
	}
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListViewContainer::MakeVisible( int nRow, bool bCenter )
{
	float        vTotRowHeight = m_cRows[nRow]->m_vHeight + m_vVSpacing;
	float        y = m_cRows[nRow]->m_vYPos;
	CRect        cBounds = Bounds();
	float        vViewHeight = cBounds.Height() + 1.0f;

	if ( m_vContentHeight <= vViewHeight )
	{
		ScrollTo( 0, 0 );
	}
	else
	{
		if ( bCenter )
		{
			float vOffset = y - vViewHeight * 0.5f + vTotRowHeight * 0.5f;
			if ( vOffset < 0.0f )
			{
				vOffset = 0.0f;
			}
			else if ( vOffset >= m_vContentHeight - 1 )
			{
				vOffset = m_vContentHeight - 1;
			}
			
			if ( vOffset < 0.0f )
			{
				vOffset = 0.0f;
			}
			else if ( vOffset > m_vContentHeight - vViewHeight )
			{
				vOffset = m_vContentHeight - vViewHeight;
			}
			ScrollTo( 0, -vOffset );
		}
		else
		{
			float vOffset;
			if ( y + vTotRowHeight * 0.5f < cBounds.top + vViewHeight * 0.5f )
			{
				vOffset = y;
			}
			else
			{
				vOffset = -(vViewHeight - (y + vTotRowHeight));
			}
			
			if ( vOffset < 0.0f )
			{
				vOffset = 0.0f;
			}
			else if ( vOffset > m_vContentHeight - vViewHeight )
			{
				vOffset = m_vContentHeight - vViewHeight;
			}
			ScrollTo( 0, -vOffset );
		}
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListViewContainer::LayoutColumns()
{
	float x = 0;

	for ( uint i = 0 ; i < m_cColMap.size() ; ++i )
	{
		CRect cFrame = GetColumn(i)->Frame();

		GetColumn(i)->MoveTo( x, 0 );
		if ( i == m_cColMap.size() - 1 )
		{
			x += GetColumn(i)->m_vContentWidth;
		} else {
			x += cFrame.Width() + 1.0f;
		}
	}
	m_vTotalWidth = x;
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

int ListViewContainer::InsertColumn( const char* pzTitle, int nWidth, int nPos )
{
	ListViewCol* pcCol = new ListViewCol( this, CRect( 0, 0, nWidth - 1, 16000000.0f ), pzTitle );
	AddChild( pcCol );
	int nIndex;
	if ( nPos == -1 )
	{
		nIndex = m_cCols.size();
		m_cCols.push_back( pcCol );
	}
	else
	{
		nIndex = nPos;
		m_cCols.insert( m_cCols.begin() + nPos, pcCol );
	}
	m_cColMap.push_back( nIndex );
	LayoutColumns();
	return( nIndex );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

int ListViewContainer::InsertRow( int nPos, ListViewRow* pcRow, bool bUpdate )
{
	int nIndex;

	if ( nPos >= 0 )
	{
		if ( nPos > int(m_cRows.size()) )
		{
			nPos = m_cRows.size();
		}
		m_cRows.insert( m_cRows.begin() + nPos, pcRow );
		nIndex = nPos;
	}
	else
	{
		if ( (m_nModeFlags & ListView::F_NO_AUTO_SORT) == 0 && m_nSortColumn >= 0 ) {
			std::vector<ListViewRow*>::iterator cIterator;

			cIterator = std::lower_bound( m_cRows.begin(),
										m_cRows.end(),
										pcRow,
										RowContentPred( m_nSortColumn ) );
			cIterator = m_cRows.insert( cIterator, pcRow );
			nIndex = cIterator - m_cRows.begin();
		}
		else
		{
			nIndex = m_cRows.size();
			m_cRows.push_back( pcRow );
		}
	}
	
	for ( uint i = 0 ; i < m_cCols.size() ; ++i )
	{
		pcRow->AttachToView( m_cCols[i], i );
	}
	
	for ( uint i = 0 ; i < m_cCols.size() ; ++i )
	{
		float vWidth = pcRow->GetWidth( m_cCols[i], i );
		if ( vWidth > m_cCols[i]->m_vContentWidth )
		{
			m_cCols[i]->m_vContentWidth = vWidth;
		}
	}
	
	pcRow->m_vHeight = pcRow->GetHeight( this );

	m_vContentHeight += pcRow->m_vHeight + m_vVSpacing;

	for ( uint i = nIndex ; i < m_cRows.size() ; ++i )
	{
		float y = 0.0f;
		if ( i > 0 )
		{
			y = m_cRows[i-1]->m_vYPos + m_cRows[i-1]->m_vHeight + m_vVSpacing;
		}
		m_cRows[i]->m_vYPos = y;
	}
	
	if ( bUpdate )
	{
		CRect cBounds = Bounds();

		if ( pcRow->m_vYPos + pcRow->m_vHeight + m_vVSpacing <= cBounds.bottom )
		{
			if ( m_nModeFlags & ListView::F_DONT_SCROLL )
			{
				cBounds.top = pcRow->m_vYPos;
				for ( uint i = 0 ; i < m_cCols.size() ; ++i )
				{
					m_cCols[i]->Invalidate( cBounds );
				}
			}
			else
			{
				cBounds.top = pcRow->m_vYPos + pcRow->m_vHeight + m_vVSpacing;
				cBounds.bottom += pcRow->m_vHeight + m_vVSpacing;
				for ( uint i = 0 ; i < m_cCols.size() ; ++i )
				{
					m_cCols[i]->ScrollRect( cBounds - CPoint( 0.0f, pcRow->m_vHeight + m_vVSpacing ), cBounds );
					m_cCols[i]->Invalidate( CRect( cBounds.left, pcRow->m_vYPos, cBounds.right, cBounds.top ) );
				}
			}
		}
	}
	return( nIndex );
}


ListViewRow* ListViewContainer::RemoveRow( int nIndex, bool bUpdate )
{
	ListViewRow* pcRow = m_cRows[nIndex];
	m_cRows.erase( m_cRows.begin() + nIndex );

	m_vContentHeight -= pcRow->m_vHeight + m_vVSpacing;

	for ( uint i = nIndex ; i < m_cRows.size() ; ++i )
	{
		float y = 0.0f;
		if ( i > 0 ) {
			y = m_cRows[i-1]->m_vYPos + m_cRows[i-1]->m_vHeight + m_vVSpacing;
		}
		m_cRows[i]->m_vYPos = y;
	}

	for ( uint i = 0 ; i < m_cCols.size() ; ++i )
	{
		m_cCols[i]->m_vContentWidth = 0.0f;
		for ( uint j = 0 ; j < m_cRows.size() ; ++j )
		{
			float vWidth = m_cRows[j]->GetWidth( m_cCols[i], i );
			if ( vWidth > m_cCols[i]->m_vContentWidth )
			{
				m_cCols[i]->m_vContentWidth = vWidth;
			}
		}
	}

	if ( bUpdate )
	{
		CRect cBounds = Bounds();

		float vRowHeight = pcRow->m_vHeight + m_vVSpacing;
		float y   = pcRow->m_vYPos;

		if ( y + vRowHeight <= cBounds.bottom )
		{
			if ( m_nModeFlags & ListView::F_DONT_SCROLL )
			{
				cBounds.top = y;
				for ( uint i = 0 ; i < m_cCols.size() ; ++i )
				{
					m_cCols[i]->Invalidate( cBounds );
				}
			}
			else
			{
				cBounds.top = y + vRowHeight;
				cBounds.bottom += vRowHeight;
				for ( uint i = 0 ; i < m_cCols.size() ; ++i )
				{
					m_cCols[i]->ScrollRect( cBounds, cBounds - CPoint( 0.0f, vRowHeight ) );
				}
			}
		}
	}
	
	bool bSelChanged = pcRow->m_bSelected;
	if ( m_nFirstSel != -1 )
	{
		if ( m_nFirstSel == m_nLastSel && nIndex == m_nFirstSel )
		{
			m_nFirstSel = m_nLastSel = -1;
			bSelChanged = true;
		}
		else
		{
			if ( nIndex < m_nFirstSel )
			{
				m_nFirstSel--;
				m_nLastSel--;
				bSelChanged = true;
			}
			else if ( nIndex == m_nFirstSel )
			{
				m_nLastSel--;
				for ( int i = nIndex ; i <= m_nLastSel ; ++i )
				{
					if ( m_cRows[i]->m_bSelected )
					{
						m_nFirstSel = i;
						bSelChanged = true;
						break;
					}
				}
			}
			else if ( nIndex < m_nLastSel )
			{
				m_nLastSel--;
				bSelChanged = true;
			}
			else if ( nIndex == m_nLastSel )
			{
				for ( int i = m_nLastSel - 1 ; i >= m_nFirstSel ; --i )
				{
					if ( m_cRows[i]->m_bSelected )
					{
						m_nLastSel = i;
						bSelChanged = true;
						break;
					}
				}
			}
		}
	}
	
	if ( m_nBeginSel != -1 && m_nBeginSel >= int(m_cRows.size()) )
	{
		m_nBeginSel = m_cRows.size() - 1;
	}
	
	if ( m_nEndSel != -1 && m_nEndSel >= int(m_cRows.size()) )
	{
		m_nEndSel = m_cRows.size() - 1;
	}
	
	if ( bSelChanged )
	{
		m_pcListView->SelectionChanged( m_nFirstSel, m_nLastSel );
	}
	
	return( pcRow );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListViewContainer::Clear()
{
	if ( m_bIsSelecting )
	{
		SetDrawingMode( DM_INVERT );
		DrawFrame( m_cSelectRect, FRAME_TRANSPARENT | FRAME_THIN );
		SetDrawingMode( DM_COPY );
		m_bIsSelecting = false;
		Flush();
	}
	m_bMouseMoved  = true;
	m_bDragIfMoved = false;

	for ( uint i = 0 ; i < m_cRows.size() ; ++i )
	{
		delete m_cRows[i];
	}

	m_cRows.clear();
	m_vContentHeight = 0.0f;

	CRect cBounds = Bounds();
	for ( uint i = 0 ; i < m_cColMap.size() ; ++i )
	{
		GetColumn(i)->Invalidate( cBounds );
	}

	for ( uint i = 0 ; i < m_cCols.size() ; ++i )
	{
		m_cCols[i]->m_vContentWidth = 0.0f;
	}

	m_nBeginSel      = -1;
	m_nEndSel        = -1;
	m_nFirstSel      = -1;
	m_nLastSel       = -1;
	m_nMouseDownTime = 0;
	m_nLastHitRow    = -1;
	ScrollTo( 0, 0 );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListViewContainer::Draw( CRect cUpdateRect )
{
	CRect cFrame = Bounds();
	cFrame.top = 0.0f;

	if ( m_cColMap.size() > 0 )
	{
		CRect cLastCol = GetColumn(m_cColMap.size()-1)->Frame();
		cFrame.left = cLastCol.right + 1.0f;
	}
	SetHighColor( 255, 255, 255 );
	cFrame.bottom = COORD_MAX;
	FillRect( cFrame );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListViewContainer::InvalidateRow( int nRow, uint32 nFlags, bool bImidiate )
{
	float vTotRowHeight = m_cRows[nRow]->m_vHeight + m_vVSpacing;
	float y = m_cRows[nRow]->m_vYPos;
	CRect cFrame( 0.0f, y, COORD_MAX, y + vTotRowHeight - 1.0f );

	if ( Bounds().Contains( cFrame ) == false )
	{
		return;
	}

	if ( bImidiate )
	{
		bool bHasFocus = m_pcListView->IsFocus();

		for ( uint i = 0 ; i < m_cColMap.size() ; ++i )
		{
			CRect cColFrame( cFrame );
			cColFrame.right = GetColumn(i)->Frame().Width();
			m_cRows[nRow]->Draw( cColFrame, GetColumn(i), i, m_cRows[nRow]->m_bSelected, m_cRows[nRow]->m_bHighlighted, bHasFocus );
		}
	}
	else
	{
		for ( uint i = 0 ; i < m_cColMap.size() ; ++i )
		{
			CRect cColFrame( cFrame );
			cColFrame.right = GetColumn(i)->Frame().Width();
			GetColumn(i)->Invalidate( cColFrame );
		}
	}
}


void ListViewContainer::InvertRange( int nStart, int nEnd )
{
	for ( int i = nStart ; i <= nEnd ; ++i )
	{
		if ( m_cRows[i]->m_bIsSelectable )
		{
			m_cRows[i]->m_bSelected = !m_cRows[i]->m_bSelected;
			InvalidateRow( i, ListView::INV_VISUAL );
		}
	}

	if ( m_nFirstSel == -1 )
	{
		m_nFirstSel = nStart;
		m_nLastSel  = nEnd;
	}
	else
	{
		if ( nEnd < m_nFirstSel )
		{
			for ( int i = nStart ; i <= nEnd ; ++i )
			{
				if ( m_cRows[i]->m_bSelected )
				{
					m_nFirstSel = i;
					break;
				}
			}
		}
		else if ( nStart > m_nLastSel )
		{
			for ( int i = nEnd ; i >= nStart ; --i )
			{
				if ( m_cRows[i]->m_bSelected )
				{
					m_nLastSel = i;
					break;
				}
			}
		}
		else
		{
			if ( nStart <= m_nFirstSel )
			{
				m_nFirstSel = m_nLastSel + 1;
				for ( int i = nStart ; i <= m_nLastSel ; ++i )
				{
					if ( m_cRows[i]->m_bSelected )
					{
						m_nFirstSel = i;
						break;
					}
				}
			}

			if ( nEnd >= m_nLastSel )
			{
				m_nLastSel = m_nFirstSel - 1;
				if ( m_nFirstSel < int(m_cRows.size()) )
				{
					for ( int i = nEnd ; i >= m_nFirstSel ; --i )
					{
						if ( m_cRows[i]->m_bSelected )
						{
							m_nLastSel = i;
							break;
						}
					}
				}
			}
		}
	}

	if ( m_nLastSel < m_nFirstSel )
	{
		m_nFirstSel = -1;
		m_nLastSel  = -1;
	}
}


bool ListViewContainer::SelectRange( int nStart, int nEnd, bool bSelect )
{
	bool bChanged = false;

	for ( int i = nStart ; i <= nEnd ; ++i )
	{
		if ( m_cRows[i]->m_bIsSelectable && m_cRows[i]->m_bSelected != bSelect )
		{
			m_cRows[i]->m_bSelected = bSelect;
			InvalidateRow( i, ListView::INV_VISUAL );
			bChanged = true;
		}
	}

	if ( bChanged )
	{
		if ( m_nFirstSel == -1 )
		{
			if ( bSelect )
			{
				m_nFirstSel = nStart;
				m_nLastSel  = nEnd;
			}
		}
		else
		{
			if ( bSelect )
			{
				if ( nStart < m_nFirstSel )
				{
					m_nFirstSel = nStart;
				}
				if ( nEnd > m_nLastSel )
				{
					m_nLastSel = nEnd;
				}
			}
			else
			{
				if ( nEnd >= m_nFirstSel )
				{
					int i = m_nFirstSel;
					m_nFirstSel = m_nLastSel + 1;
					for ( ; i <= m_nLastSel ; ++i )
					{
						if ( m_cRows[i]->m_bSelected )
						{
							m_nFirstSel = i;
							break;
						}
					}
				}
				if ( nStart <= m_nLastSel && m_nLastSel < int(m_cRows.size()) )
				{
					int i = m_nLastSel;
					m_nLastSel = m_nFirstSel - 1;
					for ( ; i >= m_nFirstSel - 1 ; --i )
					{
						if ( m_cRows[i]->m_bSelected )
						{
							m_nLastSel = i;
							break;
						}
					}
				}
			}
			if ( m_nLastSel < m_nFirstSel )
			{
				m_nFirstSel = -1;
				m_nLastSel  = -1;
			}
		}
	}
	return( bChanged );
}


void ListViewContainer::ExpandSelect( int nRow, bool bInvert, bool bClear )
{
	if ( m_cRows.empty() )
	{
		return;
	}

	if ( nRow < 0 )
	{
		nRow = 0;
	}

	if ( nRow >= int(m_cRows.size()) )
	{
		nRow = m_cRows.size() - 1;
	}

	if ( bClear ) {
		if ( m_nFirstSel != -1 ) {
			for ( int i = m_nFirstSel ; i <= m_nLastSel ; ++i ) {
				if ( m_cRows[i]->m_bSelected ) {
					m_cRows[i]->m_bSelected = false;
					InvalidateRow( i, ListView::INV_VISUAL, true );
				}
			}
		}
		m_nFirstSel = -1;
		m_nLastSel  = -1;
		if ( m_nBeginSel == -1 ) {
			m_nBeginSel = nRow;
		}
		m_nEndSel   = nRow;
		bool bChanged;
		if ( m_nBeginSel < nRow ) {
			bChanged = SelectRange( m_nBeginSel, nRow, true );
		} else {
			bChanged = SelectRange( nRow, m_nBeginSel, true );
		}
		if ( bChanged ) {
			m_pcListView->SelectionChanged( m_nFirstSel, m_nLastSel );
		}
		return;
	}
	
	if ( m_nBeginSel == -1 || m_nEndSel == -1 )
	{
		m_nBeginSel = nRow;
		m_nEndSel   = nRow;
		bool bSelChanged;
		if ( bInvert )
		{
			InvertRange( nRow, nRow );
			bSelChanged = true;
		}
		else
		{
			bSelChanged = SelectRange( nRow, nRow, true );
		}
		if ( bSelChanged )
		{
			m_pcListView->SelectionChanged( m_nFirstSel, m_nLastSel );
		}
		return;
	}

	if ( nRow == m_nEndSel )
	{
		bool bSelChanged;
		if ( bInvert )
		{
			InvertRange( nRow, nRow );
			bSelChanged = true;
		}
		else
		{
			bSelChanged = SelectRange( nRow, nRow, true );
		}
		if ( bSelChanged )
		{
			m_pcListView->SelectionChanged( m_nFirstSel, m_nLastSel );
		}
		m_nEndSel = nRow;
		return;
	}
	
	bool bSelChanged = false;
	if ( m_nBeginSel <= m_nEndSel )
	{
		if ( nRow < m_nEndSel )
		{
			if ( bInvert )
			{
				if ( nRow < m_nBeginSel )
				{
					InvertRange( m_nBeginSel + 1, m_nEndSel );
					InvertRange( nRow, m_nBeginSel - 1 );
				}
				else
				{
					InvertRange( nRow + 1, m_nEndSel );
				}
				bSelChanged = true;
			}
			else
			{
				if ( nRow < m_nBeginSel )
				{
					bSelChanged = SelectRange( m_nBeginSel + 1, m_nEndSel, false );
					bSelChanged = SelectRange( nRow, m_nBeginSel - 1, true ) || bSelChanged;
				}
				else
				{
					bSelChanged = SelectRange( nRow + 1, m_nEndSel, false );
				}
			}
		}
		else
		{
			if ( bInvert )
			{
				InvertRange( m_nEndSel + 1, nRow );
				bSelChanged = true;
			}
			else
			{
				bSelChanged = SelectRange( m_nEndSel + 1, nRow, true );
			}
		}
	}
	else
	{
		if ( nRow < m_nEndSel )
		{
			if ( bInvert )
			{
				InvertRange( nRow, m_nEndSel - 1 );
				bSelChanged = true;
			}
			else
			{
				bSelChanged = SelectRange( nRow, m_nEndSel - 1, true );
			}
		} else {
			if ( bInvert )
			{
				if ( nRow > m_nBeginSel )
				{
					InvertRange( m_nEndSel, m_nBeginSel - 1 );
					InvertRange( m_nBeginSel + 1, nRow );
				}
				else
				{
					InvertRange( m_nEndSel, nRow - 1 );
				}
				bSelChanged = true;
			}
			else
			{
				if ( nRow > m_nBeginSel )
				{
					bSelChanged = SelectRange( m_nEndSel, m_nBeginSel - 1, false );
					bSelChanged = SelectRange( m_nBeginSel + 1, nRow, true ) || bSelChanged;
				}
				else
				{
					bSelChanged = SelectRange( m_nEndSel, nRow - 1, false );
				}
			}
		}
	}

	if ( bSelChanged )
	{
		m_pcListView->SelectionChanged( m_nFirstSel, m_nLastSel );
	}
	m_nEndSel = nRow;
}


bool ListViewContainer::ClearSelection()
{
	bool bChanged = false;

	if ( m_nFirstSel != -1 ) {
		for ( int i = m_nFirstSel ; i <= m_nLastSel ; ++i )
		{
			if ( m_cRows[i]->m_bSelected )
			{
				bChanged = true;
				m_cRows[i]->m_bSelected = false;
				InvalidateRow( i, ListView::INV_VISUAL, true );
			}
		}
	}
	m_nFirstSel = -1;
	m_nLastSel  = -1;
	return( bChanged );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

ListViewHeader::ListViewHeader( ListView* pcParent, const CRect& cFrame, uint32 nModeFlags ) :
    CView( cFrame, "header_view", CF_FOLLOW_LEFT | CF_FOLLOW_TOP, 0 )
{
	m_pcParent = pcParent;
	m_nSizeColumn = -1;
	m_nDragColumn = -1;
	m_pcMainView = new ListViewContainer( pcParent, cFrame, nModeFlags );
	AddChild( m_pcMainView );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool ListViewHeader::IsFocus( void ) const
{
	if ( CView::IsFocus() )
		return( true );
	
	return( m_pcMainView->IsFocus() );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListViewHeader::DrawButton( const char* pzTitle, const CRect& cFrame, CFont* pcFont, font_height* psFontHeight )
{
	SetViewColor( get_default_color( COL_LISTVIEW_TAB ) );
	DrawFrame( cFrame, FRAME_RAISED );

	SetHighColor( get_default_color( COL_LISTVIEW_TAB_TEXT ) );
	SetLowColor( get_default_color( COL_LISTVIEW_TAB ) );

	float vFontHeight = psFontHeight->ascender + psFontHeight->descender;

	int nStrLen = pcFont->GetStringLength( pzTitle, cFrame.Width() - 9.0f );
	DrawString( pzTitle, nStrLen,
				cFrame.LeftTop() + CPoint( 5, (cFrame.Height()+1.0f) / 2 - vFontHeight / 2 + psFontHeight->ascender ) );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListViewHeader::Draw( CRect cUpdateRect )
{
	CFont* pcFont = GetFont();
	if ( pcFont == NULL )
	{
		return;
	}

	font_height sHeight;
	pcFont->GetHeight( &sHeight );

	CRect cFrame;
	for ( uint i = 0 ; i < m_pcMainView->m_cColMap.size() ; ++i )
	{
		ListViewCol* pcCol = m_pcMainView->GetColumn( i );
		cFrame = pcCol->Frame();
		cFrame.top = 0;
		cFrame.bottom = sHeight.ascender + sHeight.descender + 6 - 1;
		if ( i == m_pcMainView->m_cColMap.size() - 1 )
		{
			cFrame.right = COORD_MAX;
		}
		DrawButton( pcCol->m_cTitle.c_str(), cFrame, pcFont, &sHeight );
	}/*
	cFrame.left = cFrame.right + 1;
	cFrame.right = Bounds().right;
	if ( cFrame.IsValid() ) {
		DrawButton( "", cFrame, pcFont, &sHeight );
	}*/
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListViewHeader::MouseMoved( CPoint cNewPos, uint32 nCode, CMessage* pcData )
{
	if ( pcData != NULL )
	{
		CView::MouseMoved( cNewPos, nCode, pcData );
		return;
	}

	if ( m_nDragColumn != -1 )
	{
		if ( m_nDragColumn < int(m_pcMainView->m_cColMap.size()) )
		{
			CMessage cData( (uint32)0 );
			cData.AddInt32( "column", m_pcMainView->m_cColMap[m_nDragColumn] );
			CRect cFrame( m_pcMainView->GetColumn( m_nDragColumn )->Frame() );
//            if ( m_nDragColumn == int(m_pcMainView->m_cColMap.size()) - 1 )
//			  {
//                cFrame.right = cFrame.left + m_pcMainView->GetColumn( m_nDragColumn )->m_vContentWidth;
//            }
			m_pcMainView->ConvertToParent( &cFrame );
			ConvertFromParent( &cFrame );
			cFrame.top = 0.0f;
			cFrame.bottom = m_vHeaderHeight - 1.0f;

			BeginDrag( &cData, cNewPos - cFrame.LeftTop() - GetScrollOffset(), cFrame.Bounds(), this );
		}
		m_nDragColumn = -1;
	}

	if ( m_nSizeColumn == -1 )
	{
		return;
	}
	float vColWidth = m_pcMainView->GetColumn( m_nSizeColumn )->Frame().Width();
	float vDeltaSize = cNewPos.x - m_cHitPos.x;
	if ( vColWidth + vDeltaSize < 4.0f )
	{
		vDeltaSize = 4.0f - vColWidth;
	}
	m_pcMainView->GetColumn( m_nSizeColumn )->ResizeBy( vDeltaSize, 0 );

	ListViewCol* pcCol = m_pcMainView->m_cCols[m_pcMainView->m_cColMap[m_pcMainView->m_cColMap.size() - 1]];
	CRect cFrame = pcCol->Frame();
	cFrame.right = m_pcParent->Bounds().right - GetScrollOffset().x;
	pcCol->SetFrame( cFrame );

	m_pcMainView->LayoutColumns();
	m_pcParent->AdjustScrollBars( false );
	Draw( Bounds() );
	Flush();
	m_cHitPos = cNewPos;
}


void ListViewHeader::ViewScrolled( const CPoint& cDelta )
{
	if ( m_pcMainView->m_cColMap.empty() )
	{
		return;
	}
	ListViewCol* pcCol = m_pcMainView->m_cCols[m_pcMainView->m_cColMap[m_pcMainView->m_cColMap.size() - 1]];
	CRect cFrame = pcCol->Frame();
	cFrame.right = m_pcParent->Bounds().right - GetScrollOffset().x;
	pcCol->SetFrame( cFrame );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListViewHeader::MouseDown( CPoint cPosition )
{
	if ( m_pcMainView->m_cColMap.empty() )
	{
		return;
	}

	if ( cPosition.y >= m_pcMainView->Frame().top )
	{
		return;
	}

	CPoint cMVPos = m_pcMainView->ConvertFromParent( cPosition );
	for ( uint i = 0 ; i < m_pcMainView->m_cColMap.size() ; ++i )
	{
		CRect cFrame( m_pcMainView->GetColumn(i)->Frame() );

		if ( cMVPos.x >= cFrame.left && cMVPos.x <= cFrame.right )
		{
			if ( i > 0 && cMVPos.x >= cFrame.left && cMVPos.x <= cFrame.left + 5 )
			{
				m_nSizeColumn = i - 1;
				m_cHitPos = cPosition;
				MakeFocus( true );
				break;
			}
			if ( cMVPos.x >= cFrame.right - 5 && cMVPos.x <= cFrame.right )
			{
				m_nSizeColumn = i;
				m_cHitPos = cPosition;
				MakeFocus( true );
				break;
			}
			m_nDragColumn = i;
			break;
		}
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListViewHeader::MouseUp( CPoint cPosition )
{
	m_nSizeColumn = -1;

	if ( m_nDragColumn != -1 && m_nDragColumn < int(m_pcMainView->m_cColMap.size()) ) {
		m_pcMainView->m_nSortColumn = m_pcMainView->m_cColMap[m_nDragColumn];

		if ( (m_pcMainView->m_nModeFlags & ListView::F_NO_AUTO_SORT) == 0 )
		{
			m_pcParent->Sort();
			CRect cBounds = m_pcMainView->Bounds();
			for ( uint j = 0 ; j < m_pcMainView->m_cColMap.size() ; ++j )
			{
				if ( m_pcMainView->GetColumn(j) != NULL )
				{
					m_pcMainView->GetColumn(j)->Invalidate( cBounds );
				}
			}
			Flush();
		}
	}
	
	m_nDragColumn = -1;

	CMessage pcData((uint32)0);
	
	if ( Window()->CurrentMessage()->FindMessage( "_drag_message", &pcData ) == EOK )
	{
		int nColumn;
		if ( pcData.FindInt( "column", &nColumn ) == 0 )
		{
			CPoint cMVPos = m_pcMainView->ConvertFromParent( cPosition );
			for ( uint i = 0 ; i < m_pcMainView->m_cColMap.size() ; ++i )
			{
				CRect cFrame( m_pcMainView->GetColumn(i)->Frame() );

				if ( cMVPos.x >= cFrame.left && cMVPos.x <= cFrame.right )
				{
					int j;
					if ( cMVPos.x < cFrame.left + cFrame.Width() * 0.5f )
					{
						j = i;
					}
					else
					{
						j = i + 1;
					}
					
					m_pcMainView->m_cColMap.insert( m_pcMainView->m_cColMap.begin() + j, nColumn );
					for ( int k = 0 ; k < int(m_pcMainView->m_cColMap.size()) ; ++k ) {
						if ( k != j && m_pcMainView->m_cColMap[k] == nColumn )
						{
							m_pcMainView->m_cColMap.erase( m_pcMainView->m_cColMap.begin() + k );
							break;
						}
					}
					m_pcMainView->LayoutColumns();
					Draw( Bounds() );
					Flush();
				}
			}
		}
		else
		{
			CView::MouseUp( cPosition );
		}
		return;
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListViewHeader::FrameResized( float inWidth, float inHeight )
{
	CRect cBounds( Bounds() );
	bool  bNeedFlush = false;

	if ( inWidth != 0.0f )
	{
		CRect cDamage = cBounds;

		cDamage.left = cDamage.right - max( 2.0f, inWidth + 1.0f );
		Invalidate( cDamage );
		bNeedFlush = true;
	}

	if ( inHeight != 0.0f )
	{
		CRect cDamage = cBounds;

		cDamage.top = cDamage.bottom - max( 2.0f, inHeight + 1.0f );
		Invalidate( cDamage );
		bNeedFlush = true;
	}

	Layout();

	if ( bNeedFlush )
		Flush();
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListViewHeader::Layout()
{
	font_height sHeight;
	GetFontHeight( &sHeight );

	if ( m_pcMainView->m_nModeFlags & ListView::F_NO_HEADER )
	{
		m_vHeaderHeight = 0.0f;
	}
	else
	{
		m_vHeaderHeight = sHeight.ascender + sHeight.descender + 6.0f;
	}

	CRect cFrame = Frame().Bounds();
	cFrame.Floor();
	cFrame.top += m_vHeaderHeight;
	cFrame.right = COORD_MAX;
	m_pcMainView->SetFrame( cFrame );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

ListView::ListView( CRect inFrame, const char* inTitle, uint32 inModeFlags, uint32 inResizeMask, uint32 inFlags ) :
        CControl( inFrame, inTitle, "", NULL, inResizeMask, inFlags )
{
	m_pcHeaderView   = new ListViewHeader( this, CRect( 0, 0, 1, 1 ), inModeFlags );
	m_pcMainView     = m_pcHeaderView->m_pcMainView;
	m_pcVScroll      = NULL;
	m_pcHScroll      = NULL;
	m_pcSelChangeMsg = NULL;
	m_pcInvokeMsg    = NULL;
//    m_nModeFlags     = inModeFlags;

	AddChild( m_pcHeaderView );
	Layout();
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

ListView::~ListView()
{
    delete m_pcSelChangeMsg;
    delete m_pcInvokeMsg;
}


void ListView::LabelChanged( const std::string& cNewLabel )
{
}


void ListView::EnableStatusChanged( bool bIsEnabled )
{
}


bool ListView::Invoked( CMessage* pcMessage )
{
    CControl::Invoked( pcMessage );
    return( true );
}


bool ListView::IsMultiSelect() const
{
    return( m_pcMainView->m_nModeFlags & F_MULTI_SELECT );
}


void ListView::SetMultiSelect( bool bMulti )
{
	if ( bMulti )
	{
		m_pcMainView->m_nModeFlags |= F_MULTI_SELECT;
	}
	else
	{
		m_pcMainView->m_nModeFlags &= ~F_MULTI_SELECT;
	}
}


bool ListView::IsAutoSort() const
{
	return( (m_pcMainView->m_nModeFlags & F_NO_AUTO_SORT) == 0 );
}


void ListView::SetAutoSort( bool bAuto )
{
	if ( bAuto )
	{
		m_pcMainView->m_nModeFlags &= ~F_NO_AUTO_SORT;
	}
	else
	{
		m_pcMainView->m_nModeFlags |= F_NO_AUTO_SORT;
	}
}


bool ListView::HasBorder() const
{
	return( m_pcMainView->m_nModeFlags & F_RENDER_BORDER );
}


void ListView::SetRenderBorder( bool bRender )
{
	if ( bRender )
	{
		m_pcMainView->m_nModeFlags |= F_RENDER_BORDER;
	}
	else
	{
		m_pcMainView->m_nModeFlags &= ~F_RENDER_BORDER;
	}

	Layout();
	Flush();
}


bool ListView::HasColumnHeader() const
{
	return( (m_pcMainView->m_nModeFlags & F_NO_HEADER) == 0 );
}


void ListView::SetHasColumnHeader( bool bFlag )
{
	if ( bFlag )
	{
		m_pcMainView->m_nModeFlags &= ~F_NO_HEADER;
	}
	else
	{
		m_pcMainView->m_nModeFlags |= F_NO_HEADER;
	}

	Layout();
	Flush();
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool ListView::IsFocus( void ) const
{
	if ( CView::IsFocus() )
	{
		return( true );
	}

	return( m_pcHeaderView->IsFocus() );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListView::AllAttached()
{
	CView::AllAttached();
	SetTarget( Window() );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListView::KeyDown( const char* pzString, const char* pzRawString, uint32 nQualifiers )
{
	if ( m_pcMainView->HandleKey( pzString[0], nQualifiers ) == false )
	{
		CView::KeyDown( pzString, pzRawString, nQualifiers );
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListView::SetSelChangeMsg( CMessage* pcMsg )
{
	if ( pcMsg != m_pcSelChangeMsg )
	{
		delete m_pcSelChangeMsg;
		m_pcSelChangeMsg = pcMsg;
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListView::SetInvokeMsg( CMessage* pcMsg )
{
	if ( pcMsg != m_pcInvokeMsg )
	{
		delete m_pcInvokeMsg;
		m_pcInvokeMsg = pcMsg;
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CMessage* ListView::GetSelChangeMsg() const
{
	return( m_pcSelChangeMsg );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CMessage* ListView::GetInvokeMsg() const
{
	return( m_pcInvokeMsg );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
//        Calculate and set the range and proportion of scroll bars.
//        Scroll the view so upper/left corner
//        is at or above/left of 0,0 if needed
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListView::AdjustScrollBars( bool bOkToHScroll )
{
	if ( m_pcMainView->m_cRows.size() == 0 )
	{
		if ( m_pcVScroll != NULL )
		{
			RemoveChild( m_pcVScroll );
			delete m_pcVScroll;
		}

		if ( m_pcHScroll != NULL )
		{
			RemoveChild( m_pcHScroll );
			delete m_pcHScroll;
		}

		if ( m_pcVScroll != NULL || m_pcHScroll != NULL )
		{
			m_pcVScroll = NULL;
			m_pcHScroll = NULL;
			m_pcMainView->ScrollTo( 0, 0 );
			m_pcHeaderView->ScrollTo( 0, 0 );
			Layout();
		}
	}
	else
	{
		float vViewHeight    = m_pcMainView->Bounds().Height() + 1.0f;
		float vViewWidth     = m_pcHeaderView->Bounds().Width() + 1.0f;
		float vContentHeight = m_pcMainView->m_vContentHeight;

		float vProportion;
		if ( vContentHeight > 0 && vContentHeight > vViewHeight  )
		{
			vProportion = vViewHeight / vContentHeight;
		}
		else
		{
			vProportion = 1.0f;
		}

		if ( vContentHeight > vViewHeight )
		{
			if ( m_pcVScroll == NULL )
			{
				m_pcVScroll  = new CScrollBar( CRect( -1, -1, 0, 0 ), "v_scroll", NULL, 0, 1000, VERTICAL );
				AddChild( m_pcVScroll );
				m_pcVScroll->SetTarget( m_pcMainView );
				Layout();
			}
			else
			{
				m_pcVScroll->SetSteps( ceil( vContentHeight / float(m_pcMainView->m_cRows.size()) ), ceil( vViewHeight * 0.8f ) );
				m_pcVScroll->SetProportion( vProportion );
				m_pcVScroll->SetRange( 0, max( vContentHeight - vViewHeight, 0.0f ) );
			}
		}
		else
		{
			if ( m_pcVScroll != NULL )
			{
				RemoveChild( m_pcVScroll );
				delete m_pcVScroll;
				m_pcVScroll = NULL;
				m_pcMainView->ScrollTo( 0, 0 );
				Layout();
			}
		}

		if ( m_pcVScroll != NULL )
		{
			vViewWidth -= m_pcVScroll->Frame().Width();
		}

		if ( m_pcMainView->m_vTotalWidth > 0 && m_pcMainView->m_vTotalWidth > vViewWidth ) {
			vProportion = vViewWidth / m_pcMainView->m_vTotalWidth;
		}
		else
		{
			vProportion = 1.0f;
		}

		if ( m_pcMainView->m_vTotalWidth > vViewWidth )
		{
			if ( m_pcHScroll == NULL )
			{
				m_pcHScroll  = new CScrollBar( CRect( -1, -1, 0, 0 ), "h_scroll", NULL, 0, 1000, HORIZONTAL );
				AddChild( m_pcHScroll );
				m_pcHScroll->SetTarget( m_pcHeaderView );
				Layout();
			}
			else
			{
				m_pcHScroll->SetSteps( 15.0f, ceil( vViewWidth * 0.8f ) );
				m_pcHScroll->SetProportion( vProportion );
				m_pcHScroll->SetRange( 0, m_pcMainView->m_vTotalWidth - vViewWidth );
			}
		}
		else
		{
			if ( m_pcHScroll != NULL )
			{
				RemoveChild( m_pcHScroll );
				delete m_pcHScroll;
				m_pcHScroll = NULL;
				m_pcHeaderView->ScrollTo( 0, 0 );
				Layout();
			}
		}

		if ( bOkToHScroll )
		{
			float nOff = m_pcHeaderView->GetScrollOffset().x;
			if ( nOff < 0 )
			{
				vViewWidth  = m_pcHeaderView->Bounds().Width() + 1.0f;
				if ( vViewWidth - nOff > m_pcMainView->m_vTotalWidth )
				{
					float nDeltaScroll = min( (vViewWidth - nOff) - m_pcMainView->m_vTotalWidth, -nOff );
					m_pcHeaderView->ScrollBy( nDeltaScroll, 0 );
				}
				else if ( vViewWidth > m_pcMainView->m_vTotalWidth )
				{
					m_pcHeaderView->ScrollBy( -nOff, 0 );
				}
			}
		}

		float nOff = m_pcMainView->GetScrollOffset().y;
		if ( nOff < 0 )
		{
			vViewHeight = m_pcMainView->Bounds().Height() + 1.0f;
			if ( vViewHeight - nOff > vContentHeight )
			{
				float nDeltaScroll = min( (vViewHeight - nOff) - vContentHeight, -nOff );
				m_pcMainView->ScrollBy( 0, nDeltaScroll );
			}
			else if ( vViewHeight > vContentHeight )
			{
				m_pcMainView->ScrollBy( 0, -nOff );
			}
		}
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListView::Layout()
{
	font_height sHeight;

	GetFontHeight( &sHeight );

	float nTopHeight = m_pcHeaderView->m_vHeaderHeight;

	CRect cFrame( Bounds().Bounds() );
	cFrame.Floor();
	if ( m_pcMainView->m_nModeFlags & F_RENDER_BORDER )
	{
		cFrame.Resize( 2.0f, 2.0f, -2.0f, -2.0f );
	}

	CRect cHeaderFrame( cFrame );

	if ( m_pcHScroll != NULL )
	{
		cHeaderFrame.bottom -= 16.0f;
	}

	if ( m_pcVScroll != NULL )
	{
		CRect cVScrFrame( cFrame );

		cVScrFrame.top    += nTopHeight;
//        cVScrFrame.bottom -= 2.0f;
		if ( m_pcHScroll != NULL ) {
			cVScrFrame.bottom -= 16.0f;
		}
		cVScrFrame.left = cVScrFrame.right - 16.0f;

		m_pcVScroll->SetFrame( cVScrFrame );
	}

	if ( m_pcHScroll != NULL )
	{
		CRect cVScrFrame( cFrame );
		cHeaderFrame.bottom = floor( cHeaderFrame.bottom );
		if ( m_pcVScroll != NULL ) {
			cVScrFrame.right -= 16.0f;
		}
//        cVScrFrame.left  += 2.0f;
//        cVScrFrame.right -= 2.0f;
		cVScrFrame.top = cVScrFrame.bottom - 16.0f;
		cHeaderFrame.bottom = cVScrFrame.top;

		m_pcHScroll->SetFrame( cVScrFrame );
	}
	m_pcHeaderView->SetFrame( cHeaderFrame );
	AdjustScrollBars();
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListView::FrameResized( float inWidth, float inHeight )
{
	if ( m_pcMainView->m_cColMap.empty() == false )
	{
		ListViewCol* pcCol = m_pcMainView->m_cCols[m_pcMainView->m_cColMap[m_pcMainView->m_cColMap.size() - 1]];
		CRect cFrame = pcCol->Frame();
		cFrame.right = Bounds().right - m_pcHeaderView->GetScrollOffset().x;

		pcCol->SetFrame( cFrame );
	}

	Layout();
	Flush();
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListView::Sort()
{
	if ( uint(m_pcMainView->m_nSortColumn) >= m_pcMainView->m_cCols.size() )
	{
		return;
	}

	if ( m_pcMainView->m_cRows.empty() )
	{
		return;
	}
	std::sort( m_pcMainView->m_cRows.begin(),
			m_pcMainView->m_cRows.end(),
			RowContentPred( m_pcMainView->m_nSortColumn ) );

	float y = 0.0f;
	for ( uint i = 0 ; i < m_pcMainView->m_cRows.size() ; ++i )
	{
		m_pcMainView->m_cRows[i]->m_vYPos = y;
		y += m_pcMainView->m_cRows[i]->m_vHeight + m_pcMainView->m_vVSpacing;
	}

	if ( m_pcMainView->m_nFirstSel != -1 )
	{
		for ( uint i = 0 ; i < m_pcMainView->m_cRows.size() ; ++i )
		{
			if ( m_pcMainView->m_cRows[i]->m_bSelected )
			{
				m_pcMainView->m_nFirstSel = i;
				break;
			}
		}
		for ( int i = m_pcMainView->m_cRows.size() - 1 ; i >= 0 ; --i )
		{
			if ( m_pcMainView->m_cRows[i]->m_bSelected ) {
				m_pcMainView->m_nLastSel = i;
				break;
			}
		}
	}

	for ( uint i = 0 ; i < m_pcMainView->m_cCols.size() ; ++i )
	{
		m_pcMainView->m_cCols[i]->Invalidate();
	}
	Flush();
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListView::MakeVisible( int nRow, bool bCenter )
{
	m_pcMainView->MakeVisible( nRow, bCenter );
	Flush();
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

int ListView::InsertColumn( const char* pzTitle, int nWidth, int nPos )
{
	int nColumn = m_pcMainView->InsertColumn( pzTitle, nWidth, nPos );
	Layout();
	Flush();
	return( nColumn );
}

const std::vector<int>& ListView::GetColumnMapping() const
{
	return( m_pcMainView->m_cColMap );
}

void ListView::SetColumnMapping( const std::vector<int>& cMap )
{
	for ( int i = 0 ; i < int(m_pcMainView->m_cCols.size()) ; ++i )
	{
		if ( std::find( cMap.begin(), cMap.end(), i ) == cMap.end() )
		{
			if ( std::find( m_pcMainView->m_cColMap.begin(), m_pcMainView->m_cColMap.end(), i ) != m_pcMainView->m_cColMap.end() )
			{
				m_pcMainView->m_cCols[i]->Show( false );
			}
		}
		else
		{
			if ( std::find( m_pcMainView->m_cColMap.begin(), m_pcMainView->m_cColMap.end(), i ) == m_pcMainView->m_cColMap.end() )
			{
				m_pcMainView->m_cCols[i]->Show( true );
			}
		}
	}
	m_pcMainView->m_cColMap = cMap;
	m_pcMainView->LayoutColumns();
	m_pcHeaderView->Invalidate();
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListView::InsertRow( int nPos, ListViewRow* pcRow, bool bUpdate )
{
	m_pcMainView->InsertRow( nPos, pcRow, bUpdate );
	AdjustScrollBars();
	if ( bUpdate )
	{
		Flush();
	}
}

void ListView::InsertRow( ListViewRow* pcRow, bool bUpdate )
{
	InsertRow( -1, pcRow, bUpdate );
}

ListViewRow* ListView::RemoveRow( int nIndex, bool bUpdate )
{
	ListViewRow* pcRow = m_pcMainView->RemoveRow( nIndex, bUpdate );
	AdjustScrollBars();
	if ( bUpdate )
	{
		Flush();
	}
	return( pcRow );
}

void ListView::InvalidateRow( int nRow, uint32 nFlags )
{
	m_pcMainView->InvalidateRow( nRow, nFlags );
	Flush();
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

uint ListView::GetRowCount() const
{
	return( m_pcMainView->m_cRows.size() );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

ListViewRow* ListView::GetRow( uint nIndex ) const
{
	if ( nIndex < m_pcMainView->m_cRows.size() )
	{
		return( m_pcMainView->m_cRows[nIndex] );
	}

	return( NULL );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

ListViewRow* ListView::GetRow( const CPoint& cPos ) const
{
	int nHitRow = m_pcMainView->GetRowIndex( cPos.y );
	if ( nHitRow >= 0 )
	{
		return( m_pcMainView->m_cRows[nHitRow] );
	}

	return( NULL );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

int ListView::HitTest( const CPoint& cPos ) const
{
	CPoint cParentPos = ConvertToScreen( m_pcMainView->ConvertFromScreen( cPos ) );
	int nHitRow = m_pcMainView->GetRowIndex( cParentPos.y );
	if ( nHitRow >= 0 )
	{
		return( nHitRow );
	}

	return( -1 );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

float ListView::GetRowPos( int nRow )
{
	if ( nRow >= 0 && nRow < int(m_pcMainView->m_cRows.size()) )
	{
		return( m_pcMainView->m_cRows[nRow]->m_vYPos + m_pcMainView->Frame().top + m_pcMainView->GetScrollOffset().y );
	}

	return( 0.0f );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListView::Select( int nFirst, int nLast, bool bReplace, bool bSelect )
{
	if ( m_pcMainView->m_cRows.empty() )
	{
		return;
	}

	if ( (bReplace || (m_pcMainView->m_nModeFlags & F_MULTI_SELECT) == 0) && m_pcMainView->m_nFirstSel != -1 )
	{
		for ( int i = m_pcMainView->m_nFirstSel ; i <= m_pcMainView->m_nLastSel ; ++i )
		{
			if ( m_pcMainView->m_cRows[i]->m_bSelected )
			{
				m_pcMainView->m_cRows[i]->m_bSelected = false;
				m_pcMainView->InvalidateRow( i, INV_VISUAL );
			}
		}
		m_pcMainView->m_nFirstSel = -1;
		m_pcMainView->m_nLastSel  = -1;
	}

	if ( m_pcMainView->SelectRange( nFirst, nLast, bSelect ) )
	{
		SelectionChanged( m_pcMainView->m_nFirstSel, m_pcMainView->m_nLastSel );
	}

	Flush();
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListView::Select( int nRow, bool bReplace, bool bSelect )
{
	Select( nRow, nRow, bReplace, bSelect );
}

void ListView::ClearSelection()
{
	if ( m_pcMainView->ClearSelection() )
	{
		SelectionChanged( -1, -1 );
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListView::Highlight( int nFirst, int nLast, bool bReplace, bool bHighlight )
{
	if ( nLast < 0 )
	{
		return;
	}

	if ( nFirst < 0 )
	{
		nFirst = 0;
	}

	if ( nFirst >= int(m_pcMainView->m_cRows.size()) )
	{
		return;
	}

	if ( nLast >= int(m_pcMainView->m_cRows.size()) )
	{
		nLast = m_pcMainView->m_cRows.size() - 1;
	}

	if ( bReplace )
	{
		for ( int i = 0 ; i < int( m_pcMainView->m_cRows.size() ) ; ++i )
		{
			bool bHigh = (i >= nFirst && i <= nLast);
			if ( m_pcMainView->m_cRows[i]->m_bHighlighted != bHigh )
			{
				m_pcMainView->m_cRows[i]->m_bHighlighted = bHigh;
				m_pcMainView->InvalidateRow( i, ListView::INV_VISUAL );
			}
		}
	}
	else
	{
		for ( int i = nFirst ; i <= nLast ; ++i )
		{
			if ( m_pcMainView->m_cRows[i]->m_bHighlighted != bHighlight )
			{
				m_pcMainView->m_cRows[i]->m_bHighlighted = bHighlight;
				m_pcMainView->InvalidateRow( i, ListView::INV_VISUAL );
			}
		}
	}

	Flush();
}

void ListView::SetCurrentRow( int nRow )
{
	if ( m_pcMainView->m_cRows.empty() )
	{
		return;
	}

	if ( nRow < 0 )
	{
		nRow = 0;
	}
	else if ( nRow >= int(m_pcMainView->m_cRows.size()) )
	{
		nRow = m_pcMainView->m_cRows.size() - 1;
	}

	m_pcMainView->m_nEndSel = nRow;
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListView::Highlight( int nRow, bool bReplace, bool bHighlight )
{
	Highlight( nRow, nRow, bReplace, bHighlight );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool ListView::IsSelected( uint nRow ) const
{
	if ( nRow < m_pcMainView->m_cRows.size() )
	{
		return( m_pcMainView->m_cRows[nRow]->m_bSelected );
	}

	return( false );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListView::Clear()
{
	m_pcMainView->Clear();
	AdjustScrollBars();
	Flush();
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListView::Draw( CRect cUpdateRect )
{
	if ( m_pcVScroll != NULL && m_pcHScroll != NULL )
	{
		CRect cHFrame = m_pcHScroll->Frame();
		CRect cVFrame = m_pcVScroll->Frame();
		CRect cFrame( cHFrame.right + 1, cVFrame.bottom + 1, cVFrame.right, cHFrame.bottom );
		FillRect( cFrame, get_default_color( COL_NORMAL ) );
	}

	if ( m_pcMainView->m_nModeFlags & F_RENDER_BORDER )
	{
		DrawFrame( CRect(Bounds()), FRAME_RECESSED | FRAME_TRANSPARENT );
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListView::Invoked( int nFirstRow, int nLastRow )
{
	if ( m_pcInvokeMsg != NULL )
	{
		Invoke( m_pcInvokeMsg );
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListView::SelectionChanged( int nFirstRow, int nLastRow )
{
	if ( m_pcSelChangeMsg != NULL )
	{
		Invoke( m_pcSelChangeMsg );
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

int ListView::GetFirstSelected() const
{
	return( m_pcMainView->m_nFirstSel );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListView::StartScroll( scroll_direction eDirection, bool bSelect )
{
	m_pcMainView->StartScroll( eDirection, bSelect );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void ListView::StopScroll()
{
	m_pcMainView->StopScroll();
}
  
//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

int ListView::GetLastSelected() const
{
	return( m_pcMainView->m_nLastSel );
}

bool ListView::DragSelection( const CPoint& cPos )
{
	return( false );
}


ListView::const_iterator ListView::begin() const
{
	return( m_pcMainView->m_cRows.begin() );
}

ListView::const_iterator ListView::end() const
{
	return( m_pcMainView->m_cRows.end() );
}

void ListView::__reserved1__() {}
void ListView::__reserved2__() {}
void ListView::__reserved3__() {}
void ListView::__reserved4__() {}
