/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */

 
#include <stdio.h>
#include <assert.h>
#include <interface/Screen.h>
#include <interface/Window.h>
#include <interface/Menu.h>
#include <interface/MenuItem.h>
#include <interface/Font.h>
#include <app/Message.h>
#include <support/Autolock.h>

#include <cosmoe/kernel.h>



namespace os {

class MenuWindow : public CWindow
{
public:
	MenuWindow( CRect inFrame, CMenu* pcMenu );
	~MenuWindow();

	void ShutDown()
	{
		if ( m_pcMenu != NULL )
		{
			RemoveChild( m_pcMenu );
			m_pcMenu = NULL;
		}
		PostMessage( M_TERMINATE );
	}
private:
	CMenu* m_pcMenu;
};

}

using namespace os;


CLocker CMenu::s_cMutex( "menu_mutex" );







class CMenu::Private
{
public:
	Private() : m_cMutex( "menu_mutex" ) {}
	std::string m_cTitle;
	CMessage    m_cCloseMsg;
	CMessenger  m_cCloseMsgTarget;
	CLocker     m_cMutex;
};




//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

MenuWindow::MenuWindow( CRect inFrame, CMenu* pcMenu ) :
    CWindow( inFrame, "menu", WT_UNTYPED_WINDOW, WND_NO_BORDER | WND_SYSTEM )
{
	// If the machine is very loaded the previous window hosting the menu
	// might still own it, so we must wait until it terminates.
	while( pcMenu->Window() != NULL )
	{
		snooze( 50000 );
	}

	m_pcMenu = pcMenu;
	AddChild( pcMenu );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

MenuWindow::~MenuWindow()
{
	if ( m_pcMenu != NULL )
	{
		RemoveChild( m_pcMenu );
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CMenu::CMenu( const char* inTitle,
			  MenuLayout_e inLayout )
    : CView( CRect(0,0,10,10), "_menu_", 0, 0 )
{
	m = new Private;
	mIsTracking       = false;
	mItemMargins      = CRect( 3, 2, 3, 2 );
	mMenuLayout       = inLayout;
	mMenuWindow       = NULL;
	mFirstMenuItem    = NULL;
	mItemCount        = 0;
	mSuperItem        = NULL;
	mHasOpenChildren  = false;
	mCloseOnMouseUp   = false;
	m_hTrackPort      = -1;
	mIsRadioMode      = false;

	if ( NULL != inTitle )
	{
		m->m_cTitle = inTitle;
	}
	else
	{
		m->m_cTitle = "";
	}

	mRootMenu = this;
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CMenu::CMenu( CRect inFrame,
			  const char* inTitle,
			  uint32 nResizeMask,
			  uint32 inFlags,
			  MenuLayout_e inLayout,
			  bool inResizeToFit )
    : CView( inFrame, "_menu_", nResizeMask, inFlags & ~WID_FULL_UPDATE_ON_RESIZE )
{
	m = new Private;
	mIsTracking       = false;
	mItemMargins      = CRect( 3, 2, 3, 2 );
	mMenuLayout       = inLayout;
	mMenuWindow       = NULL;
	mFirstMenuItem    = NULL;
	mItemCount        = 0;
	mSuperItem        = NULL;
	mHasOpenChildren  = false;
	mCloseOnMouseUp   = false;
	m_hTrackPort      = -1;
	mIsRadioMode      = false;

	if ( NULL != inTitle )
	{
		m->m_cTitle = inTitle;
	}
	else
	{
		m->m_cTitle = "";
	}

	mRootMenu = this;
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CMenu::~CMenu()
{
	while( mFirstMenuItem != NULL )
	{
		CMenuItem* pcItem = mFirstMenuItem;
		mFirstMenuItem = pcItem->m_pcNext;
		delete pcItem;
	}

	delete m;
}

bool  CMenu::Lock( void ) const
{
	return( mRootMenu->m->m_cMutex.Lock() );
}


void CMenu::Unlock( void ) const
{
	mRootMenu->m->m_cMutex.Unlock();
}


std::string CMenu::Label() const
{
	return( m->m_cTitle );
}


MenuLayout_e CMenu::Layout() const
{
	return( mMenuLayout );
}


void CMenu::FrameResized( float inWidth, float inHeight )
{
	if ( inWidth > 0.0f )
	{
		CRect cBounds = Bounds();

		cBounds.left = cBounds.right - inWidth;
		Invalidate( cBounds );
		Flush();
	}
	else if ( inWidth < 0.0f )
	{
		CRect cBounds = Bounds();

		cBounds.left = cBounds.right - 2.0f;
		Invalidate( cBounds );
		Flush();
	}
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CMenu::SetRoot( CMenu* pcRoot )
{
	mRootMenu = pcRoot;

	CMenuItem* pcTmp;

	for ( pcTmp = mFirstMenuItem ; NULL != pcTmp ; pcTmp = pcTmp->m_pcNext )
	{
		if ( pcTmp->mSubMenu != NULL  )
		{
			pcTmp->mSubMenu->SetRoot( pcRoot );
		}
	}
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CMenu::StartOpenTimer( bigtime_t nDelay )
{
	CLooper* pcLooper = Looper();
	if ( pcLooper != NULL )
	{
		pcLooper->Lock();
		pcLooper->AddTimer( this, 123, nDelay, true );
		pcLooper->Unlock();
	}
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CMenu::OpenSelection()
{
	CMenuItem* pcItem = FindHighlighted();

	if ( pcItem == NULL )
	{
		return;
	}

	CMenuItem* pcTmp;

	for ( pcTmp = mFirstMenuItem ; NULL != pcTmp ; pcTmp = pcTmp->m_pcNext )
	{
		if ( pcTmp == pcItem )
		{
			continue;
		}

		if ( pcTmp->mSubMenu != NULL && pcTmp->mSubMenu->mMenuWindow != NULL )
		{
			pcTmp->mSubMenu->Close( true, false );
			break;
		}
	}

	if ( pcItem->mSubMenu == NULL || pcItem->mSubMenu->mMenuWindow != NULL )
	{
		return;
	}
	
	mHasOpenChildren = true;
	pcItem->mSubMenu->Open( ScreenLocation() );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CPoint CMenu::ScreenLocation()
{
	CMenuItem*	pcItem = FindHighlighted();
	CRect		cFrame = pcItem->Frame();
	CPoint		cScrPos;
	CPoint		cScreenRes( CScreen().GetResolution() );
	CPoint		cLeftTop(0,0);

	ConvertToScreen( &cLeftTop );

	if ( mMenuLayout == ITEMS_IN_ROW )
	{
		cScrPos.x = cLeftTop.x + cFrame.left;

		float nHeight = pcItem->mSubMenu->m_cSize.y;
		if ( cLeftTop.y + Bounds().Height() + 1.0f + nHeight <= cScreenRes.y )
		{
			cScrPos.y = cLeftTop.y + Bounds().Height()+1.0f;
		}
		else
		{
			cScrPos.y = cLeftTop.y - nHeight;
		}
	}
	else
	{
		float nHeight = pcItem->mSubMenu->m_cSize.y;

		cScrPos.x = cLeftTop.x + Bounds().Width() - 3.0f;
		if ( cLeftTop.y + cFrame.top - mItemMargins.top + nHeight <= cScreenRes.y )
		{
			cScrPos.y = cLeftTop.y + cFrame.top - mItemMargins.top;
		}
		else
		{
			cScrPos.y = cLeftTop.y + cFrame.bottom - nHeight + mItemMargins.top + pcItem->mSubMenu->mItemMargins.bottom;
		}
	}
	
	return cScrPos;
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CMenu::TimerTick( int nID )
{
	CAutolock __lock__( &mRootMenu->m->m_cMutex );

	OpenSelection();
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CMenu::SelectItem( CMenuItem* pcItem )
{
	CMenuItem* pcPrev = FindHighlighted();

	if ( pcItem == pcPrev )
	{
		return;
	}

	if ( pcPrev != NULL )
	{
		pcPrev->Highlight( false );
	}

	if ( pcItem != NULL )
	{
		pcItem->Highlight( true );
	}
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CMenu::SelectPrev()
{
	CMenuItem* pcTmp;
	CMenuItem* pcPrev = NULL;

	for ( pcTmp = mFirstMenuItem ; NULL != pcTmp ; pcTmp = pcTmp->m_pcNext )
	{
		if ( pcTmp->mIsHighlighted )
		{
			if ( pcPrev != NULL )
			{
				SelectItem( pcPrev );
				Flush();
			}
			return;
		}
		pcPrev = pcTmp;
	}

	if ( mMenuLayout == ITEMS_IN_COLUMN )
	{
		CMenu* pcSuper = Supermenu();
		if ( pcSuper != NULL && pcSuper->mMenuLayout == ITEMS_IN_ROW )
		{
			Close( false, false );
		}
	}
	else
	{
		SelectItem( mFirstMenuItem );
		Flush();
	}
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CMenu::SelectNext()
{
	CMenuItem* pcTmp;

	for ( pcTmp = mFirstMenuItem ; NULL != pcTmp ; pcTmp = pcTmp->m_pcNext )
	{
		if ( pcTmp->mIsHighlighted || pcTmp->m_pcNext == NULL )
		{
			if ( pcTmp->m_pcNext != NULL )
			{
				SelectItem( pcTmp->m_pcNext );
			}
			else
			{
				SelectItem( pcTmp );
			}
			Flush();
			return;
		}
	}
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CMenu::Open( CPoint cScrPos )
{
	CAutolock __lock__( &mRootMenu->m->m_cMutex );

	if ( mMenuWindow != NULL )
	{
		return;
	}

	float width, height;

	GetPreferredSize( &width, &height );
	CRect   cBounds( 0, 0, width - 1, height - 1 );

	mMenuWindow = new MenuWindow( cBounds + cScrPos, this );

	mRootMenu->m->m_cMutex.Lock();
	mMenuWindow->SetMutex( &mRootMenu->m->m_cMutex );
	SetFrame( cBounds );
	mIsTracking = true;
	MakeFocus( true );

	SelectItem( mFirstMenuItem );
	mMenuWindow->Show();

	Flush();
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CMenu::Close( bool bCloseChildren, bool bCloseParent )
{
	CAutolock __lock__( &mRootMenu->m->m_cMutex );
	CMenuItem* pcTmp;

	for ( pcTmp = mFirstMenuItem ; NULL != pcTmp ; pcTmp = pcTmp->m_pcNext )
	{
		if ( pcTmp->mIsHighlighted )
		{
			pcTmp->Highlight( false );
		}
	}

	if ( bCloseChildren )
	{
		for ( pcTmp = mFirstMenuItem ; NULL != pcTmp ; pcTmp = pcTmp->m_pcNext )
		{
			if ( pcTmp->mSubMenu != NULL /*&& pcTmp->mSubMenu->mMenuWindow != NULL*/ )
			{
				pcTmp->mSubMenu->Close( bCloseChildren, bCloseParent );
			}
		}
	}

	if ( mMenuWindow != NULL )
	{
		MenuWindow* pcWnd = mMenuWindow;
		mMenuWindow = NULL; // Must clear this before we shutdown.
		pcWnd->ShutDown();

		CMenu* pcParent = Supermenu();
		if ( pcParent != NULL )
		{
			// NOTE: We activate the parents window even if was not
			//       opened by the menu-system (Window() vs mMenuWindow)
			CWindow* pcWnd = pcParent->Window();
			if ( pcWnd != NULL )
			{
				pcParent->MakeFocus( true );
			}
			pcParent->mHasOpenChildren = false;
			pcParent->mIsTracking = true;
		}
	}

	if ( m->m_cCloseMsgTarget.IsValid() )
	{
		m->m_cCloseMsgTarget.SendMessage( &m->m_cCloseMsg );
	}

	if ( bCloseParent )
	{
		CMenu* pcParent = Supermenu();
		if ( pcParent != NULL )
		{
			pcParent->Close( false, true );
		}
	}
	mIsTracking = false;
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CMenu::AttachedToWindow( void )
{
	CAutolock __lock__( &mRootMenu->m->m_cMutex );

	InvalidateLayout();
//  SetTargetForItems( Window() );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CMenu::DetachedFromWindow( void )
{
	CAutolock __lock__( &mRootMenu->m->m_cMutex );

	Close( true, false );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CMenu::WindowActivated( bool bIsActive )
{
	CAutolock __lock__( &mRootMenu->m->m_cMutex );

	if ( false == bIsActive && mHasOpenChildren == false )
	{
		Close( false, true );

		if ( m_hTrackPort != -1 )
		{
			CMenuItem* pcItem = NULL;
			if ( send_msg( m_hTrackPort, 1, &pcItem, sizeof(pcItem) ) < 0 )
			{
				dbprintf( "Error: Menu::WindowActivated() failed to send message to m_hTrackPort\n" );
			}
		}
	}
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CMenu::MouseDown( CPoint cPosition )
{
	CAutolock __lock__( &mRootMenu->m->m_cMutex );

	if ( Bounds().Contains( cPosition ) )
	{
		CMenuItem* pcItem = ItemAt( cPosition );
		SetCloseOnMouseUp( false );

		if ( pcItem != NULL )
		{
			MakeFocus( true );
			mIsTracking = true;
			SelectItem( pcItem );
			if ( mMenuLayout == ITEMS_IN_ROW )
			{
				OpenSelection();
			}
			else
			{
				StartOpenTimer( 200000 );
			}
			Flush();
		}

		OpenSelection();
	}
	else
	{
		if ( mHasOpenChildren == false )
		{
			Close( false, true );
			if ( m_hTrackPort != -1 )
			{
				CMenuItem* pcItem = NULL;
				if ( send_msg( m_hTrackPort, 1, &pcItem, sizeof(pcItem) ) < 0 )
				{
					dbprintf( "Error: Menu::WindowActivated() failed to send message to m_hTrackPort\n" );
				}
			}
		}
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CMenu::EndSession( CMenuItem* pcSelItem )
{
	CMenu* pcSuper = Supermenu();

	if ( NULL != pcSuper )
	{
		CWindow* pcWindow = pcSuper->Window();

		if ( NULL != pcWindow )
		{
			CMessage        cMsg( M_MENU_EVENT );

			cMsg.AddPointer( "_item", pcSelItem /* GetSuperItem()*/ );

			pcWindow->PostMessage( &cMsg, pcSuper );
		}
	}
	if ( m_hTrackPort != -1 )
	{
		if ( send_msg( m_hTrackPort, 1, &pcSelItem, sizeof(&pcSelItem) ) < 0 )
		{
			dbprintf( "Error: Menu::EndSession() failed to send message to m_hTrackPort\n" );
		}
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CMenu::MouseUp( CPoint cPosition )
{
	CAutolock __lock__( &mRootMenu->m->m_cMutex );

	if ( Bounds().Contains( cPosition ) )
	{
		CMenuItem* pcItem = FindHighlighted();

		// If we have a valid item, it's the item under the cursor, and it
		// doesn't have a submenu...

		if ( NULL != pcItem && pcItem == ItemAt( cPosition ) && NULL == pcItem->mSubMenu )
		{
			// If we are in radio mode, mark this (and only this) item

			if (mIsRadioMode)
			{
				CMenuItem* unhiliteItem;

				for ( unhiliteItem = mFirstMenuItem ;
					  NULL != unhiliteItem ;
					  unhiliteItem = unhiliteItem->m_pcNext )
				{
					unhiliteItem->SetMarked(unhiliteItem == pcItem);
				}
			}

			// Do the action associated with this menu item then close the menu

			pcItem->Invoke();
			Close( false, true );
		}
	}
	else
	{
		if ( mCloseOnMouseUp )
		{
			Close( false, true );
			if ( m_hTrackPort != -1 )
			{
				CMenuItem* pcItem = NULL;
				if ( send_msg( m_hTrackPort, 1, &pcItem, sizeof(pcItem) ) < 0 )
				{
					dbprintf( "Error: Menu::WindowActivated() failed to send message to m_hTrackPort\n" );
				}
			}
		}
	}
	SetCloseOnMouseUp( false );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CMenu::MouseMoved( CPoint cPosition, uint32 nCode, const CMessage* pcData )
{
	CAutolock __lock__( &mRootMenu->m->m_cMutex );

	int32       nButtons = 0x01;
	CMessage*	currentMessage = Window()->CurrentMessage();

	if (currentMessage)
		currentMessage->FindInt32("buttons", &nButtons);

	if ( nButtons & 0x01 )
	{
		SetCloseOnMouseUp( true );
	}

	if ( Bounds().Contains( cPosition ) == false )
	{
		CMenu* pcParent = Supermenu();
		if ( pcParent != NULL )
		{
			pcParent->MouseMoved( pcParent->ConvertFromScreen( ConvertToScreen( cPosition ) ), nCode, pcData );
		}
	}
	else
	{
		if ( mIsTracking )
		{
			CMenuItem*        pcItem = ItemAt( cPosition );

			if ( pcItem != NULL )
			{
				SelectItem( pcItem );
				if ( mMenuLayout == ITEMS_IN_ROW )
				{
					OpenSelection();
				}
				else
				{
					StartOpenTimer( 200000 );
				}
				Flush();
			}
		}
	}
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CMenu::KeyDown( const char* pzString, const char* pzRawString, uint32 nQualifiers )
{
	CAutolock __lock__( &mRootMenu->m->m_cMutex );

	switch( pzString[0] )
	{
		case VK_UP_ARROW:
			if ( mMenuLayout == ITEMS_IN_ROW )
			{
			}
			else
			{
				SelectPrev();
				StartOpenTimer( 1000000 );
			}
			break;

		case VK_DOWN_ARROW:
			if ( mMenuLayout == ITEMS_IN_ROW )
			{
				OpenSelection();
			}
			else
			{
				SelectNext();
				StartOpenTimer( 1000000 );
			}
			break;

		case VK_LEFT_ARROW:
			if ( mMenuLayout == ITEMS_IN_ROW )
			{
				SelectPrev();
				OpenSelection();
//        StartOpenTimer( 1000000 );
			}
			else
			{
				CMenu* pcSuper = Supermenu();
				if ( pcSuper != NULL )
				{
					if ( pcSuper->mMenuLayout == ITEMS_IN_COLUMN )
					{
						Close( false, false );
					}
					else
					{
						pcSuper->SelectPrev();
						pcSuper->OpenSelection();
					}
				}
			}
			break;

		case VK_RIGHT_ARROW:
			if ( mMenuLayout == ITEMS_IN_ROW )
			{
				SelectNext();
				OpenSelection();
//        StartOpenTimer( 1000000 );
			}
			else
			{
				CMenu* pcSuper = Supermenu();
				if ( pcSuper != NULL ) {
					CMenuItem* pcItem = FindHighlighted();
					if ( pcItem != NULL && pcItem->mSubMenu != NULL )
					{
						OpenSelection();
					}
					else if ( pcSuper->mMenuLayout == ITEMS_IN_ROW )
					{
						pcSuper->SelectNext();
						pcSuper->OpenSelection();
					}
				}
			}
			break;

		case VK_ENTER:
		{
			CMenuItem* pcItem = FindHighlighted();
			if ( pcItem != NULL ) {
				if ( pcItem->mSubMenu != NULL )
				{
					OpenSelection();
				}
				else
				{
					pcItem->Invoke();
					Close( false, true );
				}
			}
			break;
		}

		case 27: // ESC
			Close( false, false );
			Flush();
			break;

		default:
			CView::KeyDown( pzString, pzRawString, nQualifiers );
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CMenu::Draw( CRect inUpdateRect )
{
	CAutolock	__lock__( &mRootMenu->m->m_cMutex );
	CMenuItem*	pcItem;
	
	DrawBackground(inUpdateRect);

	for ( pcItem = mFirstMenuItem ; NULL != pcItem ; pcItem = pcItem->m_pcNext )
	{
		pcItem->Draw();
	}

	DrawFrame( Bounds(), FRAME_RAISED | FRAME_THIN | FRAME_TRANSPARENT );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CMenu::DrawBackground( CRect inUpdateRect )
{
	SetHighColor( get_default_color( COL_MENU_BACKGROUND ) );
	FillRect( Bounds() );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool CMenu::AddItem( const char* pzLabel, CMessage* pcMessage )
{
	return( AddItem( new CMenuItem( pzLabel, pcMessage ) ) );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool CMenu::AddItem( CMenuItem* pcItem )
{
	return( AddItem( pcItem, mItemCount ) );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool CMenu::AddItem( CMenuItem* pcItem, int inIndex )
{
	bool bResult = false;

	pcItem->mSuperMenu = this;

	if ( inIndex == 0 )
	{
		pcItem->m_pcNext = mFirstMenuItem;
		mFirstMenuItem    = pcItem;
		bResult = true;
	}
	else
	{
		CMenuItem*        pcPrev;
		int               i;

		for ( i = 1, pcPrev = mFirstMenuItem ; i < inIndex && NULL != pcPrev ; ++i )
		{
			pcPrev = pcPrev->m_pcNext;
		}
		if ( pcPrev )
		{
			pcItem->m_pcNext = pcPrev->m_pcNext;
			pcPrev->m_pcNext = pcItem;
			bResult = true;
		}
	}

	if ( bResult )
	{
		if ( pcItem->mSubMenu != NULL )
		{
			pcItem->mSubMenu->SetRoot( mRootMenu );
		}
		mItemCount++;
		InvalidateLayout();
	}

	return( bResult );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool CMenu::AddItem( CMenu* pcMenu )
{
	return( AddItem( pcMenu, mItemCount ) );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool CMenu::AddItem( CMenu* pcMenu, int inIndex )
{
	pcMenu->SetRoot( mRootMenu );
	return( AddItem( new CMenuItem( pcMenu, NULL ), inIndex ) );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool CMenu::AddSeparatorItem()
{
	return( AddItem( new CSeparatorItem(), mItemCount ) );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CMenuItem* CMenu::RemoveItem( int inIndex )
{
	CMenuItem** ppcTmp;
	CMenuItem*  pcItem = NULL;

	for ( ppcTmp = &mFirstMenuItem ; *ppcTmp != NULL ; ppcTmp = &(*ppcTmp)->m_pcNext )
	{
		if ( inIndex == 0 )
		{
			pcItem = *ppcTmp;
			*ppcTmp = pcItem->m_pcNext;
			if ( pcItem->mSubMenu != NULL )
			{
				pcItem->mSubMenu->SetRoot( pcItem->mSubMenu );
			}
			mItemCount--;
			InvalidateLayout();
			break;
		}
		inIndex--;
	}
	return( pcItem );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool CMenu::RemoveItem( CMenuItem* pcItem )
{
	CMenuItem** ppcTmp;

	for ( ppcTmp = &mFirstMenuItem ; *ppcTmp != NULL ; ppcTmp = &(*ppcTmp)->m_pcNext )
	{
		if ( *ppcTmp == pcItem )
		{
			*ppcTmp = pcItem->m_pcNext;
			mItemCount--;
			InvalidateLayout();
			if ( pcItem->mSubMenu != NULL )
			{
				pcItem->mSubMenu->SetRoot( pcItem->mSubMenu );
			}
			return( true );
		}
	}
	return( false );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool CMenu::RemoveItem( CMenu* pcMenu )
{
	CMenuItem** ppcTmp;
	CMenuItem*  pcItem = NULL;

	for ( ppcTmp = &mFirstMenuItem ; *ppcTmp != NULL ; ppcTmp = &(*ppcTmp)->m_pcNext )
	{
		pcItem = *ppcTmp;
		if ( pcItem->mSubMenu == pcMenu )
		{
			*ppcTmp = pcItem->m_pcNext;
			pcMenu->SetRoot( pcMenu );
			mItemCount--;
			InvalidateLayout();
			return( true );
		}
	}
	return( false );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CMenuItem* CMenu::ItemAt( CPoint cPosition ) const
{
	CMenuItem* pcItem;

	for ( pcItem = mFirstMenuItem ; NULL != pcItem ; pcItem = pcItem->m_pcNext )
	{
		if ( pcItem->mFrame.Contains( cPosition ) )
		{
			break;
		}
	}
	return( pcItem );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CMenuItem* CMenu::ItemAt( int inIndex ) const
{
	CMenuItem*        pcItem = NULL;

	if ( inIndex >= 0 && inIndex < mItemCount )
	{
		for ( pcItem = mFirstMenuItem ; NULL != pcItem ; pcItem = pcItem->m_pcNext )
		{
			if ( 0 == inIndex )
			{
				return( pcItem );
			}
			inIndex--;
		}
	}

	return( NULL );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CMenu* CMenu::SubmenuAt( int inIndex ) const
{
	CMenuItem* pcItem = ItemAt( inIndex );

	if ( NULL != pcItem )
		return( pcItem->mSubMenu );

	return( NULL );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

int CMenu::CountItems( void ) const
{
	return( mItemCount );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

int CMenu::IndexOf( CMenuItem* pcItem ) const
{
	CMenuItem* pcPtr;
	int        nIndex = 0;

	for ( pcPtr = mFirstMenuItem; NULL != pcPtr; pcPtr = pcPtr->m_pcNext )
	{
		if ( pcPtr == pcItem )
		{
			return( nIndex );
		}
		nIndex++;
	}

	return( -1 );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

int CMenu::IndexOf( CMenu* pcMenu ) const
{
	CMenuItem*	pcPtr;
	int			nIndex = 0;

	for ( pcPtr = mFirstMenuItem; NULL != pcPtr; pcPtr = pcPtr->m_pcNext )
	{
		if ( pcPtr->mSubMenu == pcMenu )
		{
			return( nIndex );
		}
		nIndex++;
	}
	return( -1 );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CMenuItem* CMenu::FindItem( uint32 nCode ) const
{
	CMenuItem* pcItem = NULL;

	for ( pcItem = mFirstMenuItem ; NULL != pcItem ; pcItem = pcItem->m_pcNext )
	{
		if ( pcItem->mSubMenu != NULL )
		{
			CMenuItem* pcTmp = pcItem->mSubMenu->FindItem( nCode );
			if ( pcTmp != NULL )
			{
				return( pcTmp );
			}
		}
		else
		{
			CMessage* pcMsg = pcItem->Message();
			if ( pcMsg != NULL && pcMsg->what == nCode )
			{
				return( pcItem );
			}
		}
	}
	return( NULL );
}


void CMenu::SetCloseOnMouseUp( bool bFlag )
{
	CMenuItem* pcItem;

	mCloseOnMouseUp = bFlag;

	for ( pcItem = mFirstMenuItem ; NULL != pcItem ; pcItem = pcItem->m_pcNext )
	{
		if ( pcItem->mSubMenu != NULL )
		{
			pcItem->mSubMenu->SetCloseOnMouseUp( bFlag );
		}
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CMenuItem* CMenu::FindItem( const char* pzName ) const
{
	CMenuItem* pcItem;

	for ( pcItem = mFirstMenuItem ; NULL != pcItem ; pcItem = pcItem->m_pcNext )
	{
		if ( pcItem->mSubMenu != NULL )
		{
			CMenuItem* pcTmp = pcItem->mSubMenu->FindItem( pzName );
			if ( pcTmp != NULL )
			{
				return( pcTmp );
			}
		}
		else
		{
			if ( NULL != pcItem->mLabel && strcmp( pcItem->mLabel, pzName ) == 0 )
			{
				break;
			}
		}
	}

	return( pcItem );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

status_t CMenu::SetTargetForItems( CHandler* pcTarget )
{
	CMenuItem*        pcItem;

	for ( pcItem = mFirstMenuItem ; NULL != pcItem ; pcItem = pcItem->m_pcNext )
	{
		if ( NULL == pcItem->mSubMenu )
		{
			pcItem->SetTarget( pcTarget );
		}
		else
		{
			pcItem->mSubMenu->SetTargetForItems( pcTarget );
		}
	}

	return( 0 );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

status_t CMenu::SetTargetForItems( CMessenger cMessenger )
{
	CMenuItem*        pcItem;

	for ( pcItem = mFirstMenuItem ; NULL != pcItem ; pcItem = pcItem->m_pcNext )
	{
		if ( NULL == pcItem->mSubMenu )
		{
			pcItem->SetTarget( cMessenger );
		}
		else
		{
			pcItem->mSubMenu->SetTargetForItems( cMessenger );
		}
	}

	return( 0 );
}

void CMenu::SetCloseMessage( const CMessage& cMsg )
{
	m->m_cCloseMsg = cMsg;
}

void CMenu::SetCloseMsgTarget( const CMessenger& cTarget )
{
    m->m_cCloseMsgTarget = cTarget;
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CMenuItem* CMenu::FindHighlighted( void ) const
{
	CMenuItem* pcItem;

	for ( pcItem = mFirstMenuItem ; NULL != pcItem ; pcItem = pcItem->m_pcNext )
	{
		if ( pcItem->mIsHighlighted )
		{
			break;
		}
	}
	return( pcItem );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CMenuItem* CMenu::FindMarked( void ) const
{
	CMenuItem* pcItem;

	for ( pcItem = mFirstMenuItem ; NULL != pcItem ; pcItem = pcItem->m_pcNext )
	{
		if ( pcItem->mIsMarked )
		{
			break;
		}
	}
	return( pcItem );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CMenu* CMenu::Supermenu( void ) const
{
	if ( NULL != mSuperItem )
		return( mSuperItem->mSuperMenu );

	return( NULL );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CMenuItem* CMenu::Superitem( void ) const
{
	return( mSuperItem );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CMenu::GetPreferredSize(float *outWidth, float *outHeight)
{
	if (outWidth)
		*outWidth = m_cSize.x + 2.0f;

	if (outHeight)
		*outHeight = m_cSize.y + 2.0f;
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CMenu::InvalidateLayout( void )
{
	CMenuItem*        pcItem;

	m_cSize = CPoint( 0.0f, 0.0f );

	switch( mMenuLayout )
	{
		case ITEMS_IN_ROW:
			for ( pcItem = mFirstMenuItem ; NULL != pcItem ; pcItem = pcItem->m_pcNext )
			{
				float width, height;

				pcItem->GetContentSize(&width, &height);


				pcItem->mFrame = CRect( m_cSize.x,
										mItemMargins.top - 1,
										m_cSize.x + width + mItemMargins.left + 1,
										height + mItemMargins.bottom );

				pcItem->mFrame += CPoint( 1.0f, 1.0f );        // Menu border

				m_cSize.x += width + mItemMargins.left + mItemMargins.right;

				if ( height + mItemMargins.top + mItemMargins.bottom + 2 > m_cSize.y )
				{
					m_cSize.y = height + mItemMargins.top + mItemMargins.bottom + 2;
				}
			}
			break;

		case ITEMS_IN_COLUMN:
			for ( pcItem = mFirstMenuItem ; NULL != pcItem ; pcItem = pcItem->m_pcNext )
			{
				float width, height;

				pcItem->GetContentSize(&width, &height);

				pcItem->mFrame = CRect( 1.0f,
										m_cSize.y + mItemMargins.top - 1,
										1.0f,	// will be calculated later
										m_cSize.y + height + mItemMargins.top + 2 );

				m_cSize.y += height + mItemMargins.top + mItemMargins.bottom;

				// If this is the widest item we've seen, remember the width

				if ( width + mItemMargins.left + mItemMargins.right > m_cSize.x )
				{
					m_cSize.x = width + mItemMargins.left + mItemMargins.right;
				}
			}

			// Now that we know the widest item, make all items this wide

			for ( pcItem = mFirstMenuItem ; NULL != pcItem ; pcItem = pcItem->m_pcNext )
			{
				pcItem->mFrame.right = m_cSize.x - 1;
			}
			break;

		case ITEMS_CUSTOM_LAYOUT:
			dbprintf( "Error: ITEMS_CUSTOM_LAYOUT not supported\n" );
			break;
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CMenuItem* CMenu::Track( const CPoint& cScreenPos )
{
	m_hTrackPort = create_port( "menu_track_port", 1 );
	if ( m_hTrackPort < 0 )
	{
		return( NULL );
	}

	float width, height;

	GetPreferredSize( &width, &height );
	CRect   cBounds( 0, 0, width - 1, height - 1 );

	mMenuWindow = new MenuWindow( cBounds + cScreenPos, this );
	mRootMenu->m->m_cMutex.Lock();
	mMenuWindow->SetMutex( &mRootMenu->m->m_cMutex );

	SetFrame( cBounds );

	mMenuWindow->Show();

//    if ( mMenuWindow->Lock() )	//NLS
//    {
		MakeFocus( true );
//        mMenuWindow->Unlock();
//    }
	mIsTracking = true;

	uint32 nCode;
	CMenuItem* pcSelItem;
	get_msg( m_hTrackPort, &nCode, &pcSelItem, sizeof( &pcSelItem ) );

	assert( mMenuWindow == NULL );
	delete_port( m_hTrackPort );
	m_hTrackPort = -1;
	return( pcSelItem );
}


void CMenu::GetItemMargins(	float *outLeft, 
							float *outTop, 
							float *outRight, 
							float *outBottom) const
{
	if (outLeft)
		*outLeft = mItemMargins.left;
		
	if (outTop)
		*outLeft = mItemMargins.top;
		
	if (outRight)
		*outLeft = mItemMargins.right;
		
	if (outBottom)
		*outLeft = mItemMargins.bottom;
}


void CMenu::SetItemMargins(	float inLeft, 
							float inTop, 
							float inRight, 
							float inBottom)
{
	mItemMargins.Set(inLeft, inTop, inRight, inBottom);
}
