/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */

 
#include <interface/MenuItem.h>
#include <interface/Font.h>
#include "stdbitmaps.h"

using namespace os;

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CMenuItem::CMenuItem( const char* inLabel,
					  CMessage*   inMsg,
					  char        inShortcut,
					  uint32      inModifiers ) : CInvoker( inMsg, NULL, NULL )
{
	m_pcNext		= NULL;
	mSuperMenu		= NULL;
	mSubMenu		= NULL;
	mIsHighlighted	= false;
	mIsMarked		= false;
	mIsEnabled		= true;
	mShortcut		= inShortcut;
	mModifiers		= inModifiers;

	if ( NULL != inLabel )
	{
		mLabel = new char[ strlen( inLabel ) + 1 ];
		strcpy( mLabel, inLabel );
	}
	else
	{
		mLabel = NULL;
	}
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CMenuItem::CMenuItem( CMenu* pcMenu, CMessage* pcMsg ) : CInvoker( pcMsg, NULL, NULL )
{
	m_pcNext		= NULL;
	mSuperMenu		= NULL;
	mSubMenu		= pcMenu;
	mIsHighlighted	= false;
	mIsMarked		= false;
	mIsEnabled		= true;
	mShortcut		= 0;
	mModifiers		= 0UL;

	pcMenu->mSuperItem = this;

	std::string cLabel = pcMenu->Label();

	if ( cLabel.empty() == false )
	{
		mLabel = new char[ cLabel.size() + 1 ];
		strcpy( mLabel, cLabel.c_str() );
	}
	else
	{
		mLabel = NULL;
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CMenuItem::~CMenuItem()
{
	delete[] mLabel;
	delete   mSubMenu;
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool CMenuItem::Invoked( CMessage* pcMessage )
{
	return( true );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CMenuItem::SetShortcut( char inShortcut, uint32 inModifiers )
{
	mModifiers = inModifiers;
	mShortcut  = inShortcut;
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

char CMenuItem::Shortcut( uint32 *outModifiers ) const
{
	if (outModifiers)
		*outModifiers = mModifiers;

	return( mShortcut );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CMenuItem::GetContentSize( float *outWidth, float *outHeight )
{
	float width, height;

	if ( NULL != mLabel && NULL != mSuperMenu )
	{
		font_height sHeight;
		mSuperMenu->GetFontHeight( &sHeight );

		width = mSuperMenu->StringWidth( mLabel ) + 4.0f;
		height = sHeight.ascender + sHeight.descender + sHeight.line_gap;

		// Make room for the column-item-specific things

		if (mSuperMenu->Layout() == ITEMS_IN_COLUMN)
		{
			// Make room for the menu mark

			width += 10.0f;

			// Make room for the submenu triangle or shortcut character

			if (mSubMenu != NULL)
				width += 10.0f;
		}
    }
	else
	{
		width = height = 0.0f;
	}

	if (outWidth)
		*outWidth = width;

	if (outHeight)
		*outHeight = height;
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CMenuItem::SetLabel( const char* inLabel )
{
	delete[] mLabel;
	mLabel = new char[ strlen( inLabel ) + 1 ];
	strcpy( mLabel, inLabel );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CPoint CMenuItem::ContentLocation( void ) const
{
	float leftOffset;

	if ( mSuperMenu && (mSuperMenu->Layout() == ITEMS_IN_COLUMN))
		leftOffset = 10.0f;
	else
		leftOffset = 3.0f;

	return( CPoint( mFrame.left + leftOffset, mFrame.top ) );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CMenuItem::Draw( void )
{
	if ( mSuperMenu == NULL )
		return;

	if ( mLabel == NULL )
		return;

	float frameHeight = Frame().Height();

	font_height sHeight;

	mSuperMenu->GetFontHeight( &sHeight );

	if ( mIsEnabled && mIsHighlighted )
	{
		mSuperMenu->SetHighColor( get_default_color( COL_SEL_MENU_BACKGROUND ) );
	}
	else
	{
		mSuperMenu->SetHighColor( get_default_color( COL_MENU_BACKGROUND ) );
	}

	mSuperMenu->FillRect( mFrame );

	if ( mIsEnabled && mIsHighlighted )
	{
		mSuperMenu->SetHighColor( get_default_color( COL_SEL_MENU_TEXT ) );
		mSuperMenu->SetLowColor( get_default_color( COL_SEL_MENU_BACKGROUND ) );
	}
	else
	{
		mSuperMenu->SetHighColor( get_default_color( COL_MENU_TEXT ) );
		mSuperMenu->SetLowColor( get_default_color( COL_MENU_BACKGROUND ) );
	}

	CPoint	contentLocation = ContentLocation();

	float vCharHeight = sHeight.ascender + sHeight.descender + sHeight.line_gap;
	float y =	mFrame.top
				+ 1.0f
				+ (frameHeight / 2)
				- (vCharHeight / 2)
				+ sHeight.ascender
				+ (sHeight.line_gap * 0.5f);

	// Draw the menu label

	mSuperMenu->DrawString( mLabel, CPoint( contentLocation.x + 2.0f, y ) );

	// Do column-menu-specific drawing

	if (mSuperMenu->Layout() == ITEMS_IN_COLUMN)
	{
		// Draw the menu mark

		if ( mIsMarked )
		{
			mSuperMenu->FillRect(CRect( mFrame.left + 2.0f,
										mFrame.top + ((frameHeight + 1.0f) / 2) - 2.0f,
										mFrame.left + 6.0f,
										mFrame.top + ((frameHeight + 1.0f) / 2) + 2.0f) );
		}

		// Draw any object which may be to the right of the label,
		// such as the shortcut character or submenu triangle

		if ( mSubMenu != NULL )
		{
			CRect	TriangleRect;
			CBitmap* TriangleBitmap;

			TriangleBitmap = get_std_bitmap( BMID_ARROW_RIGHT, 0, &TriangleRect );

			mSuperMenu->SetDrawingMode( DM_OVER );
			
			mSuperMenu->DrawBitmap( TriangleBitmap, TriangleRect,
									TriangleRect.Bounds() +
										CPoint( mFrame.right - 6.0f,
												mFrame.top + (frameHeight / 2) - 3.0f) );

			mSuperMenu->SetDrawingMode( DM_COPY );
		}
		else if ( mShortcut != 0 )
		{
			// FIXME: we need to draw the alt or command key icon as well
			//		  This probably needs a new standard bitmap.
			
			mSuperMenu->DrawChar(mShortcut, CPoint(mFrame.right - 10.0f, y));
		}
	}
}


void	CMenuItem::DrawContent( void )
{
}


void	CMenuItem::Highlight( bool inHighlight )
{
	mIsHighlighted = inHighlight;
	Draw();
}


void	CMenuItem::SetMarked( bool inMarked )
{
	mIsMarked = inMarked;
	Draw();
}



void	CMenuItem::SetEnabled( bool inEnabled )
{
	mIsEnabled = inEnabled;
	Draw();
}


CSeparatorItem::CSeparatorItem() : CMenuItem( static_cast<char*>(NULL), NULL )
{
	mIsEnabled = false;
}

CSeparatorItem::~CSeparatorItem()
{
}

void CSeparatorItem::GetContentSize(float *outWidth, float *outHeight)
{
	*outWidth = 0.0f;
	*outHeight = 0.0f;

	if ( mSuperMenu != NULL )
	{
		if ( mSuperMenu->Layout() == ITEMS_IN_ROW )
		{
			*outWidth = 6.0f;
		}
		else
		{
			*outHeight = 6.0f;
		}
	}
}

void CSeparatorItem::Draw()
{
	if ( mSuperMenu == NULL )
	{
		return;
	}

	CRect cFrame = Frame();
	CRect cMFrame = mSuperMenu->Bounds();

	if ( mSuperMenu->Layout() == ITEMS_IN_ROW )
	{
		cFrame.top    = cMFrame.top;
		cFrame.bottom = cMFrame.bottom;
	}
	else
	{
		cFrame.left  = cMFrame.left;
		cFrame.right = cMFrame.right;
	}

	mSuperMenu->SetHighColor( get_default_color( COL_MENU_BACKGROUND ) );
	mSuperMenu->FillRect( cFrame );

	if ( mSuperMenu->Layout() == ITEMS_IN_ROW )
	{
		float x = floor( cFrame.left + (cFrame.Width() + 1.0f) * 0.5f );
		mSuperMenu->SetHighColor( get_default_color( COL_SHADOW ) );
		mSuperMenu->StrokeLine( CPoint( x, cFrame.top + 2.0f ), CPoint( x, cFrame.bottom - 2.0f ) );
		x += 1.0f;
		mSuperMenu->SetHighColor( get_default_color( COL_SHINE ) );
		mSuperMenu->StrokeLine( CPoint( x, cFrame.top + 2.0f ), CPoint( x, cFrame.bottom - 2.0f ) );
	}
	else
	{
		float y = floor( cFrame.top + (cFrame.Height() + 1.0f) * 0.5f );
		mSuperMenu->SetHighColor( get_default_color( COL_SHADOW ) );
		mSuperMenu->StrokeLine( CPoint( cFrame.left + 3.0f, y ), CPoint( cFrame.right - 3.0f, y ) );
		y += 1.0f;
		mSuperMenu->SetHighColor( get_default_color( COL_SHINE ) );
		mSuperMenu->StrokeLine( CPoint( cFrame.left + 3.0f, y ), CPoint( cFrame.right - 3.0f, y ) );
	}
}

void CSeparatorItem::Highlight( bool inHighlight )
{
}



