/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */


#include <interface/ProgressBar.h>
#include <interface/Font.h>
#include <cosmoe/macros.h>

using namespace os;


ProgressBar::ProgressBar( CRect inFrame, const char* inTitle, orientation eOrientation, uint32 inResizeMask, uint32 inFlags ) :
        CView( inFrame, inTitle, inResizeMask, inFlags )
{
	m_vProgress = 0.0f;
	m_eOrientation = eOrientation;
}


void ProgressBar::SetProgress( float vValue )
{
	m_vProgress = vValue;
	Invalidate();
	Flush();
}


float ProgressBar::GetProgress() const
{
	return( m_vProgress );
}


void ProgressBar::AttachedToWindow()
{
	CView* pcParent = Parent();

	if ( pcParent != NULL )
	{
		SetLowColor( pcParent->LowColor() );
	}
	else
	{
		SetLowColor( get_default_color( COL_NORMAL ) );
	}
}

void ProgressBar::Draw( CRect cUpdateRect )
{
	CRect cBounds = GetNormalizedBounds();
	cBounds.Floor();

	DrawFrame( cBounds, FRAME_RECESSED | FRAME_TRANSPARENT );

	float vBarLength = (m_eOrientation == HORIZONTAL) ? cBounds.Width() : cBounds.Height();
	vBarLength = ceil( (vBarLength - 3.0f) * m_vProgress );

	CRect cBarFrame = cBounds;
	cBarFrame.Resize( 2.0f, 2.0f, -2.0f, -2.0f );
	CRect cRestFrame = cBarFrame;

	if ( vBarLength < 1.0f )
	{
		FillRect( cRestFrame, LowColor() );
	}
	else
	{
		if ( m_eOrientation == HORIZONTAL )
		{
			cBarFrame.right = cBarFrame.left + vBarLength;
			if ( cBarFrame.right > cBounds.right - 2.0f )
			{
				cBarFrame.right = cBounds.right - 2.0f;
			}
			cRestFrame.left = cBarFrame.right + 1.0f;
		}
		else
		{
			cBarFrame.top = cBarFrame.bottom - vBarLength;
			if ( cBarFrame.right < cBounds.right + 2.0f ) {
				cBarFrame.top = cBounds.top + 2.0f;
			}
			cRestFrame.bottom = cBarFrame.top - 1.0f;
		}

		if ( vBarLength >= 1.0f )
		{
			FillRect( cBarFrame, Color32_s( 0x66, 0x88, 0xbb ) );
		}

		if ( cRestFrame.IsValid() )
		{
			FillRect( cRestFrame, LowColor() );
		}
	}
}


void ProgressBar::GetPreferredSize( float* outWidth, float* outHeight )
{
	font_height sHeight;
	GetFontHeight( &sHeight );
	float vHeight = sHeight.ascender + sHeight.descender + sHeight.line_gap;

	if (outWidth)
		*outWidth = vHeight * 3;

	if (outHeight)
		*outHeight = vHeight * 4;
}


void ProgressBar::FrameResized( float inWidth, float inHeight )
{
	CRect cBounds( Bounds() );
	bool bNeedFlush = false;
	if ( (m_eOrientation == HORIZONTAL && inWidth != 0.0f) || (m_eOrientation == VERTICAL && inHeight != 0.0f) )
	{
		CRect cDamage = cBounds;
		cDamage.Resize( 2.0f, 2.0f, -2.0f, -2.0f );
		Invalidate( cDamage );
		bNeedFlush = true;
	}

	if ( inWidth != 0.0f )
	{
		CRect cDamage = cBounds;

		cDamage.left = cDamage.right - max( 4.0f, inWidth + 3.0f );
		Invalidate( cDamage );
		bNeedFlush = true;
	}

	if ( inHeight != 0.0f )
	{
		CRect cDamage = cBounds;

		cDamage.top = cDamage.bottom - max( 3.0f, inHeight + 2.0f );
		Invalidate( cDamage );
		bNeedFlush = true;
	}

	if ( bNeedFlush )
		Flush();
}
