/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */


#include <assert.h>
#include <support/SupportDefs.h>

#include <interface/ScrollBar.h>
#include <app/Looper.h>
#include <app/Message.h>

#include "stdbitmaps.h"

#include <macros.h>

using namespace os;

enum { HIT_NONE, HIT_KNOB, HIT_ARROW };

enum { ID_SCROLL_TIMER = 1 };

class CScrollBar::Private
{
public:
	Private() {
		m_nHitState    = HIT_NONE;
		m_nHitButton   = 0;
		m_vProportion  = 0.1f;
		m_nOrientation = VERTICAL;
		m_vMin         = 0.0f;
		m_vMax         = FLT_MAX;
		m_vSmallStep   = 1.0f;
		m_vBigStep     = 10.0f;
		m_bChanged     = false;
		m_pcTarget     = NULL;
		m_bFirstTick   = true;
		memset( m_abArrowStates, 0, sizeof(m_abArrowStates) );
	}
	CView*       m_pcTarget;
	float        m_vMin;
	float        m_vMax;
	float        m_vProportion;
	float        m_vSmallStep;
	float        m_vBigStep;
	orientation  m_nOrientation;
	CRect        m_acArrowRects[4];
	bool         m_abArrowStates[4];
	CRect        m_cKnobArea;
	bool         m_bChanged;
	bool         m_bFirstTick;
	CPoint       m_cHitPos;
	int          m_nHitButton;
	int          m_nHitState;
};

static Color32_s Tint( const Color32_s& sColor, float vTint )
{
	int r = int( (float(sColor.red) * vTint + 127.0f * (1.0f - vTint)) );
	int g = int( (float(sColor.green) * vTint + 127.0f * (1.0f - vTint)) );
	int b = int( (float(sColor.blue) * vTint + 127.0f * (1.0f - vTint)) );
	
	if ( r < 0 ) r = 0;
	else if (r > 255) r = 255;
	
	if ( g < 0 ) g = 0;
	else if (g > 255) g = 255;
	
	if ( b < 0 ) b = 0;
	else if (b > 255) b = 255;
	
	return( Color32_s( r, g, b, sColor.alpha ) );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CScrollBar::CScrollBar( CRect inFrame,
						const char* inName,
						CMessage* inMessage,
						float inMin,
						float inMax,
						orientation inOrientation,
						uint32 inResizeMask )
        : CControl( inFrame, inName, "", inMessage, inResizeMask, WID_WILL_DRAW | WID_FULL_UPDATE_ON_RESIZE )
{
	m = new Private;
	m->m_nOrientation = inOrientation;
	m->m_vMin         = inMin;
	m->m_vMax         = inMax;
	FrameResized( 0.0f, 0.0f );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CScrollBar::~CScrollBar()
{
	if ( m->m_pcTarget != NULL )
	{
		if ( m->m_nOrientation == HORIZONTAL )
		{
			m->m_pcTarget->_SetHScrollBar( NULL );
		}
		else
		{
			m->m_pcTarget->_SetVScrollBar( NULL );
		}
	}

	delete m;
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CScrollBar::SetProportion( float inProportion )
{
	m->m_vProportion = inProportion;
	Invalidate( m->m_cKnobArea );
	Flush();
}


float CScrollBar::Proportion() const
{
	return( m->m_vProportion );
}


void CScrollBar::SetSteps( float inSmall, float inBig )
{
	m->m_vSmallStep = inSmall;
	m->m_vBigStep   = inBig;
}


void CScrollBar::GetSteps( float* outSmall, float* outBig ) const
{
	if (outSmall)
		*outSmall = m->m_vSmallStep;
	
	if (outBig)
		*outBig   = m->m_vBigStep;
}


void CScrollBar::SetRange( float inMinimum, float inMaximum )
{
	m->m_vMin = inMinimum;
	m->m_vMax = inMaximum;
}


void CScrollBar::GetRange( float* outMinimum, float* outMaximum ) const
{
	if (outMinimum)
		*outMinimum = m->m_vMin;
	
	if (outMaximum)
		*outMaximum = m->m_vMax;
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CScrollBar::SetTarget( CView* inTarget )
{
	if ( m->m_pcTarget != NULL )
	{
		if ( m->m_nOrientation == HORIZONTAL )
		{
			assert( m->m_pcTarget->GetScrollBar(HORIZONTAL) == this );
			m->m_pcTarget->_SetHScrollBar( NULL );
		}
		else
		{
			assert( m->m_pcTarget->GetScrollBar(VERTICAL) == this );
			m->m_pcTarget->_SetVScrollBar( NULL );
		}
	}

	m->m_pcTarget = inTarget;

	if ( m->m_pcTarget != NULL )
	{
		if ( m->m_nOrientation == HORIZONTAL )
		{
			assert( m->m_pcTarget->GetScrollBar(HORIZONTAL) == NULL );
			m->m_pcTarget->_SetHScrollBar( this );
			SetValue( m->m_pcTarget->GetScrollOffset().x );
		}
		else
		{
			assert( m->m_pcTarget->GetScrollBar(VERTICAL) == NULL );
			m->m_pcTarget->_SetVScrollBar( this );
			SetValue( m->m_pcTarget->GetScrollOffset().y );
		}
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CView* CScrollBar::Target( void ) const
{
	return( m->m_pcTarget );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

orientation CScrollBar::Orientation( void ) const
{
	return( m->m_nOrientation );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CScrollBar::GetPreferredSize( float* outWidth, float* outHeight )
{
	if (outWidth)
		*outWidth = ( m->m_nOrientation == HORIZONTAL ) ? 0.0f : SB_MINSIZE;

	if (outHeight)
		*outHeight = ( m->m_nOrientation == HORIZONTAL ) ? SB_MINSIZE : 0.0f;
}


void CScrollBar::PostValueChange( const Variant& cNewValue )
{
	if ( m->m_pcTarget != NULL )
	{
		CPoint cPos = m->m_pcTarget->GetScrollOffset();

		if ( m->m_nOrientation == HORIZONTAL )
		{
			cPos.x = -floor( cNewValue.AsFloat() );
		}
		else
		{
			cPos.y = -floor( cNewValue.AsFloat() );
		}
		m->m_pcTarget->ScrollTo( cPos );
		m->m_pcTarget->Flush();
	}

	Invalidate( m->m_cKnobArea );
	Flush();
}


void CScrollBar::LabelChanged( const std::string& cNewLabel )
{
}


void CScrollBar::EnableStatusChanged( bool bIsEnabled )
{
}


bool CScrollBar::Invoked( CMessage* pcMessage )
{
	CControl::Invoked( pcMessage );
	if ( m->m_nHitState == HIT_KNOB )
	{
		pcMessage->AddBool( "final", false );
	}
	else
	{
		pcMessage->AddBool( "final", true );
	}

	return( true );
}


void CScrollBar::TimerTick( int nID )
{
	if ( nID == ID_SCROLL_TIMER )
	{
		float vValue = Value();
		if ( m->m_nHitButton & 0x01 )
		{
			vValue += m->m_vSmallStep;
		}
		else
		{
			vValue -= m->m_vSmallStep;
		}

		if ( vValue < m->m_vMin )
		{
			vValue = m->m_vMin;
		}
		else if ( vValue > m->m_vMax )
		{
			vValue = m->m_vMax;
		}
		
		SetValue( vValue );
		if ( m->m_bFirstTick )
		{
			Looper()->AddTimer( this, ID_SCROLL_TIMER, 30000LL, false );
			m->m_bFirstTick = false;
		}
	}
	else
	{
		CControl::TimerTick( nID );
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CScrollBar::MouseDown( CPoint cPosition )
{
	MakeFocus( true );

	m->m_nHitState  = HIT_NONE;
	m->m_bChanged = false;

	for ( int i = 0 ; i < 4 ; ++i )
	{
		if ( m->m_acArrowRects[i].Contains( cPosition ) )
		{
			m->m_abArrowStates[i] = true;
			m->m_nHitButton = i;
			m->m_nHitState  = HIT_ARROW;
			Invalidate( m->m_acArrowRects[i] );
			Flush();
			m->m_bFirstTick = true;
			Looper()->AddTimer( this, ID_SCROLL_TIMER, 300000LL, true );
			return;
		}
	}

	if ( m->m_cKnobArea.Contains( cPosition ) )
	{
		CRect cKnobFrame = GetKnobFrame();
		if ( cKnobFrame.Contains( cPosition ) )
		{
			m->m_cHitPos = cPosition - cKnobFrame.LeftTop();
			m->m_nHitState  = HIT_KNOB;
			return;
		}

		float vValue = Value();

		if ( m->m_nOrientation == HORIZONTAL )
		{
			if ( cPosition.x < cKnobFrame.left )
			{
				vValue -= m->m_vBigStep;
			}
			else if ( cPosition.x > cKnobFrame.right )
			{
				vValue += m->m_vBigStep;
			}
		}
		else
		{
			if ( cPosition.y < cKnobFrame.top )
			{
				vValue -= m->m_vBigStep;
			}
			else if ( cPosition.y > cKnobFrame.bottom )
			{
				vValue += m->m_vBigStep;
			}
		}

		if ( vValue < m->m_vMin )
		{
			vValue = m->m_vMin;
		}
		else if ( vValue > m->m_vMax )
		{
			vValue = m->m_vMax;
		}
		SetValue( vValue );
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CScrollBar::MouseUp( CPoint cPosition )
{
	if ( m->m_nHitState == HIT_ARROW )
	{
		float vValue = Value();
		bool bChanged = false;
		if ( m->m_bFirstTick )
		{
			for ( int i = 0 ; i < 4 ; ++i )
			{
				if ( m->m_acArrowRects[i].Contains( cPosition ) )
				{
					if ( i == m->m_nHitButton )
					{
						bChanged = true;
						if ( i & 0x01 )
						{
							vValue += m->m_vSmallStep;
						}
						else
						{
							vValue -= m->m_vSmallStep;
						}
					}
					break;
				}
			}
		}

		if ( m->m_abArrowStates[m->m_nHitButton] )
		{
			Looper()->RemoveTimer( this, ID_SCROLL_TIMER );
			m->m_abArrowStates[m->m_nHitButton] = false;
		}

		Invalidate( m->m_acArrowRects[m->m_nHitButton] );
		if ( bChanged )
		{
			if ( vValue < m->m_vMin )
			{
				vValue = m->m_vMin;
			}
			else if ( vValue > m->m_vMax )
			{
				vValue = m->m_vMax;
			}
			SetValue( vValue );
		}
		Flush();
	}
	else if ( m->m_nHitState == HIT_KNOB )
	{
		if ( m->m_bChanged )
		{
			Invoke();        // Send a 'final' message
			m->m_bChanged = false;
		}
	}

	m->m_nHitState  = HIT_NONE;
	MakeFocus( false );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CScrollBar::MouseMoved( CPoint cNewPos, uint32 nCode, const CMessage* pcData )
{
	if ( m->m_nHitState == HIT_ARROW )
	{
		int i;
		for ( i = 0 ; i < 4 ; ++i )
		{
			if ( m->m_acArrowRects[i].Contains( cNewPos ) )
			{
				break;
			}
		}

		bool bHit = m->m_nHitButton == i;
		if ( bHit != m->m_abArrowStates[m->m_nHitButton] )
		{
			if ( m->m_abArrowStates[m->m_nHitButton] )
			{
				Looper()->RemoveTimer( this, ID_SCROLL_TIMER );
			}
			else
			{
				if ( m->m_bFirstTick )
				{
					Looper()->AddTimer( this, ID_SCROLL_TIMER, 300000LL, true );
				}
				else
				{
					Looper()->AddTimer( this, ID_SCROLL_TIMER, 30000LL, false );
				}
			}
			m->m_abArrowStates[m->m_nHitButton] = bHit;
			Invalidate( m->m_acArrowRects[m->m_nHitButton] );
			Flush();
		}

	}
	else if ( m->m_nHitState == HIT_KNOB )
	{
		SetValue( _PosToVal( cNewPos ) );
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CScrollBar::Draw( CRect cUpdateRect )
{
	CRect cBounds    = Bounds();
	CRect cKnobFrame = GetKnobFrame();

//  SetViewColor( get_default_color( COL_SCROLLBAR_BG ) );

//    DrawFrame( cBounds, FRAME_RECESSED | FRAME_TRANSPARENT | FRAME_THIN );

	SetViewColor( get_default_color( COL_SCROLLBAR_KNOB ) );
	DrawFrame( cKnobFrame, FRAME_RAISED );

	if ( m->m_nOrientation == HORIZONTAL )
	{
		if ( cKnobFrame.Width() > 30.0f )
		{
			const float vDist = 5.0f;
			float vCenter = cKnobFrame.left + (cKnobFrame.Width() + 1.0f) * 0.5f;
			SetHighColor( get_default_color( COL_SHINE ) );
			StrokeLine( CPoint( vCenter - vDist, cKnobFrame.top + 4.0f ), CPoint( vCenter - vDist, cKnobFrame.bottom - 4.0f ) );
			StrokeLine( CPoint( vCenter, cKnobFrame.top + 4.0f ), CPoint( vCenter, cKnobFrame.bottom - 4.0f ) );
			StrokeLine( CPoint( vCenter + vDist, cKnobFrame.top + 4.0f ), CPoint( vCenter + vDist, cKnobFrame.bottom - 4.0f ) );

			SetHighColor( get_default_color( COL_SHADOW ) );
			vCenter -= 1.0f;
			StrokeLine( CPoint( vCenter - vDist, cKnobFrame.top + 4.0f ), CPoint( vCenter - vDist, cKnobFrame.bottom - 4.0f ) );
			StrokeLine( CPoint( vCenter, cKnobFrame.top + 4.0f ), CPoint( vCenter, cKnobFrame.bottom - 4.0f ) );
			StrokeLine( CPoint( vCenter + vDist, cKnobFrame.top + 4.0f ), CPoint( vCenter + vDist, cKnobFrame.bottom - 4.0f ) );
		}
	}
	else
	{
		if ( cKnobFrame.Height() > 30.0f )
		{
			const float vDist = 5.0f;
			float vCenter = cKnobFrame.top + (cKnobFrame.Height() + 1.0f) * 0.5f;
			SetHighColor( get_default_color( COL_SHINE ) );
			StrokeLine( CPoint( cKnobFrame.left + 4.0f, vCenter - vDist ), CPoint( cKnobFrame.right - 4.0f, vCenter - vDist ) );
			StrokeLine( CPoint( cKnobFrame.left + 4.0f, vCenter ), CPoint( cKnobFrame.right - 4.0f, vCenter ) );
			StrokeLine( CPoint( cKnobFrame.left + 4.0f, vCenter + vDist ), CPoint( cKnobFrame.right - 4.0f, vCenter + vDist ) );

			SetHighColor( get_default_color( COL_SHADOW ) );
			vCenter -= 1.0f;
			StrokeLine( CPoint( cKnobFrame.left + 4.0f, vCenter - vDist ), CPoint( cKnobFrame.right - 4.0f, vCenter - vDist ) );
			StrokeLine( CPoint( cKnobFrame.left + 4.0f, vCenter ), CPoint( cKnobFrame.right - 4.0f, vCenter ) );
			StrokeLine( CPoint( cKnobFrame.left + 4.0f, vCenter + vDist ), CPoint( cKnobFrame.right - 4.0f, vCenter + vDist ) );
		}
	}

	SetHighColor( get_default_color( COL_SCROLLBAR_BG ) );

	cBounds.left   += 1;
	cBounds.top    += 1;
	cBounds.right  -= 1;
	cBounds.bottom -= 1;

	CRect cFrame( m->m_cKnobArea );
	if ( m->m_nOrientation == HORIZONTAL )
	{
		cFrame.right = cKnobFrame.left - 1.0f;
	}
	else
	{
		cFrame.bottom = cKnobFrame.top - 1.0f;
	}
	FillRect( cFrame );

	cFrame = m->m_cKnobArea;
	if ( m->m_nOrientation == HORIZONTAL )
	{
		cFrame.left = cKnobFrame.right + 1.0f;
	}
	else
	{
		cFrame.top = cKnobFrame.bottom + 1.0f;
	}

	FillRect( cFrame );
	if ( m->m_nOrientation == HORIZONTAL )
	{
		SetHighColor( Tint( get_default_color( COL_SHADOW ), 0.5f ) );
		StrokeLine( CPoint( m->m_cKnobArea.left, m->m_cKnobArea.top - 1.0f ), CPoint( m->m_cKnobArea.right, m->m_cKnobArea.top - 1.0f ) );
		SetHighColor( Tint( get_default_color( COL_SHINE ), 0.6f ) );
		StrokeLine( CPoint( m->m_cKnobArea.left, m->m_cKnobArea.bottom + 1.0f ), CPoint( m->m_cKnobArea.right, m->m_cKnobArea.bottom + 1.0f ) );
	}
	else
	{
		SetHighColor( Tint( get_default_color( COL_SHADOW ), 0.5f ) );
		StrokeLine( CPoint( m->m_cKnobArea.left - 1.0f, m->m_cKnobArea.top ), CPoint( m->m_cKnobArea.left - 1.0f, m->m_cKnobArea.bottom ) );
		SetHighColor( Tint( get_default_color( COL_SHINE ), 0.6f ) );
		StrokeLine( CPoint( m->m_cKnobArea.right + 1.0f, m->m_cKnobArea.top ), CPoint( m->m_cKnobArea.right + 1.0f, m->m_cKnobArea.bottom ) );
	}

	for ( int i = 0 ; i < 4 ; ++i )
	{
		if ( m->m_acArrowRects[i].IsValid() )
		{
			CRect cBmRect;
			CBitmap* pcBitmap;

			if ( m->m_nOrientation == HORIZONTAL )
			{
				pcBitmap = get_std_bitmap( ((i & 0x01) ? BMID_ARROW_RIGHT : BMID_ARROW_LEFT), 0, &cBmRect );
			}
			else
			{
				pcBitmap = get_std_bitmap( ((i & 0x01) ? BMID_ARROW_DOWN : BMID_ARROW_UP), 0, &cBmRect );
			}

			DrawFrame( m->m_acArrowRects[i], (m->m_abArrowStates[i] ? FRAME_RECESSED : FRAME_RAISED) | FRAME_THIN );

			SetDrawingMode( DM_OVER );

			DrawBitmap( pcBitmap, cBmRect, cBmRect.Bounds() + CPoint( (m->m_acArrowRects[i].Width() + 1.0f) * 0.5 - (cBmRect.Width() + 1.0f) * 0.5,
															(m->m_acArrowRects[i].Height() + 1.0f) * 0.5 - (cBmRect.Height() + 1.0f) * 0.5 ) +
						m->m_acArrowRects[i].LeftTop() );

			SetDrawingMode( DM_COPY );

		}
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

float CScrollBar::_PosToVal( CPoint cPos ) const
{
	float vValue = m->m_vMin;

	cPos -= m->m_cHitPos;

	if ( m->m_nOrientation == HORIZONTAL )
	{
		float        vSize = (m->m_cKnobArea.Width()+1.0f) * m->m_vProportion;
		float        vLen  = (m->m_cKnobArea.Width()+1.0f) - vSize;

		if ( vLen > 0.0f )
		{
			float vPos = (cPos.x - m->m_cKnobArea.left) / vLen;
			vValue = m->m_vMin + (m->m_vMax - m->m_vMin) * vPos;
		}
	}
	else
	{
		float vSize = (m->m_cKnobArea.Height()+1.0f) * m->m_vProportion;
		float vLen  = (m->m_cKnobArea.Height()+1.0f) - vSize;

		if ( vLen > 0.0f )
		{
			float vPos = (cPos.y - m->m_cKnobArea.top) / vLen;
			vValue = m->m_vMin + (m->m_vMax - m->m_vMin) * vPos;
		}
	}

	return( min( max( vValue, m->m_vMin ), m->m_vMax ) );
}


void CScrollBar::FrameResized( float inWidth, float inHeight )
{
	CRect cBounds = Bounds();
	CRect cArrowRect = cBounds;

	m->m_cKnobArea = cBounds;
	if ( m->m_nOrientation == HORIZONTAL )
	{
		cArrowRect.right = ceil( (cArrowRect.Height() + 1.0f) * 0.7 ) - 1.0f;
		float vWidth = cArrowRect.Width() + 1.0f;

		m->m_acArrowRects[0] = cArrowRect;
		m->m_acArrowRects[1] = cArrowRect + CPoint( vWidth, 0.0f );
		m->m_acArrowRects[2] = cArrowRect + CPoint( cBounds.Width() - vWidth * 2.0f + 1.0f, 0.0f );
		m->m_acArrowRects[3] = cArrowRect + CPoint( cBounds.Width() - vWidth + 1.0f, 0.0f );

		if ( m->m_acArrowRects[0].right > m->m_acArrowRects[3].left )
		{
			m->m_acArrowRects[0].right = floor( (cBounds.Width() + 1.0f) * 0.5f - 1.0f );
			m->m_acArrowRects[3].left  = m->m_acArrowRects[0].right + 1.0f;
			m->m_acArrowRects[1].left  = m->m_acArrowRects[0].right + 1.0f;
			m->m_acArrowRects[1].right = m->m_acArrowRects[1].left - 1.0f;
			m->m_acArrowRects[2].right = m->m_acArrowRects[3].left - 1.0f;
			m->m_acArrowRects[2].left  = m->m_acArrowRects[2].right + 1.0f;
		}
		else if ( m->m_acArrowRects[1].right + 16.0f > m->m_acArrowRects[2].left )
		{
			m->m_acArrowRects[1].right = m->m_acArrowRects[1].left - 1.0f;
			m->m_acArrowRects[2].left = m->m_acArrowRects[2].right + 1.0f;
		}
		m->m_cKnobArea.left    = m->m_acArrowRects[1].right + 1.0f;
		m->m_cKnobArea.right   = m->m_acArrowRects[2].left  - 1.0f;
		m->m_cKnobArea.top    += 1.0f;
		m->m_cKnobArea.bottom -= 1.0f;
	}
	else
	{
		cArrowRect.bottom = ceil( (cArrowRect.Width() + 1.0f) * 0.7 ) - 1.0f;

		float vWidth = cArrowRect.Height() + 1.0f;

		m->m_acArrowRects[0] = cArrowRect;
		m->m_acArrowRects[1] = cArrowRect + CPoint( 0.0f, vWidth );
		m->m_acArrowRects[2] = cArrowRect + CPoint( 0.0f, cBounds.Height() - vWidth * 2.0f + 1.0f );
		m->m_acArrowRects[3] = cArrowRect + CPoint( 0.0f, cBounds.Height() - vWidth + 1.0f );

		if ( m->m_acArrowRects[0].bottom > m->m_acArrowRects[3].top )
		{
			m->m_acArrowRects[0].bottom = floor( (cBounds.Width() + 1.0f) * 0.5f - 1.0f );
			m->m_acArrowRects[3].top    = m->m_acArrowRects[0].bottom + 1.0f;
			m->m_acArrowRects[1].top    = m->m_acArrowRects[0].bottom + 1.0f;
			m->m_acArrowRects[1].bottom = m->m_acArrowRects[1].top - 1.0f;
			m->m_acArrowRects[2].bottom = m->m_acArrowRects[3].top - 1.0f;
			m->m_acArrowRects[2].top    = m->m_acArrowRects[2].bottom + 1.0f;
		}
		else if ( m->m_acArrowRects[1].bottom + 16.0f > m->m_acArrowRects[2].top )
		{
			m->m_acArrowRects[1].bottom = m->m_acArrowRects[1].top - 1.0f;
			m->m_acArrowRects[2].top    = m->m_acArrowRects[2].bottom + 1.0f;
		}
		m->m_cKnobArea.top    = m->m_acArrowRects[1].bottom + 1.0f;
		m->m_cKnobArea.bottom = m->m_acArrowRects[2].top  - 1.0f;
		m->m_cKnobArea.left  += 1.0f;
		m->m_cKnobArea.right -= 1.0f;
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CRect CScrollBar::GetKnobFrame( void ) const
{
	CRect        cBounds = Bounds();
	CRect        cRect;

	cBounds.left   += 1.0f;
	cBounds.top    += 1.0f;
	cBounds.right  -= 1.0f;
	cBounds.bottom -= 1.0f;

	cRect = m->m_cKnobArea;

	float vValue = Value().AsFloat();

	if ( vValue > m->m_vMax )
	{
		vValue = m->m_vMax;
	}
	else if ( vValue < m->m_vMin )
	{
		vValue = m->m_vMin;
	}

	if ( m->m_nOrientation == HORIZONTAL )
	{
		float vViewLength = m->m_cKnobArea.Width() + 1.0f;
		float vSize = max( SB_MINSIZE, vViewLength * m->m_vProportion );
		float vLen  = vViewLength - vSize;
		float vDelta = m->m_vMax - m->m_vMin;

		if ( vDelta > 0.0f )
		{
			cRect.left  = cRect.left + vLen * (vValue - m->m_vMin) / vDelta;
			cRect.right = cRect.left + vSize - 1.0f;
		}
	}
	else
	{
		float vViewLength = m->m_cKnobArea.Height() + 1.0f;
		float vSize  = max( SB_MINSIZE, vViewLength * m->m_vProportion );
		float vLen   = vViewLength - vSize;
		float vDelta = m->m_vMax - m->m_vMin;

		if ( vDelta > 0.0f )
		{
			cRect.top = cRect.top + vLen * (vValue - m->m_vMin) / vDelta;
			cRect.bottom = cRect.top + vSize - 1.0f;
		}
	}
	return( cRect & m->m_cKnobArea );
}

