/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */


#include <assert.h>
#include <stdio.h>

#include <support/SupportDefs.h>

#include <interface/Slider.h>
#include <interface/Font.h>
#include <interface/Bitmap.h>

#include <app/Message.h>

#include <macros.h>

using namespace os;

#define TICK_LENGTH        4.0f
#define TICK_SPACING       3.0f
#define VLABEL_SPACING     3.0f
#define HLABEL_SPACING     2.0f

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CSlider::CSlider( CRect inFrame, const char* inName, CMessage* pcMsg, uint32 nTickFlags, int nTickCount,
                knob_mode eKnobMode, orientation eOrientation, uint32 nResizeMask )
    : CControl( inFrame, inName, "", pcMsg, nResizeMask, WID_WILL_DRAW | WID_FULL_UPDATE_ON_RESIZE )
{
	m_eOrientation = eOrientation;
	m_eKnobMode    = eKnobMode;
	m_vMin         = 0.0f;
	m_vMax         = 1.0f;
	m_vSmallStep   = 0.05f;
	m_vBigStep     = 0.1f;
//    m_vValue           = m_vMin;
	m_bTrack       = false;
	m_bChanged     = false;
	m_vSliderSize  = 5.0f;
	m_nNumSteps    = 0;
	m_nNumTicks    = nTickCount;
	m_nTickFlags   = nTickFlags;

	mFillColor = get_default_color( COL_SCROLLBAR_BG );;
	mBarColor = mFillColor;

	SetValue( m_vMin, false );

	m_pcRenderView   = new CView( inFrame.Bounds(), inName );
	if ( inFrame.IsValid() )
	{
		m_pcRenderBitmap = new CBitmap( int(inFrame.Width()) + 1, int(inFrame.Height()) + 1, CS_RGB16, true );
		m_pcRenderBitmap->AddChild( m_pcRenderView );

		_RefreshDisplay();
	}
	else
	{
		m_pcRenderBitmap = NULL;
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CSlider::~CSlider()
{
	if ( m_pcRenderBitmap != NULL )
	{
		delete m_pcRenderBitmap;
	}
	else
	{
		delete m_pcRenderView;
	}
}

/** Sets the background color to the one of the parent view
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CSlider::AttachedToWindow()
{
	SetLowColor( Parent()->LowColor() );
}


void CSlider::FrameResized( float inWidth, float inHeight )
{
	Sync();
	if ( m_pcRenderBitmap != NULL )
	{
		m_pcRenderBitmap->RemoveChild( m_pcRenderView );
		delete m_pcRenderBitmap;
	}

	CRect cBounds(GetNormalizedBounds());

	if ( cBounds.IsValid() )
	{
		m_pcRenderView->SetFrame( cBounds );
		m_pcRenderBitmap = new CBitmap( int(cBounds.Width()) + 1, int(cBounds.Height()) + 1, CS_RGB16, true );
		m_pcRenderBitmap->AddChild( m_pcRenderView );

		RenderSlider( m_pcRenderView );
		m_pcRenderBitmap->Sync();
		Invalidate();
	}
	else
	{
		m_pcRenderBitmap = NULL;
	}
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CSlider::GetPreferredSize( float* outWidth, float* outHeight )
{
	font_height	sHeight;
	CPoint		cSize;

	GetFontHeight( &sHeight );

	if ( m_eOrientation == HORIZONTAL )
	{
		cSize.Set( 100.0f, BarThickness() );
		CRect cSliderFrame = BarFrame();
		CRect cKnobFrame   = ThumbFrame();

		float vCenter = ceil(cSliderFrame.top + (cSliderFrame.Height()+1.0f)*0.5f);
		if ( m_cMinLabel.size() > 0 && m_cMaxLabel.size() > 0 )
		{
			float vStrWidth = StringWidth( m_cMinLabel ) + StringWidth( m_cMaxLabel ) + 20.0f;
			if ( vStrWidth > cSize.x )
			{
				cSize.x = vStrWidth;
			}
			cSliderFrame.bottom += sHeight.ascender + sHeight.descender + TICK_SPACING + TICK_LENGTH + HLABEL_SPACING;
		}
		else if ( m_nTickFlags & TICKS_BELOW )
		{
			cSliderFrame.bottom += TICK_SPACING + TICK_LENGTH;
		}

		if ( vCenter + (cKnobFrame.Height()+1.0f)*0.5f > cSliderFrame.bottom )
		{
			cSliderFrame.bottom = vCenter + (cKnobFrame.Height()+1.0f)*0.5f;
		}

		cSize.y = cSliderFrame.bottom + 1.0f;
	}
	else
	{
		cSize.Set( BarThickness(), 100.0f );

		if ( m_nTickFlags & TICKS_LEFT )
		{
			cSize.x += TICK_SPACING + TICK_LENGTH + 1.0f;
		}

		if ( m_nTickFlags & TICKS_RIGHT )
		{
			cSize.x += TICK_SPACING + TICK_LENGTH + 1.0f;
		}

		if ( m_cMinLabel.size() > 0 )
		{
			float vStrWidth = StringWidth( m_cMinLabel );
			if ( vStrWidth > cSize.x )
			{
				cSize.x = vStrWidth;
			}
			cSize.y += sHeight.ascender + sHeight.descender + VLABEL_SPACING;
		}

		if ( m_cMaxLabel.size() > 0 )
		{
			float vStrWidth = StringWidth( m_cMaxLabel );
			if ( vStrWidth > cSize.x )
			{
				cSize.x = vStrWidth;
			}
			cSize.y += sHeight.ascender + sHeight.descender + VLABEL_SPACING;
		}

		if ( cSize.x < ThumbFrame().Width() + 1.0f )
		{
			cSize.x = ThumbFrame().Width() + 1.0f;
		}
	}

	if (outWidth)
		*outWidth = cSize.x;

	if (outHeight)
		*outHeight = cSize.y;
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

bool CSlider::Invoked( CMessage* pcMsg )
{
	pcMsg->AddBool( "final", !m_bTrack );

	return( true );
/*
    if ( NULL == a_pcMessage )
	{
        a_pcMessage = Message();
    }
	
    if ( NULL != a_pcMessage )
    {
        CMessage* pcMsg = new CMessage( *a_pcMessage );

        if ( m_bTrack ) {
            pcMsg->AddBool( "final", false );
        } else {
            pcMsg->AddBool( "final", true );
        }
        return( Control::Invoke( pcMsg ) );
    }
    return( -1 );*/
}


void CSlider::PostValueChange( const Variant& cNewValue )
{
	m_bChanged = true;
	_RefreshDisplay();
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------
/*
void CSlider::SetCurValue( float vValue, bool bInvoke )
{
    if ( GetCurValue() == vValue ) {
        return;
    }
//    Control::SetValue( nValue );

    m_vValue = vValue;
    m_bChanged = true;
    _RefreshDisplay();

    if ( bInvoke == false ) {
        return;
    }

    Invoke();
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CSlider::SetCurValue( float vValue )
{
    SetCurValue( vValue, true );
}

float CSlider::GetCurValue() const
{
    return( m_vValue );
}
*/


/** Modify the base color of the slider bar.
 * \par Description:
 *        Set the base color used to draw the slider bar.
 *        This color is used for the "unfilled" part of the slider,
 *        or for the whole slider if a fill color is not used.
 * \param inColor - The color to use as the base color for the slider bar
 * \sa BarColor(), UseFillColor(), FillColor()
 * \author        Bill Hayden (hayden@haydentech.com)
 *****************************************************************************/

void CSlider::SetBarColor(Color32_s inColor)
{
	mBarColor = inColor;
	_RefreshDisplay();
}


/** Modify the fill color of the slider bar.
 * \par Description:
 *        Set the fill color used to draw the slider bar.
 *        This color is used for the "filled" part of the slider.
 *        The use of a fill color can be turned off with this function.
 * \param inUseIt - Should we use a fill color at all?
 * \param inColor - The color to use as a fill color for the slider bar
 * \sa BarColor(), SetBarColor(), FillColor()
 * \author        Bill Hayden (hayden@haydentech.com)
 *****************************************************************************/
 
void CSlider::UseFillColor(bool inUseIt, const Color32_s* inColor)
{
	if (inUseIt && inColor != NULL)
		mFillColor = *inColor;
	else
		mFillColor = mBarColor;

	_RefreshDisplay();
}

bool CSlider::FillColor(rgb_color* outColor) const
{
	if (outColor)
		*outColor = mFillColor;
	
	return true;
}


/** Set the thickness of the slider bar.
 * \par Description:
 *        Set the thickness of the slider bar in pixels. For a horizontal
 *        slider it sets the height and for a vertical one it sets the width.
 * \param inSize - The desired slider bar thickness in pixels.
 * \sa BarThickness(), BarFrame()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CSlider::SetBarThickness( float inSize )
{
	m_vSliderSize = inSize;
	_RefreshDisplay();
}


/** Get the current slider thickness.
 * \par Description:
 *        Returns the width for a horizontal and the height for a vertical
 *        slider of the slider-bar.
 * \return The thickness of the slider-bar in pixels.
 * \sa SetBarThickness(), BarFrame()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

float CSlider::BarThickness() const
{
	return( m_vSliderSize );
}


/** Set format string for the progress label.
 * \par Description:
 *        Set a printf() style string that will be used to format the
 *        "progress" label printed above the slider. The format function
 *        will be passed the sliders value as a parameter so if you throw
 *        in an %f somewhere it will be replaced with the current value.
 * \par Note:
 *        If you need some more advanced formatting you can inherit a new
 *        class from os::Slider and overload the GetProgressString() and
 *        return the string to be used as a progress label.
 * \param cFormat - The printf() style format string used to generate the
 *                    progress label.
 * \sa GetProgStrFormat(), GetProgressString(), SetLimitLabels(), GetLimitLabels()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CSlider::SetProgStrFormat( const std::string& cFormat )
{
	m_cProgressFormat = cFormat;
	_RefreshDisplay();
}


std::string CSlider::GetProgStrFormat() const
{
	return( m_cProgressFormat );
}


std::string CSlider::GetProgressString() const
{
	if ( m_cProgressFormat.size() == 0 )
	{
		return( "" );
	}
	else
	{
		char* pzString = new char[m_cProgressFormat.size() + 32];
		sprintf( pzString, m_cProgressFormat.c_str(), Value().AsDouble() );
		std::string cLabel( pzString );
		delete[] pzString;
		return( cLabel );
	}
}


/** Set the number of possible knob positions.
 * \par Description:
 *        The step count is used to limit the number of possible thumb
 *        positions. The points are equaly spread out over the length
 *        of the slider and the thumb will "snap" to these points.
 *        If set to the same value as SetHashMarkCount() the thumb will
 *        snap to the ticks.
 *
 *        If the value is less than 2 the thumb is not snapped.
 * \param nCound - Number of possible thumb positions. 0 to disable snapping.
 * \sa GetStepCount(), SetHashMarkCount(), HashMarkCount()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CSlider::SetStepCount( int nCount )
{
	m_nNumSteps = nCount;
}


/** Obtain the step-count as set by SetStepCount()
 * \return The number of possible knob positions.
 * \sa SetStepCount()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

int CSlider::GetStepCount() const
{
	return( m_nNumSteps );
}

/** Set number of "ticks" rendered along the slider.
 * \par Description:
 *        The slider can render "ticks" along both sides of the slider-bar.
 *        With SetHashMarkCount() you can set how many such ticks should be rendered
 *        and with SetTickFlags() you can set on which side (if any) the ticks
 *        should be rendered. To make the knob "snap" to the ticks you can set
 *        the same count with SetStepCount().
 * \param nCount - The number of ticks to be rendered.
 * \sa HashMarkCount(), SetTickFlags(), GetTickFlags(), SetStepCount()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/


void CSlider::SetHashMarkCount( int nCount )
{
	m_nNumTicks = nCount;
	_RefreshDisplay();
}

/** Obtain the tick count as set with SetHashMarkCount()
 * \return The number of ticks to be rendered
 * \sa SetHashMarkCount()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/


int CSlider::HashMarkCount() const
{
	return( m_nNumTicks );
}

/** Configure where the slider ticks should be rendered.
 * \par Description:
 *        The flag can be any combination of TICKS_ABOVE and TICKS_BELOW for
 *        horizontal sliders and TICKS_LEFT and TICKS_RIGHT for vertical
 *        sliders. The render ticks on both side you can bitwize "or" the
 *        flags together.
 * \par Warning:
 * \param nFlags - The new tick flags.
 * \sa GetTickFlags(), SetHashMarkCount(), HashMarkCount()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/


void CSlider::SetTickFlags( uint32 nFlags )
{
	m_nTickFlags   = nFlags;
	_RefreshDisplay();
}

/** Obtain the tick-flags as set with SetTickFlags()
 * \return The current tick-flags
 * \sa SetTickFlags()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/


uint32 CSlider::GetTickFlags() const
{
	return( m_nTickFlags );
}

/** Set the static labels rendere at each end of the slider.
 * \par Description:
 *        The min/max labels are rendered below a horizontal slider
 *        and to the right of a vertical slider.
 * \par Note:
 *        Both label must be set for the labels to be rendered.
 * \param cMin - The label to be rendered at the "smaller" end of the slider.
 * \param cMax - The label to be rendered at the "greater" end of the slider.
 * \sa GetLimitLabels(), SetProgStrFormat()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CSlider::SetLimitLabels( const char* cMinLabel, const char* cMaxLabel )
{
	m_cMinLabel = cMinLabel;
	m_cMaxLabel = cMaxLabel;
	_RefreshDisplay();
}


const char* CSlider::MinLimitLabel()
{
	return( m_cMinLabel.c_str() );
}


const char* CSlider::MaxLimitLabel()
{
	return( m_cMaxLabel.c_str() );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CSlider::EnableStatusChanged( bool bIsEnabled )
{
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CSlider::MouseDown( CPoint cPosition )
{
	m_cHitPos  = cPosition - ValToPos( Value() );;
	m_bTrack   = true;
	m_bChanged = false;
	MakeFocus( true );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CSlider::MouseUp( CPoint cPosition )
{
	m_bTrack = false;
	if ( m_bChanged )
	{
		Invoke();        // Send a 'final' message
		m_bChanged = false;
	}

	MakeFocus( false );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CSlider::MouseMoved( CPoint cNewPos, uint32 nCode, const CMessage* pcData )
{
	if ( m_bTrack )
	{
		SetValue( ValueForPoint( cNewPos - m_cHitPos ) );
	}
}


void CSlider::RenderSlider( CView* pcRenderView )
{
	CRect cBounds = GetNormalizedBounds();

	pcRenderView->FillRect( cBounds, LowColor() );

	pcRenderView->SetViewColor( get_default_color( COL_SCROLLBAR_KNOB ) );

	pcRenderView->SetHighColor( get_default_color( COL_SCROLLBAR_BG ) );

	CRect cSliderBounds = BarFrame();

	pcRenderView->DrawFrame( cSliderBounds, FRAME_RECESSED | FRAME_TRANSPARENT );

	cSliderBounds.Resize( 2.0f, 2.0f, -2.0f, -2.0f );

	if ( cSliderBounds.IsValid() )
	{
		CPoint cPos = ValToPos( Value() );

		CRect cSliderBounds2 = cSliderBounds;
		if ( m_eOrientation == HORIZONTAL ) {
			cSliderBounds2.left = cPos.x;
			if ( cPos.x + 1.0f < cSliderBounds.right ) {
				cSliderBounds.right = cPos.x + 1.0f;
			}
			if ( cSliderBounds2.left < cSliderBounds.left )
			{
				cSliderBounds2.left = cSliderBounds.left;
			}
		}
		else
		{
			if ( cPos.y + 1.0f > cSliderBounds2.top )
			{
				cSliderBounds.top = cPos.y + 1.0f;
			}
			if ( cPos.y < cSliderBounds2.bottom )
			{
				cSliderBounds2.bottom = cPos.y;
			}
		}

		if ( cSliderBounds.IsValid() )
		{
			pcRenderView->FillRect( cSliderBounds, mFillColor );
		}

		if ( cSliderBounds2.IsValid() )
		{
			pcRenderView->FillRect( cSliderBounds2, mBarColor );
		}
	}

	DrawHashMarks( pcRenderView );
	RenderLabels( pcRenderView );
	DrawThumb( pcRenderView );
}


void CSlider::DrawThumb( CView* pcRenderView )
{
	CRect cKnobFrame = ThumbFrame() + ValToPos( Value() );

	if ( m_eKnobMode == KNOB_SQUARE )
	{
		pcRenderView->SetViewColor( get_default_color( COL_SCROLLBAR_KNOB ) );
		pcRenderView->DrawFrame( cKnobFrame, FRAME_RAISED );
	}
	else
	{
		CPoint cCenter = ValToPos( Value() );

		if ( m_eOrientation == HORIZONTAL )
		{
			pcRenderView->SetHighColor( LowColor() );
			for ( int i = 1 ; i < 6 ; ++i )
			{
				pcRenderView->StrokeLine( CPoint( cCenter.x - i + 1.0f, cCenter.y + i ), CPoint( cCenter.x + i - 1.0f, cCenter.y + i ) );
			}
			pcRenderView->SetHighColor( 0, 0, 0 );
			pcRenderView->StrokeLine( cCenter, CPoint( cCenter.x + 7.0f, cCenter.y + 7.0f ) );
			pcRenderView->StrokeLine( CPoint( cCenter.x - 7.0f, cCenter.y + 7.0f ) );
			pcRenderView->StrokeLine( cCenter );
			pcRenderView->StrokeLine( CPoint( cCenter.x - 6.0f, cCenter.y + 6 ), CPoint( cCenter.x + 6.0f, cCenter.y + 6 ) );
		}
		else
		{
			pcRenderView->SetHighColor( LowColor() );
			for ( int i = 1 ; i < 6 ; ++i )
			{
				pcRenderView->StrokeLine( CPoint( cCenter.x - i, cCenter.y - i + 1.0f ), CPoint( cCenter.x - i, cCenter.y + i - 1.0f ) );
			}
			pcRenderView->SetHighColor( 0, 0, 0 );
			pcRenderView->StrokeLine( cCenter, CPoint( cCenter.x - 7.0f, cCenter.y - 7.0f ) );
			pcRenderView->StrokeLine( CPoint( cCenter.x - 7.0f, cCenter.y + 7.0f ) );
			pcRenderView->StrokeLine( cCenter );
		}
	}
}


void CSlider::RenderLabels( CView* pcRenderView )
{
	CRect cBounds = GetNormalizedBounds();
	CRect cSliderFrame = BarFrame();
	CPoint cCenter( cBounds.Width() * 0.5f, cBounds.Height() * 0.5f );
	font_height sHeight;
	pcRenderView->GetFontHeight( &sHeight );

	pcRenderView->SetHighColor( 0, 0, 0 );
	if ( m_eOrientation == HORIZONTAL )
	{
		if ( m_cMinLabel.size() > 0 && m_cMaxLabel.size() > 0 )
		{
			pcRenderView->MovePenTo( 0.0f, cSliderFrame.bottom + sHeight.ascender + TICK_SPACING + TICK_LENGTH + HLABEL_SPACING );
			pcRenderView->DrawString( m_cMinLabel.c_str() );
			pcRenderView->MovePenTo( cBounds.right - StringWidth( m_cMaxLabel.c_str() ),
									cSliderFrame.bottom + sHeight.ascender + TICK_SPACING + TICK_LENGTH + HLABEL_SPACING );
			pcRenderView->DrawString( m_cMaxLabel.c_str() );
		}
		std::string cProgress = GetProgressString();

		if ( cProgress.size() > 0 )
		{
			pcRenderView->MovePenTo( 0.0f, cSliderFrame.top - sHeight.descender - TICK_SPACING - TICK_LENGTH - HLABEL_SPACING );
			pcRenderView->DrawString( cProgress.c_str(), cProgress.size() );
		}
	}
	else
	{
		CRect cSliderBounds = BarFrame();
		float vOffset = ceil( ThumbFrame().Height() * 0.5f ) + VLABEL_SPACING;
		if ( m_cMaxLabel.size() > 0 )
		{
			pcRenderView->MovePenTo( cCenter.x - StringWidth( m_cMaxLabel ) * 0.5f, cSliderBounds.top + sHeight.descender - vOffset );
			pcRenderView->DrawString( m_cMaxLabel.c_str() );
		}

		if ( m_cMinLabel.size() > 0 )
		{
			pcRenderView->MovePenTo( cCenter.x - StringWidth( m_cMinLabel ) * 0.5f, cSliderBounds.bottom + sHeight.ascender + vOffset );
			pcRenderView->DrawString( m_cMinLabel.c_str() );
		}
	}
}


void CSlider::DrawHashMarks( CView* pcRenderView )
{
	if ( m_nNumTicks > 0 )
	{
		CRect cSliderFrame( BarFrame() );

		float vScale = 1.0f/float(m_nNumTicks - 1);
		if ( m_eOrientation == HORIZONTAL )
		{
			float vWidth = floor(cSliderFrame.Width() - 1.0f);

			float y1 = floor( cSliderFrame.top - TICK_SPACING );
			float y2 = y1 - TICK_LENGTH;
			float y4 = floor( cSliderFrame.bottom + TICK_SPACING );
			float y3 = y4 + TICK_LENGTH;

			pcRenderView->SetHighColor( get_default_color( COL_SHADOW ) );
			for ( float i = 0 ; i < m_nNumTicks ; i += 1.0f )
			{
				float x = floor( cSliderFrame.left + vWidth * i * vScale );
				if ( m_nTickFlags & TICKS_ABOVE )
				{
					pcRenderView->StrokeLine( CPoint( x, y1 ), CPoint( x, y2 ) );
				}
				if ( m_nTickFlags & TICKS_BELOW )
				{
					pcRenderView->StrokeLine( CPoint( x, y3 ), CPoint( x, y4 ) );
				}
			}
			pcRenderView->SetHighColor( get_default_color( COL_SHINE ) );
			for ( float i = 0 ; i < m_nNumTicks ; i += 1.0f )
			{
				float x = floor( cSliderFrame.left + vWidth * i * vScale ) + 1.0f;
				if ( m_nTickFlags & TICKS_ABOVE )
				{
					pcRenderView->StrokeLine( CPoint( x, y1 ), CPoint( x, y2 ) );
				}
				if ( m_nTickFlags & TICKS_BELOW )
				{
					pcRenderView->StrokeLine( CPoint( x, y3 ), CPoint( x, y4 ) );
				}
			}
		}
		else
		{
			float vHeight = cSliderFrame.Height() - 1.0f;

			float x1 = floor( cSliderFrame.left - TICK_SPACING );
			float x2 = x1 - TICK_LENGTH;
			float x4 = floor( cSliderFrame.right + TICK_SPACING );
			float x3 = x4 + TICK_LENGTH;

			pcRenderView->SetHighColor( get_default_color( COL_SHADOW ) );
			for ( float i = 0 ; i < m_nNumTicks ; i += 1.0f )
			{
				float y = floor( cSliderFrame.top + vHeight * i * vScale );
				if ( m_nTickFlags & TICKS_ABOVE )
				{
					pcRenderView->StrokeLine( CPoint( x1, y ), CPoint( x2, y ) );
				}
				if ( m_nTickFlags & TICKS_BELOW )
				{
					pcRenderView->StrokeLine( CPoint( x3, y ), CPoint( x4, y ) );
				}
			}
			pcRenderView->SetHighColor( get_default_color( COL_SHINE ) );
			for ( float i = 0 ; i < m_nNumTicks ; i += 1.0f )
			{
				float y = floor( cSliderFrame.top + vHeight * i * vScale ) + 1.0f;
				if ( m_nTickFlags & TICKS_ABOVE )
				{
					pcRenderView->StrokeLine( CPoint( x1, y ), CPoint( x2, y ) );
				}
				if ( m_nTickFlags & TICKS_BELOW )
				{
					pcRenderView->StrokeLine( CPoint( x3, y ), CPoint( x4, y ) );
				}
			}
		}
	}
}


float CSlider::ValueForPoint( CPoint inPosition ) const
{
	CRect cSliderFrame = BarFrame();
	float vValue;

	CPoint cPos = inPosition - cSliderFrame.LeftTop();

	if ( m_eOrientation == HORIZONTAL )
	{
		if ( m_nNumSteps > 1 )
		{
			float vAlign = (cSliderFrame.Width()+1.0f) / float(m_nNumSteps-1);
			cPos.x = floor((cPos.x + vAlign*0.5f) / vAlign) * vAlign;
		}
		vValue = m_vMin + (cPos.x / (cSliderFrame.Width()+1.0f)) * (m_vMax - m_vMin);
	}
	else
	{
		if ( m_nNumSteps > 1 )
		{
			float vAlign = (cSliderFrame.Height()+1.0f) / float(m_nNumSteps-1);
			cPos.y = floor((cPos.y + vAlign*0.5f) / vAlign) * vAlign;
		}
		vValue = m_vMax + (cPos.y / (cSliderFrame.Height()+1.0f)) * (m_vMin - m_vMax);
	}

	if ( vValue < m_vMin )
	{
		return( m_vMin );
	}
	else if ( vValue > m_vMax )
	{
		return( m_vMax );
	}

	return( vValue );
}


CPoint CSlider::ValToPos( float vVal ) const
{
	CRect  cBounds = BarFrame();
	CRect  cSliderFrame = BarFrame();
	float vPos = (vVal - m_vMin) / (m_vMax - m_vMin);

	if ( m_eOrientation == HORIZONTAL )
	{
		return( CPoint( cSliderFrame.left + (cSliderFrame.Width()) * vPos, cBounds.top + (cBounds.Height()+1.0f) * 0.5f ) );
	}
	else
	{
		return( CPoint( cBounds.left + (cBounds.Width()+1.0f) * 0.5f, cSliderFrame.top + (cSliderFrame.Height()) * (1.0f-vPos) ) );
	}
}


CRect CSlider::ThumbFrame() const
{
	if ( m_eOrientation == HORIZONTAL )
	{
		return( CRect( -5.0f, -7.0f, 5.0f, 7.0f ) );
	}
	else
	{
		return( CRect( -7.0f, -5.0f, 7.0f, 5.0f ) );
	}
}


CRect CSlider::BarFrame() const
{
	CRect  cBounds( GetNormalizedBounds() );
	CRect  cKnobFrame = ThumbFrame();
	CRect  cSliderBounds;
	CPoint cCenter( (cBounds.Width()+1.0f) * 0.5f, (cBounds.Height()+1.0f) * 0.5f );

	float vSize = ceil(BarThickness() * 0.5f);

	if ( m_eOrientation == HORIZONTAL )
	{
		float vKWidth  = ceil((cKnobFrame.Width()+1.0f) * 0.5f );
		float vKHeight = ceil((cKnobFrame.Height()+1.0f) * 0.5f );
		float vCenter = vSize;

		if ( m_nTickFlags & TICKS_ABOVE ) {
			vCenter += TICK_SPACING + TICK_LENGTH;
		}


		cSliderBounds.top = 0;
		std::string cProgress = GetProgressString();
		if ( cProgress.size() > 0 ) {
			font_height sHeight;
			GetFontHeight( &sHeight );

			float vHeight = vSize + sHeight.ascender + sHeight.descender + TICK_SPACING + TICK_LENGTH + HLABEL_SPACING;
			if ( vHeight > vCenter )
			{
				vCenter = vHeight;
			}
		}

		if ( vCenter < vKHeight )
		{
			vCenter = vKHeight;
		}
//        cSliderBounds.bottom = cSliderBounds.top + ceil(BarThickness() * 0.5f);
//        cSliderBounds.top -= ceil( BarThickness() * 0.5f );
		cSliderBounds = CRect( vKWidth, vCenter - vSize, cBounds.right - vKWidth, vCenter + vSize );
	}
	else
	{
		float vKHeight = ceil( (cKnobFrame.Height()+1.0f) * 0.5f );

		float vTFHeight = 0.0f;
		float vBFHeight = 0.0f;

		font_height sHeight;
		GetFontHeight( &sHeight );

		if ( m_cMinLabel.size() > 0 )
		{
			vBFHeight = sHeight.ascender + sHeight.descender + VLABEL_SPACING;
		}

		if ( m_cMaxLabel.size() > 0 )
		{
			vTFHeight = sHeight.ascender + sHeight.descender + VLABEL_SPACING;
		}
		cSliderBounds = CRect( cCenter.x - vSize, vKHeight + vTFHeight, cCenter.x + vSize, cBounds.bottom - vKHeight - vBFHeight );
	}
	cSliderBounds.Floor();
	return( cSliderBounds );
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CSlider::Draw( CRect cUpdateRect )
{
	if ( m_pcRenderBitmap != NULL )
	{
		DrawBitmap( m_pcRenderBitmap, cUpdateRect, cUpdateRect );
	}
}


void CSlider::_RefreshDisplay()
{
	if ( Window() == NULL )
	{
		return;
	}
	Sync();
	RenderSlider( m_pcRenderView );
	m_pcRenderBitmap->Sync();
	Invalidate();
	Flush();
}

