/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */

 
#include <cosmoe/Debug.h>
#include <assert.h>
#include <interface/Window.h>
#include <interface/TableView.h>
#include <interface/Font.h>
#include <app/Message.h>

using namespace os;

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

TableCell::TableCell()
{
	m_eHorAlign     = ALIGN_LEFT;
	m_eVerAlign     = ALIGN_CENTER;
	m_vHorWeight    = 1.0f;
	m_vVerWeight    = 1.0f;
	m_pcView        = NULL;
	m_nLeftBorder   = m_nRightBorder = m_nTopBorder = m_nBottomBorder = 3;
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

TableCell::~TableCell()
{
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void TableCell::SetView( CView* pcView )
{
	m_pcView = pcView;
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CView* TableCell::GetView( void ) const
{
	return( m_pcView );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void TableCell::SetWeights( float vHor, float vVer )
{
	m_vHorWeight = vHor;
	m_vVerWeight = vVer;
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void TableCell::GetPreferredSize( float* outWidth, float* outHeight )
{
	float width, height;

	if ( NULL != m_pcView )
	{
		m_pcView->GetPreferredSize( &width, &height );
		width += m_nLeftBorder + m_nRightBorder;
		height += m_nTopBorder + m_nBottomBorder;
	}
	else
	{
		width = 50.0f;
		height = 20.0f;
	}

	if (outWidth)
		*outWidth = width;

	if (outHeight)
		*outHeight = height;
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void TableCell::SetAlignment( alignment eHor, alignment eVer )
{
	m_eHorAlign = eHor;
	m_eVerAlign = eVer;
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void TableCell::SetWidth( int nWidth )
{
	m_cSize.x = nWidth;
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void TableCell::SetHeight( int nHeight )
{
	m_cSize.y = nHeight;
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void TableCell::SetSize( CPoint cSize )
{
	m_cSize = cSize;
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void TableCell::SetPos( CPoint cPos )
{
	m_cPos = cPos;
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void TableCell::Layout( void )
{
	if ( NULL != m_pcView )
	{
		float  width, height;

		m_pcView->GetPreferredSize(&width, &height);

		CRect cBounds =
			CRect( 0, 0, width - 1, height - 1 ) &
			CRect( 0, 0, m_cSize.x - m_nLeftBorder - m_nRightBorder - 1,
						 m_cSize.y - m_nTopBorder - m_nBottomBorder - 1 );

		CPoint cPos = m_cPos; // + CPoint( m_nLeftBorder, m_nTopBorder );

		switch( m_eHorAlign )
		{
			case ALIGN_LEFT:
				cPos.x += m_nLeftBorder;
				break;

			case ALIGN_CENTER:
				cPos.x += m_cSize.x / 2 - (cBounds.Width()+1.0f) / 2;
				break;

			case ALIGN_RIGHT:
				cPos.x += m_cSize.x - cBounds.Width() + 1.0f - m_nRightBorder;
				break;

			default:
				dbprintf( "Error: TableCell::Layout() invalid horizontal alignment %d\n", m_eHorAlign);
				break;
		}

		switch( m_eVerAlign )
		{
			case ALIGN_TOP:
				cPos.y += m_nTopBorder;
				break;

			case ALIGN_CENTER:
				cPos.y += m_cSize.y / 2 - (cBounds.Height()+1.0f) / 2;
				break;

			case ALIGN_BOTTOM:
				cPos.y += m_cSize.y - cBounds.Height() + 1.0f - m_nBottomBorder;
				break;

			default:
				dbprintf( "Error: TableCell::Layout() invalid vertical alignment %d\n", m_eHorAlign);
				break;
		}
		m_pcView->SetFrame( (cBounds + cPos).Floor() );
	}
}




//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

TableView::TableView( CRect inFrame, const char* pzName, const char* pzTitle,
                      int nNumCols, int nNumRows, uint32 nResizeMask )
    : CView( inFrame, pzName, nResizeMask, WID_WILL_DRAW )
{
	if ( NULL != pzTitle )
	{
		m_pzTitle = new char[ strlen( pzTitle ) + 1 ];
		strcpy( m_pzTitle, pzTitle );
		m_nLeftBorder = m_nRightBorder = m_nBottomBorder = 4;
		m_nTopBorder  = 14;        // Will be updated when added to window;
	}
	else
	{
		m_nLeftBorder = m_nRightBorder = m_nTopBorder = m_nBottomBorder = 0;
		m_pzTitle = NULL;
	}

	m_pacCells          = new TableCell[ nNumCols * nNumRows ];
	m_panMinWidths  = new float[ nNumCols ];
	m_panMaxWidths  = new float[ nNumCols ];
	m_panMinHeights = new float[ nNumRows ];
	m_panMaxHeights = new float[ nNumRows ];

	m_pavColWeights = new float[ nNumCols ];
	m_pavRowWeights = new float[ nNumRows ];

	m_nNumCols        = nNumCols;
	m_nNumRows        = nNumRows;
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

TableView::~TableView()
{
	delete [] m_pzTitle;
	delete [] m_pacCells;
	delete [] m_panMinWidths;
	delete [] m_panMaxWidths;
	delete [] m_panMinHeights;
	delete [] m_panMaxHeights;

	delete [] m_pavColWeights;
	delete [] m_pavRowWeights;
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void TableView::AttachedToWindow( void )
{
	CWindow* pcWindow = Window();

	assert( pcWindow != NULL );

	SetLowColor( get_default_color( COL_NORMAL ) );

	if ( NULL != m_pzTitle )
	{
		if ( '\0' == m_pzTitle[ 0 ] )
		{
			m_nTopBorder = m_nBottomBorder;
		}
		else
		{
			font_height sHeight;
			GetFontHeight( &sHeight );
			m_nTopBorder = sHeight.ascender + sHeight.descender + 2.0f;
		}
	}
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void TableView::Layout()
{
	Layout( Bounds() );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void TableView::AllAttached( void )
{
	Layout( Bounds() );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void TableView::FrameResized( float inWidth, float inHeight )
{
	Layout( Bounds() );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void TableView::SetColAlignment( int nCol, alignment eAlign )
{
	int        y;

	for ( y = 0 ; y < m_nNumRows ; ++y )
	{
		TableCell*        pcCell = GetCell( nCol, y );
		if ( NULL != pcCell )
		{
			pcCell->SetAlignment( eAlign, ALIGN_CENTER );
		}
	}
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void TableView::SetCellAlignment( int x, int y, alignment eHor, alignment eVer )
{
	TableCell*        pcCell = GetCell( x, y );

	if ( NULL != pcCell )
	{
		pcCell->SetAlignment( eHor, eVer );
	}
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void TableView::SetCellBorders( int x, int y, int nLeft, int nTop, int nRight, int nBottom )
{
	TableCell*        pcCell = GetCell( x, y );

	if ( NULL != pcCell )
	{
		pcCell->m_nLeftBorder         = nLeft;
		pcCell->m_nTopBorder         = nTop;
		pcCell->m_nRightBorder         = nRight;
		pcCell->m_nBottomBorder = nBottom;
	}
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void TableView::SetCellBorders( int x, int nLeft, int nTop, int nRight, int nBottom )
{
	int y;

	for ( y = 0 ; y < m_nNumRows ; ++y )
	{
		SetCellBorders( x, y, nLeft, nTop, nRight, nBottom );
	}
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void TableView::SetCellBorders( int nLeft, int nTop, int nRight, int nBottom )
{
	int x;

	for ( x = 0 ; x < m_nNumCols ; ++x )
	{
		SetCellBorders( x, nLeft, nTop, nRight, nBottom );
	}
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

TableCell* TableView::GetCell( int x, int y ) const
{
	if ( x < m_nNumCols && y < m_nNumRows )
	{
		return( &m_pacCells[ x + y * m_nNumCols ] );
	}

	return( NULL );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

CView* TableView::SetChild( CView* pcView, int x, int y, float vHorWeight, float vVerWeight )
{
	CView* pcPrevView = NULL;

	TableCell*        pcCell = GetCell( x, y );

	if ( NULL != pcCell )
	{
		pcPrevView = pcCell->GetView();

		if ( NULL != pcPrevView )
		{
			RemoveChild( pcPrevView );
		}

		pcCell->SetView( pcView );
		pcCell->SetWeights( vHorWeight, vVerWeight );
		if ( NULL != pcView )
		{
			AddChild( pcView );
		}
	}
	return( pcPrevView );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void TableView::CalcSizes( void ) const
{
	int        x;
	int        y;

	m_nMinWidth  = 0;
	m_nMaxWidth  = 0;
	m_nMinHeight = 0;
	m_nMaxHeight = 0;

	for ( x = 0 ; x < m_nNumCols ; ++x )
	{
		m_panMaxWidths[x]         = 0;
		m_panMinWidths[x]         = 0;
		m_pavColWeights[x] = 0.0f;
	}

	for ( y = 0 ; y < m_nNumRows ; ++y )
	{
		m_panMinHeights[y]         = 0;
		m_panMaxHeights[y]         = 0;
		m_pavRowWeights[y] = 0.0f;
	}

	for ( x = 0 ; x < m_nNumCols ; ++x )
	{
		for ( y = 0 ; y < m_nNumRows ; ++y )
		{
			/* FIXME
			TableCell*        pcCell = GetCell( x, y );

			CPoint        cMinSize = pcCell->GetPreferredSize( false );
			CPoint        cMaxSize = pcCell->GetPreferredSize( true );

			m_pavColWeights[ x ] += pcCell->m_vVerWeight;
			m_pavRowWeights[ y ] += pcCell->m_vHorWeight;


			if ( cMinSize.x > m_panMinWidths[x] )
			{
				m_panMinWidths[x] = cMinSize.x;
			}
			if ( cMaxSize.x > m_panMaxWidths[x] )
			{
				m_panMaxWidths[x] = cMaxSize.x;
			}
			if ( cMinSize.y > m_panMinHeights[y] )
			{
				m_panMinHeights[y] = cMinSize.y;
			}
			if ( cMaxSize.y > m_panMaxHeights[y] )
			{
				m_panMaxHeights[y] = cMaxSize.y;
			}
			*/
		}
	}
	for ( x = 0 ; x < m_nNumCols ; ++x )
	{
		m_nMinWidth += m_panMinWidths[x];
		m_nMaxWidth += m_panMaxWidths[x];
	}

	for ( y = 0 ; y < m_nNumRows ; ++y )
	{
		m_nMinHeight += m_panMinHeights[y];
		m_nMaxHeight += m_panMaxHeights[y];
	}
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void TableView::Layout( CRect cBounds )
{
	int        x;
	int        y;

	CalcSizes();

	cBounds.left   += m_nLeftBorder;
	cBounds.top           += m_nTopBorder;
	cBounds.right  -= m_nRightBorder;
	cBounds.bottom -= m_nBottomBorder;

/*
float        vHorAspect = (float) cBounds.Width() / (float) nWidth;
float        vVerAspect = (float) cBounds.Height() / (float) nHeight;
*/

	CPoint        cPos( 0, 0 );

	float        nDeltaW        = cBounds.Width() + 1.0f - m_nMinWidth;
	float        nDeltaH        = cBounds.Height() + 1.0f - m_nMinHeight;

	float nXSpacing = (cBounds.Width()+1.0f > m_nMaxWidth) ? ((cBounds.Width() + 1.0f - m_nMaxWidth) / m_nNumCols) : 0;
	float nYSpacing = (cBounds.Height()+1.0f > m_nMaxHeight) ? ((cBounds.Height() + 1.0f - m_nMaxHeight)        / m_nNumRows) : 0;

	for ( x = 0 ; x < m_nNumCols ; ++x )
	{
		float        nDeltaX = 0.0f;
		float        nDeltaY;

		cPos.y = 0;

		for ( y = 0 ; y < m_nNumRows ; ++y )
		{
			TableCell* pcCell = GetCell( x, y );

			if ( (m_nMaxWidth - m_nMinWidth) != 0 ) {
				nDeltaX = nDeltaW * (m_panMaxWidths[x] - m_panMinWidths[x]) / (m_nMaxWidth - m_nMinWidth);
			} else {
				nDeltaX = 0;
			}
			if ( (m_nMaxHeight - m_nMinHeight) != 0 ) {
				nDeltaY = nDeltaH * (m_panMaxHeights[y] - m_panMinHeights[y]) / (m_nMaxHeight - m_nMinHeight);
			} else {
				nDeltaY        = 0;
			}

			pcCell->SetSize( CPoint( m_panMinWidths[x] + nDeltaX, m_panMinHeights[y] + nDeltaY ) );
			pcCell->SetPos( cPos + CPoint( m_nLeftBorder, m_nTopBorder ) );
			pcCell->Layout();

			cPos.y += m_panMinHeights[y] + nDeltaY + nYSpacing;
		}
		cPos.x += m_panMinWidths[x] + nDeltaX + nXSpacing;
	}
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void TableView::GetPreferredSize( float* outWidth, float* outHeight )
{
    CalcSizes();
/*
    if ( bLargest )
	{
        return( CPoint( 10000, 10000 ) );
//                return( CPoint( m_nMaxWidth, m_nMaxHeight ) );
    }
*/
	if (outWidth)
	{
		*outWidth = m_nMinWidth + m_nLeftBorder + m_nRightBorder;
	}

	if (outHeight)
	{
		*outHeight = m_nMinHeight + m_nTopBorder + m_nBottomBorder;
	}

}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void TableView::Draw( CRect cUpdateRect )
{
	CWindow*        pcWindow = Window();

	if ( NULL != pcWindow )
	{
		CRect cBounds = Bounds();

		SetHighColor( get_default_color( COL_NORMAL ) );
		FillRect( Bounds() );

		if ( NULL != m_pzTitle )
		{
			SetHighColor( 0, 0, 0, 0 );

			CFont* pcFont = GetFont();

			if ( NULL != pcFont )
			{
				float        x = 20.0f;
				float        vStrWidth = 0.0f;

				if ( '\0' != m_pzTitle[0] )
				{
					font_height sHeight;

					pcFont->GetHeight( &sHeight );
					float vStrHeight = sHeight.ascender + sHeight.descender;

					vStrWidth = pcFont->GetStringWidth( m_pzTitle );

					cBounds.top = vStrHeight / 2.0f;
					DrawString( m_pzTitle, CPoint( x, sHeight.ascender ) );
				}
				// Left
				StrokeLine( CPoint( cBounds.left, cBounds.bottom ), CPoint( cBounds.left, cBounds.top ) );
				// Top
				if ( '\0' == m_pzTitle[0] )
				{
					StrokeLine( CPoint( cBounds.left, cBounds.top ), CPoint( cBounds.right, cBounds.top ) );
				}
				else
				{
					StrokeLine( CPoint( cBounds.left, cBounds.top ), CPoint( x - 3, cBounds.top ) );
					StrokeLine( CPoint( x + vStrWidth + 3.0f, cBounds.top ), CPoint( cBounds.right, cBounds.top ) );
				}
				// Right
				StrokeLine( CPoint( cBounds.right - 1, cBounds.top ), CPoint( cBounds.right - 1, cBounds.bottom ) );
				// Bottom
				StrokeLine( CPoint( cBounds.left + 2, cBounds.bottom - 1 ), CPoint( cBounds.right - 1, cBounds.bottom - 1 ) );


				SetHighColor( 255, 255, 255 );
				// Left
				StrokeLine( CPoint( cBounds.left + 1, cBounds.bottom ), CPoint( cBounds.left + 1, cBounds.top + 1 ) );
				// Top
				if ( '\0' == m_pzTitle[0] )
				{
					StrokeLine( CPoint( 1, cBounds.top + 1 ), CPoint( cBounds.right, cBounds.top + 1 ) );
				}
				else
				{
					StrokeLine( CPoint( 1, cBounds.top + 1 ), CPoint( x - 3, cBounds.top + 1 ) );
					StrokeLine( CPoint( x + vStrWidth + 3.0f, cBounds.top + 1 ), CPoint( cBounds.right, cBounds.top + 1 ) );
				}
				// Right
				StrokeLine( CPoint( cBounds.right, cBounds.top + 1 ), CPoint( cBounds.right, cBounds.bottom ) );
				// Bottom
				StrokeLine( CPoint( cBounds.left + 1, cBounds.bottom ), CPoint( cBounds.right, cBounds.bottom ) );
			}
		}
	}
}
