/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */

 
#include <stdio.h>

#include <interface/TabView.h>
#include <interface/Window.h>
#include <app/Message.h>

inline float max( float a, float b )
{
    return( (a < b) ? b : a );
}

using namespace os;

/** Contruct the TabView
 * \par Description:
 *        Not much happens here. All paramaters are sent to the View() constructor
 * \author        Kurt Skauen (kurt.skauen@c2i.net)
 *//////////////////////////////////////////////////////////////////////////////

TabView::TabView( CRect inFrame, const char* inTitle, uint32 nResizeMask, uint32 nFlags ) : CView( inFrame, inTitle, nResizeMask, nFlags & ~WID_FULL_UPDATE_ON_RESIZE )
{
	GetFontHeight( &m_sFontHeight );
	m_vGlyphHeight = m_sFontHeight.ascender + m_sFontHeight.descender + m_sFontHeight.line_gap;
	m_vTabHeight = ceil( m_vGlyphHeight * 1.2f + 6.0f );
	m_nSelectedTab  = 0;
	m_vScrollOffset = 0;
	m_bMouseClicked = false;
	m_vTotalWidth   = 4.0f;

	m_pcTopView = new TopView( CRect( 2, 0, inFrame.Width() + 1.0f - 3.0f, m_vTabHeight - 3.0f ), this );
	AddChild( m_pcTopView );
}


void TabView::FrameResized( float inWidth, float inHeight )
{
	CRect cBounds = Bounds();
	bool bNeedFlush = false;

	if ( m_nSelectedTab < m_cTabList.size() )
	{
		CView* pcView = m_cTabList[m_nSelectedTab].m_pcView;
		if ( pcView != NULL )
		{
			CRect cClientBounds = cBounds;
			cClientBounds.Resize( 2.0f, m_vTabHeight, -2.0f, -2.0f );
			cClientBounds.Floor();
			pcView->SetFrame( cClientBounds );
		}
	}

	if ( inWidth != 0.0f )
	{
		CRect cDamage = cBounds;

		cDamage.left = cDamage.right - max( 2.0f, inWidth + 8.0f );
		Invalidate( cDamage );
		bNeedFlush = true;
	}

	if ( inHeight != 0.0f )
	{
		CRect cDamage = cBounds;

		cDamage.top = cDamage.bottom - max( 2.0f, inHeight + 1.0f );
		Invalidate( cDamage );
		bNeedFlush = true;
	}

	cBounds.Floor();
	float vWidth = cBounds.Width() + 1.0f;

	float nOldOffset = m_vScrollOffset;
	if ( vWidth < m_vTotalWidth )
	{
		m_pcTopView->SetFrame( CRect( cBounds.left + 8.0f, cBounds.top, cBounds.right - 8.0f, m_vTabHeight - 3.0f ) );
		vWidth -= 16.0f;
		if ( m_vScrollOffset > 0 )
		{
			m_vScrollOffset = 0;
		}

		if ( m_vTotalWidth + m_vScrollOffset < vWidth )
		{
			m_vScrollOffset = vWidth - m_vTotalWidth;
		}

		if ( vWidth + 16.0f - inWidth >= m_vTotalWidth )
		{
			m_pcTopView->Invalidate();
			CRect cDamage( cBounds );
			cDamage.bottom = cDamage.top + m_vTabHeight;
			Invalidate( cDamage );
			bNeedFlush = true;
		}
	}
	else
	{
		m_pcTopView->SetFrame( CRect( cBounds.left + 2, cBounds.top, cBounds.right - 2, m_vTabHeight - 3 ) );
		if ( m_vScrollOffset > vWidth - m_vTotalWidth )
		{
			m_vScrollOffset = vWidth - m_vTotalWidth;
		}

		if ( m_vScrollOffset <  0 )
		{
			m_vScrollOffset = 0;
		}

		if ( vWidth - inHeight < m_vTotalWidth )
		{
			m_pcTopView->Invalidate();
			CRect cDamage( cBounds );
			cDamage.bottom = cDamage.top + m_vTabHeight;
			Invalidate( cDamage );
			bNeedFlush = true;
		}
	}

	if ( m_vScrollOffset != nOldOffset )
	{
		m_pcTopView->ScrollBy( m_vScrollOffset - nOldOffset, 0 );
		m_pcTopView->Invalidate();
		CRect cDamage( cBounds );
		cDamage.bottom = cDamage.top + m_vTabHeight;
		Invalidate( cDamage );
		bNeedFlush = true;
	}

	if ( bNeedFlush )
		Flush();
}

/**
 * \par Description:
 *        TabView overloades AllAttached to set the window as it's message target.
 * \sa Invoker::SetTarget(), Invoker::SetMessage(), CView::AllAttached()
 * \author        Kurt Skauen (kurt.skauen@c2i.net)
 *//////////////////////////////////////////////////////////////////////////////


void TabView::AllAttached()
{
	SetTarget( Window() );
	CView::AllAttached();
}

/** Add a tab at the end of the list.
 * \par Description:
 *        Add a new tab to the TabView. The new tab will be appended to the list
 *        and will appear rightmost. Each tab is associated with title that is
 *        printed inside it, and a View that will be made visible when the tab
 *        is selected or hidden when the tab is inactive. If you want to use
 *        the TabView to something else than flipping between views, you can
 *        pass a NULL pointer, or you can pass the same ViewPointer for all tabs.
 *        Then you can associate a message with the TabView that will be sent
 *        to it's target everytime the selection change.
 * \par Note:
 *        Views associated with tabs should *NOT* be attached to any other
 *        view when or after the tab is created. It will automatically be a
 *        child of the TabView when added to a tab.
 *
 *        Views associated with tabs *MUST* have the CF_FOLLOW_ALL flag set to
 *        be properly resized when the TabView is resized.
 *
 *        When a view is associated with a tab, the view will be resized to
 *        fit the interiour of the TabView.
 *
 * \param pcTitle - The string that should appear inside the tab.
 * \param pcView  - The View to be accosiated with the tab, or NULL.
 *
 * \return The zero based index of the new tab.
 * \sa InsertTab(), DeleteTab(), Invoker::SetMessage(), Invoker::SetTarget()
 * \author        Kurt Skauen (kurt.skauen@c2i.net)
 *//////////////////////////////////////////////////////////////////////////////

int TabView::AppendTab( const char* pzTitle, CView* pcView )
{
    return( InsertTab( m_cTabList.size(), pzTitle, pcView ) );
}

/** Insert tabs at a given position
 * \par Description:
 *        Same as AppendTab() except that InsertTab() accept a zero based index
 *        at which the tab will be inserted. Look at AppendTab() to get the full
 *        thruth.
 * \par Note:
 *        If the rightmost tab is selected, the selection will be moved to
 *        the previous tab, and a notification message (if any) is sendt.
 * \par Warning:
 *        The index *MUST* be between 0 and the current number of tabs!
 *
 * \param nIndex  - The zero based position where the tab is inserted.
 * \param pcTitle - The string that should appear inside the tab.
 * \param pcView  - The View to be accosiated with the tab, or NULL.
 *
 * \return The zero based index of the new tab (Same as nIndex).
 * \sa AppendTab(), DeleteTab(), Invoker::SetMessage(), Invoker::SetTarget()
 * \author        Kurt Skauen (kurt.skauen@c2i.net)
 *//////////////////////////////////////////////////////////////////////////////

int TabView::InsertTab( uint nIndex, const char* pzTitle, CView* pcView )
{
	m_cTabList.insert( m_cTabList.begin() + nIndex, Tab( pzTitle, pcView ) );
	if ( pcView != NULL && pcView->Parent() != this )
	{
		AddChild( pcView );
		CRect cBounds = GetNormalizedBounds();
		cBounds.Resize( 2, m_vTabHeight, -2, -2 );
		cBounds.Floor();
		pcView->SetFrame( cBounds );
		if ( nIndex != m_nSelectedTab )
		{
			pcView->Show( false );
		}
	}
	m_cTabList[nIndex].m_vWidth = ceil(StringWidth( pzTitle ) * 1.1f) + 4.0f;
	m_vTotalWidth += m_cTabList[nIndex].m_vWidth;
	m_pcTopView->Invalidate( Bounds() );
	m_pcTopView->Flush();
	return( nIndex );
}

/** Delete a given tab
 * \par Description:
 *        The tab at position nIndex is deleted, and the associated view is
 *        removed from the child list unless it is still associated with any
 *        of the remaining tabs. The view associated with the deleted tab
 *        is returned, and must be deleted by the caller if it is no longer
 *        needed.
 * \param nIndex - The zero based index of the tab to delete.
 * \return Pointer to the view associated with the deleted tab.
 * \sa AppendTab(), InsertTab(), GetTabView(), GetTabTitle()
 * \author        Kurt Skauen (kurt.skauen@c2i.net)
 *//////////////////////////////////////////////////////////////////////////////


CView* TabView::DeleteTab( uint nIndex )
{
	uint        nOldSelection = m_nSelectedTab;
	CView* pcView = m_cTabList[nIndex].m_pcView;

	if ( m_nSelectedTab == m_cTabList.size() - 1 )
	{
		SetSelection( m_nSelectedTab - 1 );
	}

	m_vTotalWidth -= m_cTabList[nIndex].m_vWidth;

	m_cTabList.erase( m_cTabList.begin() + nIndex );

	bool bStillMember = false;
	for ( uint i = 0 ; i < m_cTabList.size() ; ++i )
	{
		if ( m_cTabList[i].m_pcView == pcView )
		{
			bStillMember = true;
			break;
		}
	}
	if ( bStillMember == false )
	{
		RemoveChild( pcView );
		if ( nIndex != nOldSelection )
		{
			pcView->Show(); // Leave the view as we found it.
		}
	}
	m_pcTopView->Invalidate();
	Invalidate();
	m_pcTopView->Flush();
	return( pcView );
}

/** Get the View associated with a given tab.
 * \param nIndex - The zero based index of the tab.
 * \return Pointer to the View associated with the tab.
 * \sa GetTabTitle(), AppendTab(), InsertTab()
 * \author        Kurt Skauen (kurt.skauen@c2i.net)
 *//////////////////////////////////////////////////////////////////////////////


CView* TabView::GetTabView( uint nIndex ) const
{
	return( m_cTabList[ nIndex ].m_pcView );
}

/** Get the title of a given tab.
 * \param nIndex - The zero based index of the tab.
 * \return const reference to a STL string containing the title.
 * \sa GetTabView(), AppendTab(), InsertTab()
 * \author        Kurt Skauen (kurt.skauen@c2i.net)
 *//////////////////////////////////////////////////////////////////////////////

const std::string& TabView::GetTabTitle( uint nIndex ) const
{
	return( m_cTabList[ nIndex ].m_cTitle );
}


int TabView::SetTabTitle( uint nIndex, const std::string& cTitle )
{
	m_cTabList[ nIndex ].m_cTitle = cTitle;

	float vOldWidth = m_cTabList[nIndex].m_vWidth;
	m_cTabList[nIndex].m_vWidth = ceil(StringWidth( cTitle ) * 1.1f) + 4.0f;
	m_vTotalWidth += m_cTabList[nIndex].m_vWidth - vOldWidth;

	Invalidate( Bounds() );
	m_pcTopView->Invalidate( Bounds() );
	m_pcTopView->Flush();

	return( 0 );
}


/** Get the current selection
 * \par Description:
 *        Returns the zero based index of the selected tab.
 * \par Note:
 *        This index is also added to the notification sendt when the selection
 *        change. The selection is then added under the name "selection".
 * \return The zero based index of the selected tab.
 * \sa SetSelection(), Invoker::SetMessage(), Invoker::SetTarget()
 * \author        Kurt Skauen (kurt.skauen@c2i.net)
 *//////////////////////////////////////////////////////////////////////////////

uint TabView::GetSelection()
{
    return( m_nSelectedTab );
}

/** Select a tab, and optionally notify the target.
 * \par Description:
 *        Selects the given tab. If bNotify is true, and a messages has been
 *        assigned through Invoker::SetMessage() the message will be sendt
 *        to the target.
 * \param nIndex - The zero based index of the tab to select.
 * \param bNotify - Set to true if Invoker::Invoke() should be called.
 * \sa GetSelection(), Invoker::Invoke(), Invoker::SetMessage(), Invoker::SetTarget()
 * \author        Kurt Skauen (kurt.skauen@c2i.net)
 *//////////////////////////////////////////////////////////////////////////////

void TabView::SetSelection( uint nIndex, bool bNotify )
{
	if ( nIndex != m_nSelectedTab && nIndex < m_cTabList.size() )
	{
		if ( m_cTabList[m_nSelectedTab].m_pcView != NULL )
		{
			m_cTabList[m_nSelectedTab].m_pcView->Show( false );
		}
		m_nSelectedTab = nIndex;
		if ( m_cTabList[m_nSelectedTab].m_pcView != NULL )
		{
			CView* pcView = m_cTabList[m_nSelectedTab].m_pcView;
			CRect cClientBounds = GetNormalizedBounds();
			cClientBounds.Resize( 2.0f, m_vTabHeight, -2.0f, -2.0f );
			cClientBounds.Floor();
			pcView->SetFrame( cClientBounds );
			pcView->Show( true );
			pcView->MakeFocus( true );
		}
		Invalidate( Bounds() );
		m_pcTopView->Invalidate( m_pcTopView->Bounds() );
		m_pcTopView->Flush();
		if ( bNotify )
		{
			CMessage* pcMsg = Message();
			if ( pcMsg != NULL )
			{
				CMessage cMsg( *pcMsg );
				cMsg.AddInt32( "selection", nIndex );
				Invoke( &cMsg );
			}
		}
	}
}

void TabView::MouseDown( CPoint cPosition )
{
	if ( m_cTabList.size() == 0 )
	{
		CView::MouseDown( cPosition );
		return;
	}

	if ( cPosition.y >= m_vTabHeight )
	{
		return;
	}
	m_bMouseClicked = true;
	m_cHitPos = cPosition;
	float x = m_vScrollOffset;

	for ( uint i = 0 ; i < m_cTabList.size() ; ++i )
	{
		float vWidth = m_cTabList[i].m_vWidth;
		if ( cPosition.x > x && cPosition.x < x + vWidth - 1 )
		{
			SetSelection( i );
			break;
		}
		x += vWidth;
	}
}

void TabView::MouseUp( CPoint cPosition )
{
	m_bMouseClicked = false;
}

void TabView::MouseMoved( CPoint cNewPos, uint32 nCode, const CMessage* pcData )
{
	int32       nButton = 0x01;

	Window()->CurrentMessage()->FindInt32("buttons", &nButton);

	if ( (nButton & 0x01) && m_bMouseClicked )
	{
		float nOldOffset = m_vScrollOffset;

		float vWidth = Width() + 1.0f;

		if ( m_vTotalWidth <= vWidth && (GetQualifiers() & QUAL_SHIFT) == 0 )
		{
			return;
		}

		m_vScrollOffset += cNewPos.x - m_cHitPos.x;

		if ( m_vTotalWidth <= vWidth )
		{
			if ( m_vScrollOffset > vWidth - m_vTotalWidth )
			{
				m_vScrollOffset = vWidth - m_vTotalWidth;
			}
			if ( m_vScrollOffset <  0 )
			{
				m_vScrollOffset = 0;
			}
		} else {
			vWidth -= 16;
			if ( m_vScrollOffset > 0 )
			{
				m_vScrollOffset = 0;
			}
			
			if ( m_vTotalWidth + m_vScrollOffset < vWidth )
			{
				m_vScrollOffset = vWidth - m_vTotalWidth;
			}
		}
		if ( m_vScrollOffset != nOldOffset )
		{
			m_cHitPos = cNewPos;
			if ( m_vTotalWidth <= vWidth )
			{
				Invalidate( Bounds() );
			}
			else
			{
				CRect cBounds = Bounds();
				Invalidate( CRect( cBounds.left, m_vTabHeight - 2, cBounds.right, m_vTabHeight ) );
			}
			m_pcTopView->ScrollBy( m_vScrollOffset - nOldOffset, 0 );
			Flush();
		}
	}
}


void TabView::KeyDown( const char* pzString, const char* pzRawString, uint32 nQualifiers )
{
    switch( pzString[0] )
    {
        case VK_LEFT_ARROW:
            if ( m_nSelectedTab > 0 ) {
                SetSelection( m_nSelectedTab - 1 );
            }
            break;
        case VK_RIGHT_ARROW:
            if ( m_nSelectedTab < m_cTabList.size() - 1 ) {
                SetSelection( m_nSelectedTab + 1 );
            }
            break;
        default:
            CView::KeyDown( pzString, pzRawString, nQualifiers );
            break;
    }
}

/** 
 * \par Description:
 *        TabView overloads GetPreferredSize() and return the largest size returned
 *        by any of the views associated with the different tabs.
 * \return The largest preffered size returned by it's childs.
 * \sa CView::GetPreferredSize()
 * \author        Kurt Skauen (kurt.skauen@c2i.net)
 *//////////////////////////////////////////////////////////////////////////////

void TabView::GetPreferredSize( float* outWidth, float* outHeight )
{
	float	width, height;
	float	minWidth, minHeight;

	width  = m_vTotalWidth;
	height = 0.0f;

	for ( uint i = 0 ; i < m_cTabList.size() ; ++i )
	{
		if ( m_cTabList[i].m_pcView != NULL )
		{
			m_cTabList[i].m_pcView->GetPreferredSize( &minWidth, &minHeight );

			if ( minWidth > width )
				width = minWidth;

			if ( minHeight > height )
				height = minHeight;
		}
	}

	height += m_vTabHeight + 3.0f;
	width  += 5.0f;

	if (outWidth)
		*outWidth = width;

	if (outHeight)
		*outHeight = height;
}


static Color32_s Tint( const Color32_s& sColor, float vTint )
{
	return( Color32_s(  int( (float(sColor.red) * vTint + 127.0f * (1.0f - vTint)) * 0.5f ),
						int( (float(sColor.green) * vTint + 127.0f * (1.0f - vTint)) * 0.5f ),
						int( (float(sColor.blue) * vTint + 127.0f * (1.0f - vTint)) * 0.5f ),
						sColor.alpha ) );
}


void TabView::TopView::Draw( CRect cUpdateRect )
{
	CRect cBounds = Bounds();
	cBounds.Floor();

	Color32_s sFgCol = get_default_color( COL_SHINE );
	Color32_s sBgCol = get_default_color( COL_SHADOW );
	Color32_s sFgShadowCol = Tint( sFgCol, 0.9f );
	Color32_s sBgShadowCol = Tint( sBgCol, 0.8f );

	float x;

	if ( m_pcParent->m_vTotalWidth > m_pcParent->Width()+1.0f ) {
		x = 0;
	} else {
		x = -2;
	}

	float       vTabHeight = m_pcParent->m_vTabHeight;
	font_height sFontHeight = m_pcParent->m_sFontHeight;
	float       vGlyphHeight = m_pcParent->m_vGlyphHeight;

	FillRect( cBounds, get_default_color( COL_NORMAL ) );
	for ( uint i = 0 ; i < m_pcParent->m_cTabList.size() ; ++i )
	{
		float vWidth = m_pcParent->m_cTabList[i].m_vWidth;
		CRect cTabFrame( x + 2, 2, x + vWidth - 1 + 2, m_pcParent->m_vTabHeight - 1 );
		cTabFrame.Floor();
		if ( i == m_pcParent->m_nSelectedTab )
		{
			cTabFrame.Resize( -2, -2, 2, 0 );
		}

		SetHighColor( sFgShadowCol );
		if ( i != m_pcParent->m_nSelectedTab + 1 )
		{
			StrokeLine( CPoint( cTabFrame.left, cTabFrame.bottom - 2 ), CPoint( cTabFrame.left, cTabFrame.top + 3 ) );
			StrokeLine( CPoint( cTabFrame.left + 3, cTabFrame.top ) );
			StrokeLine( CPoint( cTabFrame.right, cTabFrame.top ) );

			SetHighColor( sFgCol );

			StrokeLine( CPoint( cTabFrame.left + 1, cTabFrame.bottom - 1 ), CPoint( cTabFrame.left + 1, cTabFrame.top + 4 ) );
			StrokeLine( CPoint( cTabFrame.left + 4, cTabFrame.top + 1 ) );
			StrokeLine( CPoint( cTabFrame.right - 1, cTabFrame.top + 1 ) );
		}
		else
		{
			StrokeLine( CPoint( cTabFrame.left + 2, cTabFrame.top ), CPoint( cTabFrame.right, cTabFrame.top ) );
			SetHighColor( sFgCol );
			StrokeLine( CPoint( cTabFrame.left + 2, cTabFrame.top + 1 ), CPoint( cTabFrame.right - 1, cTabFrame.top + 1 ) );
		}

		if ( i != m_pcParent->m_nSelectedTab - 1 )
		{
			SetHighColor( sBgShadowCol );
			StrokeLine( CPoint( cTabFrame.right, cTabFrame.top + 2 ), CPoint( cTabFrame.right, cTabFrame.bottom - 2 ) );

			SetHighColor( sBgCol );
			StrokeLine( CPoint( cTabFrame.right - 1, cTabFrame.top + 2 ), CPoint( cTabFrame.right - 1, cTabFrame.bottom - 1 ) );
		}
		SetHighColor( 0, 0, 0 );

		MovePenTo( x + (vWidth - 4) * 5 / 100 + 4,
				cTabFrame.top + vTabHeight / 2 + sFontHeight.ascender - sFontHeight.line_gap / 2 - vGlyphHeight / 2 + 2 );

		DrawString( m_pcParent->m_cTabList[i].m_cTitle.c_str() );
		x += vWidth;
	}
}

void TabView::Draw( CRect cUpdateRect )
{
	CRect cBounds = Bounds();

	cBounds.Floor();

	float nViewWidth = cBounds.Width() + 1.0f;

	CRect cTabRect = cBounds;
	cTabRect.bottom = m_vTabHeight - 3;
	cTabRect.Floor();

	cBounds.top += m_vTabHeight - 2;

	cBounds.Floor();

	Color32_s sFgCol( get_default_color( COL_SHINE ) );
	Color32_s sBgCol( get_default_color( COL_SHADOW ) );

	Color32_s sFgShadowCol = Tint( sFgCol, 0.9f );
	Color32_s sBgShadowCol = Tint( sBgCol, 0.8f );

	SetHighColor( sBgShadowCol );
	StrokeLine( CPoint( cBounds.right, cBounds.top + 2 ), CPoint( cBounds.right, cBounds.bottom ) );
	StrokeLine( CPoint( cBounds.left, cBounds.bottom ) );

	SetHighColor( sBgCol );

	StrokeLine( CPoint( cBounds.right - 1, cBounds.top + 2 ), CPoint( cBounds.right - 1, cBounds.bottom - 1 ) );
	StrokeLine( CPoint( cBounds.left + 1, cBounds.bottom - 1 ) );

	SetHighColor( sFgShadowCol );
	StrokeLine( CPoint( cBounds.left, cBounds.top - 1 ), CPoint( cBounds.left, cBounds.bottom - 2 ) );

	SetHighColor( sFgCol );
	StrokeLine( CPoint( cBounds.left + 1, cBounds.top + 1 ), CPoint( cBounds.left + 1, cBounds.bottom - 2 ) );

	float x = m_vScrollOffset;

	if ( m_vTotalWidth > nViewWidth )
	{
		x += 8;
	}

	if ( m_vTotalWidth <= nViewWidth )
	{
		SetHighColor( get_default_color( COL_NORMAL ) );
		FillRect( CRect( 0, 0, 1, m_vTabHeight - 3 ) );
		FillRect( CRect( cTabRect.right - 1, 0, cTabRect.right, m_vTabHeight - 3 ) );
	}

	for ( uint i = 0 ; i < m_cTabList.size() ; ++i )
	{
		float vWidth = m_cTabList[i].m_vWidth;
		CRect cTabFrame( x + 2, 2, x + vWidth - 1 + 2, m_vTabHeight - 1 );
		cTabFrame.Floor();
		if ( i == m_nSelectedTab )
		{
			cTabFrame.Resize( -2, -2, 2, 0 );
		}

		if ( m_vTotalWidth <= nViewWidth )
		{
			if ( i != m_nSelectedTab + 1 )
			{
				SetHighColor( sFgShadowCol );
				StrokeLine( CPoint( cTabFrame.left, cTabFrame.bottom - 2 ), CPoint( cTabFrame.left, cTabFrame.top + 3 ) );
				StrokeLine( CPoint( cTabFrame.left + 3, cTabFrame.top ) );
				SetHighColor( sFgCol );
				StrokeLine( CPoint( cTabFrame.left + 1, cTabFrame.bottom - 1 ), CPoint( cTabFrame.left + 1, cTabFrame.top + 4 ) );
				StrokeLine( CPoint( cTabFrame.left + 4, cTabFrame.top + 1 ) );
			}
		}
		else
		{
			if ( i == uint(m_nSelectedTab) )
			{
				StrokeLine( CPoint( cTabFrame.left + 1, cTabFrame.bottom - 1 ), CPoint( cTabFrame.left + 1, cTabFrame.bottom - 1 ) );
			}
		}

		if ( i == uint(m_nSelectedTab) )
		{
			SetHighColor( get_default_color( COL_NORMAL ) );
			float x1 = x + 2.0f;
			float x2 = x + vWidth + 2.0f;

			if ( m_vTotalWidth > nViewWidth )
			{
				if ( x1 < 8 ) {
					x1 = 8;
				}
				if ( x2 > cBounds.right - 9.0f )
				{
					x2 = cBounds.right - 9.0f;
				}
			}
			StrokeLine( CPoint( x1, cTabFrame.bottom - 1 ), CPoint( x2, cTabFrame.bottom - 1 ) );
			StrokeLine( CPoint( x1, cTabFrame.bottom ), CPoint( x2, cTabFrame.bottom ) );

			SetHighColor( sFgShadowCol );

			if ( x >= 0 ) {
				StrokeLine( CPoint( 0, cTabFrame.bottom - 1 ), CPoint( x, cTabFrame.bottom - 1 ) );
			}
			StrokeLine( CPoint( x + vWidth + 2, cTabFrame.bottom - 1 ), CPoint( cBounds.right, cTabFrame.bottom - 1 ) );

			SetHighColor( sFgCol );
			if ( x >= 1 )
			{
				StrokeLine( CPoint( 1, cTabFrame.bottom ), CPoint( x + 1, cTabFrame.bottom ) );
			}
			StrokeLine( CPoint( x + vWidth + 2, cTabFrame.bottom ), CPoint( cBounds.right, cTabFrame.bottom ) );
		}
		x += vWidth;
	}

	if ( m_vTotalWidth > nViewWidth )
	{
		SetHighColor( get_default_color( COL_NORMAL ) );
		FillRect( CRect( 0, 0, 7, m_vTabHeight - 3 ) );
		FillRect( CRect( cTabRect.right - 7, 0, cTabRect.right, m_vTabHeight - 3 ) );
		SetHighColor( sBgCol );
		for ( int x = 2 ; x < 6 ; ++x )
		{
			StrokeLine( CPoint( x, m_vTabHeight / 2 - (x-2) ), CPoint( x, m_vTabHeight / 2 + (x-2) ) );
			StrokeLine( CPoint( cBounds.right - x, m_vTabHeight / 2 - (x-2) ), CPoint( cBounds.right - x, m_vTabHeight / 2 + (x-2) ) );
		}

		SetHighColor( sFgShadowCol );
		StrokeLine( CPoint( 0, m_vTabHeight - 2 ), CPoint( 8, m_vTabHeight - 2 ) );
		StrokeLine( CPoint( cBounds.right - 8, m_vTabHeight - 2 ), CPoint( cBounds.right, m_vTabHeight - 2 ) );
		SetHighColor( sFgCol );
		StrokeLine( CPoint( 1, m_vTabHeight - 1 ), CPoint( 8, m_vTabHeight - 1 ) );
		StrokeLine( CPoint( cBounds.right - 8, m_vTabHeight - 1 ), CPoint( cBounds.right, m_vTabHeight - 1 ) );
	}
}



