/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */


#include <assert.h>
#include <stdio.h>

#include <support/SupportDefs.h>

#include <appserver/protocol.h>
#include <app/Application.h>
#include <app/Message.h>
#include <interface/Window.h>
#include <interface/Bitmap.h>
#include <interface/Menu.h>
#include <support/List.h>
#include <exceptions.h>
#include <macros.h>

#define kWindowPulseTimerID		-1972

using namespace os;

enum { FOCUS_STACK_SIZE = 8 };

class TopView : public CView
{
public:
					TopView( const CRect& cFrame, CWindow* pcWindow );

	virtual void	FrameMoved( CPoint cDelta );
	virtual void	FrameResized( float inWidth, float inHeight );
private:
	CWindow*		m_pcWindow;
};

class CWindow::Private
{
public:
	Private( const char* cTitle ) : m_cTitle( cTitle )
	{
		m_psRenderPkt     = NULL;
		m_nRndBufSize     = RENDER_BUFFER_SIZE;
		m_nFlags          = 0;
		m_hLayer          = -1;
		m_hLayerPort      = -1;
		m_hReplyPort      = -1;
		m_bIsActive       = false;
		m_nButtons        = 0;
		m_pcTopView       = NULL;
		m_pcLastMouseView = NULL;
		m_pcMenuBar       = NULL;
		m_bMenuOpen       = false;
		m_nMouseMoveRun   = 1;
		m_nNextTabOrder   = 0;
		m_pcDefaultButton = NULL;
		m_pcDefaultWheelView = NULL;
		m_bDidScrollRect  = false;
		mPulseRate		  = 500000LL;
		memset( m_apcFocusStack, 0,  FOCUS_STACK_SIZE * sizeof(CView*) );
	}
    
	std::string  m_cTitle;
	WR_Render_s* m_psRenderPkt;
	uint32       m_nRndBufSize;

	uint32       m_nFlags;
	int          m_nMouseMoveRun;
	int32        m_nMouseTransition;
	int          m_hLayer;
	int          m_nNextTabOrder;
	port_id      m_hLayerPort;
	port_id      m_hReplyPort;        // We direct replies to us to this port

	bool         m_bIsActive;
	uint32       m_nButtons;
	CPoint       m_cMousePos;

	TopView*     m_pcTopView;
	CView*       m_pcLastMouseView;
	CView*       m_apcFocusStack[FOCUS_STACK_SIZE];
	CView*       m_pcDefaultButton;
	CView*       m_pcDefaultWheelView;
	CMenu*       m_pcMenuBar;
	bool         m_bMenuOpen;
	bool         m_bIsRunning;
	bool         m_bDidScrollRect;
	window_type  m_windowType;
	window_feel  m_windowFeel;
	window_look  m_windowLook;

	bigtime_t	 mPulseRate;
	CList		 mPulsingViewsList;
};

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CWindow::_Cleanup()
{
	if ( m->m_pcTopView != NULL )
	{
		CMessage cReq( AR_CLOSE_WINDOW );

		cReq.AddInt32( "top_view", m->m_pcTopView->_GetHandle() );

		m->m_pcTopView->_Detached( true, 0 );
		delete m->m_pcTopView;
		if ( CMessenger( m->m_hLayerPort ).SendMessage( &cReq ) < 0 )
		{
			dbprintf( "Error: Window::_Cleanup() failed to send AR_CLOSE_WINDOW request to server\n" );
		}
	}
	delete m->m_psRenderPkt;

	if ( m->m_hReplyPort >= 0 )
	{
		delete_port( m->m_hReplyPort );
	}
	CApplication::GetInstance()->RemoveWindow( this );
	delete m;
}

/** Initialize the window.
 * \par Description:
 *        The window contructor initialize the local object, and
 *        initiate the windows connection to the appserver. The window
 *        is invisible and locked when the contructor returns. The first
 *        call to Show() will unlock the widow, start the looper thread
 *        and make it visible. If you want the windows looper to start
 *        handling messages before the window is made visible, you can
 *        call Start(). This will unlock the window, and start the
 *        looper thread without making the window visible. Simply
 *        calling Run()/Unlock() will not work, since the internal state
 *        of the Window will not be updated.
 * \param cFrame - The size and position of the window.
 * \sa Show(), Start(), CLooper::Lock(), CLooper::Unlock()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CWindow::CWindow( CRect inFrame,
				  const char* inTitle,
				  window_type inType,
				  uint32 inFlags,
				  uint32 inWorkspaces )
        : CLooper( inTitle, DISPLAY_PRIORITY )
{
	window_look	aLook;
	window_feel aFeel;
	
	_InterpretWindowType(inType, &aLook, &aFeel);
	_Init(inTitle, inFrame, aLook, aFeel, inFlags, inWorkspaces);
}


CWindow::CWindow( CRect inFrame,
				  const char* inTitle,
				  window_look inLook,
				  window_feel inFeel,
				  uint32 inFlags,
				  uint32 inWorkspaces )
        : CLooper( inTitle, DISPLAY_PRIORITY )
{
	_Init(inTitle, inFrame, inLook, inFeel, inFlags, inWorkspaces);
}

/** Construct a window attached to a Bitmap
 * \par Description:
 *        This constructor is used internally by the GUI toolkit to create
 *        a window to handle rendering in a bitmap.
 * \param pcBitmap - The bitmap to which the window will be attached.
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CWindow::CWindow( CBitmap* pcBitmap ) : CLooper( "bitmap", DISPLAY_PRIORITY )
{
	assert( CApplication::GetInstance() != NULL );

	m = new Private( "" );

	m->m_nFlags = WND_NO_BORDER;

	try
	{
		m->m_hReplyPort = create_port( "bwindow_reply", DEFAULT_PORT_SIZE );

		CRect cWndBounds = pcBitmap->mBounds;

		m->m_pcTopView = new TopView( cWndBounds, this );

		int hTopView;
		m->m_hLayerPort = CApplication::GetInstance()->CreateWindow( m->m_pcTopView, pcBitmap->mHandle, &hTopView );

		m->m_psRenderPkt             = new WR_Render_s;
		m->m_psRenderPkt->m_hTopView = hTopView;
		m->m_psRenderPkt->nCount     = 0;
		m->m_nRndBufSize             = 0;
		
		m->m_pcTopView->_Attached( this, NULL, hTopView, 0 );
		CApplication::GetInstance()->AddWindow( this );
	}
	catch(...)
	{
		_Cleanup();
		throw;
	}
}

/** Destruct a window.
 * \par Description:
 *        The window destructor will delete all views still attached to
 *        it, and shut down the connection to the appserver.
 * \sa CWindow::CWindow()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CWindow::~CWindow()
{
	_Cleanup();
}


void CWindow::_Init(const char* inTitle,
					CRect inFrame,
					window_look inLook,
					window_feel inFeel,
					uint32 inFlags,
					uint32 inWorkspaces)
{
	assert( CApplication::GetInstance() != NULL );
	m = new Private( inTitle );

	m->m_nFlags = inFlags;
	m->m_bIsRunning = false;

	try
	{
		m->m_hReplyPort = create_port( "window_reply", DEFAULT_PORT_SIZE );
		if ( m->m_hReplyPort < 0 )
		{
			throw( GeneralFailure( "Failed to create message port", ENOMEM ) );
		}

		CRect cWndBounds   = inFrame.Bounds();

		m->m_pcTopView = new TopView( inFrame, this );


		int hTopView;
		m->m_hLayerPort = CApplication::GetInstance()->CreateWindow( m->m_pcTopView, inFrame, inTitle, inFlags,
																	 inWorkspaces, GetMsgPort(), &hTopView );

		if ( m->m_hLayerPort < 0 )
		{
			throw( GeneralFailure( "Failed to open window", ENOMEM ) );
		}

		m->m_psRenderPkt				= new WR_Render_s;
		m->m_psRenderPkt->m_hTopView	= hTopView;
		m->m_psRenderPkt->nCount		= 0;
		m->m_nRndBufSize				= 0;
		m->m_pcMenuBar					= NULL;
		
		SetLook(inLook);
		SetFeel(inFeel);
		
		m->m_pcTopView->_Attached( this, NULL, hTopView, 1 );
		CApplication::GetInstance()->AddWindow( this );
//        Lock();
	}
	catch(...)
	{
		_Cleanup();
		throw;
	}
}


void CWindow::_InterpretWindowType(window_type inType, window_look* outLook, window_feel* outFeel)
{
	window_look	aLook;
	window_feel aFeel;
	
	switch (inType)
	{
		case WT_DOCUMENT_WINDOW:
			aFeel = WF_NORMAL;
			aLook = WL_DOCUMENT_WINDOW;
			break;
			
		case WT_MODAL_WINDOW:
			aFeel = WF_APP_MODAL;
			aLook = WL_MODAL_WINDOW;
			break;

		case WT_FLOATING_WINDOW:
			aFeel = WF_APP_FLOATING;
			aLook = WL_FLOATING_WINDOW;
			break;
			
		case WT_BORDERED_WINDOW:
			aFeel = WF_NORMAL;
			aLook = WL_BORDERED_WINDOW;
			break;

		default:	// titled or unknown goes here
			aFeel = WF_NORMAL;
			aLook = WL_TITLED_WINDOW;
			break;
	}
	
	if (outFeel)
		*outFeel = aFeel;
	
	if (outLook)
		*outLook = aLook;
}


/** Set the appearance and "feel" of a window via a predefined type
 * \par Description:
 *        The window's appearance and "feel" can be set to one of several
 *        predefined values with this function.  A window_type is really
 *        a predefined combination of a window_feel and a window_look.
 * \sa Type()
 * \author        Bill Hayden (hayden@haydentech.com)
 *****************************************************************************/

status_t CWindow::SetType( window_type inType )
{
	status_t err;
	window_look	aLook;
	window_feel aFeel;
	
	// FIXME we don't actually make functional changes yet
	m->m_windowType = inType;
	
	_InterpretWindowType(inType, &aLook, &aFeel);

	err = SetLook(aLook);
	
	if (err == EOK)
		SetFeel(aFeel);
	
	return err;
}


status_t CWindow::SetFeel( window_feel inFeel )
{
	// FIXME we don't actually make functional changes yet
	m->m_windowFeel = inFeel;
	return EOK;
}


status_t CWindow::SetLook( window_look inLook )
{
	// FIXME we don't actually make functional changes yet
	m->m_windowLook = inLook;
	return EOK;
}


/** Return the type of this window
 * \par Description:
 *        This function returns the predefined type of this window as was set
 *        in the constructor or via SetType().
 * \sa SetType()
 * \author        Bill Hayden (hayden@haydentech.com)
 *****************************************************************************/

window_type CWindow::Type() const
{
	return m->m_windowType;
}


/** Return the feel of this window
 * \par Description:
 *        This function returns the predefined feel of this window as was set
 *        in the constructor or via SetFeel().
 * \sa SetFeel()
 * \author        Bill Hayden (hayden@haydentech.com)
 *****************************************************************************/

window_feel CWindow::Feel() const
{
	return m->m_windowFeel;
}


/** Return the look of this window
 * \par Description:
 *        This function returns the predefined look of this window as was set
 *        in the constructor or via SetLook().
 * \sa SetLook()
 * \author        Bill Hayden (hayden@haydentech.com)
 *****************************************************************************/

window_look CWindow::Look() const
{
	return m->m_windowLook;
}


void CWindow::SetFlags( uint32 nFlags )
{
	Flush();
	if ( m->m_hLayerPort >= 0 )
	{
		CMessage cReq( WR_SET_FLAGS );
		cReq.AddInt32( "flags", nFlags );
		if ( CMessenger( m->m_hLayerPort ).SendMessage( &cReq ) < 0 )
		{
			dbprintf( "Error: Window::SetFlags() failed to send request to server\n" );
		}
		else
		{
			m->m_nFlags = nFlags;
		}
	}
}

uint32 CWindow::GetFlags() const
{
	return( m->m_nFlags );
}

/** Assign a default button.
 * \par Description:
 *        Set the default button. The default button will receive KeyDown()/KeyUp()
 *        events generated by the <ENTER> key even when not having focus. Since the
 *        os::Button class will invoke itself when the <ENTER> key is hit, it will
 *        allow the user to activate the default button by simply hitting the
 *        <ENTER> key without making the button active first.
 * \param pcView - Pointer to the View that should receive all events generated
 *                   by the <ENTER> key. This will typically be an instance of
 *                   the os::Button class.
 * \sa DefaultButton(), os::CButton, CView::KeyUp(), CView::KeyDown()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CWindow::SetDefaultButton( CView* pcView )
{
	m->m_pcDefaultButton = pcView;
}

/** Return the view assigned as the default button.
 * \par Description:
 *        Return the pointer last set through SetDefaultButton() or NULL
 *        if not default button is assigned.
 *        Note that this is a os::View pointer, and if you want to
 *        call any os::CButton spesific member you must static_cast it
 *        to a os::CButton. If you are not 100% sure that the View is in fact
 *        a os::CButton, you should use dynamic_cast, and test the result
 *        before touching it.
 * \return Pointer to the default button, or NULL if no default button is assigned.
 * \sa SetDefaultButton()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CView* CWindow::DefaultButton() const
{
	return( m->m_pcDefaultButton );
}


void CWindow::SetDefaultWheelView( CView* pcView )
{
	m->m_pcDefaultWheelView = pcView;
}


CView* CWindow::GetDefaultWheelView() const
{
	return( m->m_pcDefaultWheelView );
}


/** Activate a view.
 * \par Description:
 *        Give a view focus, or remove the focus from the current active
 *        view by passing in a NULL pointer. This has the same effect as
 *        calling pcView->MakeFocus(true).
 * \param pcView - The View that should receive focus, or a NULL pointer to
 *                   remove focus from the currently active view.
 * \return Pointer to the previously active view, or NULL if no view
 *           had focus.
 * \sa CView::MakeFocus()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/


CView* CWindow::SetFocusChild( CView* pcView )
{
	CView* pcPrevView = CurrentFocus();

	if ( pcPrevView == pcView )
	{
		return( pcPrevView );
	}

	if ( pcView == NULL )
	{
		pcView = m->m_apcFocusStack[0];
		memmove( &m->m_apcFocusStack[0], &m->m_apcFocusStack[1], sizeof( CView* ) * (FOCUS_STACK_SIZE - 1) );
		m->m_apcFocusStack[FOCUS_STACK_SIZE-1] = NULL;
	}
	else
	{
		if ( pcPrevView != NULL )
		{
			memmove( &m->m_apcFocusStack[1], &m->m_apcFocusStack[0], sizeof( CView* ) * (FOCUS_STACK_SIZE - 1) );
			m->m_apcFocusStack[0] = pcPrevView;
		}
	}

	SetPreferredHandler( pcView );

	if ( pcPrevView != NULL )
	{
		pcPrevView->Activated( false );
	}

	if ( pcView != NULL )
	{
		pcView->Activated( true );
	}

	return( pcPrevView );
}

/** Change the window title.
 * \par Description:
 *        SetTitle() will change the title rendered in the window border.
 * \param pzTitle
 *        The new window title.
 * \sa Title()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CWindow::SetTitle( const char* inTitle )
{
	m->m_cTitle = inTitle;

	if ( m->m_hLayerPort >= 0 )
	{
		CMessage cReq( WR_SET_TITLE );
		cReq.AddString( "title", m->m_cTitle.c_str() );

		if ( CMessenger( m->m_hLayerPort ).SendMessage( &cReq ) < 0 )
		{
			dbprintf( "Error: CWindow::SetTitle() failed to send request to server\n" );
		}
	}
}

/** Limit the minimum and maximum window size.
 * \par Description:
 *        SetSizeLimits() sets the maximum and minimum size the window
 *        can be resized to by the user. The limits are not enforced by
 *        ResizeBy(), ResizeTo(), and SetFrame()
 * \par Note:
 *        If you want to prevent the user from resizing the window at
 *        all it is normally better to set the \b WND_NOT_RESIZABLE flag
 *        in the constructor or with the SetFlags() member than to set
 *        the same size as minimum and maximum. Setting the \b
 *        WND_NOT_RESIZABLE flag will cause visual changes to notify the
 *        user that the window can not be resized.
 *
 *        Also think twice before making a window non-resizable it is
 *        normally better to make the window layout dynamic through
 *        the builtin layout engine.
 *
 * \param cMinSize
 *        Minimum size in pixels.
 * \param cMaxSize
 *        Maximum size in pixels.
 *        
 * \sa SetFlags()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CWindow::SetSizeLimits( float minX, float minY, float maxX, float maxY )
{
    Flush();
    if ( m->m_hLayerPort >= 0 )
	{
        CMessage cReq( WR_SET_SIZE_LIMITS );
        cReq.AddPoint( "min_size", CPoint(minX, minY));
        cReq.AddPoint( "max_size", CPoint(maxX, maxY));

        if ( CMessenger( m->m_hLayerPort ).SendMessage( &cReq ) < 0 )
		{
            dbprintf( "Error: CWindow::SetSizeLimits() failed to send request to server\n" );
        }
    }
}

void CWindow::SetAlignment( const IPoint& cSize, const IPoint& cSizeOffset, const IPoint& cPos, const IPoint& cPosOffset )
{
    Flush();
    if ( m->m_hLayerPort >= 0 )
	{
        CMessage cReq( WR_SET_ALIGNMENT );
        cReq.AddIPoint( "size", cSize );
        cReq.AddIPoint( "size_off", cSizeOffset );
        cReq.AddIPoint( "pos", cPos );
        cReq.AddIPoint( "pos_off", cPosOffset );

        if ( CMessenger( m->m_hLayerPort ).SendMessage( &cReq ) < 0 )
		{
            dbprintf( "Error: CWindow::SetAlignment() failed to send request to server\n" );
        }
    }
}

/** Obtain the current window title.
 * \return A const pointer to the internal window title.
 * \sa SetTitle()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

const char* CWindow::Title( void ) const
{
    return( m->m_cTitle.c_str() );
}

/** Activate/Deactivate the window.
 * \par Description:
 *        If bFocus is true the window will gain focus, and start receiving
 *        input events generated by the mouse and keyboard. If bFocus is false
 *        the window will lose focus to the window that previously had it. 
 * \par Note:
 *        The appserver keeps a small stack of previously active
 *        windows. When a window give up focus voluntarily it will then
 *        pop one window off the stack and if the stack was not empty
 *        that window will receive focus.
 * \param bFocus - True to activate the window, false to deactivate it.
 * \sa CView::MakeFocus(), SetFocusChild()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CWindow::Activate( bool bFocus )
{
    CMessage cReq( WR_MAKE_FOCUS );

    cReq.AddBool( "focus", bFocus );
    if ( CMessenger( m->m_hLayerPort ).SendMessage( &cReq ) < 0 )
	{
        dbprintf( "Error: CWindow::MakeFocus() failed to send WR_MAKE_FOCUS to server\n" );
    }
}

/** Hide/Unhide the window.
 * \par Description:
 *        When a window is first constructed, it is not made visible on
 *        the screen. You must first call Show(true), The first time
 *        Show() is called it will also unlock the window, and start the
 *        looper thread if it is not done already by calling Start().
 *
 *        You can nest calls to Show(false). It will then require the same
 *        numbers of Show(true) calls to make the window visible.
 * \par Note:
 *        The Window visibility state is global across all desktops.
 * \param bMakeVisible - Set to true to make the window visible, and false to hide it.
 * \sa Start(), CWindow::Window()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CWindow::Show( bool bMakeVisible )
{
    m->m_pcTopView->Show( bMakeVisible );
    if ( m->m_bIsRunning == false )
	{
        m->m_bIsRunning = true;
        Run();
//        Unlock();
    }
}


/** Is this window hidden (i.e. not visible)?.
 * \par Note:
 *        The window visibility state is global across all desktops.
 * \sa Show()
 * \author        Bill Hayden (hayden@haydentech.com)
 *****************************************************************************/
bool CWindow::IsHidden() const
{
	return m->m_pcTopView->IsHidden();
}


/** Unlock the window and start the looper thread.
 * \par Description:
 *        If you would like to unlock the window and start the looper thread
 *        without making the window visible, you can call Start() instead of
 *        Show(). This will invert the locking done by the contructor and
 *        make the looper run. It will also prevent Show() from doing the
 *        same when it is called later to make the window visible.
 *
 * \sa Show(), CWindow::Window()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CWindow::Start()
{
	if ( m->m_bIsRunning == false )
	{
		m->m_bIsRunning = true;
		Run();
		Unlock();
	}
}

/** Set the window's position and size.
 * \par Description:
 *        SetFrame() will set the window's client position and size on the
 *        current desktop.
 * \param cRect - The new frame rectangle of the window's client area.
 * \sa Frame(), Bounds(), MoveBy(), MoveTo(), ResizeBy(), ResizeTo()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CWindow::SetFrame( const CRect& cRect, bool bNotifyServer )
{
	m->m_pcTopView->SetFrame( cRect, bNotifyServer );
	Flush();
}


/** Move the window relative to its current position.
 * \param nDeltaX - The horizontal distance to move the window.
 * \param nDeltaY - The vertical distance to move the window.
 * \sa MoveTo(), ResizeBy(), ResizeTo(), SetFrame()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CWindow::MoveBy( float vDeltaX, float vDeltaY )
{
	m->m_pcTopView->MoveBy( vDeltaX, vDeltaY );
}

/** Move the window to an absolute position
 * \par Description:
 *        Move the window so the upper left corner of the client area is
 *        at the given position.
 * \param x - The new horizontal position.
 * \param y - The new vertical position.
 * \sa MoveBy(), ResizeBy(), ResizeTo(), SetFrame()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CWindow::MoveTo( float x, float y )
{
	m->m_pcTopView->MoveTo( x, y );
}


/** Resize the window relative to its current size
 * \param nDeltaW - The horizontal distance to move the lower-right corner of the window.
 * \param nDeltaH - The vertical distance to move the lower-right corner of the window.
 * \sa ResizeTo(), MoveBy(), MoveTo(), SetFrame()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CWindow::ResizeBy( float vDeltaW, float vDeltaH )
{
	m->m_pcTopView->ResizeBy( vDeltaW, vDeltaH );
}


/** Resize the window to a new absolute size.
 * \param W - The new width of the windows client area.
 * \param W - The new heigth of the windows client area.
 * \sa ResizeBy(), MoveBy(), MoveTo(), SetFrame()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CWindow::ResizeTo( float w, float h )
{
    m->m_pcTopView->ResizeTo( w, h );
}


/** Add a view to the window.
 * \par Description:
 *        AddChild() adds a view to the window's top-view.
 *        The view's CView::AttachedToWindow() and CView::AllAttached()
 *        will be called to let the view know about its new status.
 *
 *        If bAssignTabOrder is true, the view will be assigned
 *        a tab order making it possible to activate the view with
 *        the keyboard.
 * \par Note:
 *        The window share the tab-order counter with the CView class.
 *        
 *        When the window is later closed, it will automatically delete
 *        all remaining child views.
 * \param pcChild - The View to add to the window.
 * \param bAssignTabOrder - If true the view will be assigned a tab-order
 *                            one higher than the previously added view.
 * \return
 * \sa RemoveChild(), CView::AddChild(), CView::AttachedToWindow(), CView::SetTabOrder()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CWindow::AddChild( CView* pcChild, bool bAssignTabOrder )
{
	m->m_pcTopView->AddChild( pcChild, bAssignTabOrder );

	if (pcChild->Flags() & WID_PULSE_NEEDED)
	{
		_AddPulser(pcChild);
	}

}


/** Remove a view from the window.
 * \par Description:
 *        Unlink the view from the windows top-view and call the view's
 *        CView::DetachedFromWindow() and CView::AllDetached()
 * \param pcChild - The view to remove.
 * \sa AddChild(), CView::RemoveChild()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CWindow::RemoveChild( CView* pcChild )
{
	m->m_pcTopView->RemoveChild( pcChild );
}


void CWindow::_ViewDeleted( CView* pcView )
{
	if ( pcView == m->m_pcLastMouseView )
	{
		m->m_pcLastMouseView = NULL;
	}

	if ( pcView == m->m_pcDefaultButton )
	{
		m->m_pcDefaultButton = NULL;
	}
	
	if ( pcView == m->m_pcDefaultWheelView )
	{
		 m->m_pcDefaultWheelView = NULL;
	}

	if (pcView->Flags() & WID_PULSE_NEEDED)
	{
		_RemovePulser(pcView);
	}

	for ( int i = 0 ; i < FOCUS_STACK_SIZE ; ++i )
	{
		if ( m->m_apcFocusStack[i] == pcView )
		{
			for ( int j = i ; j < FOCUS_STACK_SIZE - 1 ; ++j )
			{
				m->m_apcFocusStack[j] = m->m_apcFocusStack[j+1];
			}
			m->m_apcFocusStack[FOCUS_STACK_SIZE-1] = NULL;
		}
	}
}


/** Get the window's position and size on the current desktop.
 * \return The windows client rectangle.
 * \sa SetFrame(), Bounds()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CRect CWindow::Frame() const
{
	return( m->m_pcTopView->Frame() );
}


/** Get the window boundary.
 * \par Description:
 *        Same as Frame() except that the rectangle is moved so
 *        the left/top corner is located at 0,0.
 * \return The windows boundary.
 * \sa Frame(), SetFrame()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CRect CWindow::Bounds() const
{
	return( m->m_pcTopView->Bounds() );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CWindow::_HandleActivate( bool bIsActive, const CPoint& cMousePos )
{
	m->m_bIsActive = bIsActive;

	if ( NULL == m->m_pcTopView )
	{
		dbprintf( "Error: CWindow::_HandleActivate() m->m_pcTopView == NULL\n" );
	}

	m->m_pcTopView->_WindowActivated( m->m_bIsActive );

	m->m_nMouseTransition = MOUSE_INSIDE;
	_MouseEvent( cMousePos, m->m_nButtons, NULL, false ); // FIXME: Get buttons from message
}


/** Flush the window's render queue.
 * \par Description:
 *        When a view is rendering graphics, the render commands are buffered
 *        in the window, and sent to the appserver in batches to reduce
 *        the overhead of message passing. When the CView::Draw() member is
 *        called by the system, the render queue will be automatically flushed
 *        when CView::Draw() returnes. If the view however decides to render
 *        anything between draw messages, the render queue must be flushed
 *        manually to execute the rendering.
 * \par Note:
 *        Flush() does not guarantee that the pixels are indeed on-screen when
 *        it returns, it only guarantees that the rendering will take place
 *        as soon as possible. If you want to make sure all the pixels are
 *        rendered, you must use Sync() instead.
 * \sa Sync(), CBitmap::Sync(), CBitmap::Flush(), CView::Flush(), CView::Sync()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CWindow::Flush()
{
	if ( m->m_psRenderPkt->nCount == 0 )
	{
		return;
	}
	m->m_psRenderPkt->hReply = -1;
	if ( send_msg( m->m_hLayerPort, WR_RENDER, m->m_psRenderPkt,
				sizeof( WR_Render_s ) + m->m_nRndBufSize - RENDER_BUFFER_SIZE ) )
	{
		dbprintf( "Error: CWindow::Flush() failed to send WR_RENDER request to server\n" );
	}
	m->m_psRenderPkt->nCount = 0;
	m->m_nRndBufSize         = 0;
}

/** Flush the render queue, and wait til the rendering is done.
 * \par Description:
 *        Sync() will, like the Flush() member, flush the render queue.
 *        The difference is that Sync() will wait for the appserver to
 *        finish the rendering and send a reply informing that the
 *        rendering is done. This means that when Sync() returnes you
 *        have a guarantee that all the previously issued rendering operations
 *        are indeed executed.
 * \par Note:
 *        Sync() is normally not used on normal windows, since it have a higher
 *        overhead than Flush(), and you normally don't need to wait for the
 *        appserver to finish. It is only useful when the window is attached
 *        to a bitmap, and you want to assure that the rendering indeed has
 *        taken place before you start using the bitmap.
 * \sa Flush(), Bitmap::Sync(), Bitmap::Flush(), CView::Sync(), CView::Flush()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CWindow::Sync()
{
	m->m_psRenderPkt->hReply = m->m_hReplyPort;
	if ( send_msg( m->m_hLayerPort, WR_RENDER, m->m_psRenderPkt,
				sizeof( WR_Render_s ) + m->m_nRndBufSize - RENDER_BUFFER_SIZE ) == 0 )
	{
		if ( get_msg( m->m_hReplyPort, NULL, NULL, 0 ) < 0 )
		{
			dbprintf( "Error: CWindow::Sync() failed to get reply\n" );
		}
	}
	m->m_psRenderPkt->nCount = 0;
	m->m_nRndBufSize         = 0;
}

//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void* CWindow::_AllocRenderCmd( uint32 nCmd, CView* pcView, uint32 nSize )
{
	void* pObj = NULL;

	if ( nSize <= RENDER_BUFFER_SIZE )
	{
		if ( m->m_nRndBufSize + nSize > RENDER_BUFFER_SIZE )
		{
			Flush();
		}
		pObj = (void*) &m->m_psRenderPkt->aBuffer[ m->m_nRndBufSize ];

		GRndHeader_s* psHdr = (GRndHeader_s*) pObj;

		psHdr->nSize      = nSize;
		psHdr->nCmd       = nCmd;
		psHdr->hViewToken = pcView->_GetHandle();

		m->m_psRenderPkt->nCount++;
		m->m_nRndBufSize += nSize;

		if ( nCmd == DRC_COPY_RECT )
		{
			m->m_bDidScrollRect = true;
		}
	}
	else
	{
		dbprintf( "Error: CView::_AllocRenderCmd() packet to big!\n" );
	}
	return( pObj );
}


/** Find the view covering a given position on the window.
 * \par Description:
 *        FindView( const CPoint& cPos ) will do a recursive search through
 *        all attached views to find the topmost view covering the given
 *        point. If no view is intersecting the point NULL is returned.
 * \return Pointer to the View, or NULL.
 * \sa FindView( const char* pzName )
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CView* CWindow::FindView( const CPoint& cPos ) const
{
	CPoint cViewPos = cPos;
	CView*  pcView  = NULL;

	for ( CView* pcTmpWid = m->m_pcTopView->ChildAt( cViewPos ) ;
		  pcTmpWid != NULL ;
		  pcTmpWid = pcTmpWid->ChildAt( cViewPos ) )
	{
		CRect cChildFrame = pcTmpWid->Frame();

		pcView = pcTmpWid;

		cViewPos -= cChildFrame.LeftTop();
		cViewPos -= pcTmpWid->GetScrollOffset();
	}
	return( pcView );
}


void CWindow::_CallMouseMoved( CView* pcView, uint32 nButtons, int nWndTransit, const CMessage* pcData )
{
	while ( pcView != NULL && pcView->_GetMouseMoveRun() == m->m_nMouseMoveRun )
	{
		pcView = pcView->Parent();
	}

	if ( pcView == NULL )
	{
		return;
	}
	
	CPoint cPos = pcView->ConvertFromScreen( m->m_cMousePos );
	int    nCode;

	if ( (nWndTransit == MOUSE_ENTERED || nWndTransit == MOUSE_INSIDE) &&
		pcView->Bounds().Contains( cPos ) )
	{
		if ( pcView->_GetMouseMode() == MOUSE_OUTSIDE )
		{
			nCode = MOUSE_ENTERED;
			pcView->_SetMouseMode( MOUSE_INSIDE );
		}
		else
		{
			nCode = MOUSE_INSIDE;
		}
	}
	else
	{
		if ( pcView->_GetMouseMode() == MOUSE_INSIDE )
		{
			nCode = MOUSE_EXITED;
			pcView->_SetMouseMode( MOUSE_OUTSIDE );
		}
		else
		{
			nCode = MOUSE_OUTSIDE;
		}
	}

	pcView->_SetMouseMoveRun( m->m_nMouseMoveRun );
	pcView->MouseMoved( cPos, nCode, pcData );
}



//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CWindow::_MouseEvent( const CPoint& cNewPos, uint32 nButtons, const CMessage* pcData, bool bReEntered )
{
	CView* pcFocusChild;
	CView* pcMouseView;

	pcMouseView = FindView( m->m_pcTopView->ConvertFromScreen( m->m_cMousePos ) );
	pcFocusChild = CurrentFocus();

	if ( m->m_pcLastMouseView != pcMouseView )
	{
		if ( m->m_pcLastMouseView != NULL )
		{
			_CallMouseMoved( m->m_pcLastMouseView, nButtons, m->m_nMouseTransition, pcData );
		}
		m->m_pcLastMouseView = pcMouseView;
	}

	if ( pcMouseView != NULL )
	{
		_CallMouseMoved( pcMouseView, nButtons, m->m_nMouseTransition, pcData );
	}

	if ( pcFocusChild != NULL )
	{
		_CallMouseMoved( pcFocusChild, nButtons, m->m_nMouseTransition, pcData );
	}
}


CView* CWindow::_GetPrevTabView( CView* pcCurrent )
{
	int nFocusOrder = pcCurrent->GetTabOrder();
	const CLooper::handler_map& cMap = GetHandlerMap();
	CLooper::handler_map::const_iterator i;

	int nHighestOrder = INT_MIN;
	int nClosest      = -1;
	CView* pcClosest      = NULL;
	CView* pcHighestOrder = NULL;

	for ( i = cMap.begin() ; i != cMap.end() ; ++i )
	{
		CView* pcView = dynamic_cast<CView*>((*i).second);
		if ( pcView == NULL || pcView == pcCurrent )
		{
			continue;
		}

		int nOrder = pcView->GetTabOrder();
		if ( nOrder < 0 )
		{
			continue;
		}

		if ( nOrder <= nFocusOrder )
		{
			if ( nFocusOrder - nOrder < nFocusOrder - nClosest )
			{
				pcClosest = pcView;
				nClosest = nOrder;
			}
		}

		if ( nOrder > nHighestOrder )
		{
			nHighestOrder  = nOrder;
			pcHighestOrder = pcView;
		}
	}

	return( (pcClosest != NULL) ? pcClosest : pcHighestOrder );
}


CView* CWindow::_GetNextTabView( CView* pcCurrent )
{
	int nFocusOrder = pcCurrent->GetTabOrder();
	const CLooper::handler_map& cMap = GetHandlerMap();
	CLooper::handler_map::const_iterator i;

	int nLowestOrder = INT_MAX;
	int nClosest     = INT_MAX;
	CView* pcClosest     = NULL;
	CView* pcLowestOrder = NULL;

	for ( i = cMap.begin() ; i != cMap.end() ; ++i )
	{
		CView* pcView = dynamic_cast<CView*>((*i).second);
		if ( pcView == NULL || pcView == pcCurrent )
		{
			continue;
		}

		int nOrder = pcView->GetTabOrder();
		if ( nOrder < 0 )
		{
			continue;
		}

		if ( nOrder >= nFocusOrder )
		{
			if ( nOrder - nFocusOrder < nClosest - nFocusOrder )
			{
				pcClosest = pcView;
				nClosest = nOrder;
			}
		}

		if ( nOrder < nLowestOrder )
		{
			nLowestOrder = nOrder;
			pcLowestOrder = pcView;
		}
	}
	return( (pcClosest != NULL) ? pcClosest : pcLowestOrder );
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CWindow::DispatchMessage( CMessage* pcMsg, CHandler* pcHandler )
{
	CView*        pcView;

	if ( pcMsg->FindPointer( "_widget", (void**) &pcView ) == 0 && pcView != NULL && _FindHandler( pcView->m_nToken ) == pcView )
	{
		switch( pcMsg->what )
		{
			case M_MOUSE_DOWN:
			{
				int32 nButtons;

				if ( pcMsg->FindInt32( "buttons", &nButtons ) == 0 )
				{
					CView* pcChild;

					m->m_nButtons |= 1 << (nButtons-1);
					/*
						if ( pcChild = CurrentFocus() ) {
						pcChild->MouseDown( pcChild->ConvertFromScreen( m->m_cMousePos ) );
						}
						*/
					pcChild = FindView( m->m_pcTopView->ConvertFromScreen( m->m_cMousePos ) );
					if ( pcChild != NULL /*  && pcChild != CurrentFocus()*/ )
					{
						pcChild->MouseDown( pcChild->ConvertFromScreen( m->m_cMousePos ) );
					}
					else
					{
						CView* pcFocusView = CurrentFocus();
						if ( pcFocusView != NULL )
						{
							pcFocusView->MouseDown( pcFocusView->ConvertFromScreen( m->m_cMousePos ) );
						}
					}
				}
				else
				{
					dbprintf( "Error: buttons not found\n" );
				}
				break;
			}

			case M_MOUSE_UP:
			{
				int32 nButtons;

				if ( pcMsg->FindInt32( "buttons", &nButtons ) == 0 )
				{
					CMessage* pcData = NULL;
					CMessage  cData((uint32)0);

					CView*        pcChild;

					if ( pcMsg->FindMessage( "_drag_message", &cData ) == 0 )
					{
						pcData = &cData;
					}

					m->m_nButtons &= ~( 1 << (nButtons - 1) );
					CPoint cPos = m->m_pcTopView->ConvertFromScreen( m->m_cMousePos );
					pcChild = FindView( cPos );
					
					if ( pcChild != NULL )
					{
						pcChild->MouseUp( pcChild->ConvertFromScreen( m->m_cMousePos )/*, pcData*/ );
					}
					
					if ( pcChild != CurrentFocus() )
					{
						pcChild = CurrentFocus();
						if ( pcChild != NULL )
						{
							pcChild->MouseUp( pcChild->ConvertFromScreen( m->m_cMousePos )/*, NULL*/ );
						}
					}
				}
				break;
			}
			case M_MOUSE_MOVED:
			{
				CMessageQueue*        pcQueue = MessageQueue();

				if ( pcMsg->FindInt32( "_transit", &m->m_nMouseTransition ) != 0 ) {
					dbprintf( "Warning: CWindow::DispatchMessage() could not find '_transit' in M_MOUSE_MOVED message\n" );
				}
				bool bDoProcess = true;

				pcQueue->Lock();
				CMessage* pcNextMsg = pcQueue->FindMessage( M_MOUSE_MOVED, 0 );
				if ( pcNextMsg != NULL ) {
					int32 nTransition;
					if ( pcMsg->FindInt32( "_transit", &nTransition ) != 0 )
					{
						dbprintf( "Warning: CWindow::DispatchMessage() could not find '_transit' in M_MOUSE_MOVED message\n" );
					}
					if ( nTransition == m->m_nMouseTransition )
					{
						bDoProcess = false;
					}
				}
				pcQueue->Unlock();
				if ( bDoProcess )         // Wait for the last mouse move message
				{
					if ( pcMsg->FindPoint( "_scr_pos", &m->m_cMousePos ) == 0 )
					{

						CMessage* pcData = NULL;
						CMessage  cData((uint32)0);

						if ( pcMsg->FindMessage( "_drag_message", &cData ) == 0 )
						{
							pcData = &cData;
						}
						m->m_nMouseMoveRun++;
						_MouseEvent( m->m_cMousePos, m->m_nButtons, pcData, false );
					}
					else
					{
						dbprintf( "ERROR : Could not find Point in mouse move message\n" );
					}
				}
				break;
			}
			
			case M_WHEEL_MOVED:
			{
				CView* pcView = m->m_pcDefaultWheelView;
				if ( pcView == NULL )
				{
					pcView = m->m_pcLastMouseView;
				}
				
				if ( pcView != NULL )
				{
					CPoint cDelta;
					pcMsg->FindPoint( "delta_move", &cDelta );
					pcView->WheelMoved( cDelta );
				}
				break;
			}

			case M_KEY_DOWN:
			{
				int32              nQualifier;
				const char*        pzString;
				const char*        pzRawString;

				if ( pcMsg->FindInt32( "_qualifiers", &nQualifier ) != 0 )
				{
					dbprintf( "Error: CWindow::DispatchMessage() could not find '_qualifiers' member in M_KEY_DOWN message\n" );
					break;
				}
				
				if ( pcMsg->FindString( "_string", &pzString ) != 0 )
				{
					dbprintf( "Error: CWindow::DispatchMessage() could not find '_string' member in M_KEY_DOWN message\n" );
					break;
				}
				
				if ( pcMsg->FindString( "_raw_string", &pzRawString ) != 0 )
				{
					dbprintf( "Error: CWindow::DispatchMessage() could not find '_raw_string' member in M_KEY_DOWN message\n" );
					break;
				}

				CView* pcFocusChild = CurrentFocus();

				if ( pcFocusChild != NULL )
				{
					if ( m->m_pcDefaultButton != NULL && pzString[0] == VK_ENTER && pzString[1] == '\0' )
					{
						m->m_pcDefaultButton->KeyDown( pzString, pzRawString, nQualifier );
						break;
					}
					int nFocusOrder = pcFocusChild->GetTabOrder();
					if ( nFocusOrder >= 0 && pzString[0] == '\t' )
					{
						if ( nFocusOrder != -1 )
						{
							CView* pcNext;

							if ( nQualifier & QUAL_SHIFT )
							{
								pcNext = _GetPrevTabView( pcFocusChild );
							}
							else
							{
								pcNext = _GetNextTabView( pcFocusChild );
							}

							if ( pcNext != NULL )
							{
								pcNext->MakeFocus();
							}
						}
					}
					else
					{
						pcFocusChild->KeyDown( pzString, pzRawString, nQualifier );
					}
				}
				break;
			}

			case M_KEY_UP:
			{
				CView* pcFocusChild = CurrentFocus();
				if ( pcFocusChild != NULL )
				{
					int32              nQualifier;
					const char*        pzString;
					const char*        pzRawString;

					if ( pcMsg->FindInt32( "_qualifiers", &nQualifier ) != 0 )
					{
						dbprintf( "Error: CWindow::DispatchMessage() could not find '_qualifiers' member in M_KEY_UP message\n" );
						break;
					}
					if ( pcMsg->FindString( "_string", &pzString ) != 0 )
					{
						dbprintf( "Error: CWindow::DispatchMessage() could not find '_string' member in M_KEY_UP message\n" );
						break;
					}
					if ( pcMsg->FindString( "_raw_string", &pzRawString ) != 0 )
					{
						dbprintf( "Error: CWindow::DispatchMessage() could not find '_raw_string' member in M_KEY_UP message\n" );
						break;
					}
					if ( m->m_pcDefaultButton != NULL && pzString[0] == VK_ENTER && pzString[1] == '\0' ) {
						m->m_pcDefaultButton->KeyUp( pzString, pzRawString, nQualifier );
					} else {
						pcFocusChild->KeyUp( pzString, pzRawString, nQualifier );
					}
				}
				break;
			}
			case M_PAINT:
			{
				CRect        cRect;
				if ( pcMsg->FindRect( "_rect", &cRect ) == 0 )
				{
					pcView->_BeginUpdate();
					pcView->_ConstrictRectangle( &cRect, CPoint(0,0) );
					pcView->Draw( cRect );
					pcView->_EndUpdate();
					if ( m->m_bDidScrollRect )
					{
						m->m_bDidScrollRect = false;
						Sync();
						SpoolMessages();
					}
					else
					{
						Flush();
					}
				}
				else
				{
					dbprintf( "Error: Could not find rectangle in paint message\n" );
				}
				break;
			}
			case M_FONT_CHANGED:
				pcView->FontChanged( pcView->GetFont() );
				break;
			default:
				CLooper::DispatchMessage( pcMsg, pcHandler );
				break;
		}
	}
	else
	{
		switch( pcMsg->what )
		{
			case M_WINDOW_FRAME_CHANGED:
			{
				CRect cNewFrame;
				if ( pcMsg->FindRect( "_new_frame", &cNewFrame ) != 0 )
				{
					dbprintf( "Error: Could not find '_new_frame' in M_WINDOW_FRAME_CHANGED message\n" );
					break;
				}
				SetFrame( cNewFrame, false );
				if ( send_msg( m->m_hLayerPort, WR_WND_MOVE_REPLY, NULL, 0 ) < 0 )
				{
					dbprintf( "Error: CWindow::DispatchMessage() failed to send WR_WND_MOVE_REPLY to server\n" );
				}
				break;
			}
			case M_MOVE_WINDOW:
			{
				CPoint cDeltaPos;
				CPoint cDeltaSize;

				if ( pcMsg->FindPoint( "_delta_move", &cDeltaPos ) != 0 ) {
					dbprintf( "Error: Could not find '_delta_move' in M_MOVE_WINDOW message\n" );
					break;
				}
				if ( pcMsg->FindPoint( "_delta_size", &cDeltaSize ) != 0 ) {
					dbprintf( "Error: Could not find '_delta_size' in M_MOVE_WINDOW message\n" );
					break;
				}
				CRect cFrame = Frame();

				cFrame.right  += cDeltaSize.x + cDeltaPos.x;
				cFrame.bottom += cDeltaSize.y + cDeltaPos.y;
				cFrame.left   += cDeltaPos.x;
				cFrame.top    += cDeltaPos.y;
				SetFrame( cFrame );
				Flush();
				if ( send_msg( m->m_hLayerPort, WR_WND_MOVE_REPLY, NULL, 0 ) < 0 ) {
					dbprintf( "Error: CWindow::DispatchMessage() failed to send WR_WND_MOVE_REPLY to server\n" );
				}
				break;
			}
			
			case M_WINDOW_ACTIVATED:
			{
				bool        bIsActive;

				if ( pcMsg->FindBool( "_is_active", &bIsActive ) != 0 )
				{
					dbprintf( "Error: Failed to find '_is_active' member in M_WINDOW_ACTIVATED message\n" );
				}

				if ( pcMsg->FindPoint( "_scr_pos", &m->m_cMousePos ) != 0 )
				{
					dbprintf( "Error: Failed to find '_scr_pos' member in M_WINDOW_ACTIVATED message\n" );
				}
				_HandleActivate( bIsActive, m->m_cMousePos );
				break;
			}
			default:
				CLooper::DispatchMessage( pcMsg, pcHandler );
//        m->m_pcMouseChild = NULL;
				break;
		}
	}
}


bool PulseOneView(void* inView)
{
	((CView*)inView)->Pulse();
	return false;
}


void CWindow::MessageReceived(CMessage *message)
{
	if (message->what == M_PULSE)
	{
		m->mPulsingViewsList.DoForEach(PulseOneView);
	}
}


//----------------------------------------------------------------------------
// NAME:
// DESC:
//        Send a delete request to the appserver to get rid of the server
//        part of a view object. When the object is deleted on the server
//        we ask the CLooper to copy all unread messages into the internal
//        message queue and then filter out all messages that was targeted
//        to the delete view.
// NOTE:
// SEE ALSO:
//----------------------------------------------------------------------------

void CWindow::_DeleteViewFromServer( CView* pcView )
{
	Sync();
	if ( pcView != m->m_pcTopView )
	{
		CMessage cReq( WR_DELETE_VIEW );

		cReq.AddInt32( "top_view", m->m_pcTopView->_GetHandle() );
		cReq.AddInt32( "handle", pcView->_GetHandle() );

		if ( CMessenger( m->m_hLayerPort ).SendMessage( &cReq ) < 0 )
		{
			dbprintf( "Error: CWindow::_DeleteViewFromServer() failed to send message to server\n" );
		}
	}

	SpoolMessages(); // Copy all messages from the message port to the message queue.
	CMessageQueue*        pcQueue = MessageQueue();
	CMessageQueue cTmp;
	CMessage*     pcMsg;

	pcQueue->Lock();

	while( (pcMsg = pcQueue->NextMessage()) != NULL )
	{
		CView* pcTmpView;
		if ( pcMsg->FindPointer( "_widget", (void**) &pcTmpView ) == 0 && pcTmpView == pcView ) {
			delete pcMsg;
		}
		else
		{
			cTmp.AddMessage( pcMsg );
		}
	}

	while( (pcMsg = cTmp.NextMessage()) != NULL )
	{
		pcQueue->AddMessage( pcMsg );
	}

	pcQueue->Unlock();
}


bool CWindow::IsActive() const
{
	return( m->m_bIsActive || m->m_bMenuOpen );
}


int CWindow::ToggleDepth()
{
	return( m->m_pcTopView->ToggleDepth() );
}


CPoint CWindow::_GetPenPosition( int nViewHandle )
{
	WR_GetPenPosition_s      sReq;
	WR_GetPenPositionReply_s sReply;

	sReq.m_hTopView    = m->m_pcTopView->_GetHandle();
	sReq.m_hReply      = m->m_hReplyPort;
	sReq.m_hViewHandle = nViewHandle;

	if ( send_msg( m->m_hLayerPort, WR_GET_PEN_POSITION, &sReq, sizeof( sReq ) ) == 0 )
	{
		if ( get_msg( m->m_hReplyPort, NULL, &sReply, sizeof( sReply ) ) < 0 )
		{
			dbprintf( "Error: CWindow::_GetPenPosition() failed to get reply\n" );
		}
	}
	else
	{
		dbprintf( "Error: CWindow::_GetPenPosition() failed to send request to server\n" );
	}

	return( sReply.m_cPos );
}


void CWindow::FrameMoved( CPoint cDelta )
{
}


void CWindow::FrameResized( float inWidth, float inHeight )
{
}


void CWindow::ScreenModeChanged( const IPoint& cNewRes, color_space eColorSpace )
{
}


void CWindow::WorkspaceActivated( int32 inWorkspace, bool inActive )
{
}


void CWindow::WorkspacesChanged( uint32 inOldWorkspace, uint32 inNewWorkspace )
{
}


uint32 CWindow::Workspaces() const
{
}


void CWindow::SetWorkspaces(uint32)
{
}


TopView::TopView( const CRect& cFrame, CWindow* pcWindow ) :
    CView( cFrame, "window_top_view", CF_FOLLOW_ALL, WID_CLEAR_BACKGROUND )
{
	m_pcWindow = pcWindow;
}


void TopView::FrameMoved( CPoint cDelta )
{
	m_pcWindow->FrameMoved( cDelta );
}


void TopView::FrameResized( float inWidth, float inHeight )
{
	m_pcWindow->FrameResized( inWidth, inHeight );
}


void CWindow::_DefaultColorsChanged()
{
}


void CWindow::_SetMenuOpen( bool bOpen )
{
	m->m_bMenuOpen = bOpen;
}


CView* CWindow::_GetTopView() const
{
	return( m->m_pcTopView );
}


port_id CWindow::_GetAppserverPort() const
{
	return( m->m_hLayerPort );
}


bigtime_t CWindow::PulseRate() const
{
	return m->mPulseRate;
}


void CWindow::SetPulseRate(bigtime_t inRate)
{
	int oldRate = m->mPulseRate;

	m->mPulseRate = inRate;

	if (m->mPulsingViewsList.CountItems() > 0L)
	{
		if ((oldRate == 0LL) && (inRate > 0LL))
		{
			AddTimer( this, kWindowPulseTimerID, inRate, false );
		}
		else if ((oldRate != 0LL) && (inRate == 0LL))
		{
			RemoveTimer( this, kWindowPulseTimerID );
		}
	}
}


void CWindow::TimerTick( int inID )
{
	if (inID == kWindowPulseTimerID)
	{
		this->PostMessage( M_PULSE );
	}
}


void CWindow::_AddPulser(CView* inView)
{
	if ((m->mPulsingViewsList.IsEmpty()) && (m->mPulseRate > 0LL))
	{
		AddTimer( this, kWindowPulseTimerID, m->mPulseRate, false );
	}

	m->mPulsingViewsList.AddItem(inView);
}


void CWindow::_RemovePulser(CView* inView)
{
	if (m->mPulsingViewsList.RemoveItem((void*)inView))
	{
		if (m->mPulsingViewsList.IsEmpty())
		{
			RemoveTimer( this, kWindowPulseTimerID );
		}
	}
	else
	{
		dbprintf("Tried to remove a non-existant pulsing view\n");
	}

}

void CWindow::FBC_Window1() {}
void CWindow::FBC_Window2() {}
void CWindow::FBC_Window3() {}
void CWindow::FBC_Window4() {}
void CWindow::FBC_Window5() {}
void CWindow::FBC_Window6() {}
