/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */


#include <support/DataIO.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

using namespace os;


CDataIO::~CDataIO()
{
}


CPositionIO::~CPositionIO()
{
}


ssize_t	CPositionIO::Read( void* outBuffer, size_t inSize )
{
	size_t bytesRead = ReadAt( Position(), outBuffer, inSize );
	
	if ( bytesRead > 0 )
	{
		Seek( bytesRead, SEEK_CUR );
	}
	
	return( bytesRead );
}


ssize_t	CPositionIO::Write( const void* inBuffer, size_t inSize )
{
	size_t bytesWritten = WriteAt( Position(), inBuffer, inSize );
	
	if ( bytesWritten > 0 )
	{
		Seek( bytesWritten, SEEK_CUR );
	}
	
	return( bytesWritten );
}


off_t CPositionIO::Position() const
{
	// If the position is before the file (i.e. negative), then
	// return an error code instead of the position.
	
	return (mPosition < 0) ? -1 : mPosition;
}


// ---


CMallocIO::CMallocIO()
{
	mPosition = 0;
	mData = NULL;
	mAllocatedLength = 0;
	mLength = 0;
}


CMallocIO::~CMallocIO()
{
	if (mData != NULL)
		free(mData);
}


ssize_t		CMallocIO::ReadAt(off_t inPosition, void *outData, size_t inSize)
{
	if ((inPosition < 0) || (inPosition > mLength) || (outData == NULL))
		return 0;
		
	ssize_t bytesRead;
	
	if (mLength < inPosition + inSize)
		bytesRead = inSize;
	else
		bytesRead = inSize - ((inPosition + inSize) - mLength);
		
	memcpy(outData, mData, bytesRead);
		
	return bytesRead;
}


ssize_t		CMallocIO::WriteAt(off_t inPosition, const void *inData, size_t inSize)
{
	if (inPosition < 0)
		return 0;
		
	// If we need to grow the buffer, do so
	
	if (mAllocatedLength < inPosition + inSize)
		if (SetSize(inPosition + inSize) < 0)
			return -1;
	
	memcpy(mData, inData, inSize);
	
	return inSize;
}


off_t		CMallocIO::Seek(off_t inPosition, uint32 inSeekMode)
{
	// Note that even though a seek to an invalid offset returns an
	// error, mPosition does in fact get set to the invalid value.
	
	switch (inSeekMode)
	{
		case SEEK_SET:
			mPosition = inPosition;
			break;
		
		case SEEK_CUR:
			mPosition += inPosition;
			break;
		
		case SEEK_END:
			mPosition = mLength + inPosition;
			break;
	}
	
	return (mPosition < 0) ? -1 : mPosition;
}


status_t	CMallocIO::SetSize(off_t inSize)
{
	mLength = inSize;
	
	// Round up the allocated size to the next block size
	
	if (inSize % mBlockSize)
		inSize += (mBlockSize - (inSize % mBlockSize));
		
	mAllocatedLength = inSize;
	
	char* newData = (char*)realloc(mData, inSize);
	
	if (newData)
		mData = newData;
		
	return (newData ? EOK : -1);
}


void		CMallocIO::SetBlockSize(size_t inSize)
{
	mBlockSize = inSize;
}

void		CMallocIO::FBC_MallocIO1()	{}
void		CMallocIO::FBC_MallocIO2()	{}


// ---


CMemoryIO::CMemoryIO(void *inData, size_t inSize)
{
	mIsReadOnly = false;
	mPosition = 0;
	mData = (char*)inData;
	mLength = inSize;
}


CMemoryIO::CMemoryIO(const void *inData, size_t inSize)
{
	mIsReadOnly = true;
	mPosition = 0;
	mData = (char*)inData;
	mLength = inSize;
}


CMemoryIO::~CMemoryIO()
{
}


ssize_t	CMemoryIO::ReadAt(off_t inPosition, void *outData, size_t inSize)
{
	if ((inPosition < 0) || (inPosition > mLength) || (outData == NULL))
		return 0;
		
	ssize_t bytesRead;
	
	if (mLength < inPosition + inSize)
		bytesRead = inSize;
	else
		bytesRead = inSize - ((inPosition + inSize) - mLength);
		
	memcpy(outData, mData, bytesRead);
		
	return bytesRead;
}


ssize_t	CMemoryIO::WriteAt(off_t inPosition, const void *inData, size_t inSize)
{
	if (mIsReadOnly || (inPosition < 0) || (inPosition > mLength) || (inData == NULL))
		return 0;
		
	ssize_t bytesWritten;
	
	if (mLength < inPosition + inSize)
		bytesWritten = inSize;
	else
		bytesWritten = inSize - ((inPosition + inSize) - mLength);
		
	memcpy(mData, inData, bytesWritten);
		
	return bytesWritten;
}


off_t	CMemoryIO::Seek(off_t inPosition, uint32 inSeekMode)
{
	// Note that even though a seek to an invalid offset returns an
	// error, mPosition does in fact get set to the invalid value.
	
	switch (inSeekMode)
	{
		case SEEK_SET:
			mPosition = inPosition;
			break;
		
		case SEEK_CUR:
			mPosition += inPosition;
			break;
		
		case SEEK_END:
			mPosition = mLength + inPosition;
			break;
	}
	
	return (mPosition < 0) ? -1 : mPosition;
}

	

void	CMemoryIO::FBC_MemoryIO1() {}
void	CMemoryIO::FBC_MemoryIO2() {}
