/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */


#include <stdio.h>
#include <unistd.h>
#include <limits.h>
#include <storage/Directory.h>
#include <storage/Entry.h>
#include <storage/File.h>
#include <storage/SymLink.h>
#include <storage/Path.h>

#include <util/exceptions.h>

using namespace os;

/** Default contructor.
 * \par Description:
 *        Initiate the instance to a known but invalid state.
 *        The instance must be successfully initialized with one
 *        of the SetTo() members before it can be used.
 *
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CDirectory::CDirectory()
{
    mDirIterator = NULL;
}

/** Construct a directory from a path.
 * \par Description:
 *         See: SetTo( const std::string& cPath, int nOpenMode )
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CDirectory::CDirectory( const char* cPath, int nOpenMode ) : CNode( cPath, nOpenMode )
{
	if ( IsDirectory() == false )
	{
		throw errno_exception( "Node is not a CDirectory", ENOTDIR );
	}

	if( m_cPath[m_cPath.length()-1] != '/' )
	{
		m_cPath += "/";
	}

	mDirIterator = opendir( GetPathName().c_str() );
	if ( mDirIterator == NULL )
	{
		throw errno_exception( "Failed to create CDirectory iterator" );
	}
}

/** Construct a directory from a path.
 * \par Description:
 *         See: ( const Directory& cDir, const std::string& cPath, int nOpenMode )
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CDirectory::CDirectory( const CDirectory* cDir, const char* cPath, int nOpenMode ) : CNode( cDir, cPath, nOpenMode )
{
	if ( IsDirectory() == false )
	{
		throw errno_exception( "Node is not a CDirectory", ENOTDIR );
	}

	if( m_cPath[m_cPath.length()-1] != '/' )
	{
		m_cPath += "/";
	}

	mDirIterator = opendir( GetPathName().c_str() );
	if ( mDirIterator == NULL )
	{
		throw errno_exception( "Failed to create CDirectory iterator" );
	}
}

/** Construct a CDirectory from a path
 * \par Description:
 *        See: SetTo( const CEntry& cRef, int nOpenMode )
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CDirectory::CDirectory( const CEntry* inEntry, int nOpenMode ) : CNode( inEntry, nOpenMode )
{
	if ( IsDirectory() == false )
	{
		throw errno_exception( "Node is not a CDirectory", ENOTDIR );
	}
	if( m_cPath[m_cPath.length()-1] != '/' )
	{
		m_cPath += "/";
	}
	mDirIterator = opendir( GetPathName().c_str() );
	if ( mDirIterator == NULL ) {
		throw errno_exception( "Failed to create CDirectory iterator" );
	}
}

/** Construct a CDirectory from a path
 * \par Description:
 *        See: SetTo( const Node& cNode )
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CDirectory::CDirectory( const CNode& cNode ) : CNode( cNode )
{
	if ( IsDirectory() == false )
	{
		throw errno_exception( "Node is not a CDirectory", ENOTDIR );
	}

	if( m_cPath[m_cPath.length()-1] != '/' )
	{
		m_cPath += "/";
	}

	mDirIterator = opendir( GetPathName().c_str() );
	if ( mDirIterator == NULL )
	{
		throw errno_exception( "Failed to create CDirectory iterator" );
	}
}

/** Copy constructor
 * \par Description:
 *        Make a independent copy of another CDirectory.
 *
 *        The new CDirectory will consume a new file descriptor so the
 *        copy might fail (throwing an errno_exception exception) if the
 *        process run out of file descriptors.
 *
 * \param cDir
 *        The CDirectory to copy
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/


CDirectory::CDirectory( const CDirectory& cDir ) : CNode( cDir )
{
	mDirIterator = opendir( GetPathName().c_str() );
	if ( mDirIterator == NULL )
	{
		throw errno_exception( "Failed to create CDirectory iterator" );
	}
}

CDirectory::~CDirectory()
{
	if ( mDirIterator != NULL )
	{
		// the closedir() will close the FD so we must make sure ~Node() don't attempt to do the same.
		_SetFD( -1 ); 
		closedir( mDirIterator );
		mDirIterator = NULL;
	}
}

status_t CDirectory::FDChanged( int nNewFD )
{
	if ( nNewFD >= 0 && IsDirectory() == false )
	{
		errno = ENOTDIR;
		return( -1 );
	}

	if ( nNewFD != -1 )
	{
		DIR* hNewIter = opendir( GetPathName().c_str() );

		if ( hNewIter == NULL )
		{
			return( -1 );
		}
		if ( mDirIterator != NULL )
		{
			// the closedir() will close the FD so we must make sure Node don't attempt to do the same.
			_SetFD( -1 ); 
			closedir( mDirIterator );
		}
		mDirIterator = hNewIter;
	}
	else if ( mDirIterator != NULL )
	{
		// the closedir() will close the FD so we must make sure Node don't attempt to do the same.
		_SetFD( -1 ); 
		closedir( mDirIterator );
		mDirIterator = NULL;
	}
	return( 0 );
}

/*
void CDirectory::Unset()
{
	if ( mDirIterator != NULL )
	{
		// the closedir() will close the FD so we must make sure Node::Unset() don't attempt to do the same.
		_SetFD( -1 ); 
		closedir( mDirIterator );
		mDirIterator = NULL;
	}
	CNode::Unset();
}
*/

/** Open a CDirectory using a path.
 * \f status_t CDirectory::SetTo( const std::string& cPath, int nOpenMode = O_RDONLY )
 * \par Description:
 *        Open the CDirectory pointed to by \p cPath. The path must
 *        be valid and it must point to a CDirectory.
 *
 * \param cPath
 *        The directory to open.
 * \param nOpenMode
 *        Flags describing how to open the directory. Only O_RDONLY,
 *        O_WRONLY, and O_RDWR are relevant to directories. Take a look
 *        at the os::Node documentation for a more detailed
 *        description of open modes.
 *
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/


/** Get the absolute path of the directory
 * \par Description:
 * \par Note:
 * \par Warning:
 * \param
 * \return
 * \par Error codes:
 * \sa
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/


status_t CDirectory::GetPath( std::string* pcPath ) const
{
	*pcPath = GetPathName();
	return 0;
}


status_t CDirectory::GetNextEntryName( std::string* pcName )
{
	struct dirent* psEntry = readdir( mDirIterator );
	if ( psEntry == NULL )
		return( -1 );

	*pcName = psEntry->d_name;
	return 0;
}


status_t CDirectory::GetNextEntry( CEntry* inRef, bool inTraverse )
{
	struct dirent* psEntry = readdir( mDirIterator );
	if ( psEntry == NULL )
		return( -1 );

	status_t nError = inRef->SetTo( this, psEntry->d_name, false );
	if ( nError < 0 )
		return( nError );
	
	return 0;
}


status_t CDirectory::Rewind()
{
	rewinddir( mDirIterator );
	return( 0 );
}


status_t CDirectory::CreateFile( const char* cName, CFile* pcFile, int nAccessMode )
{
	std::string cPath = GetPathName() + cName;
	int nFile = open( cPath.c_str(), O_WRONLY | O_CREAT, nAccessMode );
	if ( nFile < 0 )
		return( nFile );

	close( nFile );
	return( static_cast<CNode*>(pcFile)->SetTo( this, cName, O_RDWR ) );
}


status_t CDirectory::CreateDirectory( const char* cName, CDirectory* pcDir, int nAccessMode )
{
	std::string cDirPath;
	GetPath( &cDirPath );
	CPath cPath( cDirPath.c_str() );
	cPath.Append( cName );

	int nError = mkdir( cPath.Path(), nAccessMode );
	if ( nError < 0 )
		return( nError );

	return( pcDir->SetTo( this, cName, O_RDONLY ) );
}


status_t CDirectory::CreateSymLink( const char* cName, const char* cDestination, CSymLink* pcLink )
{
	std::string cDirPath;
	GetPath( &cDirPath );
	CPath cPath = cDirPath.c_str();
	cPath.Append( cName );

	int nError = symlink( cPath.Path(), cDestination );
	if ( nError < 0 )
		return( nError );

	return( pcLink->SetTo( this, cName, O_RDONLY ) );
}
