/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */


#include <unistd.h>
#include <stdio.h>
#include <limits.h>
#include <string.h>

#include <storage/Entry.h>
#include <storage/SymLink.h>
#include <storage/Path.h>

#include <util/exceptions.h>

using namespace os;


EntryRef::EntryRef()
{
	device = -1;
	directory = -1;
}


EntryRef::EntryRef(dev_t inDevice, ino_t inDir, const char* inName)
{
	device = inDevice;
	directory = inDir;
	name = inName;
}


EntryRef::EntryRef(const EntryRef &inRef)
{
	device = inRef.device;
	directory = inRef.directory;
	name = inRef.name;
}


EntryRef::~EntryRef()
{
}


status_t EntryRef::set_name(const char *inName)
{
	name = inName;
	return EOK;
}


bool EntryRef::operator==(const EntryRef &inRef) const
{
	return ((inRef.device == device) && (inRef.directory == directory) && (inRef.name == name));
}


bool EntryRef::operator!=(const EntryRef &inRef) const
{
	return ((inRef.device != device) || (inRef.directory != directory) || (inRef.name != name));
}


CEntry::CEntry()
{
}


CEntry::CEntry( const CEntry& cRef ) : m_cDirectory(cRef.m_cDirectory), m_cName( cRef.m_cName )
{
}


CEntry::CEntry( const char* a_cPath, bool inFollowLinks )
{
	CPath cPath( a_cPath );

	while ( inFollowLinks )
	{
		try
		{
			CSymLink cLink( cPath.Path() );
			CPath cNewPath;
			if ( cLink.ConstructPath( cPath.GetParent().Path(), &cNewPath ) < 0 )
			{
				break;
			}
			cPath = cNewPath;
		}
		catch( ... )
		{
			break;
		}
	}

	m_cName = cPath.Leaf();
	if ( m_cName.size() == 0 )
	{
		throw errno_exception( "Invalid path name", EINVAL );
	}

	if ( m_cDirectory.SetTo( cPath.GetParent().Path(), O_RDONLY ) < 0 )
	{
		throw errno_exception( "Failed to open directory" );
	}
}


CEntry::CEntry( const CDirectory* cDir, const char* cName, bool inFollowLinks )
{
	if ( m_cDirectory.SetTo( *cDir ) < 0 )
	{
		throw errno_exception( "Failed to copy directory" );
	}

	m_cName = cName;
	if ( inFollowLinks )
	{
		try
		{
			CSymLink cLink( cDir, cName );
			CPath cNewPath;
			if ( cLink.ConstructPath( cDir, &cNewPath ) >= 0 )
			{
				SetTo( cNewPath.Path(), true );
			}
		} catch( ... ) {
		}
	}
}


CEntry::CEntry( const EntryRef* inEntryRef, bool inFollowLinks )
{
	throw errno_exception( "Unimplemented" );
}


CEntry::~CEntry()
{
}


status_t CEntry::SetTo( const CEntry& cRef )
{
	if ( m_cDirectory.SetTo( cRef.m_cDirectory ) < 0 )
	{
		return( -1 );
	}
	m_cName = cRef.m_cName;
	return( 0 );
}


int CEntry::SetTo( const char* a_cPath, bool bFollowLinks )
{
	try
	{
		CEntry cTmp( a_cPath, bFollowLinks );
		return( SetTo( cTmp ) );
	}
	catch( errno_exception& cExc )
	{
		errno = cExc.error();
		return( -1 );
	}
}


int CEntry::SetTo( const CDirectory* cDir, const char* cName, bool bFollowLinks )
{
	try
	{
		CEntry cTmp( cDir, cName, bFollowLinks );
		return( SetTo( cTmp ) );
	}
	catch( errno_exception& cExc )
	{
		errno = cExc.error();
		return( -1 );
	}
}


void CEntry::Unset()
{
	m_cDirectory.Unset();
}


status_t CEntry::InitCheck() const
{
	return( m_cDirectory.InitCheck() );
}


status_t	CEntry::GetName( char *ioBuffer ) const
{
	strcpy(ioBuffer, m_cName.c_str());
	return EOK;
}


std::string	CEntry::GetName() const		// Deprecated
{
	return m_cName;
}


status_t CEntry::GetPath( CPath* ioPath ) const
{
	std::string cPath;

	if ( m_cDirectory.GetPath( &cPath ) < 0 )
		return( -1 );

	cPath += "/";
	cPath += m_cName;
	ioPath->SetTo(cPath.c_str());
	return EOK;
}


status_t CEntry::Rename( const char* cNewName )
{
	std::string cOldPath;
	status_t nError;
	nError = m_cDirectory.GetPath( &cOldPath );
	if ( nError < 0 )
	{
		return( nError );
	}

	cOldPath += "/";
	std::string cNewPath( cOldPath );
	cOldPath += m_cName;

	if ( cNewName[0] == '/' )
	{
		cNewPath = cNewName;
	}
	else
	{
		cNewPath += cNewName;
	}

	nError = rename( cOldPath.c_str(), cNewPath.c_str() );
	if ( nError < 0 )
	{
		return( nError );
	}
	else
	{
		if ( strchr( cNewName, '/' ) == NULL )
		{
			m_cName = cNewName;
		}
		else
		{
			SetTo( cNewPath.c_str(), false );
		}
	}
	return( -1 );
}


status_t CEntry::Delete()
{
	CPath		aPath;
	status_t	err;

	err = GetPath( &aPath );
	if ( err == 0 )
		err = unlink( aPath.Path() );

	return( err );
}


bool CEntry::Exists() const
{
	std::string cPath = m_cDirectory.GetPathName() + m_cName.c_str();
	int nFD = open( cPath.c_str(), O_RDONLY );
	if ( nFD < 0 )
		return( false );

	close( nFD );
	return( true );
}


status_t CEntry::GetRef( EntryRef* ioRef ) const
{
	struct stat		aStat;
	status_t		err;
	
	err = GetStat(&aStat);
	
	if (err == EOK)
	{
		ioRef->device    = aStat.st_dev;
		ioRef->directory = aStat.st_ino;
		err = ioRef->set_name(m_cName.c_str());
	}
	
	return err;
}


status_t CEntry::GetStat( struct stat* psStat ) const
{
	std::string cPath = m_cDirectory.GetPathName() + m_cName.c_str();
	int nFD = open( cPath.c_str(), O_RDONLY );
	if ( nFD < 0 )
		return( -1 );

	if ( fstat( nFD, psStat ) < 0 )
	{
		close( nFD );
		return( -1 );
	}
	else
	{
		close( nFD );
		return( 0 );
	}
}


status_t CEntry::GetParent( CEntry *ioEntry ) const
{
	status_t err = ioEntry->SetTo(*this);
	
	if (err != EOK)
	{
		ioEntry->Unset();
	}
	
	return err;
}


status_t CEntry::GetParent( CDirectory *ioDirectory ) const
{
	status_t err = ioDirectory->SetTo(this);
	
	if (err != EOK)
	{
		ioDirectory->Unset();
	}
	
	return err;
}


const CDirectory& CEntry::GetDirectory() const
{
	return( m_cDirectory );
}
