/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */

 
#include <stdio.h>
#include <limits.h>
#include <storage/File.h>
#include <storage/Entry.h>
#include <storage/Path.h>
#include <support/Autolock.h>

#include <exceptions.h>

#include <new>

using namespace os;

void CFile::_Init()
{
    m_pBuffer       = NULL;
    m_nBufferSize   = DEFAULT_BUFFER_SIZE;
    m_nValidBufSize = 0;
    mPosition       = 0;
    m_nBufPos       = 0;
    m_bDirty        = false;
}

/** Default constructor
 * \par Description:
 *        Initialize the instance to a known but invalid state. The instance
 *        must be successfully initialized with one of the SetTo() members
 *        before it can be used.
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CFile::CFile() : m_cMutex( "os::CFile" )
{
    _Init();
}

/** Construct a file from a regular path.
 * \par Description:
 *        Open the file pointed to by \p cPath. The node must be a regular file
 *        or a symlink pointing at a regular file. Anything else will cause and
 *        errno_exception(EINVAL) exception to be thrown.
 *
 * \param cPath
 *        The file to open. The path can eigther be absolute (starting
 *        with "/") or relative to the current working directory.
 * \param nOpenMode
 
 *        Flags controling how the file should be opened. The value
 *        should be one of the O_RDONLY, O_WRONLY, or O_RDWR. In addition
 *        the following flags can be bitwise or'd in to further control
 *        the operation:
 *
 *        - O_TRUNC        trunate the size to 0.
 *        - O_APPEND        automatically move the file-pointer to the end
 *                        of the file before each write (only valid for files)
 *        - O_NONBLOCK        open the file in non-blocking mode. This is only
 *                        relevant for named pipes and PTY's.
 *        - O_NOCTTY        Don't make the file controlling TTY even if \p cPath
 *                        points to a PTY.
 *
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CFile::CFile( const char* cPath, int nOpenMode ) :
        CNode( cPath, nOpenMode & ~O_NOTRAVERSE ), m_cMutex( "os::CFile" )
{
	if ( IsDirectory() )
	{
		throw errno_exception( "Node not a file", EINVAL );
	}
	_Init();
}

/** Open a file addressed as a name inside a specified directory.
 * \par Description:
 *        Look at CFile( const std::string& cPath, int nOpenMode ) for a more
 *        detailed description.
 *
 * \param cDir
 *        The directory to use as a base for \p cPath
 * \param cPath
 *        Path relative to \p cPath.
 * \param nOpenMode
 *        Flags controlling how to open the file. See
 *        CFile( const std::string& cPath, int nOpenMode )
 *        for a full description.
 *
 * \sa CFile( const std::string& cPath, int nOpenMode )
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/


CFile::CFile( const CDirectory* cDir, const char* cPath, int nOpenMode ) :
        CNode( cDir, cPath, nOpenMode & ~O_NOTRAVERSE ), m_cMutex( "os::CFile" )
{
	if ( IsDirectory() )
	{
		throw errno_exception( "Node not a file", EINVAL );
	}
	_Init();
}

/** Open a file referred to by a os::CEntry.
 * \par Description:
 *        Look at CFile( const std::string& cPath, int nOpenMode ) for a more
 *        detailed description.
 *
 * \param cRef
 *        Reference to the file to open.
 * \param nOpenMode
 *        Flags controlling how to open the file. See
 *        CFile( const std::string& cPath, int nOpenMode )
 *        for a full description.
 *
 * \sa CFile( const std::string& cPath, int nOpenMode )
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CFile::CFile( const CEntry* inEntry, int nOpenMode ) :
        CNode( inEntry, nOpenMode & ~O_NOTRAVERSE ), m_cMutex( "os::CFile" )
{
	if ( IsDirectory() )
	{
		throw errno_exception( "Node not a file", EINVAL );
	}
	_Init();
}

/** Construct a file from a Node.
 * \par Description:
 *        This constructor can be used to "downcast" an os::CNode to a
 *        os::CFile. The Node must represent a regular file, pipe, or PTY.
 *        Attempts to convert symlinks and directories will cause an
 *        errno_exception(EINVAL) exception to be thrown.
 *
 * \param cNode
 *        The Node to downcast.
 *
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CFile::CFile( const CNode& cNode ) :
        CNode( cNode ), m_cMutex( "os::CFile" )
{
	if ( IsDirectory() || IsSymLink() )
	{
		throw errno_exception( "Node not a file", EINVAL );
	}
	_Init();
}

/** Copy constructos.
 * \par Description:
 *        This constructor will make an independent copy of \p cFile.
 *        The copy and original will have their own file pointers and
 *        they will have their own attribute directory iterators (see note).
 * \par Note:
 *        The attribute directory iterator will not be cloned so when
 *        CNode::GetNextAttrName() is called for the first time it will
 *        return the first attribute name even if the iterator was not
 *        located at the beginning of the originals attribute directory.
 *
 * \param cFile
 *        The file object to copy.
 *
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CFile::CFile( const CFile& cFile ) : CNode( cFile ), m_cMutex( "os::CFile" )
{
	_Init();
	mPosition     = cFile.mPosition;
	m_nBufferSize = cFile.m_nBufferSize;
}


CFile::~CFile()
{
	if ( m_bDirty )
	{
		Flush();
	}
	delete[] m_pBuffer;
}


status_t CFile::FDChanged( int nNewFD )
{
	if ( m_bDirty && (InitCheck() == EOK) )
	{
		Flush();
	}
	m_nValidBufSize = 0;
	return( 0 );
}


status_t CFile::_FillBuffer( off_t nPos )
{
	if ( m_bDirty )
	{
		if ( Flush() < 0 )
		{
			return( -1 );
		}
	}
	m_nBufPos = nPos;

	while( true )
	{
		ssize_t nLen = read_pos( GetFileDescriptor(), nPos, m_pBuffer, m_nBufferSize );
		if ( nLen < 0 ) {
			if ( errno == EINTR )
			{
				continue;
			}
			else
			{
				return( -1 );
			}
		}
		m_nValidBufSize = nLen;
		return( 0 );
	}
}

/** Set the size of the files caching buffer.
 * \par Description:
 *        By default the os::CFile class use a 32KB internal memory
 *        buffer to cache resently read/written data. Normally you the
 *        buffer can greatly increase the performance since it reduce
 *        the number of kernel-calls when doing multiple small reads or
 *        writes on the file. There are cases however where it is
 *        beneficial to increase the buffer size or even to disable
 *        buffering entirely.
 *
 *        If you read/write large chunks of data at a time the buffer
 *        might impose more overhead than gain and it could be a good
 *        idea to disable the buffering entirely by setting the buffer
 *        size to 0. When for example streaming video the amount of data
 *        read are probably going to be much larger than the buffer
 *        anyway and each byte is only read once by the application and
 *        if the application reads the file in reasonably sized chunks
 *        the extra copy imposed by reading the data into the internal
 *        buffer and then copy it to the callers buffer will only
 *        decrease the performance.
 *
 * \param nBufferSize
 *        The buffer size in bytes. If the size is set to 0 the file
 *        will be unbuffered and each call to Read()/Write() and
 *        ReadAt()/WriteAt() will translate directly to kernel calls.
 * 
 * \return
 *        On success 0 is returned. If there was not enough memory for
 *        the new buffer -1 will be returned and the global variable
 *        "errno" will be set to ENOMEM.
 * \sa GetBufferSize()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

status_t CFile::SetBufferSize( int nBufferSize )
{
	CAutolock cLock( &m_cMutex );

	if ( nBufferSize == m_nBufferSize )
	{
		return( 0 );
	}

	if ( nBufferSize == 0 )
	{
		if ( m_bDirty ) {
			if ( Flush() < 0 )
			{
				return( -1 );
			}
		}
		m_nValidBufSize = 0;
		m_nBufferSize = 0;
		delete [] m_pBuffer;
		m_pBuffer = NULL;
	}
	else
	{
		try {
			if ( m_pBuffer != NULL ) {
				uint8* pNewBuffer = new uint8[nBufferSize];
				if ( m_nValidBufSize > nBufferSize )
				{
					m_nValidBufSize = nBufferSize;
				}
				if ( m_nValidBufSize > 0 )
				{
					memcpy( pNewBuffer, m_pBuffer, m_nValidBufSize );
				}
				delete[] m_pBuffer;
				m_pBuffer = pNewBuffer;
			}
			m_nBufferSize = nBufferSize;
		}
		catch( std::bad_alloc& cExc )
		{
			errno = ENOMEM;
			return( -1 );
		}
	}
	return( 0 );
}

/** Obtain the files buffer size.
 * \return
 *        The files buffer size in bytes. A value of 0 means the file is
 *        unbuffered.
 * \sa SetBufferSize()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

int CFile::GetBufferSize() const
{
	return( m_nBufferSize );
}

/** Write unwritten data to the underlying file.
 * \par Note:
 *        Flush() will be called automatically by the destructor if there is
 *        unwritten data in the buffer.
 * \return
 *        On success 0 will be returned. On failure -1 will be returned and
 *        a error code will be stored in the global variable "errno".
 *
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

status_t CFile::Flush()
{
	CAutolock cLock( &m_cMutex );

	if ( m_bDirty )
	{
		ssize_t nSize = m_nValidBufSize;
		int     nOffset = 0;
		while ( nSize > 0 )
		{
			ssize_t nLen = write_pos( GetFileDescriptor(), m_nBufPos + nOffset, m_pBuffer + nOffset, nSize );
			if ( nLen < 0 )
			{
				if ( errno == EINTR )
				{
					continue;
				}
				else
				{
					return( -1 );
				}
			}
			nSize   -= nLen;
			nOffset += nLen;
		}
		m_bDirty = false;
	}
	return( 0 );
}


ssize_t CFile::Read( void* outBuffer, size_t inSize )
{
	CAutolock cLock( &m_cMutex );
	
	return CPositionIO::Read(outBuffer, inSize);
}


ssize_t CFile::Write( const void* inBuffer, size_t inSize )
{
	CAutolock cLock( &m_cMutex );
	
	return CPositionIO::Write(inBuffer, inSize);
}


ssize_t CFile::ReadAt( off_t nPos, void* pBuffer, size_t nSize )
{
	CAutolock cLock( &m_cMutex );
	if ( m_nBufferSize == 0 )
	{
		return( read_pos( GetFileDescriptor(), nPos, pBuffer, nSize ) );
	}
	if ( m_pBuffer == NULL )
	{
		try
		{
			m_pBuffer = new uint8[m_nBufferSize];
		} catch( std::bad_alloc& cExc ) {
			errno = ENOMEM;
			return( -1 );
		}
	}
	ssize_t nBytesRead = 0;

	if ( nSize > m_nBufferSize * 2 )
	{
		ssize_t nPreSize = nSize - m_nBufferSize;
		if ( m_nValidBufSize == 0 || nPos + nPreSize <= m_nBufPos || nPos > m_nBufPos + m_nValidBufSize )
		{
			ssize_t nLen = read_pos( GetFileDescriptor(), nPos, pBuffer, nPreSize );
			if ( nLen < 0 )
			{
				return( nLen );
			}
			nBytesRead += nLen;
			nPos += nLen;
			nSize -= nLen;
			pBuffer = ((uint8*)pBuffer) + nLen;
		}
	}

	while ( nSize > 0 )
	{
		if ( nPos < m_nBufPos || nPos >= m_nBufPos + m_nValidBufSize )
		{
			if ( _FillBuffer( nPos ) < 0 )
			{
				return( (nBytesRead > 0) ? nBytesRead : -1 );
			}
			else if ( m_nValidBufSize == 0 )
			{
				return( nBytesRead );
			}
		}
		int nBufOffset = nPos - m_nBufPos;
		ssize_t nCurLen = m_nValidBufSize - nBufOffset;
		if ( nCurLen > nSize )
		{
			nCurLen = nSize;
		}
		
		memcpy( pBuffer, m_pBuffer + nBufOffset, nCurLen );
		nBytesRead += nCurLen;
		nPos += nCurLen;
		nSize -= nCurLen;
		pBuffer = ((uint8*)pBuffer) + nCurLen;
	}

	return( nBytesRead );
}


ssize_t CFile::WriteAt( off_t nPos, const void* pBuffer, size_t nSize )
{
	CAutolock cLock( &m_cMutex );
	if ( m_nBufferSize == 0 )
	{
		return( write_pos( GetFileDescriptor(), nPos, pBuffer, nSize ) );
	}

	if ( m_pBuffer == NULL )
	{
		try
		{
			m_pBuffer = new uint8[m_nBufferSize];
		} catch( std::bad_alloc& cExc )
		{
			errno = ENOMEM;
			return( -1 );
		}
	}
	ssize_t nBytesWritten = 0;

	if ( nSize > m_nBufferSize * 2 )
	{
		ssize_t nPreSize = nSize - m_nBufferSize;
		if ( m_nValidBufSize == 0 || nPos + nPreSize <= m_nBufPos || nPos > m_nBufPos + m_nValidBufSize )
		{
			ssize_t nLen = write_pos( GetFileDescriptor(), nPos, pBuffer, nPreSize );
			if ( nLen < 0 )
			{
				return( nLen );
			}
			nBytesWritten += nLen;
			nPos += nLen;
			nSize -= nLen;
			pBuffer = ((const uint8*)pBuffer) + nLen;
		}
	}

	if ( nPos < m_nBufPos || nPos > m_nBufPos + m_nValidBufSize )
	{
		if ( m_bDirty )
		{
			if ( Flush() < 0 )
			{
				return( -1 );
			}
		}
		m_nBufPos = nPos;
		m_nValidBufSize = 0;
	}

	while ( nSize > 0 )
	{
		if ( nPos >= m_nBufPos + m_nBufferSize )
		{
			if ( m_bDirty )
			{
				if ( Flush() < 0 )
				{
					return( -1 );
				}
			}
			m_nBufPos = nPos;
			m_nValidBufSize = 0;
		}
		int nBufOffset = nPos - m_nBufPos;
		ssize_t nCurLen = m_nBufferSize - nBufOffset;
		if ( nCurLen > nSize )
		{
			nCurLen = nSize;
		}
		
		memcpy( m_pBuffer + nBufOffset, pBuffer, nCurLen );
		if ( nBufOffset + nCurLen > m_nValidBufSize )
		{
			m_nValidBufSize = nBufOffset + nCurLen;
		}
		
		nBytesWritten += nCurLen;
		nPos += nCurLen;
		nSize -= nCurLen;
		pBuffer = ((const uint8*)pBuffer) + nCurLen;
		m_bDirty = true;
	}
	return( nBytesWritten );
}


status_t CFile::GetSize(off_t* outSize) const
{
	CAutolock cLock( &m_cMutex );
	status_t	err;

	err = CNode::GetSize( outSize );

	if ( (err == EOK) &&
		 (outSize != NULL) &&
		 (*outSize != -1) &&
		 (m_nValidBufSize > 0) &&
		 (m_nBufPos + m_nValidBufSize > *outSize) )
	{
		*outSize = m_nBufPos + m_nValidBufSize;
	}

	return( err );
}

/** Move the file pointer.
 * \par Description:
 * \par Note:
 * \par Warning:
 * \param
 * \return
 * \par Error codes:
 * \sa
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/


off_t CFile::Seek( off_t inPosition, uint32 inSeekMode )
{
	CAutolock cLock( &m_cMutex );

	switch( inSeekMode )
	{
		case SEEK_SET:
			if ( inPosition < 0 )
			{
				errno = EINVAL;
				return( -1 );
			}
			mPosition = inPosition;
			return( mPosition );

		case SEEK_CUR:
			if ( mPosition + inPosition < 0 )
			{
				errno = EINVAL;
				return( -1 );
			}
			mPosition += inPosition;
			return( mPosition );

		case SEEK_END:
		{
			off_t nSize;
			GetSize(&nSize);
			if ( nSize + inPosition < 0  )
			{
				errno = EINVAL;
				return( -1 );
			}
			mPosition = nSize + inPosition;
			return( mPosition );
		}

		default:
			errno = EINVAL;
			return( -1 );
	}
}

