/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Copyright (C) 2001-2002 Bill Hayden
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */

 
#include <storage/Mime.h>

using namespace os;

#define validSupertypeCount	7

static const char*	sValidSupertypes[validSupertypeCount] =
{
	"text/",
	"application/",
	"image/",
	"audio/",
	"video/",
	"multipart/",
	"message/"
};

static const char*	sIllegalCharacters = "/<>@,;:\"()[]?=\0";


CMimeType::CMimeType()
{
	Unset();
}


CMimeType::CMimeType(const char *inMimeType)
{
	SetTo(inMimeType);
}


CMimeType::~CMimeType()
{
}


status_t	CMimeType::SetTo(const char *inMimeType)
{
	if (CMimeType::IsValid(inMimeType))
	{
		mType = inMimeType;
		mStatus = EOK;
	}
	else
	{
		mStatus = -1; 	// FIXME what should the error be here?
	}

	return mStatus;
}


void		CMimeType::Unset()
{
	mType.erase();
	mStatus = -1;	// FIXME what should the error be here?
}


status_t	CMimeType::InitCheck() const
{
	return mStatus;
}


const char*	CMimeType::Type() const
{
	return mType.c_str();
}


bool		CMimeType::IsValid() const
{
	return CMimeType::IsValid(mType.c_str());
}


bool		CMimeType::IsSupertypeOnly() const
{
	bool isOnlySupertype = false;
	int  i;

	for (i = 0; i < validSupertypeCount; i++)
	{
		if (mType.compare(sValidSupertypes[i]) == 0)
		{
			isOnlySupertype = true;
			break;
		}
	}

	return isOnlySupertype;
}


status_t	CMimeType::GetSupertype(CMimeType *ioSuperType) const
{
	int  i;

	for (i = 0; i < validSupertypeCount; i++)
	{
		if (mType.find(sValidSupertypes[i]) <= 1)
		{
			ioSuperType->SetTo(sValidSupertypes[i]);
			break;
		}
	}

	return ioSuperType->InitCheck();
}


bool		CMimeType::operator==(const CMimeType &inMimeType) const
{
	return (inMimeType.mType == mType);
}


bool		CMimeType::operator==(const char *inMimeType) const
{
	return (mType.compare(inMimeType) == 0);
}


bool		CMimeType::Contains(const CMimeType *inMimeType) const
{
	bool contains = false;

	if (inMimeType->mType == mType)
	{
		contains = true;
	}
	else if (IsSupertypeOnly())
	{
		CMimeType	supertypeOfParam;

		if (inMimeType->GetSupertype(&supertypeOfParam) == 0)
		{
			if (supertypeOfParam.mType == mType)
			{
				contains = true;
			}
		}
	}

	return contains;
}


bool	CMimeType::IsValid(const char *inString)
{
	bool hasValidSupertype = false;
	bool hasValidSubtype = true;
	int  subtypeStart;
	int  i;
	int  badCharacterCount;

	std::string mimeType(inString);

	for (i = 0; i < validSupertypeCount; i++)
	{
		if (mimeType.find(sValidSupertypes[i]) <= 1)
		{
			hasValidSupertype = true;
			break;
		}
	}

	subtypeStart = mimeType.find('/') + 1;
	badCharacterCount = strlen(sIllegalCharacters);

	for (i = 0; i < badCharacterCount; i++)
	{
		if (mimeType.find(sIllegalCharacters[i], subtypeStart) != std::string::npos)
		{
			hasValidSubtype = false;
			break;
		}
	}

	return (hasValidSupertype && hasValidSubtype);
}

