/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */


#include <unistd.h>
#include <assert.h>
#include <storage/Node.h>
#include <storage/Entry.h>
#include <storage/Directory.h>
#include <storage/Path.h>

#include <util/exceptions.h>

using namespace os;


NodeRef::NodeRef()
{
	device = -1;
	node = -1;
}


NodeRef::NodeRef(const NodeRef &inNodeStruct)
{
	device = inNodeStruct.device;
	node = inNodeStruct.node;
}


bool NodeRef::operator==(const NodeRef& inRef) const
{
	return ((inRef.device == device) && (inRef.node == node));
}


bool NodeRef::operator!=(const NodeRef& inRef) const
{
	return ((inRef.device != device) || (inRef.node != node));
}


/** Default contructor.
 * \par Description:
 *        Initiate the node to a known but "invalid" state.
 *        The node must be initialized with one of the SetTo()
 *        members before any other members can be called.
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CNode::CNode()
{
	mInitCheck = EINVAL;		// FIXME: fix error type
	m_nFD = -1;
	m_hAttrDir = NULL;
}

/** Construct a node from a file path.
 * \par Description:
 *        See: SetTo( const std::string& cPath, int nOpenMode )
 * \par Note:
 *        Since constructors can't return error codes it will throw an
 *        os::errno_exception in the case of failure. The error code can
 *        be retrieved from the exception object.
 *
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CNode::CNode( const char* cPath, int nOpenMode )
{
	mInitCheck = EOK;

	m_cPath = cPath;
	m_nFD = open( cPath, nOpenMode );
	m_hAttrDir = NULL;
	if ( m_nFD < 0 )
	{
		mInitCheck = EINVAL;	// FIXME: fix error type
		throw errno_exception( "Failed to open node" );
	}
}

/** Construct a Node from directory and a name inside that directory.
 * \par Description:
 *        See: SetTo( const Directory& cDir, const std::string& cPath, int nOpenMode )
 * \par Note:
 *        Since constructors can't return error codes it will throw an
 *        os::errno_exception in the case of failure. The error code can
 *        be retrieved from the exception object.
 *
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/


CNode::CNode( const CDirectory* inDirectory, const char* cPath, int nOpenMode )
{
	mInitCheck = EOK;

	if ( inDirectory->InitCheck() != EOK )
	{
		mInitCheck = EINVAL;	// FIXME: fix error type
		throw errno_exception( "Invalid directory", EINVAL );
	}

	m_cPath = inDirectory->GetPathName() + cPath;
	m_nFD = open( m_cPath.c_str(), nOpenMode );
	m_hAttrDir = NULL;
	if ( m_nFD < 0 )
	{
		mInitCheck = EINVAL;	// FIXME: fix error type
		throw errno_exception( "Failed to open node" );
	}
}

/** Construct a node from a file reference.
 * \par Description:
 *        See: SetTo( const CEntry* cRef, int nOpenMode )
 * \par Note:
 *        Since constructors can't return error codes it will throw an
 *        os::errno_exception in the case of failure. The error code can
 *        be retrieved from the exception object.
 *
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CNode::CNode( const CEntry* inEntry, int nOpenMode )
{
	mInitCheck = inEntry->InitCheck();

	if ( mInitCheck != EOK )
	{
		throw errno_exception( "Invalid directory", EINVAL );
	}

	m_cPath = inEntry->GetDirectory().GetPathName() + inEntry->GetName();
	m_nFD = open( m_cPath.c_str(), nOpenMode );
	m_hAttrDir = NULL;
	if ( m_nFD < 0 )
	{
		mInitCheck = EINVAL;	// FIXME: fix error type
		throw errno_exception( "Failed to open node" );
	}
}


/** Copy contructor
 * \par Description:
 *        Copy an existing node. If the node can't be copied an
 *        os::errno_exception will be thrown. Each node consumes
 *        a file-descriptor so running out of FD's will cause
 *        the copy to fail.
 * \par Note:
 *        The attribute directory iterator will not be cloned so when
 *        CNode::GetNextAttrName() is called for the first time it will
 *        return the first attribute name even if the iterator was not
 *        located at the beginning of the originals attribute directory.
 * \param cNode
 *        The node to copy.
 *
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CNode::CNode( const CNode& cNode )
{
	mInitCheck = EOK;

	m_cPath = cNode.m_cPath;
	m_nFD = dup( cNode.m_nFD ); // fcntl( cNode.m_nFD, F_COPYFD, -1 );
	m_hAttrDir = NULL;
	if ( m_nFD < 0 )
	{
		mInitCheck = EINVAL;	// FIXME: fix error type
		throw errno_exception( "Failed to duplicate FD" );
	}
}


/** Destructor
 * \par Description:
 *        Will close the file descriptor and release other resources may
 *        consumed by the Node.
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

CNode::~CNode()
{
	if ( m_hAttrDir != NULL )
	{
		close_attrdir( m_hAttrDir );
	}

	if ( m_nFD != -1 )
	{
		close( m_nFD );
	}
}


void CNode::_SetFD( int nFD )
{
	m_nFD = nFD;
}


status_t CNode::InitCheck() const
{
    if( mInitCheck == EOK && m_nFD < 0 )
	{
		return EINVAL;	// FIXME: just a temp error
	}

	return mInitCheck;
}

/** Open a node using a path.
 * \par Description:
 *        Open a node by path. The path must be valid and the
 *        process must have access to it but it can point to
 *        any kind of FS-node (file, directory, symlink).
 *
 *        The \p nOpenMode should be a compination of any of the O_*
 *        flags defined in <fcntl.h>. Their meaning is the same as when
 *        opening a file with the open() POSIX function except you can
 *        not create a file by setting the O_CREAT flag.
 *
 *        The following flags are accepted:
 *
 *        - O_RDONLY        open the node read-only
 *        - O_WRONLY        open the node write-only
 *        - O_RDWR        open the node for both reading and writing
 *        - O_TRUNC        trunate the size to 0 (only valid for files)
 *        - O_APPEND        automatically move the file-pointer to the end
 *                        of the file before each write (only valid for files)
 *        - O_NONBLOCK        open the file in non-blocking mode
 *        - O_DIRECTORY        throw an exception if \p cPath don't point at
 *                        a directory
 *        - O_NOTRAVERSE        open the symlink it self rather than its target
 *                        if \p cPath points at a symlink
 *
 *        If the node can not be opened a os::errno_exception is thrown.
 * \param cPath
 *        Path pointing at the node to open.
 * \param nOpenMode
 *        Flags controlling how to open the node.
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

status_t CNode::SetTo( const char* cPath, int nOpenMode )
{
	m_cPath = cPath;
	int nNewFD = open( cPath, nOpenMode );
	if ( nNewFD < 0 )
	{
		return( -1 );
	}

	status_t nError = FDChanged( nNewFD );
	if ( nError < 0 )
	{
		int nSavedErrno = errno;
		close( nNewFD );
		errno = nSavedErrno;
		return( nError );
	}

	if ( m_nFD >= 0 )
	{
		close( m_nFD );
	}
	m_nFD = nNewFD;
	return( 0 );
}


/** Open a node using a dir/path pair.
 * \par Description:
 *        Open a node by using a directory and a path relative to
 *        that directory.
 *
 *        The path can either be absolute (\p cDir will then be
 *        ignored) or it can be relative to \p cDir. This have much the
 *        same semantics as setting the current working directory to \p
 *        cDir and then open the node by calling SetTo( const
 *        std::string& cPath, int nOpenMode ) with the path. The main
 *        advantage with this function is that it is thread-safe. You
 *        don't get any races while temporarily changing the current
 *        working directory.
 *
 *        For a more detailed description look at:
 *        SetTo( const std::string& cPath, int nOpenMode )
 *
 * \param cDir
 *        A valid directory from which the \p cPath is relative to.
 * \param cPath
 *        The file path relative to \p cDir. The path can eighter be
 *        absoulute (in which case \p cDir is ignored) or it can
 *        be relative to \p cDir.
 * \param nOpenMode
 *        Flags controlling how to open the node. See
 *        SetTo( const std::string& cPath, int nOpenMode )
 *        for a full description of the various flags.
 *
 * \par Note:
 *        If this call fail the old state of the Node will remain
 *        unchanged
 * \return
 *        On success 0 is returned. On error -1 is returned and a
 *        error code is assigned to the global variable "errno".
 *        The error code can be any of the errors returned by
 *        the open() POSIX function.
 *
 * \sa Node( const std::string& cPath, int nOpenMode )
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

status_t CNode::SetTo( const CDirectory* cDir, const char* cPath, int nOpenMode )
{
	if ( cDir->InitCheck() != EOK )
	{
		errno = EINVAL;
		return( -1 );
	}

	m_cPath = cDir->GetPathName();
	m_cPath += cPath;
	int nNewFD = open( m_cPath.c_str(), nOpenMode );
	if ( nNewFD < 0 )
	{
		return( -1 );
	}

	status_t nError = FDChanged( nNewFD );
	if ( nError < 0 )
	{
		int nSavedErrno = errno;
		close( nNewFD );
		errno = nSavedErrno;
		return( nError );
	}

	if ( m_nFD >= 0 )
	{
		close( m_nFD );
	}
	m_nFD = nNewFD;
	return( 0 );
}


/** Open the node referred to by the given os::CEntry.
 * \par Description:
 *        Same semantics SetTo( const std::string& cPath, int nOpenMode )
 *        except that the node to open is targeted by a file reference
 *        rather than a regular path.
 * \par Note:
 *        If this call fail the old state of the Node will remain
 *        unchanged
 * \return
 *        On success 0 is returned. On error -1 is returned and a
 *        error code is assigned to the global variable "errno".
 *        The error code can be any of the errors returned by
 *        the open() POSIX function.
 * \sa SetTo( const char* cPath, int nOpenMode )
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/


status_t CNode::SetTo( const CEntry* inEntry, int nOpenMode )
{
	if ( inEntry->InitCheck() != EOK )
	{
		errno = EINVAL;
		return( -1 );
	}

	m_cPath = inEntry->GetDirectory().GetPathName() + inEntry->GetName();
	int nNewFD = open( m_cPath.c_str(), nOpenMode );
	if ( nNewFD < 0 )
	{
		return( -1 );
	}

	status_t nError = FDChanged( nNewFD );
	if ( nError < 0 )
	{
		int nSavedErrno = errno;
		close( nNewFD );
		errno = nSavedErrno;
		return( nError );
	}

	if ( m_nFD >= 0 )
	{
		close( m_nFD );
	}

	m_nFD = nNewFD;
	return( 0 );
}

/** Copy another Node
 * \par Description:
 *        Make this node a clone of \p cNode.
 * \par Note:
 *        If this call fail the old state of the Node will remain
 *        unchanged
 * \param cNode
 *        The Node to copy.
 *
 * \return
 *        On success 0 is returned. On error -1 is returned and a
 *        error code is assigned to the global variable "errno".
 *        The error code can be any of the errors returned by
 *        the open() POSIX function.
 * \sa
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

status_t CNode::SetTo( const CNode& inNode )
{
	m_cPath = inNode.m_cPath;
	int nNewFD = dup( inNode.m_nFD ); // fcntl( cNode.m_nFD, F_COPYFD, -1 );
	if ( nNewFD < 0 )
	{
		return( -1 );
	}

	status_t nError = FDChanged( nNewFD );
	if ( nError < 0 )
	{
		int nSavedErrno = errno;
		close( nNewFD );
		errno = nSavedErrno;
		return( nError );
	}

	if ( m_nFD >= 0 )
	{
		close( m_nFD );
	}

	m_nFD = nNewFD;
	return( 0 );
}

/** Reset the Node
 * \par Description:
 *        Will close the file descriptor and other resources may
 *        consumed by the Node. The IsValid() member will return false
 *        until the node is reinitialized with one of the SetTo()
 *        members.
 *
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

void CNode::Unset()
{
	if ( m_hAttrDir != NULL )
	{
		close_attrdir( m_hAttrDir );
		m_hAttrDir = NULL;
	}

	FDChanged( -1 );
	if ( m_nFD != -1 )
	{
		close( m_nFD );
		m_nFD = -1;
	}
}

status_t CNode::FDChanged( int nNewFD )
{
    return( 0 );
}

status_t CNode::GetStat( struct stat* psStat ) const
{
	if (( m_nFD < 0 ) || (psStat == NULL))
	{
		errno = EINVAL;
		return( -1 );
	}

	status_t nError = ::fstat( m_nFD, psStat );
	if ( nError < 0 )
	{
		return( nError );
	}

	return( 0 );
}

/*
dev_t CNode::GetDev() const
{
	if ( m_nFD < 0 )
	{
		errno = EINVAL;
		return( -1 );
	}
	return( m_sStatCache.st_dev );
}
*/


/** Read the node's attribute directory.
 * \par Description:
 *        Iterate over the node's "attribute directory". Call this
 *        member in sequence until it return "0" to iterate over all the
 *        attributes associated with the node. The attribute iterator
 *        can be reset to the first attribute with the RewindAttrdir()
 *        member.
 *
 *        More info about the returned attributes can be obtain with the
 *        StatAttr() member and the content of an attribute can be read with
 *        the ReadAttr() member.
 * \par Note:
 *        Attributes are only stored on AFS or BFS volumes, neither of which
 *        are support on the Linux kernel themselves.  While functional, this
 *        code currently has nowhere to function.
 *
 * \param pcName
 *        Pointer to an STL string that will receive the name.
 * \return
 *        If a new name was successfully obtained 1 will be returned. If
 *        we reach the end of the attribute directory 0 will be
 *        returned.  Any other errors will cause -1 to be returned and a
 *        error code will be stored in the global "errno" variable.
 * \sa StatAttr(), ReadAttr(), WriteAttr()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

status_t CNode::GetNextAttrName( std::string* pcName )
{
	if ( m_nFD < 0 )
	{
		errno = EINVAL;
		return( -1 );
	}

	if ( m_hAttrDir == NULL )
	{
		m_hAttrDir = open_attrdir( m_nFD );
		if ( m_hAttrDir == NULL )
		{
			return( -1 );
		}
	}

	struct dirent* psEntry = read_attrdir( m_hAttrDir );
	if ( psEntry != NULL )
	{
		*pcName = psEntry->d_name;
		return( 1 );
	}

	return( 0 );
}

/** Reset the attribute directory iterator
 * \par Description:
 *        RewindAttrdir() will cause the next GetNextAttrName() call to
 *        return the name of the first attribute associated with this
 *        node.
 * \par Note:
 *        Attributes are only stored on AFS or BFS volumes, neither of which
 *        are support on the Linux kernel themselves.  While functional, this
 *        code currently has nowhere to function.
 * \return
 *        0 on success. On failure -1 is returned and a error code is stored
 *        in the global variable "errno".
 * \sa GetNextAttrName()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

status_t CNode::RewindAttrdir()
{
    if ( m_nFD < 0 )
	{
        errno = EINVAL;
        return( -1 );
    }

    if ( m_hAttrDir != NULL )
	{
        rewind_attrdir( m_hAttrDir );
    }

    return( 0 );
}

/** Add/update an attribute
 * \par Description:
 *        WriteAttr() is used to create new attributes and update
 *        existing attributes. A attribute is simply a chunc of data
 *        that is associated with the file but that is not part of the
 *        files regular data-stream. Attributes can be added to all
 *        kind's of FS nodes like regular files, directories, and
 *        symlinks.
 *
 *        A attribute can contain a untyped stream of data of an
 *        arbritary length or it can contain typed data like integers,
 *        floats, strings, etc etc. The reason for having typed data
 *        is to be able to make a search index that can be used to
 *        for efficient search for files based on their attributes.
 *        The indexing system is not implemented at this time.
 *
 * \par Note:
 *        Attributes are only stored on AFS or BFS volumes, neither of which
 *        are supported on the Linux kernel themselves.  While functional,
 *        this code currently has nowhere to function.
 * \param cAttrName
 *        The name of the attribute. The name must be unique inside the
 *        scope of the node it belongs to. If an attribute already exists
 *        with that name it will be overwritten.
 * \param nFlags
 *        Currently only O_TRUNC is accepted. If you pass in O_TRUNC and a
 *        attribute with the same name already exists it will be truncated
 *        to a size of 0 before the new attribute data is written. By passing
 *        in 0 you can update parts of or extend an existing attribute.
 * \param nType
 *        The data-type of the attribute. This should be one of the ATTR_TYPE_*
 *        types defined in <cosmoe/fs_attr.h>.
 *
 *        - \b ATTR_TYPE_NONE,        Untyped "raw" data of any size.
 *        - \b ATTR_TYPE_INT32,        32-bit integer value (the size must be exactly 4 bytes).
 *        - \b ATTR_TYPE_INT64,        64-bit integer value often used for time-values (the size must be exactly 8 bytes).
 *        - \b ATTR_TYPE_FLOAT,        32-bit floating point value (the size must be exactly 4 bytes).
 *        - \b ATTR_TYPE_DOUBLE,        64-bit floating point value (the size must be exactly 8 bytes).
 *        - \b ATTR_TYPE_STRING,        UTF8 string. The string should not be NUL terminated.
 * \param pBuffer
 *        Pointer to the data to be written.
 * \param nPos
 *        The offset into the attribute where the data will be written.
 * \param nLen
 *        Number of bytes to be written.
 *
 * \return
 *        On success the number of bytes actually written is
 *        returned. On failure -1 is returned and the error code is
 *        stored in the global variable "errno"
 * 
 * \sa ReadAttr(), StatAttr()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

ssize_t CNode::WriteAttr( const char* cAttrName, int nFlags, int nType,
                           const void* pBuffer, off_t nPos, size_t nLen )
{
	if ( m_nFD < 0 )
	{
		errno = EINVAL;
		return( -1 );
	}
	return( write_attr( m_nFD, cAttrName, nFlags, nType, pBuffer, nPos, nLen ) );
}

/** Read the data held by an attribute.
 * \par Description:
 *        Read an arbritary chunk of an attributes data. Both the name
 *        and the type must match for the operation so succede. If you
 *        don't know the type in advance it can be retrieved with the
 *        StatAttr() member.
 * 
 * \par Note:
 *        Attributes are only stored on AFS or BFS volumes, neither of which
 *        are supported on the Linux kernel themselves.  While functional,
 *        this code currently has nowhere to function.
 * \param cAttrName
 *        The name of the attribute to read.
 * \param nType
 *        The expected attribute type. The attribute must be of this type
 *        for the function to succede. See WriteAttr() for a more detailed
 *        descritpion of attribute types.
 *
 * \return
 *        On success the number of bytes actually read is returned. On
 *        failure -1 is returned and the error code is stored in the
 *        global variable "errno".
 * 
 * \sa WriteAttr(), StatAttr()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/


ssize_t CNode::ReadAttr( const char* cAttrName, int nType,
                          void* pBuffer, off_t nPos, size_t nLen )
{
	if ( m_nFD < 0 )
	{
		errno = EINVAL;
		return( -1 );
	}

	return( read_attr( m_nFD, cAttrName, nType, pBuffer, nPos, nLen ) );
}

/** Remove an attribute from an FS node.
 * \par Description:
 *        This will remove the named attribute from the node itself and
 *        if the attribute has been indexed it will also be removed from
 *        the index.
 * \par Note:
 *        Attributes are only stored on AFS or BFS volumes, neither of which
 *        are supported on the Linux kernel themselves.  While functional,
 *        this code currently has nowhere to function.
 * \param cName
 *        Name of the attribute to remove.
 * \return
 *        On success 0 is returned. On failure -1 is returned and the
 *        error code is stored in the global variable "errno".
 * \sa WriteAttr(), ReadAttr()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/


status_t CNode::RemoveAttr( const char* cName )
{
	if ( m_nFD < 0 )
	{
		errno = EINVAL;
		return( -1 );
	}

	return( remove_attr( m_nFD, cName ) );
}

/** Get extended info about an attribute.
 * \par Description:
 *        Call this function to retrieve the size and type of an
 *        attribute.  For a detailed description of the attribute type
 *        take a look at WriteAttr().
 * \par Note:
 *        Attributes are only stored on AFS or BFS volumes, neither of which
 *        are supported on the Linux kernel.  While functional, this code
 *        currently has nowhere to function.
 * \param cName
 *        The name of the attribute to examine.
 * \return
 *        On success 0 is returned. On failure -1 is returned and the
 *        error code is stored in the global variable "errno".
 * \sa WriteAttr(), ReadAttr()
 * \author        Kurt Skauen (kurt@atheos.cx)
 *****************************************************************************/

status_t CNode::StatAttr( const char* cName, struct attr_info* psBuffer )
{
	if ( m_nFD < 0 )
	{
		errno = EINVAL;
		return( -1 );
	}

	return( stat_attr( m_nFD, cName, psBuffer ) );
}


int CNode::GetFileDescriptor() const
{
	return( m_nFD );
}




