/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */


#include <unistd.h>
#include <fcntl.h>

#include <storage/NodeMonitor.h>
#include <storage/Entry.h>
#include <storage/Node.h>
#include <storage/Directory.h>
#include <app/Looper.h>
#include <app/Messenger.h>

using namespace os;




int NodeMonitor::_CreateMonitor( const std::string& cPath, uint32 nFlags, const CMessenger& cTarget )
{
	int nFile = open( cPath.c_str(), O_RDONLY );
	if ( nFile < 0 )
	{
		return( -1 );
	}

	int nMonitor = watch_node( nFile, cTarget.m_hPort, nFlags, (void*)cTarget.m_hHandlerID );
	close( nFile );
	return( nMonitor );
}


int NodeMonitor::_CreateMonitor( const CDirectory& cDir, const std::string& cPath, uint32 nFlags, const CMessenger& cTarget )
{
	if ( cDir.InitCheck() != EOK )
	{
		return( -1 );
	}

	std::string cFullPath = cDir.GetPathName() + cPath;
	int nFile = open( cFullPath.c_str(), O_RDONLY, nFlags );
	if ( nFile < 0 )
	{
		return( -1 );
	}

	int nMonitor = watch_node( nFile, cTarget.m_hPort, nFlags, (void*)cTarget.m_hHandlerID );
	close( nFile );
	return( nMonitor );
}


int NodeMonitor::_CreateMonitor( const CEntry& cRef, uint32 nFlags, const CMessenger& cTarget )
{
	if ( cRef.InitCheck() != EOK )
	{
		return( -1 );
	}

	std::string cFullPath = cRef.GetDirectory().GetPathName() + cRef.GetName();
	int nFile = open( cFullPath.c_str(), O_RDONLY );
	if ( nFile < 0 )
	{
		return( -1 );
	}

	int nMonitor = watch_node( nFile, cTarget.m_hPort, nFlags, (void*)cTarget.m_hHandlerID );
	close( nFile );
	return( nMonitor );
}


int NodeMonitor::_CreateMonitor( const CNode* pcNode, uint32 nFlags, const CMessenger& cTarget )
{
	if ( pcNode->InitCheck() != EOK )
	{
		errno = EINVAL;
		return( -1 );
	}

	int nMonitor = watch_node( pcNode->GetFileDescriptor(), cTarget.m_hPort, nFlags, (void*)cTarget.m_hHandlerID );
	return( nMonitor );
}



NodeMonitor::NodeMonitor()
{
	m_nMonitor = -1;
}


NodeMonitor::NodeMonitor( const std::string& cPath, uint32 nFlags, const CHandler* pcHandler, const CLooper* pcLooper )
{
	m_nMonitor = _CreateMonitor( cPath, nFlags, CMessenger( pcHandler, pcLooper ) );
	if ( m_nMonitor < 0 )
	{
		throw errno_exception( "Failed to create monitor" );
	}
}


NodeMonitor::NodeMonitor( const std::string& cPath, uint32 nFlags, const CMessenger& cTarget )
{
	m_nMonitor = _CreateMonitor( cPath, nFlags, cTarget );
	if ( m_nMonitor < 0 )
	{
		throw errno_exception( "Failed to create monitor" );
	}
}


NodeMonitor::NodeMonitor( const CDirectory& cDir, const std::string& cPath, uint32 nFlags, const CHandler* pcHandler, const CLooper* pcLooper )
{
	if ( cDir.InitCheck() != EOK )
	{
		throw errno_exception( "Invalid directory", EINVAL );
	}

	m_nMonitor = _CreateMonitor( cDir, cPath, nFlags, CMessenger( pcHandler, pcLooper ) );
	if ( m_nMonitor < 0 )
	{
		throw errno_exception( "Failed to create monitor" );
	}
}


NodeMonitor::NodeMonitor( const CDirectory& cDir, const std::string& cPath, uint32 nFlags, const CMessenger& cTarget )
{
	if ( cDir.InitCheck() != EOK )
	{
		throw errno_exception( "Invalid directory", EINVAL );
	}

	m_nMonitor = _CreateMonitor( cDir, cPath, nFlags, cTarget );
	if ( m_nMonitor < 0 )
	{
		throw errno_exception( "Failed to create monitor" );
	}
}


NodeMonitor::NodeMonitor( const CEntry& cRef, uint32 nFlags, const CHandler* pcHandler, const CLooper* pcLooper )
{
	if ( cRef.InitCheck() != EOK )
	{
		throw errno_exception( "Invalid directory", EINVAL );
	}

	m_nMonitor = _CreateMonitor( cRef, nFlags, CMessenger( pcHandler, pcLooper ) );
	if ( m_nMonitor < 0 )
	{
		throw errno_exception( "Failed to create monitor" );
	}
}


NodeMonitor::NodeMonitor( const CEntry& cRef, uint32 nFlags, const CMessenger& cTarget )
{
	if ( cRef.InitCheck() != EOK )
	{
		throw errno_exception( "Invalid directory", EINVAL );
	}

	m_nMonitor = _CreateMonitor( cRef, nFlags, cTarget );
	if ( m_nMonitor < 0 )
	{
		throw errno_exception( "Failed to create monitor" );
	}
}


NodeMonitor::NodeMonitor( const CNode* pcNode, uint32 nFlags, const CHandler* pcHandler, const CLooper* pcLooper )
{
	if ( pcNode->InitCheck() != EOK )
	{
		throw errno_exception( "Invalid node", EINVAL );
	}

	m_nMonitor = _CreateMonitor( pcNode, nFlags, CMessenger( pcHandler, pcLooper ) );
	if ( m_nMonitor < 0 )
	{
		throw errno_exception( "Failed to create monitor" );
	}
}


NodeMonitor::NodeMonitor( const CNode* pcNode, uint32 nFlags, const CMessenger& cTarget )
{
	if ( pcNode->InitCheck() != EOK )
	{
		throw errno_exception( "Invalid node", EINVAL );
	}

	m_nMonitor = _CreateMonitor( pcNode, nFlags, cTarget );
	if ( m_nMonitor < 0 )
	{
		throw errno_exception( "Failed to create monitor" );
	}
}


NodeMonitor::~NodeMonitor()
{
	if ( m_nMonitor >= 0 )
		delete_node_monitor( m_nMonitor );
}


status_t NodeMonitor::Unset()
{
	if ( m_nMonitor >= 0 )
		return( delete_node_monitor( m_nMonitor ) );

	errno = EINVAL;
	return( -1 );
}


status_t NodeMonitor::SetTo( const std::string& cPath, uint32 nFlags, const CMessenger& cTarget )
{
	int nNewMonitor = _CreateMonitor( cPath, nFlags, cTarget );
	if ( nNewMonitor < 0 )
	{
		return( -1 );
	}

	if ( m_nMonitor >= 0 )
	{
		delete_node_monitor( m_nMonitor );
	}
	m_nMonitor = nNewMonitor;
	return( 0 );
}


status_t NodeMonitor::SetTo( const std::string& cPath, uint32 nFlags, const CHandler* pcHandler, const CLooper* pcLooper )
{
	return( SetTo( cPath, nFlags, CMessenger( pcHandler, pcLooper ) ) );
}


status_t NodeMonitor::SetTo( const CDirectory& cDir, const std::string& cPath, uint32 nFlags, const CMessenger& cTarget )
{
	if ( cDir.InitCheck() != EOK )
	{
		return( -1 );
	}

	int nNewMonitor = _CreateMonitor( cDir, cPath, nFlags, cTarget );
	if ( nNewMonitor < 0 )
	{
		return( -1 );
	}

	if ( m_nMonitor >= 0 )
	{
		delete_node_monitor( m_nMonitor );
	}

	m_nMonitor = nNewMonitor;
	return( 0 );
}


status_t NodeMonitor::SetTo( const CDirectory& cDir, const std::string& cPath, uint32 nFlags, const CHandler* pcHandler, const CLooper* pcLooper )
{
	return( SetTo( cDir, cPath, nFlags, CMessenger( pcHandler, pcLooper ) ) );
}


status_t NodeMonitor::SetTo( const CEntry& cRef, uint32 nFlags, const CMessenger& cTarget )
{
	if ( cRef.InitCheck() != EOK )
	{
		return( -1 );
	}

	int nNewMonitor = _CreateMonitor( cRef, nFlags, cTarget );
	if ( nNewMonitor < 0 )
	{
		return( -1 );
	}

	if ( m_nMonitor >= 0 )
	{
		delete_node_monitor( m_nMonitor );
	}

	m_nMonitor = nNewMonitor;
	return( 0 );
}


status_t NodeMonitor::SetTo( const CEntry& cRef, uint32 nFlags, const CHandler* pcHandler, const CLooper* pcLooper )
{
    return( SetTo( cRef, nFlags, CMessenger( pcHandler, pcLooper ) ) );
}


status_t NodeMonitor::SetTo( const CNode* pcNode, uint32 nFlags, const CMessenger& cTarget )
{
    if ( pcNode->InitCheck() != EOK )
	{
        return( -1 );
    }

    int nNewMonitor = _CreateMonitor( pcNode, nFlags, cTarget );
    if ( nNewMonitor < 0 )
	{
        return( -1 );
    }

    if ( m_nMonitor >= 0 )
	{
        delete_node_monitor( m_nMonitor );
    }
    m_nMonitor = nNewMonitor;
    return( 0 );
}


status_t NodeMonitor::SetTo( const CNode* pcNode, uint32 nFlags, const CHandler* pcHandler, const CLooper* pcLooper )
{
    return( SetTo( pcNode, nFlags, CMessenger( pcHandler, pcLooper ) ) );
}


int NodeMonitor::GetMonitor() const
{
    return( m_nMonitor );
}
