/*  libcosmoe.so - the interface to the Cosmoe UI
 *  Portions Copyright (C) 2001-2002 Bill Hayden
 *  Portions Copyright (C) 1999-2001 Kurt Skauen
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA
 */

 
#include <unistd.h>
#include <string.h>
#include <assert.h>

#include <storage/Path.h>
#include <storage/Entry.h>

using namespace os;


CPath::CPath()
{
	m_pzPath = NULL;
	m_nMaxSize = 0;
	mStatus = -1;
}


CPath::CPath( const CPath& cPath )
{
	m_pzPath = NULL;
	m_nMaxSize = 0;
	SetTo( cPath.m_pzPath );
}


CPath::CPath( const CEntry* inEntry )
{
	m_pzPath = NULL;
	m_nMaxSize = 0;
	mStatus = inEntry->GetPath( this );
}


CPath::CPath( const char* pzPath )
{
	m_pzPath = NULL;
	m_nMaxSize = 0;
	SetTo( pzPath );
}


CPath::CPath( const char* pzPath, int nLen )
{
	m_pzPath = NULL;
	m_nMaxSize = 0;
	SetTo( pzPath, nLen );
}


CPath::~CPath()
{
	delete[] m_pzPath;
	m_pzPath = NULL;
}


status_t	CPath::InitCheck() const
{
	return mStatus;
}


void CPath::operator =( const CPath& inPath )
{
	SetTo( inPath.m_pzPath );
}


void CPath::operator =( const char* inPath )
{
	SetTo( inPath );
}


bool CPath::operator==(const CPath& inPath) const
{
	return( strcmp( m_pzPath, inPath.m_pzPath ) == 0 );
}


bool CPath::operator!=(const char* inPath) const
{
	return( strcmp( m_pzPath, inPath ) != 0 );
}


bool CPath::operator!=(const CPath& inPath) const
{
	return( strcmp( m_pzPath, inPath.m_pzPath ) != 0 );
}


bool CPath::operator==(const char* inPath) const
{
	return( strcmp( m_pzPath, inPath ) == 0 );
}


void CPath::SetTo( const char* inPath, bool inNormalize )
{
	SetTo( inPath, strlen( inPath ), inNormalize );
}


void CPath::SetTo( const char* pzPath, int nLen, bool inNormalize )
{
	char zCWD[4096];

	uint nMaxSize = nLen;

	if ( pzPath[0] != '/' )
	{
		getcwd( zCWD, 4095 );
		strcat( zCWD, "/" );
		nMaxSize += strlen( zCWD );
	}
	else
	{
		zCWD[0] = '/';
		zCWD[1] = '\0';
		nMaxSize++;
	}
	nMaxSize++; // Zero terminator

	if ( nMaxSize > m_nMaxSize )
	{
		delete[] m_pzPath;
		m_pzPath = new char[nMaxSize];
		m_nMaxSize = nMaxSize;
	}
	strcpy( m_pzPath, zCWD );
	memcpy( m_pzPath + strlen( zCWD ), pzPath, nLen );
	(m_pzPath + strlen( zCWD ) + nLen)[0] = '\0';

	assert( strlen( m_pzPath ) < m_nMaxSize );

	if (inNormalize)
	{
		_Normalize();
		assert( strlen( m_pzPath ) < m_nMaxSize );
	}

	mStatus = EOK;
}


status_t CPath::Append( const char* inPath, bool inNormalize )
{
	if ( inPath == NULL )
		return -1;

	uint nNewLen = strlen( m_pzPath ) + strlen( inPath ) + 2;

	if ( nNewLen > m_nMaxSize )
	{
		char* pzNewBuf = new char[nNewLen];

		strcpy( pzNewBuf, m_pzPath );
		delete[] m_pzPath;
		m_pzPath = pzNewBuf;
		m_nMaxSize = nNewLen;
	}

	strcat( m_pzPath, "/" );
	strcat( m_pzPath, inPath );
	assert( strlen( m_pzPath ) < m_nMaxSize );

	if (inNormalize)
	{
		_Normalize();
		assert( strlen( m_pzPath ) < m_nMaxSize );
	}

	return EOK;
}


const char* CPath::Leaf() const
{
	char* pzLeaf = strrchr( m_pzPath, '/' );
	if ( pzLeaf == NULL )
	{
		pzLeaf = m_pzPath;
	}

	return( pzLeaf + 1 );
}


CPath CPath::GetParent() const
{
	char* pzLeaf = strrchr( m_pzPath, '/' );
	if ( pzLeaf == NULL )
	{
		return( CPath( "." ) );
	}
	else if ( pzLeaf == m_pzPath )
	{
		return( "/" );
	}

	return( CPath( m_pzPath, pzLeaf - m_pzPath ) );
}


status_t CPath::GetParent(CPath* ioPath)
{
	char* pzLeaf = strrchr( m_pzPath, '/' );

	if ( pzLeaf == NULL )
	{
		ioPath->SetTo( "." );
	}
	else if ( pzLeaf == m_pzPath )
	{
		ioPath->SetTo( "/" );
	}
	else
	{
		ioPath->SetTo( m_pzPath, pzLeaf - m_pzPath );
	}

	return mStatus;
}


const char* CPath::Path() const
{
	return( m_pzPath );
}

void CPath::_Normalize()
{
	int   i;
	char* pzPath  = m_pzPath;
	char* pzDst   = m_pzPath;
	char* pzCurName;
	char  c;
	int   nLen;

	if ( '/' == pzPath[0] )
	{
		pzDst++;
		pzPath++;
	}

	pzCurName = pzPath;

	for ( i = 0, nLen = 0 ; (c = pzPath[i]) ; ++i )
	{
		if ( '/' == c )
		{
			if ( 0 != nLen || (1 == nLen && c == '.') )
			{
				if ( nLen == 2 && pzCurName[0] == '.' && pzCurName[1] == '.' )
				{
					while( pzDst != pzPath ) {
						pzDst--;
						if ( '/' == pzDst[-1] ) {
							break;
						}
					}
				}
				else
				{
					memcpy( pzDst, pzCurName, nLen );
					pzDst[nLen] = '/';
					pzDst        += nLen + 1;
				}

				nLen = 0;
			}
			pzCurName        = &pzPath[i + 1];
		}
		else
		{
			nLen++;
		}
	}

	if ( nLen == 2 && pzCurName[0] == '.' && pzCurName[1] == '.' )
	{
		while( pzDst != pzPath )
		{
			pzDst--;
			if ( '/' == pzDst[-1] )
			{
				break;
			}
		}
	}
	else
	{
		memcpy( pzDst, pzCurName, nLen );
		pzDst += nLen;
	}
	pzDst[0]        =        0;

	nLen = strlen( pzPath ) - 1;

	while ( '/' == pzPath[ nLen ] && nLen > 0 )
	{
		pzPath[ nLen ] = 0;
		nLen--;
	}
}

CPath::operator std::string() const
{
	if ( m_pzPath == NULL )
	{
		return( "" );
	}

	return( m_pzPath );
}
