#include <cosmoe/kernel.h>
#include <cosmoe/image.h>


#include <translation/translator.h>
#include <app/Message.h>
#include <app/Messenger.h>
#include <storage/Directory.h>

#include <vector>

using namespace os;


static TranslatorFactory* g_pcDefaultFactory = NULL;


struct Plugin
{
    int	m_nLibrary;
    std::vector<TranslatorNode*> m_cNodes;
};

struct TranslatorFactory::Internal
{
    std::vector<Plugin*>	 m_cPlugins;
    std::vector<TranslatorNode*> m_cNodes;
};

TranslatorFactory::TranslatorFactory()
{
    m = new Internal;
}

TranslatorFactory::~TranslatorFactory()
{
    delete m;
}

typedef int get_translator_count();
typedef TranslatorNode* get_translator_node( int nIndex );

void TranslatorFactory::LoadAll()
{
	fprintf( stderr, "TranslatorFactory::LoadAll()\n" );
    try
	{
		CDirectory cDir( "translators" );

		std::string cName;
		while( cDir.GetNextEntry( &cName ) == 1 )
		{
			if ( cName == "." || cName == ".." )
				continue;

			std::string cPath( "translators/" );
			cPath += cName;

			int nLib = load_library( cPath.c_str(), 0 );
			if ( nLib < 0 )
			{
				dbprintf( "failed to load translator %s\n", cPath.c_str() );
				continue;
			}
			get_translator_count* pGetTransCount;
			get_translator_node* pGetTrans;

			int nError;

			nError = get_symbol_address( nLib, "get_translator_count", -1, (void**) &pGetTransCount );
			if ( nError < 0 )
			{
				dbprintf( "Translator %s does not export entry point get_translator_count()\n", cName.c_str() );
				unload_library( nLib );
				continue;
			}
			nError = get_symbol_address( nLib, "get_translator_node", -1, (void**) &pGetTrans );
			if ( nError < 0 )
			{
				dbprintf( "Translator %s does not export entry point get_translator_node()\n", cName.c_str() );
				unload_library( nLib );
				continue;
			}

			int nTransCount = pGetTransCount();
			if ( nTransCount == 0 )
			{
				unload_library( nLib );
				continue;
			}
			Plugin* pcPlugin = new Plugin;
			pcPlugin->m_nLibrary = nLib;
			m->m_cPlugins.push_back( pcPlugin );

			for ( int i = 0 ; i < nTransCount ; ++i )
			{
				TranslatorNode* pcNode = pGetTrans( i );
				if ( pcNode != NULL )
				{
					pcPlugin->m_cNodes.push_back( pcNode );
					m->m_cNodes.push_back( pcNode );
				}
			}
		}
    }
	catch(...)
	{
    }
}

status_t TranslatorFactory::FindTranslator( const std::string& cSrcType, const std::string& cDstType,
					    const void* pData, size_t nLen, Translator** ppcTrans )
{
    float vBestQuality = -1000.0f;
    TranslatorNode* pcBestTrans = NULL;
    int nError = ERR_UNKNOWN_FORMAT;

    for ( uint i = 0 ; i < m->m_cNodes.size() ; ++i )
	{
		TranslatorNode* pcNode = m->m_cNodes[i];
		TranslatorInfo sInfo   = pcNode->GetTranslatorInfo();

		if ( sInfo.dest_type != cDstType )
		{
			continue;
		}

		int nResult = pcNode->Identify( cSrcType, cDstType, pData, nLen );

		if ( nResult == 0 )
		{
			if ( sInfo.quality > vBestQuality )
			{
			vBestQuality = sInfo.quality;
			pcBestTrans = pcNode;
			}
		} else if ( nResult == ERR_NOT_ENOUGH_DATA )
		{
			nError = ERR_NOT_ENOUGH_DATA;
		}
    }

    if ( pcBestTrans != NULL )
	{
		*ppcTrans = pcBestTrans->CreateTranslator();
		if ( *ppcTrans != NULL )
		{
			nError = 0;
		}
		else
		{
			nError = ERR_NO_MEM;
		}
    }

    return( nError );
}

int TranslatorFactory::GetTranslatorCount()
{
    return( m->m_cNodes.size() );
}

TranslatorNode* TranslatorFactory::GetTranslatorNode( int nIndex )
{
    return( m->m_cNodes[nIndex] );
}

TranslatorInfo TranslatorFactory::GetTranslatorInfo( int nIndex )
{
    return( m->m_cNodes[nIndex]->GetTranslatorInfo() );
}

Translator* TranslatorFactory::CreateTranslator( int nIndex )
{
    return( m->m_cNodes[nIndex]->CreateTranslator() );
}




TranslatorFactory* TranslatorFactory::GetDefaultFactory()
{
    if ( g_pcDefaultFactory == NULL )
	{
		g_pcDefaultFactory = new TranslatorFactory();
		g_pcDefaultFactory->LoadAll();
    }

    return( g_pcDefaultFactory );
}




struct Translator::Internal
{
    Internal() {
	m_pcDataTarget = NULL;
    }
    CMessenger	       m_cMsgTarget;
    CMessage	       m_cMessage;
    DataReceiver*      m_pcDataTarget;
    bool	           m_bSendData;
    //std::vector<uint8> m_cBuffer;
	uint8*				m_cBuffer;
	size_t				mBufferSize;
};




Translator::Translator()
{
	m = new Internal;
	m->mBufferSize = 0;
}

Translator::~Translator()
{
	delete m;
}

void Translator::SetMessage( const CMessage& cMsg )
{
	m->m_cMessage = cMsg;
}

void Translator::SetTarget( const CMessenger& cTarget, bool bSendData )
{
	m->m_cMsgTarget = cTarget;
	m->m_bSendData = bSendData;
}


void Translator::SetTarget( DataReceiver* pcTarget )
{
	m->m_pcDataTarget = pcTarget;
}


status_t Translator::AddData( const void* pData, size_t nLen, bool bFinal )
{
	size_t nOldSize = m->mBufferSize;
	void*	newBuffer;
	newBuffer = realloc( m->m_cBuffer, nOldSize + nLen );
	if (newBuffer != NULL)
	{
		m->mBufferSize = nOldSize + nLen;
		m->m_cBuffer = (uint8*)newBuffer;
		memcpy( m->m_cBuffer + nOldSize, pData, nLen );

		return( DataAdded( m->m_cBuffer, m->mBufferSize, bFinal ) );
	}

	return -1;
}


status_t Translator::DataAdded( void* pData, size_t nLen, bool bFinal )
{
	return( 0 );
}


void Translator::Invoke( void* pData, size_t nSize, bool bFinal )
{
    if ( m->m_cMsgTarget.IsValid() )
	{
		CMessage cMsg( m->m_cMessage );

		cMsg.AddInt32( "size", nSize );
		cMsg.AddBool( "final", bFinal );

		if ( m->m_bSendData )
		{
			cMsg.AddData( "data", T_ANY_TYPE, pData, nSize );
		}

		m->m_cMsgTarget.SendMessage( &cMsg );
    }

    if ( m->m_pcDataTarget != NULL )
	{
		m->m_pcDataTarget->DataReady( pData, nSize, bFinal );
    }
}


TranslatorNode::TranslatorNode()
{
    m = NULL;
}


TranslatorNode::~TranslatorNode()
{
}

